Package: BindingSiteFinder
Type: Package
Title: Binding site defintion based on iCLIP data
Version: 2.11.0
Authors@R: c(
    person("Mirko", "Brüggemann",
        email="mirko.brueggemann@mail.de",
        role = c("aut", "cre"),
        comment = c(ORCID = "0000-0002-1778-0248")),
    person("Melina", "Klostermann",
        email="melinaklostermann@googlemail.com",
        role = c("aut"),
        comment = c(ORCID = "0000-0003-3122-1095")),
    person("Kathi", "Zarnack",
        email="kathi.zarnack@bmls.de",
        role = c("aut"),
        comment = c(ORCID = "0000-0003-3527-3378")))
Description: Precise knowledge on the binding sites of an RNA-binding 
    protein (RBP) is key to understand (post-) transcriptional regulatory 
    processes. Here we present a workflow that describes how exact 
    binding sites can be defined from iCLIP data. The package provides functions
    for binding site definition and result visualization. For details please
    see the vignette. 
License: Artistic-2.0
Encoding: UTF-8
VignetteBuilder: knitr
Imports: tidyr, tibble, plyr, matrixStats, stats, ggplot2, methods,
        rtracklayer, S4Vectors, ggforce, GenomeInfoDb, ComplexHeatmap,
        RColorBrewer, lifecycle, rlang, forcats, dplyr,
        GenomicFeatures, IRanges, kableExtra, ggdist
Depends: GenomicRanges, R (>= 4.2)
Suggests: testthat, BiocStyle, knitr, rmarkdown, GenomicAlignments,
        scales, Gviz, xlsx, GGally, patchwork, viridis, ggplotify,
        SummarizedExperiment, DESeq2, ggpointdensity, ggrastr, ashr,
        txdbmaker, ggrepel, stringr
RoxygenNote: 7.3.2
Collate: 'AllClasses.R' 'AllGenerics.R' 'Functions.R' 'methods.R'
        'bindingsites.R' 'helper.R' 'PlotFunction.R'
        'CoverageFunctions.R' 'workflow.R' 'helperSpecific.R'
        'exports.R' 'coveragePlots.R' 'bsfind.R' 'helperPlots.R'
        'differentialFunctions.R' 'differentialPlots.R'
biocViews: Sequencing, GeneExpression, GeneRegulation,
        FunctionalGenomics, Coverage, DataImport
BugReports: https://github.com/ZarnackGroup/BindingSiteFinder/issues
git_url: https://git.bioconductor.org/packages/BindingSiteFinder
git_branch: devel
git_last_commit: aec001f
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 21:30:59 UTC; biocbuild
Author: Mirko Brüggemann [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1778-0248>),
  Melina Klostermann [aut] (ORCID:
    <https://orcid.org/0000-0003-3122-1095>),
  Kathi Zarnack [aut] (ORCID: <https://orcid.org/0000-0003-3527-3378>)
Maintainer: Mirko Brüggemann <mirko.brueggemann@mail.de>

# Generated by roxygen2: do not edit by hand

export(BSFDataSet)
export(BSFDataSetFromBigWig)
export(BSFind)
export(annotateWithScore)
export(assignToGenes)
export(assignToTranscriptRegions)
export(bindingSiteCoveragePlot)
export(bindingSiteDefinednessPlot)
export(calculateBsBackground)
export(calculateBsFoldChange)
export(calculateSignalToFlankScore)
export(clipCoverage)
export(collapseReplicates)
export(combineBSF)
export(coverageOverRanges)
export(duplicatedSitesPlot)
export(estimateBsWidth)
export(estimateBsWidthPlot)
export(exportTargetGenes)
export(exportToBED)
export(filterBsBackground)
export(geneOverlapsPlot)
export(geneRegulationPlot)
export(getMeta)
export(getName)
export(getRanges)
export(getSignal)
export(getSummary)
export(globalScorePlot)
export(imputeBsDifferencesForTestdata)
export(makeBindingSites)
export(makeBsSummaryPlot)
export(mergeCrosslinkDiagnosticsPlot)
export(mergeSummaryPlot)
export(plotBsBackgroundFilter)
export(plotBsMA)
export(plotBsVolcano)
export(processingStepsFlowChart)
export(processingStepsTable)
export(pureClipGeneWiseFilter)
export(pureClipGlobalFilter)
export(pureClipGlobalFilterPlot)
export(quickFigure)
export(rangeCoveragePlot)
export(reproducibilityCutoffPlot)
export(reproducibilityFilter)
export(reproducibilityFilterPlot)
export(reproducibilitySamplesPlot)
export(reproducibilityScatterPlot)
export(setMeta)
export(setName)
export(setRanges)
export(setSignal)
export(setSummary)
export(supportRatio)
export(supportRatioPlot)
export(targetGeneSpectrumPlot)
export(transcriptRegionOverlapsPlot)
export(transcriptRegionSpectrumPlot)
exportMethods("+")
exportMethods("[")
exportMethods(getMeta)
exportMethods(getName)
exportMethods(getRanges)
exportMethods(getSignal)
exportMethods(getSummary)
exportMethods(setMeta)
exportMethods(setName)
exportMethods(setRanges)
exportMethods(setSignal)
exportMethods(setSummary)
exportMethods(show)
exportMethods(summary)
import(ComplexHeatmap)
import(GenomicRanges)
import(ggplot2)
import(lifecycle)
import(methods)
import(rtracklayer)
import(tidyr)
importFrom(GenomeInfoDb,keepSeqlevels)
importFrom(GenomeInfoDb,seqlevels)
importFrom(GenomicFeatures,cds)
importFrom(GenomicFeatures,fiveUTRsByTranscript)
importFrom(GenomicFeatures,genes)
importFrom(GenomicFeatures,intronsByTranscript)
importFrom(GenomicFeatures,threeUTRsByTranscript)
importFrom(GenomicRanges,findOverlaps)
importFrom(IRanges,extractList)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,subjectHits)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,count)
importFrom(dplyr,desc)
importFrom(dplyr,everything)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,lead)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_all)
importFrom(dplyr,n)
importFrom(dplyr,pull)
importFrom(dplyr,reframe)
importFrom(dplyr,rename)
importFrom(dplyr,rename_with)
importFrom(dplyr,row_number)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,slice_head)
importFrom(dplyr,slice_tail)
importFrom(dplyr,summarise)
importFrom(dplyr,summarize)
importFrom(dplyr,tally)
importFrom(dplyr,ungroup)
importFrom(ggdist,stat_halfeye)
importFrom(ggforce,geom_mark_rect)
importFrom(grDevices,colorRampPalette)
importFrom(grid,gpar)
importFrom(kableExtra,kable)
importFrom(matrixStats,colSds)
importFrom(plyr,rbind.fill)
importFrom(rtracklayer,import)
importFrom(stats,density)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,relevel)
importFrom(stats,sd)
importFrom(tibble,add_row)
importFrom(tibble,rownames_to_column)
importFrom(tidyr,pivot_longer)
importFrom(utils,head)
importFrom(utils,tail)
importFrom(utils,write.csv)

# BindingSiteFinder

Precise knowledge on the binding sites of an RNA-binding protein (RBP) is key to understand (post-) transcriptional regulatory processes.

The `BindingSiteFinder` software package provides a full workflow for RBP binding site definition. The package provides function as well as rich visualizations and is well integrated in state-of-the-art and widely used Bioconductor classes, such as `GenomicRanges` and `SummarizedExperiments`. For details please see the [vignette](https://bioconductor.org/packages/release/bioc/html/BindingSiteFinder.html).

For latest changes and updates please have a look at the [devel-page](https://www.bioconductor.org/packages/devel/bioc/html/BindingSiteFinder.html).

![](https://github.com/ZarnackGroup/BindingSiteFinder/blob/devel/vignettes/images/bsf_promotion.png)

# Prepare BSFDataSet for testing

# load packages
library(BindingSiteFinder)

# signal reduction function
# -> this function is part of the package, but not exported to the user
# -> we need it here once ot reduce the file size of the test dataset
.reduceSignalToFrame <- function(object, frame, quiet = FALSE) {
    rng = getRanges(object)
    # extend ranges by desired frame on which signal should be kept
    rng = rng + frame
    newObj = setRanges(object, rng, quiet = quiet)
    # drop signal outside of protective frame
    newObj = newObj[seq_along(rng), drop=TRUE]
    # put original sized ranges back in place
    rng = rng-frame
    newObj = setRanges(newObj, rng, quiet = quiet)

    return(newObj)
}

# load all datasets
TestSets = readRDS(file = "/Users/mirko/Projects/Datasets/01_data/TEST_DATASETS_FULL.rds")
# select the example case
this.set = TestSets$SRSF6_Diabetes

# reduce crosslink sites to chr22 for size
cs = getRanges(this.set)
cs = subset(cs, seqnames(cs) == "chr22")
seqlevels(cs) = "chr22"
this.new = setRanges(this.set, cs)

# reduce signal for size
this.new = .reduceSignalToFrame(this.new, frame = 50)

# export object
bds = setName(bds, "Test set")
save(bds, file = "./bds.rda")



# --- old
# import example crosslinks
# csFile = "/Users/mirko/Projects/clip_methods/04_pureclip/PureCLIP.crosslink_sites.bed"
# cs = import(con = csFile, format = "BED")
#
# # reduce sites to chr22 for size
# cs = subset(cs, seqnames(cs) == "chr22")
# seqlevels(cs) = "chr22"
#
# # pureclip score filter -> reduce file size
# quants = quantile(cs$score, probs = seq(0,1, by = 0.05))
# csFilter = cs[cs$score >= quants[2]]
#
# # make meta data
# clipFiles = "/Users/mirko/Projects/clip_methods/01_cov/lujh32-combined___bw/DR/raw/"
# clipFilesP <- list.files(clipFiles, pattern = "plus.bw$", full.names = TRUE)
# clipFilesM <- list.files(clipFiles, pattern = "minus.bw$", full.names = TRUE)
# meta = data.frame(id = c(1,2,3,4),
#                   condition = c("WT", "WT", "KD", "KD"),
#                   clPlus = clipFilesP, clMinus = clipFilesM)
#
# # make BSFDataSet
# bdsFull = BSFDataSetFromBigWig(ranges = csFilter, meta = meta)
#
# # reduce signal for size
# bds = .reduceSignalToFrame(bdsFull, frame = 50)
# bds = setName(bds, "Test set")
#
# # export object
# save(bds, file = "./bds.rda")

## ----setup, echo=FALSE, results="hide"----------------------------------------
knitr::knit_hooks$set(optipng = knitr::hook_optipng)
knitr::opts_chunk$set(tidy = FALSE,
                      cache = FALSE,
                      dev = "png",
                      message = FALSE, error = FALSE, warning = TRUE,
                      optipng = '-o7')
this.dpi = 100

## ----echo=FALSE, results="hide", warning=FALSE--------------------------------
suppressPackageStartupMessages({
    library(GenomicRanges)
    library(GenomicAlignments)
    library(rtracklayer)
    library(ggplot2)
    library(tidyr)
    library(ComplexHeatmap)
    library(BindingSiteFinder)
    library(forcats)
    library(dplyr)
})

## ----BiocManager, eval=FALSE--------------------------------------------------
# if (!require("BiocManager"))
#     install.packages("BiocManager")
# BiocManager::install("BindingSiteFinder")

## -----------------------------------------------------------------------------
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds

## -----------------------------------------------------------------------------
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
gns

## -----------------------------------------------------------------------------
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
regions

## ----echo=TRUE, eval=FALSE----------------------------------------------------
# # run BSFind to compute binding sites
# bdsQuick = BSFind(object = bds, anno.genes = gns, anno.transcriptRegionList = regions, est.subsetChromosome = "chr22")
# # export output as .bed
# exportToBED(bdsQuick, con = "./myBindingSites.bed")

## ----echo=TRUE, eval=FALSE----------------------------------------------------
# # not run
# quickFigure(bdsQuick, what = "main", save.filename = "./my_fig_main.pdf", save.device = "pdf")

## ----echo=TRUE, eval=FALSE----------------------------------------------------
# # not run
# quickFigure(bdsQuick, what = "supp", save.filename = "./my_fig_supp.pdf", save.device = "pdf")

## -----------------------------------------------------------------------------
# load crosslink sites
csFile <- system.file("extdata", "PureCLIP_crosslink_sites_examples.bed", 
                      package="BindingSiteFinder")
cs = rtracklayer::import(con = csFile, format = "BED", extraCols=c("additionalScores" = "character"))
cs$additionalScores = NULL
cs

## -----------------------------------------------------------------------------
# Load clip signal files and define meta data object
files <- system.file("extdata", package="BindingSiteFinder")
clipFilesP <- list.files(files, pattern = "plus.bw$", full.names = TRUE)
clipFilesM <- list.files(files, pattern = "minus.bw$", full.names = TRUE)

## -----------------------------------------------------------------------------
# make the meta table
meta = data.frame(
  id = c(1:4),
  condition = factor(rep("WT", 4)), 
  clPlus = clipFilesP, clMinus = clipFilesM)
meta

## ----eval=FALSE---------------------------------------------------------------
# library(BindingSiteFinder)
# bds = BSFDataSetFromBigWig(ranges = cs, meta = meta, silent = TRUE)

## ----eval=TRUE----------------------------------------------------------------
exampleFile <- list.files(files, pattern = ".rda$", full.names = TRUE)
load(exampleFile)
bds

## ----eval=TRUE, echo=FALSE----------------------------------------------------
# exampleFile <- list.files(files, pattern = ".rda$", full.names = TRUE)
# load(exampleFile)
# meta = data.frame(
#   id = c(1:4),
#   condition = factor(rep("WT", 4)), 
#   clPlus = clipFilesP, clMinus = clipFilesM)
# bds@meta = meta
# names(bds@signal$signalPlus) = c("1_WT", "2_WT", "3_WT", "4_WT")
# names(bds@signal$signalMinus) = c("1_WT", "2_WT", "3_WT", "4_WT")

## ----eval=FALSE---------------------------------------------------------------
# # exampleFile <- list.files(files, pattern = ".rda$", full.names = TRUE)
# # load(exampleFile)

## ----echo=TRUE, eval=FALSE----------------------------------------------------
# library(txdbmaker)
# # Make annotation database from gff3 file
# annoFile = "./gencode_v37_annotation.gff3"
# 
# annoDb = txdbmaker::makeTxDbFromGFF(file = annoFile, format = "gff3")
# annoInfo = rtracklayer::import(annoFile, format = "gff3")
# # Get genes as GRanges
# gns = genes(annoDb)
# idx = match(gns$gene_id, annoInfo$gene_id)
# meta = cbind(elementMetadata(gns),
#              elementMetadata(annoInfo)[idx,])
# meta = meta[,!duplicated(colnames(meta))]
# elementMetadata(gns) = meta

## -----------------------------------------------------------------------------
# Load GRanges with genes
geneFile <- list.files(files, pattern = "gns.rds$", full.names = TRUE)
load(geneFile)
gns

## ----eval=FALSE---------------------------------------------------------------
# # Count the overlaps of each binding site for each region of the transcript.
# cdseq = cds(annoDb)
# intrns = unlist(intronsByTranscript(annoDb))
# utrs3 = unlist(threeUTRsByTranscript(annoDb))
# utrs5 = unlist(fiveUTRsByTranscript(annoDb))
# regions = GRangesList(CDS = cdseq, Intron = intrns, UTR3 = utrs3, UTR5 = utrs5)

## -----------------------------------------------------------------------------
# Load list with transcript regions
regionFile <- list.files(files, pattern = "regions.rds$", full.names = TRUE)
load(regionFile)

## -----------------------------------------------------------------------------
bdsOut = BSFind(object = bds, anno.genes = gns, anno.transcriptRegionList = regions,
              est.subsetChromosome = "chr22", veryQuiet = TRUE, est.maxBsWidth = 29)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="processingStepsFlowChart"-------
processingStepsFlowChart(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="pureClipGlobalFilterPlot. The diagnostic shows the distribution of PureCLIP scores on a log2 scale with a 10% step increased coloring for easier visualization and comparability. The selected cutoff is indicated by the dashed line. ", fig.small=TRUE----
pureClipGlobalFilterPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="estimateBsWidthPlot. The diagnostic show the ratio between the crosslink events within binding sites and the crosslink events in adjacent windows to the binding sites. This is effectively a 'percent-bound' type ratio, with the higher the better the associated binding site width captures the distribution of the underlying crosslink events. The plot shows how that ratio behaves for different binding site width under varying levels of the gene-wise filter.", fig.small=TRUE----
estimateBsWidthPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="duplicatedSitesPlot. Diagnostic plot that show how many crosslink sites overlap how many different genes in the annotation. The grey area for 2 and 3 overlaps indicates what proportion the option overlaps='keepSingle' removes in order to keep crosslink site numbers stable. ", fig.small=TRUE----
duplicatedSitesPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="mergeCrosslinkDiagnosticsPlot. Diagnostic plot with number and region size at each fitting itteration loop. ", fig.small=TRUE----
mergeCrosslinkDiagnosticsPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="makeBsSummaryPlot. Diagnostic bar chart indicating the number of binding site after each of the indicated filter options is applied given their input values.", fig.small=TRUE----
makeBsSummaryPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilityFilterPlot. Reproducibility crosslink histogram, with the number of crosslink sites over the number of crosslinks per binding site for each replicate. The dashed line indicates the replicate-specific cutoff."----
reproducibilityFilterPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilitySamplesPlot. UpSet plot indicating the degree of reproducibility overlap between each replicate. "----
reproducibilitySamplesPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, message=FALSE, fig.cap="reproducibilityScatterPlot. Bottom left) Pairwise correlations between all replicates as scatter. Upper right) Pairwise pearson correlation coefficient. Diagnoal) Coverage distribution as density."----
reproducibilityScatterPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="geneOverlapsPlot. UpSet plot with the number of binding sites overlapping a specific gene type from the annotation before overlaps were resolved. "----
geneOverlapsPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="targetGeneSpectrumPlot. Bar chart with the number of gene and binding sites of a given gene type after overlaps were resolved. ", fig.small=TRUE----
targetGeneSpectrumPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionOverlapsPlot. UpSet plot with the number of binding sites overlapping a specific transcript region from the annotation before overlaps were resolved."----
transcriptRegionOverlapsPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionSpectrumPlot. Bar chart with the number of binding sites for each transcript region after overlaps were resolved. ", fig.small=TRUE----
transcriptRegionSpectrumPlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="globalScorePlot. The diagnostic shows the distribution of PureCLIP scores on a log2 scale with a 10% step increased coloring for easier visualization and comparability.", fig.small=TRUE----
globalScorePlot(bdsOut)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="bindingSiteDefinednessPlot. Density distribution of binding site definedness as percent bound grouped by transcript region.", fig.small=TRUE----
bindingSiteDefinednessPlot(bdsOut, by = "transcript")

## -----------------------------------------------------------------------------
getRanges(bdsOut)

## ----eval=FALSE---------------------------------------------------------------
# exportToBED(bdsOut, con = "./myResults.bed")

## ----eval=FALSE---------------------------------------------------------------
# exportTargetGenes(bdsOut, format = "xslx", split = "transcriptRegion")

## ----fig.retina = 1, dpi = this.dpi, fig.cap="bindingSiteCoveragePlot. Coverage plot on replicate level."----
bindingSiteCoveragePlot(bdsOut, plotIdx = 8, flankPos = 100, autoscale = TRUE)

## ----fig.retina = 1, dpi = this.dpi-------------------------------------------
rangesBeforeRepFilter = getRanges(bds)
rangesAfterRepFilter = getRanges(bdsOut)
idx = which(!match(rangesBeforeRepFilter, rangesAfterRepFilter, nomatch = 0) > 0)
rangesRemovedByFilter = rangesBeforeRepFilter[idx]
bdsRemovedRanges = setRanges(bds, rangesRemovedByFilter)

bindingSiteCoveragePlot(bdsRemovedRanges, plotIdx = 2, flankPos = 50)

## ----fig.retina = 1, dpi = this.dpi, eval=FALSE-------------------------------
# pSites = getRanges(bds)
# bindingSiteCoveragePlot(bdsOut, plotIdx = 8, flankPos = 20, autoscale = TRUE,
#                         customRange = pSites, customRange.name = "pSites", shiftPos = -10)

## ----fig.retina = 1, dpi = this.dpi, eval=FALSE-------------------------------
# bindingSiteCoverage = coverageOverRanges(bdsOut, returnOptions = "merge_positions_keep_replicates")
# idxMaxCountsBs = which.max(rowSums(as.data.frame(mcols(bindingSiteCoverage))))
# bindingSiteCoveragePlot(bdsOut, plotIdx = idxMaxCountsBs, flankPos = 100, mergeReplicates = FALSE, shiftPos = 50)

## ----fig.retina = 1, dpi = this.dpi, eval=FALSE-------------------------------
# bdsCDS = setRanges(bdsOut, regions$CDS)
# cdsWithMostBs = which.max(countOverlaps(regions$CDS, getRanges(bdsOut)))
# 
# bindingSiteCoveragePlot(bdsCDS, plotIdx = cdsWithMostBs, showCentralRange = FALSE,
#                        flankPos = 250, shiftPos = 50, mergeReplicates = TRUE,
#                        highlight = FALSE, customRange = getRanges(bdsOut),
#                        customAnnotation = regions$CDS)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionSpectrumPlot. Without region size normalization.", fig.small=TRUE, eval=FALSE----
# transcriptRegionSpectrumPlot(bdsOut, normalize = FALSE, values = "percentage")

## ----fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionSpectrumPlot. With region size normalization.", fig.small=TRUE, eval=FALSE----
# transcriptRegionSpectrumPlot(bdsOut, normalize = TRUE, values = "percentage", normalize.factor = "median")

## -----------------------------------------------------------------------------
# compute binding sites
bds1 <- makeBindingSites(object = bds, bsSize = 3)
bds2 <- makeBindingSites(object = bds, bsSize = 9)
bds3 <- makeBindingSites(object = bds, bsSize = 19)
bds4 <- makeBindingSites(object = bds, bsSize = 29)
# summarize in list
l = list(`1. bsSize = 3` = bds1, `2. bsSize = 9` = bds2, 
         `3. bsSize = 19` = bds3, `4. bsSize = 29` = bds4)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="rangeCoveragePlot. Crosslink coverage summarized at potential binding sites, with current size given by the grey area."----
rangeCoveragePlot(l, width = 20, show.samples = TRUE, subset.chromosome = "chr22") 

## -----------------------------------------------------------------------------
bdsManual = BSFind(object = bds, anno.genes = gns, anno.transcriptRegionList = regions, bsSize = 9, cutoff.geneWiseFilter = 0)

## -----------------------------------------------------------------------------
bdsSimple = makeBindingSites(object = bds, bsSize = 9)

## -----------------------------------------------------------------------------
customBSFind <- function(object) {
    this.obj = pureClipGlobalFilter(object)
    this.obj = makeBindingSites(this.obj, bsSize = 9)
    this.obj = calculateSignalToFlankScore(this.obj)
    return(this.obj)
}
bdsSimple = customBSFind(bds)

## -----------------------------------------------------------------------------
noAnnotationBSFind <- function(object) {
    this.obj = pureClipGlobalFilter(object)
    this.obj = makeBindingSites(this.obj, bsSize = 9)
    this.obj = reproducibilityFilter(this.obj)
    this.obj = annotateWithScore(this.obj, getRanges(object))
    this.obj = calculateSignalToFlankScore(this.obj)
    return(this.obj)
}
bdsSimple = noAnnotationBSFind(bds)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilityScatterPlot. Before reproducibility filter.", eval=FALSE----
# bds.before = makeBindingSites(bds, bsSize = 9)
# reproducibilityScatterPlot(bds.before)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilityScatterPlot. After reproducibility filter.", eval=FALSE----
# bds.after = reproducibilityFilter(bds.before, minCrosslinks = 2)
# reproducibilityScatterPlot(bds.after)

## ----fig.retina = 1, dpi = this.dpi, fig.small=TRUE, eval=FALSE---------------
# set.seed(1234)
# bdsSub = bds[sample(seq_along(getRanges(bds)), 100, replace = FALSE)]
# 
# cov = coverageOverRanges(bdsSub, returnOptions = "merge_positions_keep_replicates")
# df = mcols(cov) %>%
#     as.data.frame() %>%
#     pivot_longer(everything())
# 
# ggplot(df, aes(x = name, y = log2(value+1), fill = name)) +
#     geom_violin() +
#     geom_boxplot(width = 0.1, fill = "white") +
#     scale_fill_brewer(palette = "Greys") +
#     theme_bw() +
#     theme(legend.position = "none") +
#     labs(x = "Samples", y = "#Crosslinks (log2)")

## ----fig.retina = 1, dpi = this.dpi, fig.small=TRUE, eval=FALSE---------------
# bdsMerge = collapseReplicates(bds)[1:100]
# covTotal = coverageOverRanges(bdsMerge, returnOptions = "merge_positions_keep_replicates")
# 
# covRep = coverageOverRanges(bds[1:100], returnOptions = "merge_positions_keep_replicates")
# 
# df = cbind.data.frame(mcols(covTotal), mcols(covRep)) %>%
#     mutate(rep1 = round(`1_WT`/ WT, digits = 2) * 100,
#            rep2 = round(`2_WT`/ WT, digits = 2) * 100,
#            rep3 = round(`3_WT`/ WT, digits = 2) * 100,
#            rep4 = round(`4_WT`/ WT, digits = 2) * 100) %>%
#     tibble::rowid_to_column("BsID") %>%
#     dplyr::select(BsID, rep1, rep2, rep3, rep4) %>%
#     pivot_longer(-BsID) %>%
#     group_by(BsID) %>%
#     arrange(desc(value), .by_group = TRUE) %>%
#     mutate(name = factor(name, levels = name)) %>%
#     group_by(name) %>%
#     arrange(desc(value), .by_group = TRUE) %>%
#     mutate(BsID = factor(BsID, levels = BsID))
# 
# 
# ggplot(df, aes(x = BsID, y = value, fill = name)) +
#     geom_col(position = "fill", width = 1) +
#     theme_bw() +
#     scale_fill_brewer(palette = "Set3") +
#     theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1, size = 7)) +
#     labs(x = "Binding site ID",
#          y = "Percentage",
#          fill = "Replicate"
#          ) +
#     scale_y_continuous(labels = scales::percent)

## ----eval=FALSE---------------------------------------------------------------
# sgn = getSignal(bds)
# export(sgn$signalPlus$`1_WT`, con = "./WT_1_plus.bw", format = "bigwig")
# export(sgn$signalMinus$`1_WT`, con = "./WT_1_minus.bw", format = "bigwig")

## ----eval=FALSE---------------------------------------------------------------
# bdsMerge = collapseReplicates(bds)
# sgn = getSignal(bdsMerge)
# export(sgn$signalPlus$WT, con = "./sgn_plus.bw", format = "bigwig")
# export(sgn$signalPlus$WT, con = "./sgn_minus.bw", format = "bigwig")

## -----------------------------------------------------------------------------
cov = clipCoverage(bdsOut, ranges.merge = TRUE, samples.merge = FALSE, positions.merge = FALSE)
cov

## ----eval=TRUE, echo=FALSE----------------------------------------------------
makeExampleDataSets <- function() {
    # load clip data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    # make binding sites
    bds = makeBindingSites(bds, bsSize = 5, quiet = TRUE)
    bds = assignToGenes(bds, anno.genes = gns, quiet = TRUE)
    # create artificial changes between binding sites
    bds = imputeBsDifferencesForTestdata(bds)
    
    # handle ranges
    bs.all = getRanges(bds)
    set.seed(1234)
    idx = sample(c(TRUE,FALSE), size = length(bs.all), replace = TRUE)       
    bs.wt = bs.all[idx]
    bs.ko = bs.all[!idx]
    
    # handle meta
    meta.all = getMeta(bds)
    meta.wt = meta.all[c(1:2),]
    meta.ko = meta.all[c(3:4),]
    meta.ko$id = 1:2
    
    # handle signal
    signal.all = getSignal(bds)
    signal.wt.p = signal.all$signalPlus[1:2]
    signal.wt.m = signal.all$signalMinus[1:2]
    signal.wt = list(signalPlus = signal.wt.p, signalMinus = signal.wt.m)
    
    signal.all = getSignal(bds)
    signal.ko.p = signal.all$signalPlus[3:4]
    signal.ko.m = signal.all$signalMinus[3:4]
    signal.ko = list(signalPlus = signal.ko.p, signalMinus = signal.ko.m)
    
    # set objects
    bds.wt = setRanges(bds, bs.wt)
    bds.wt = setMeta(bds.wt, meta.wt)
    bds.wt = setSignal(bds.wt, signal.wt)
    bds.wt = setName(bds.wt, "Example.WT")
    
    bds.ko = setRanges(bds, bs.ko)
    bds.ko = setMeta(bds.ko, meta.ko)
    bds.ko = setSignal(bds.ko, signal.ko)
    bds.ko = setName(bds.ko, "Example.KO")
    
    # return
    l = list(wt = bds.wt, ko = bds.ko)
    return(l)
}

## -----------------------------------------------------------------------------
# make artifical data sets
myDataSets = makeExampleDataSets()

bds.wt = myDataSets$wt
bds.ko = myDataSets$ko

bds.wt
bds.ko

## -----------------------------------------------------------------------------
bds.comb = bds.wt + bds.ko
getRanges(bds.comb)

## ----warning=FALSE, message=FALSE---------------------------------------------
bds.diff = calculateBsBackground(bds.comb, anno.genes = gns)
getRanges(bds.diff)

## -----------------------------------------------------------------------------
bds.diff = filterBsBackground(bds.diff)

## -----------------------------------------------------------------------------
bds.diff = calculateBsFoldChange(bds.diff)
getRanges(bds.diff)

## ----fig.retina = 1, dpi = this.dpi, fig.cap="plotBsMA. MA plot for binding site changes.", fig.small=TRUE----
plotBsMA(bds.diff, what = "bs")

## -----------------------------------------------------------------------------
bds.diff = assignToTranscriptRegions(bds.diff, anno.transcriptRegionList = regions)
geneRegulationPlot(bds.diff, plot.geneID = "ENSG00000180957.18", anno.genes = gns)

## ----eval=FALSE---------------------------------------------------------------
# ############################
# # Construct WT binding sites
# ############################
# # Locate peak calling result
# peaks.wt = "./myFolder/wt/pureclip.bed"
# peaks.wt = import(con = peaks.wt, format = "BED", extraCols=c("additionalScores" = "character"))
# peaks.wt = keepStandardChromosomes(peaks.wt, pruning.mode = "coarse")
# # Locate bigwig files
# clipFiles = "./myFolder/wt/bigwig/"
# clipFiles = list.files(clipFiles, pattern = ".bw$", full.names = TRUE)
# clipFiles = clipFiles[grepl("WT", clipFiles)]
# clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
# clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# # Organize clip data in dataframe
# colData = data.frame(
#     name = "Example.WT", id = c(1:2),
#     condition = factor(c("WT", "WT"), levels = c("WT")),
#     clPlus = clipFilesP, clMinus = clipFilesM)
# # Construct BSFDataSet object
# bdsWT.initial = BSFDataSetFromBigWig(ranges = peaksInitial, meta = colData)
# # Define binding sites
# bdsWT = BSFind(bdsWT.initial, anno.genes = gns, anno.transcriptRegionList = trl)

## ----eval=FALSE---------------------------------------------------------------
# ############################
# # Construct KO binding sites
# ############################
# # Locate peak calling result
# peaks.ko = "./myFolder/ko/pureclip.bed"
# peaks.ko = import(con = peaks.ko, format = "BED", extraCols=c("additionalScores" = "character"))
# peaks.ko = keepStandardChromosomes(peaks.ko, pruning.mode = "coarse")
# # Locate bigwig files
# clipFiles = "./myFolder/ko/bigwig/"
# clipFiles = list.files(clipFiles, pattern = ".bw$", full.names = TRUE)
# clipFiles = clipFiles[grepl("KO", clipFiles)]
# clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
# clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# # Organize clip data in dataframe
# colData = data.frame(
#     name = "Example.KO", id = c(1:2),
#     condition = factor(c("KO", "KO"), levels = c("KO")),
#     clPlus = clipFilesP, clMinus = clipFilesM)
# # Construct BSFDataSet object
# bdsKO.initial = BSFDataSetFromBigWig(ranges = peaksInitial, meta = colData)
# # Define binding sites
# bdsKO = BSFind(bdsKO.initial, anno.genes = gns, anno.transcriptRegionList = trl)

## ----eval=FALSE---------------------------------------------------------------
# # Locate bigwig files - KO
# clipFilesKO = "./myFolder/ko/bigwig/"
# clipFilesKO = list.files(clipFilesKO, pattern = ".bw$", full.names = TRUE)
# clipFilesKO = clipFilesKO[grepl("KO", clipFilesKO)]
# # Locate bigwig files - WT
# clipFilesWT = "./myFolder/wt/bigwig/"
# clipFilesWT = list.files(clipFilesWT, pattern = ".bw$", full.names = TRUE)
# clipFilesWT = clipFilesWT[grepl("WT", clipFilesWT)]
# # Merge both conditions
# clipFiles = c(clipFilesWT, clipFilesKO)
# # Split by strand
# clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
# clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# # Organize clip data in dataframe
# colData = data.frame(
#     name = "Example.KO", id = c(1:4),
#     condition = factor(c("WT", "WT", "KO", "KO"), levels = c("WT", "KO")),
#     clPlus = clipFilesP, clMinus = clipFilesM)
# # Construct BSFDataSet object from pre-defined binding sites
# # -> myBindingSites must be a GRanges object
# bds.comb = BSFDataSetFromBigWig(ranges = myBindingSites, meta = colData)

## ----eval=FALSE---------------------------------------------------------------
# bds.complex = combineBSF(list(ctrl = bds.ctrl, wt = bds.wt, ko = bds.ko))

## ----eval=FALSE---------------------------------------------------------------
# # extract binding site ranges from complex merge
# bindingSites.complex = getRanges(bds.complex)
# # Locate bigwig files - KO
# clipFilesKO = "./myFolder/ko/bigwig/"
# clipFilesKO = list.files(clipFilesKO, pattern = ".bw$", full.names = TRUE)
# clipFilesKO = clipFilesKO[grepl("KO", clipFilesKO)]
# # Locate bigwig files - WT
# clipFilesWT = "./myFolder/wt/bigwig/"
# clipFilesWT = list.files(clipFilesWT, pattern = ".bw$", full.names = TRUE)
# clipFilesWT = clipFilesWT[grepl("WT", clipFilesWT)]
# # Merge both conditions
# clipFiles = c(clipFilesWT, clipFilesKO)
# # Split by strand
# clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
# clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# # Organize clip data in dataframe
# colData = data.frame(
#     name = "Example.KO", id = c(1:2),
#     condition = factor(c("KO", "KO"), levels = c("KO")),
#     clPlus = clipFilesP, clMinus = clipFilesM)
# # Construct BSFDataSet object from pre-defined binding sites
# bds.comb = BSFDataSetFromBigWig(ranges = bindingSites.complex, meta = colData)

## -----------------------------------------------------------------------------
sessionInfo()

library(rtracklayer)
library(GenomicFeatures)
library(txdbmaker)

################################################################################
# prepare crosslink sites
################################################################################
# import crosslinks
csFile = "/Users/mirko/Projects/clip_methods/04_pureclip/PureCLIP.crosslink_sites.bed"

cs = import(con = csFile, format = "BED")
# pureclip score filter
quants = quantile(cs$score, probs = seq(0,1, by = 0.05))
csFilter = cs[cs$score >= quants[2]]


################################################################################
# subset .bw files for those that overlap our example genes
################################################################################
olGen = subsetByOverlaps(gns, csFilter)

# file path
# f = "/Users/mirko/Projects/clip_methods/01_cov/lujh32-combined___bw/DR/raw/"
f = list.files(f, pattern = ".bw$", full.names = TRUE)

# plus strand files
b = rtracklayer::import(f[2], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep1_clip_plus.bw", format = "bw")

b = rtracklayer::import(f[4], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep2_clip_plus.bw", format = "bw")

b = rtracklayer::import(f[6], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep3_clip_plus.bw", format = "bw")

b = rtracklayer::import(f[8], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep4_clip_plus.bw", format = "bw")

# minus strand files
b = rtracklayer::import(f[1], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep1_clip_minus.bw", format = "bw")

b = rtracklayer::import(f[3], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep2_clip_minus.bw", format = "bw")

b = rtracklayer::import(f[5], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep3_clip_minus.bw", format = "bw")

b = rtracklayer::import(f[7], as = "GRanges")
b = subsetByOverlaps(b, olGen)
rtracklayer::export(b, con = "../inst/extdata/rep4_clip_minus.bw", format = "bw")



################################################################################
# prepare BindingSiteFinder example dataset
################################################################################
library(BindingSiteFinder)
# clipFiles = "/Users/mirko/Projects/clip_methods/01_cov/lujh32-combined___bw/DR/raw/"
clipFilesP <- list.files(clipFiles, pattern = "plus.bw$", full.names = TRUE)
clipFilesM <- list.files(clipFiles, pattern = "minus.bw$", full.names = TRUE)
meta = data.frame(id = c(1,2,3,4),
                  condition = c("WT", "WT", "KD", "KD"),
                  clPlus = clipFilesP, clMinus = clipFilesM)

# bdsFull = BSFDataSet(ranges = csFilter, meta = meta)
bdsFull = BSFDataSetFromBigWig(ranges = csFilter, meta = meta)
bds = .subsetByChr(bdsFull, chr = "chr22")

save(bds, file = "./bds.rda")


################################################################################
# perpare gene annotaiton data
################################################################################
# annoFile = "/Users/mirko/Projects/PackageDevelopment/BindingSiteDefinition/data/gencode.v37.annotation.gff3"
anno <- makeTxDbFromGFF(file = annoFile)
gns = genes(anno)

# Make annotation database from gff3 file
# annoFile = "/Users/mirko/Projects/PackageDevelopment/BindingSiteDefinition/data/gencode.v37.annotation.chr22.header.gff3"
annoDb = makeTxDbFromGFF(file = annoFile, format = "gff3")
annoInfo = import(annoFile, format = "gff3")
# Get genes as GRanges
gns = genes(annoDb)
idx = match(gns$gene_id, annoInfo$gene_id)
elementMetadata(gns) = cbind(elementMetadata(gns), elementMetadata(annoInfo)[idx,])
# Clean gene object for storage
mcols(gns)$ccdsid = NULL
mcols(gns)$protein_id = NULL
mcols(gns)$havana_transcript = NULL
mcols(gns)$ont = NULL
mcols(gns)$hgnc_id = NULL
mcols(gns)$exon_id = NULL
mcols(gns)$exon_number = NULL
mcols(gns)$transcript_support_level = NULL
mcols(gns)$transcript_name = NULL
mcols(gns)$transcript_id = NULL
mcols(gns)$transcript_type = NULL
mcols(gns)$score = NULL
mcols(gns)$phase = NULL
mcols(gns)$gene_id.1 = NULL
mcols(gns)$Parent = NULL
mcols(gns)$level = NULL
mcols(gns)$tag = NULL
mcols(gns)$havana_gene = NULL
mcols(gns)$source = NULL
mcols(gns)$ID = NULL
mcols(gns)$type = NULL
# Save genes object for load
save(gns, file="gns.rds")

# Count the overlaps of each binding site fore each part of the transcript.
cdseq = cds(annoDb)
intrns = unlist(intronsByTranscript(annoDb))
utrs3 = unlist(threeUTRsByTranscript(annoDb))
utrs5 = unlist(fiveUTRsByTranscript(annoDb))
regions = GRangesList(CDS = cdseq, Intron = intrns, UTR3 = utrs3, UTR5 = utrs5)
save(regions, file = "./regions.rds")



# make additional subset to the test object
# -> keep only signal in the a frame of 100nt around each binding site
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

rngExtended = getRanges(bds) + 100
bdsSub = setRanges(bds, rngExtended)
bdsSub = bdsSub[1:length(rngExtended), drop = TRUE]

rngNormal = getRanges(bdsSub) - 100
bds = setRanges(bdsSub, rngNormal)

save(bds, file = "./bds.rda")

#' Replicate reproducibility filter function
#'
#' For each replicate the number of binding sites with a certain number of
#' crosslinks is calculated. A quantile based threshold (\code{cutoff}) is
#' applied to each replicate. This indicates how many of the merged binding
#' sites are supported by crosslinks from the respective replicate. Next, one
#' can specify how many replicates need to pass the defined threshold for a
#' binding site to be considered reproducible.
#'
#' If \code{cutoff} is a single number then the indicated cutoff will be
#' applied to all replicates. If it is a vector then each element in the vector
#' is applied to all replicates of the respective condition. The order is
#' hereby given by the levels of the condition column of the meta data
#' (see \code{\link{BSFDataSet}},\code{\link{getMeta}}). If the condition
#' specific filter is applied, a meta column is added to the GRanges of the
#' \code{BSFDataSet} object, indicating the support for each condition.
#'
#' If \code{nReps} is a single number then this number is used as treshold for
#' all binding sites. If it is a vector then it is applied to the replicates of
#' the respective condition (like in \code{cutoff}). This allows the
#' application of different thresholds for experiments of different
#' experimental conditions. If the condition specific filter is applied, a meta
#' column is added to the GRanges of the \code{BSFDataSet} object,
#' indicating the support for each condition.
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a BSFDataSet object
#' @param cutoff numeric; percentage cutoff to be used for the
#' reproducibility quantile filtering
#' @param nReps numeric; number of replicates that must meet the cutoff
#' defined in \code{cutoff} for a binding site to be called reproducible.
#' Defaults to N-1.
#' @param minCrosslinks numeric; minimal number of crosslinks a binding
#' site needs to have to be called reproducible. Acts as a lower boundary for
#' \code{cutoff}. Defaults to 1.
#' @param returnType one of "BSFDataSet" or "data.frame". "BSFDataSet" is the
#' default and "matrix" can be used for easy plotting.
#' @param n.reps deprecated -> use nReps instead
#' @param min.crosslinks deprecated -> use minCrosslinks instead
#' @param quiet logical; whether to print messages
#'
#' @return an object of type BSFDataSet
#'
#' @seealso \code{\link{BSFind}},
#' \code{\link{reproducibilityFilterPlot}},
#' \code{\link{reproducibilitySamplesPlot}},
#' \code{\link{reproducibilityScatterPlot}}
#'
#' @import tidyr GenomicRanges lifecycle
#' @importFrom dplyr count select
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' # merge binding sites
#' bds <- makeBindingSites(object = bds, bsSize = 9)
#'
#' # use default return with condition specific threshold
#' bds = reproducibilityFilter(bds, cutoff = 0.1, nReps = 1)
#'
#' @export
reproducibilityFilter <- function(object,
                                  cutoff = NULL,
                                  nReps = NULL,
                                  minCrosslinks = 1,
                                  returnType = c("BSFDataSet", "data.frame"),
                                  n.reps = lifecycle::deprecated(),
                                  min.crosslinks = lifecycle::deprecated(),
                                  quiet = FALSE) {
    # initialize local vairables
    name <- value <- NULL

    # Argument deprecation warnings
    # --------------------------------------------------------------------------

    if (lifecycle::is_present(n.reps)) {
        lifecycle::deprecate_warn("2.0.0", "reproducibilityFilter(n.reps = )", "reproducibilityFilter(nReps = )")
        nReps = n.reps
    }

    if (lifecycle::is_present(min.crosslinks)) {
        lifecycle::deprecate_warn("2.0.0", "reproducibilityFilter(min.crosslinks = )", "reproducibilityFilter(minCrosslinks = )")
        minCrosslinks = min.crosslinks
    }

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    if (length(cutoff) != length(nReps)) {
        stop("Number of values for 'cutoff' does not match the number of values
             for 'nReps'. ")
    }

    metaData = getMeta(object)

    numberOfConditions = length(levels(metaData$condition))
    if (numberOfConditions != length(cutoff) && !is.null(cutoff)) {
        msg = paste0("Reproducibility filter cutoff does not match the number of conditions. You specified: ",
                     length(cutoff), ", but there are: ", numberOfConditions, "\n")
        msg2 = paste0("The specified cutoff (",  cutoff, ") ",
                      "is applied to all conditions (",
                      paste(as.character(levels(metaData$condition)), collapse = ",") ,") \n")
        warning(paste0(msg, msg2))
        defaultCutoff = rep(cutoff, numberOfConditions)
    }
    if (is.null(cutoff)) {
        msg = paste0("Reproducibility cutoff not defined. Defaults to 0.05 for each condition. \n")
        if(!quiet) message(msg)
        defaultCutoff = rep(0.05, numberOfConditions)
    }
    if (numberOfConditions == length(cutoff) && !is.null(cutoff)) {
        defaultCutoff = cutoff
    }
    # Manage parameter nReps
    if (numberOfConditions != length(nReps) && !is.null(nReps)) {
        msg = paste0("Parameter nReps does not match the number of conditions. You specified: ",
                     length(nReps), ", but there are: ", numberOfConditions, "\n")
        msg2 = paste0("nReps defaults to N-1 for each condition. \n")
        warning(paste0(msg, msg2))

        n.conditions = table(metaData$condition) %>% as.data.frame()
        defaultNreps = n.conditions$Freq -1
    }
    if (is.null(nReps)) {
        msg = paste0("Parameter nReps not defined. Defaults to N-1 for each condition. \n")
        if(!quiet) message(paste0(msg))
        n.conditions = table(metaData$condition) %>% as.data.frame()
        defaultNreps = n.conditions$Freq -1
    }
    if (numberOfConditions == length(nReps) && !is.null(nReps)) {
        defaultNreps = nReps
    }

    # ---
    # Store function parameters in list
    optstr = list(cutoff = defaultCutoff, nReps = defaultNreps, minCrosslinks = minCrosslinks)
    object@params$reproducibilityFilter = optstr

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    rngInitial = getRanges(object)
    cond = metaData$condition
    # get number of crosslinks per binding site and replicate
    df = as.data.frame(mcols(coverageOverRanges(
        object, returnOptions = "merge_positions_keep_replicates")))

    # Manage single cutoff for single condition
    if (length(defaultCutoff) == 1) {
        if(length(levels(cond)) > 1) {
            stop("Only one cutoff is given for multiple conditions.")
        }
        # calculate sample specific thresholds
        qSel = .selectQuantilesSingleCondtion(
            covDf = df,
            userCond = cond,
            userNreps = defaultNreps,
            userCutoff = defaultCutoff
        )
        # apply minimal crosslink threshold
        qSel$value = ifelse(qSel$value < minCrosslinks,
                            minCrosslinks,
                            qSel$value)
        matchIdx = match(qSel$name, colnames(df))

        # ---
        # Store data for diagnostic plot in list
        dfPlot = df %>%
            pivot_longer(everything()) %>%
            group_by(name, value) %>% dplyr::count() %>%
            separate(name, into = c(NA, "condition"), sep = "_", remove = FALSE)
        object@plotData$reproducibilityFilterPlot$data = dfPlot
        object@plotData$reproducibilityFilterPlot$cutoffs = qSel

        # calculate replicate support based on quantile cutoff
        s = apply(df, 1, function(x) {
            ifelse(x > qSel$value[matchIdx], 1, 0)
        }) %>%
            t() %>% as.data.frame()

        # ---
        # Store results for plotting
        object@plotData$reproducibilitySamplesPlot$data = s

        # Filter binding sites ranges by replicate support
        support = rowSums(s) >= defaultNreps
        newRanges = getRanges(object)
        newRanges = newRanges[support]
        newObject = setRanges(object, newRanges)
    }

    # Manage multiple cutoffs for multiple conditions
    if (length(defaultCutoff) > 1) {
        if (length(levels(cond)) == 1) {
            stop("multiple cutoffs are given but only one condition exists")
        }
        # calculate sample specific thresholds
        qSel = .selectQuantilesMultipleConditions(
            covDf = df,
            userCond = cond,
            userNreps = defaultNreps,
            userCutoff = defaultCutoff
        )
        # apply minimal crosslink threshold
        qSel$value = ifelse(qSel$value < minCrosslinks,
                            minCrosslinks,
                            qSel$value)
        matchIdx = match(qSel$name, colnames(df))

        # ---
        # Store data for diagnostic plot in list
        dfPlot = df %>%
            pivot_longer(everything()) %>%
            group_by(name, value) %>% dplyr::count() %>%
            separate(name, into = c(NA, "condition"), sep = "_", remove = FALSE)
        object@plotData$reproducibilityFilterPlot$data = dfPlot
        object@plotData$reproducibilityFilterPlot$cutoffs = qSel

        # calculate reproducibility per condition
        s = apply(df, 1, function(x) {
            ifelse(x > qSel$value[matchIdx], 1, 0)
        }) %>%
            t %>% as.data.frame()

        # ---
        # Store results for plotting
        object@plotData$reproducibilitySamplesPlot$data = s

        # Filter binding sites ranges by replicate support
        sSplit = vapply(levels(cond), function(x) {
            s %>% dplyr::select(contains(x)) %>% rowSums()
        }, FUN.VALUE = numeric(nrow(s))) %>% as.data.frame()
        idx = match(colnames(sSplit), qSel$sel)
        support = apply(sSplit, 1, function(x) {
            x >= qSel$defaultNreps[idx]
        }) %>%
            t %>% as.data.frame()

        supportAll = apply(support, 1, any)

        newRanges = getRanges(object)
        # mcols(newRanges) = support
        mcols(newRanges) = cbind.data.frame(mcols(newRanges), support)
        newRanges = newRanges[supportAll]
        newObject = setRanges(object, newRanges, quiet = quiet)
    }

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "reproducibilityFilter()", class = "binding sites",
        nIn = length(rngInitial), nOut = length(newRanges),
        per = paste0(round(length(newRanges)/ length(rngInitial), digits = 2)*100,"%"),
        options = paste0("Cutoff=", paste(optstr$cutoff, collapse = ", "),
                         ", nReps=", paste(optstr$nReps, collapse = ", "),
                         ", minCrosslinks=", paste(optstr$minCrosslinks, collapse = ", "))
    )
    newObject@results = rbind(newObject@results, resultLine)

    # Manage return options
    returnType = match.arg(returnType, choices = c("BSFDataSet", "data.frame"))
    if (returnType == "BSFDataSet") {
        retObj = newObject
    }
    if (returnType == "data.frame") {
        retObj = s
    }
    return(retObj)
}


#' Annotation function for BSFDataSet object
#'
#' This function can be used to annotate a \code{BSFDataSet} object with
#' merged binding sites with scores from the initial ranges
#' (eg. PureCLIP scores).
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a BSFDataSet object
#' @param match.ranges a GRanges object, with numeric column for the score to match
#' @param match.score character; meta column name of the crosslink site
#' \code{\link{GenomicRanges}} object that holds the score to match
#' @param match.option character; option how score should be matched
#' @param quiet logical; whether to print messages
#' @param scoreRanges deprecated -> use match.ranges instead
#' @param MatchColScore deprecated -> use match.score instead
#'
#' @return an object of class BSFDataSet with updated meta columns of the ranges
#'
#' @seealso \code{\link{BSFind}}, \code{\link{globalScorePlot}}
#'
#' @import GenomicRanges
#' @importFrom S4Vectors queryHits subjectHits
#'
#' @examples
#' if (.Platform$OS.type != "windows") {
#'     # load data
#'     csFile <- system.file("extdata", "PureCLIP_crosslink_sites_examples.bed",
#'                         package="BindingSiteFinder")
#'     cs = rtracklayer::import(con = csFile, format = "BED",
#'     extraCols=c("additionalScores" = "character"))
#'     cs$additionalScores = NULL
#'     clipFiles <- system.file("extdata", package="BindingSiteFinder")
#'     # two experimental conditions
#'     meta = data.frame(
#'     id = c(1,2,3,4),
#'     condition = factor(c("WT", "WT", "KD", "KD"),
#'     levels = c("KD", "WT")),
#'     clPlus = list.files(clipFiles, pattern = "plus.bw$", full.names = TRUE),
#'     clMinus = list.files(clipFiles, pattern = "minus.bw$",
#'      full.names = TRUE))
#'     bds = BSFDataSetFromBigWig(ranges = cs, meta = meta, silent = TRUE)
#'
#'     # merge binding sites
#'     bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#'     minCrosslinks = 2, minClSites = 1)
#'
#'     # annotate with original pureCLIP score
#'     bdsRe = annotateWithScore(bds, cs)
#' }
#' @export
annotateWithScore <- function(object, # bindingSiteFinder
                              match.ranges = NULL,
                              match.score = "score",
                              match.option = c("max", "sum", "mean"),
                              scoreRanges = lifecycle::deprecated(),
                              MatchColScore = lifecycle::deprecated(),
                              quiet = FALSE
) {

    # bind locally used variables
    qHits <- NULL

    # Argument deprecation warnings
    # --------------------------------------------------------------------------

    if (lifecycle::is_present(scoreRanges)) {
        lifecycle::deprecate_warn("2.0.0", "annotateWithScore(scoreRanges = )", "annotateWithScore(match.ranges = )")
        match.ranges = scoreRanges
    }

    if (lifecycle::is_present(MatchColScore)) {
        lifecycle::deprecate_warn("2.0.0", "annotateWithScore(MatchColScore = )", "annotateWithScore(match.score = )")
        match.score = MatchColScore
    }

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    # check input ranges
    if(is.null(match.ranges)) {
        msg = paste0("No ranges to match the score from. Please provide a GRanges object with a matching score column in the meta data.")
        stop(msg)
    }
    stopifnot(is(match.ranges, "GRanges"))

    # handle options
    match.option = match.arg(match.option, choices = c("max", "sum", "mean"))

    # match score column
    scoreColNames = colnames(mcols(match.ranges))
    if (!(match.score %in% scoreColNames)) {
        msg = paste0("Matching columns (", match.score,
                     ") is not present in the provided match.ranges. \n")
        stop(msg)
    }

    # ---
    # Store function parameters in list
    optstr = list(match.score = match.score, match.option = match.option)
    object@params$annotateWithScore = optstr

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    rngInitial = getRanges(object)
    rng = getRanges(object)
    ol = findOverlaps(rng, match.ranges)

    if (length(ol) == 0) {
        stop("Ranges of 'object' and 'match.ranges' do not overlap.")
    }

    matchDF = data.frame(qHits = queryHits(ol),
                         sHits = subjectHits(ol),
                         score = match.ranges$score[subjectHits(ol)])

    if (match.option == "max") {
        score = dplyr::group_by(matchDF, qHits) %>%
            dplyr::summarize(score = max(score), .groups = "drop") %>%
            as.data.frame()
    }
    if (match.option == "sum") {
        score = dplyr::group_by(matchDF, qHits) %>%
            dplyr::summarize(score = sum(score), .groups = "drop") %>%
            as.data.frame()
    }
    if (match.option == "mean") {
        score = dplyr::group_by(matchDF, qHits) %>%
            dplyr::summarize(score = mean(score), .groups = "drop") %>%
            as.data.frame()
    }

    # ---
    # Store results for plotting
    object@plotData$annotateWithScore$data = score$score

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "annotateWithScore()", class = "binding sites",
        nIn = length(rngInitial), nOut = length(rng),
        per = paste0(round(length(rng)/ length(rngInitial), digits = 2)*100,"%"),
        options = paste0("MatchOption=", optstr$match.option, ", match.score=", match.score)
    )
    object@results = rbind(object@results, resultLine)

    mcols(rng)$score = score$score
    newObject = setRanges(object, rng, quiet = quiet)
    return(newObject)
}




#' Support ratio function for BSFDataSet objects
#'
#' Functions that computes a ratio to determine how well a given binding site
#' with is supported by the crosslink coverage of the data. For a given
#' \code{BSFDataSet} object binding sites are computed for each width indicated
#' in the \code{bsWidths} vector (using the \code{\link{coverageOverRanges}}
#' function). These coverages are compared to the coverage of regions flanking
#' the binding sites. If not indicated in \code{bsFlank} these regions are of
#' the same width as the binding sites.
#'
#' Testing the width of 3nt for example, would result in a coverage within all
#' 3nt wide binding sites (c1) and a coverage computed on the adjacent 3nt
#' flanking the binding sites up- and downstream (f1, f2). Based on these
#'  numbers the ratio is computed by: c1/(1/2(f1+f2)).
#'
#' The median over all ratios is reported as representative value.
#'
#' @param object a BSFDataSet object
#' @param bsWidths a numeric vector indicating the different binding site
#' width to compute the ratio for
#' @param bsFlank optional; a numeric vector of the same length as
#' \code{bsWidth} used to specify the width of the flanking regions
#' @param sub.chr chromosome identifier (eg, chr1, chr2) used for subsetting the
#' BSFDataSet object.
#' @param ... further arguments passed to \code{makeBindingSites}
#'
#' @return an object of class \code{data.frame}
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' suppressWarnings(supportRatio(bds, bsWidths = c(3,7)))
#'
#' @export
supportRatio <- function(object, bsWidths, bsFlank = NA, sub.chr = NA, ...) {
    # deprecation notice
    .Deprecated("estimateBsWidth")

    stopifnot(is(object, "BSFDataSet"))
    if (!any(is.numeric(bsWidths))) {
        stop("bsWidth needs to be numeric. ")
    }
    if (any(round(bsWidths) != bsWidths)) {
        stop("bsWidth is not an integer. ")
    }
    # check bsFlank vector if set
    if (!any(is.na(bsFlank))) {
        if (!any(is.numeric(bsFlank))) {
            stop("bsFlank needs to be numeric. ")
        }
        if (any(round(bsFlank) != bsFlank)) {
            stop("bsFlank is not an integer. ")
        }
        if (length(bsWidths) != length(bsFlank)) {
            stop("bsWidths and bsFlank needs to be vectors
                 of the same length. ")
        }
    }
    # use same width for flanking regions as the binding sites are
    if (any(is.na(bsFlank))) {
        bsFlank = bsWidths
    }
    # compute binding sites
    print("make bs")
    objList = lapply(bsWidths, function(x){
        makeBindingSites(object = object, bsSize = x, sub.chr = sub.chr, ...)
    })
    # calculate ratio
    print("calc ratio")
    objScore = lapply(seq_along(bsWidths), function(x){
        .computeSupportRatio(object = objList[[x]], flankSize = bsFlank[x])
    })
    # return results
    resDf = data.frame(bsWidths = factor(bsWidths),
                       supportRatio = unlist(objScore))
    return(resDf)
}



#' Calculate signal-to-flank score
#'
#' This function calculates the signal-to-flank ratio for all present binding
#' sites.
#'
#' Each input range is treated as a binding site. For a particular binding site
#' all overlapping crosslinks are summed up and divided by the normalized sum of
#' the crosslinks in the two adjecent regions of the same size. This is done
#' for all bining sites and the ratio is reported as a score.
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a BSFDataSet object
#' @param flank character; how the flanking region shoule be set. Options are
#' 'bs', 'manual'
#' @param flank.size numeric; if flank='manual' provide the desired flanking size
#' @param quiet logical; whether to print messages
#'
#' @return an object of class \code{\link{BSFDataSet}} with signal-to-flank ratios
#' added to the meta column of the ranges.
#'
#' @seealso \code{\link{BSFind}}, \code{\link{bindingSiteDefinednessPlot}}
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' bds = makeBindingSites(bds, bsSize = 5)
#' bds = calculateSignalToFlankScore(bds)
#'
#' @export
calculateSignalToFlankScore <- function(
        object,
        flank = c("bs", "manual"),
        flank.size = NULL,
        quiet = FALSE
){
    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    flank = match.arg(flank, choices = c("bs", "manual"))

    # ---
    # Store function parameters in list
    optstr = list(flank = flank)
    object@params$calculateSignalToFlankScore = optstr

    # PREPARE TEST RANGES + SIGNAL
    # --------------------------------------------------------------------------
    # collapse signal from replicates
    sgnMerge = .collapseSamples(getSignal(object))
    # prepare ranges
    rng = getRanges(object)

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    # handle all plus ranges
    cRangePlus = subset(rng, strand == "+")
    if (length(cRangePlus) > 0) {
        bsSumPlus = sum(sgnMerge$signalPlus[cRangePlus])
        # make flanking range
        if (flank == "bs") {
            extendedRangePlus = cRangePlus + width(cRangePlus)
        }
        if (flank == "manual") {
            if (is.null(flank.size)) {
                msg = paste0("If 'flank' is set to be 'manual', then provide a valide flank.size as numeric value reflecting in nucleotides, or use option 'bs'.\n")
            } else {
                extendedRangePlus = cRangePlus + flank.size
            }
        }
        # get sum over flanking range
        exSumPlus = sum(sgnMerge$signalPlus[extendedRangePlus])
        mcols(extendedRangePlus)$signalToFlankRatio = (bsSumPlus / exSumPlus)
    } else {
        msg = paste0("No ranges on '+' strand. \n")
        if(!quiet) warning(msg)
        extendedRangePlus = cRangePlus
    }
    # handle all minus ranges
    cRangeMinus = subset(rng, strand == "-")
    if (length(cRangeMinus) > 0) {
        bsSumMinus = sum(sgnMerge$signalMinus[cRangeMinus])
        # make flanking range
        if (flank == "bs") {
            extendedRangeMinus = cRangeMinus + width(cRangeMinus)
        }
        if (flank == "manual") {
            if (is.null(flank.size)) {
                msg = paste0("If 'flank' is set to be 'manual', then provide a valide flank.size as numeric value reflecting in nucleotides, or use option 'bs'.\n")
            } else {
                extendedRangeMinus = cRangeMinus + flank.size
            }
        }
        # get sum over flanking range
        extendedRangeMinus = cRangeMinus + cRangeMinus$bsSize
        exSumMinus = sum(sgnMerge$signalMinus[extendedRangeMinus])
        mcols(extendedRangeMinus)$signalToFlankRatio = (bsSumMinus / exSumMinus)
    } else {
        msg = paste0("No ranges on '-' strand. \n")
        if(!quiet) warning(msg)
        extendedRangeMinus = cRangeMinus
    }

    rngNew = c(extendedRangePlus, extendedRangeMinus)
    rngNew = .sortRanges(rngNew)

    idx = match(rng$bsID, rngNew$bsID)
    mcols(rng) = mcols(rngNew[idx])

    object = setRanges(object, rng)

    # ---
    # Store for results
    rng = getRanges(object)
    resultLine = data.frame(
        funName = "calculateSignalToFlankScore()", class = "estimate",
        nIn = length(rng), nOut = length(rng),
        per = paste0(round(length(rng)/ length(rng), digits = 2)*100,"%"),
        options = paste0("flank=", optstr$flank)
    )
    object@results = rbind(object@results, resultLine)

    return(object)
}



#' Combine multiple \code{\link{BSFDataSet}} objects
#'
#' This function combines all \code{\link{BSFDataSet}} objects from the input
#' list into a single \code{\link{BSFDataSet}} object.
#'
#' Meta-data tables are added to each other by performing a row-wise bind,
#' basically adding all meta data tables underneath each other.
#'
#' The default way of signal combination is merging all signal lists on the level
#' of the indivdual samples. One can also force a re-load of the signal list
#' component by using \code{force.reload=TRUE}.
#' The signal can be combined by
#'
#' The ranges are combined by adding both granges objects together. With option
#' \code{overlaps.fix} one can decide if partially overlapping ranges should be
#' combined into a single range or not. If this option is FALSE one is likely to
#' have overlapping binding sites after the merge. If this option is TRUE, then
#' the combined coverage is used to guide the new center point for these cases.
#'
#' The \code{combine.bsSize} option allows one to set a unique bsSize for all
#' objects that should be combined. Although it is recommended to combine only
#' objects with the same bsSize this option can be used to ensure that the
#' merged result has the same bsSize for all ranges.
#'
#' This function is usually used to combine two datasets in the context of a
#' differntial testing analysis.
#'
#' @param list list; a list of objects from class \code{\link{BSFDataSet}} that
#' should be combined
#' @param overlaps.fix logical; if partially overlapping binding sites should be
#' re-centered
#' @param combine.bsSize numeric; the binding site size that the merged sites
#' should have. Default=NULL, then bsSize is taken from the input objects
#' in \code{list}.
#' @param combine.name character; meta table name of the combined object.
#' Default=NULL; then name is set to 'combined'
#' @param force.reload logical; whether the signal should be derived from the
#' merge of the input objects given in \code{list} or if the signal should be
#' re-loaded from the path given in the meta data.
#' @param quiet logical; whether to print messages or not
#' @param veryQuiet logical; whether to print status messages or not
#'
#' @return an object of class \code{\link{BSFDataSet}} with ranges, signal and
#' meta data resulting from the merge of the input objects.
#'
#' @seealso \code{\link{processingStepsFlowChart}}, \code{\link{calculateBsBackground}}
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' # make binding sites
#' bds = makeBindingSites(bds, bsSize = 7)
#'
#' # split ranges in two groups
#' allRanges = getRanges(bds)
#' set.seed(1234)
#' idx = sample(1:length(allRanges), size = length(allRanges)/2, replace = FALSE)
#' r1 = allRanges[idx]
#' r2 = allRanges[-idx]
#'
#' # splite meta data
#' allMeta = getMeta(bds)
#' m1 = allMeta[1:2,]
#' m2 = allMeta[3:4,]
#'
#' # create new objects
#' bds1 = setRanges(bds, r1)
#' bds2 = setRanges(bds, r2)
#' bds1 = setMeta(bds1, m1)
#' bds2 = setMeta(bds2, m2)
#' bds1 = setName(bds1, "test1")
#' bds2 = setName(bds2, "test2")
#'
#' # merge two objects with '+' operator
#' c1 = bds1 + bds2
#'
#' # merge two objects from list
#' list = list(bds1, bds2)
#' c1 = combineBSF(list = list, overlaps.fix = TRUE,
#'  combine.bsSize = NULL, combine.name = NULL, quiet = TRUE)
#'
#' @export
combineBSF <- function(list, # list of class BSFDataSet
                       overlaps.fix = TRUE,
                       combine.bsSize = NULL,
                       combine.name = NULL,
                       force.reload = FALSE,
                       quiet = TRUE,
                       veryQuiet = FALSE
){
    # initialize local variables
    this.bsSize <- resizedHit <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(all(unlist(lapply(list, is, "BSFDataSet"))))
    if (!is.null(combine.bsSize)){
        stopifnot(is(combine.bsSize, "numeric"))
    }

    # check dataset names
    list = lapply(seq_along(list), function(x){
        this.dataset = list[[x]]
        this.name = getName(list[[x]])
        if (length(this.name) == 0) {
            # No name was given to this dataset
            # -> make name and set
            new.name = paste0("bds", x)
            this.dataset = setName(list[[x]], new.name)

            # Inform user
            msg0 = paste0("No name present for dataset No: ", x, ".\n")
            msg1 = paste0("Name is set to: ", new.name, ".\n")
            msg2 = paste0("Prevent this message by setting a name using 'setNames()'.\n")
            if (!quiet) warning(c(msg0, msg1, msg2))
        }
        return(this.dataset)
    })

    # Combine meta
    # --------------------------------------------------------------------------
    comb.meta = do.call(rbind, lapply(list, getMeta))

    # check for duplicated path
    if (any(duplicated(comb.meta$clPlus))) {
        msg = paste0("Duplicated path for plus strand found. Please check input.\n")
        if (!quiet) warning(c(msg))
    }
    if (any(duplicated(comb.meta$clMinus))) {
        msg = paste0("Duplicated path for minus strand found. Please check input.\n")
        if (!quiet) warning(c(msg))
    }
    # set ids
    comb.meta$id = 1:nrow(comb.meta)

    # set name
    comb.meta$datasets = comb.meta$name
    if (is.null(combine.name)) {
        comb.meta$name = "combined"
    } else {
        comb.meta$name = combine.name
    }

    # Combine signal
    # --------------------------------------------------------------------------
    # combine signal if user wants it or not
    if (!isTRUE(force.reload)) {
        # combine plus strand
        signal.plus = lapply(list, function(x){
            this.signal = getSignal(x)
            this.signal.plus = this.signal$signalPlus
            return(this.signal.plus)
        })
        signal.plus = unlist(signal.plus)
        names(signal.plus) = paste0(comb.meta$id, "_", comb.meta$condition)
        # combine minus strand
        signal.minus = lapply(list, function(x){
            this.signal = getSignal(x)
            this.signal.minus = this.signal$signalMinus
            return(this.signal.minus)
        })
        signal.minus = unlist(signal.minus)
        names(signal.minus) = paste0(comb.meta$id, "_", comb.meta$condition)

        comb.signal = list(signalPlus = signal.plus,
                           signalMinus = signal.minus)
    }

    # Combine ranges
    # --------------------------------------------------------------------------
    comb.ranges = lapply(seq_along(list), function(x){
        this.range = getRanges(list[[x]])
        mcols(this.range)$dataset = getName(list[[x]])
        return(this.range)
    })
    comb.ranges = do.call(c, comb.ranges)

    # fix overlaps or not
    if (!isTRUE(overlaps.fix)) {
        msg = paste0("Overlaps are not resoloved. This could lead to overlapping binding site.\n")
        if (!quiet) warning(msg)
    }

    # check for identical bsSize
    if (!length(unique(width(comb.ranges))) == 1) {
        # not all ranges have the same size
        msg0 = paste0("Not all ranges are of the same size. \n")
        msg1 = paste0("Found ranges of bsSize = c(", paste(unique(width(comb.ranges)), collapse = ","), ").\n")
        # do not fix ranges
        if (!isTRUE(overlaps.fix)) {
            if (!quiet) warning(c(msg0, msg1))
        } else {
            # ranges should be fixed
            if (is.null(combine.bsSize)) {
                msg2 = paste0("Use 'combine.bsSize' to set a reference size for all ranges.\n")
                stop (c(msg0, msg1, msg2))
            } else {
                msg2 = paste0("Manual size of ", combine.bsSize, "nt is used for merging.\n")
                this.bsSize = combine.bsSize
            }
            if (!quiet) warning(c(msg0, msg1, msg2))
        }
    } else {
        # all ranges have equal size
        if (unique(width(comb.ranges)) == 1) {
            # ranges to combine are of 1 nt range
            msg0 = paste0("Range to combine have width of 1nt.\n")
            msg1 = paste0("Make sure to run binding site definition first; see `BSFind()`.\n")
            msg2 = paste0("Alternativly set `overlaps.fix=FALSE` to simply add ranges.\n")
            if (!quiet) warning(c(msg0,msg1,msg2))
        }
        if (!is.null(combine.bsSize)) {
            # set combined bsSize manually
            this.bsSize = combine.bsSize
        } else {
            # set combined bsSize from data
            this.bsSize = unique(width(comb.ranges))
        }
    }

    # ---
    # Store function parameters in list
    optstr = list(overlaps.fix = overlaps.fix, combine.bsSize = this.bsSize,
                  combine.name = combine.name, force.reload = force.reload)

    # Resize ranges overlaps
    # --------------------------------------------------------------------------
    # don't fix overlaps, keep as it is
    if (!isTRUE(overlaps.fix)) {
        # apply new bsSize for all ranges if needed
        if (!is.null(this.bsSize)) {
            # turn ranges into midpoint representations
            comb.ranges = resize(comb.ranges, fix = "center", width = this.bsSize)
        }
        # create combined BSFDataSet
        if (!veryQuiet) message("Creating merged object...")
        # use reload option or not
        if (isTRUE(force.reload)){
            resize.bds = BSFDataSetFromBigWig(ranges = comb.ranges, meta = comb.meta, silent = quiet)
        } else {
            resize.bds = BSFDataSet(ranges = comb.ranges, meta = comb.meta, signal = comb.signal, silent = quiet)
        }
    } else {
        # turn ranges into midpoint representations
        resize.ranges = resize(comb.ranges, fix = "center", width = 1)
        # create combined BSFDataSet
        if (!veryQuiet) message("Creating merged object...")
        # use reload option or not
        if (isTRUE(force.reload)) {
            resize.bds = BSFDataSetFromBigWig(ranges = resize.ranges, meta = comb.meta, silent = quiet)
        } else {
            resize.bds = BSFDataSet(ranges = resize.ranges, meta = comb.meta, signal = comb.signal, silent = quiet)
        }
        # resolve overlaps by merging
        if (!veryQuiet) message("Fixing partial overlaps...")
        resize.bds = makeBindingSites(object = resize.bds, bsSize = this.bsSize,
                                      minWidth = 0, minCrosslinks = 0, minClSites = 0,
                                      centerIsSummit = FALSE, centerIsClSite = FALSE,
                                      quiet = TRUE)
        # remove stats from makeBiningSites
        resize.bds@results = data.frame()
        resize.bds@params = list()
        resize.bds@plotData = list()
    }

    # Add meta information from original ranges
    # --------------------------------------------------------------------------
    # match resized binding sites with all input binding sites
    if (!veryQuiet) message("Merging meta info...")

    resized.ranges = getRanges(resize.bds)
    ols = findOverlaps(resized.ranges, comb.ranges)

    matched.ranges = resized.ranges[queryHits(ols)]
    matched.ranges$combHit = subjectHits(ols)
    matched.ranges$resizedHit = queryHits(ols)

    mcols(matched.ranges) = cbind(combHit = matched.ranges$combHit,
                                  resizedHit = matched.ranges$resizedHit,
                                  mcols(comb.ranges[matched.ranges$combHit]))

    # group by matching
    grouped.meta = mcols(matched.ranges) %>%
        as.data.frame(, row.names = NULL) %>%
        group_by(resizedHit)

    # match meta data from overlapping binding sites
    total.meta = .matchMetaData(grouped.meta)

    # attach meta data to ranges
    fix.ranges = resized.ranges
    mcols(fix.ranges) = cbind(mcols(fix.ranges), total.meta)

    # manage other columns
    mcols(fix.ranges)$bsID = paste0("BS", seq_along(fix.ranges))
    if ("bsSize" %in% colnames(mcols(resized.ranges))) {
        mcols(fix.ranges)$bsSize = max(resized.ranges$bsSize)
    }
    mcols(fix.ranges)$resizedHit = NULL

    # set final object
    comb.bds = setRanges(resize.bds, fix.ranges)

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "combineBSF()", class = "transform",
        nIn = length(comb.ranges), nOut = length(fix.ranges),
        per = paste0(round(length(fix.ranges)/ length(comb.ranges), digits = 2)*100,"%"),
        options = paste0("overlaps.fix=", optstr$overlaps.fix, ", combine.bsSize=", optstr$combine.bsSize)
    )
    comb.bds@results = rbind(comb.bds@results, resultLine)
    comb.bds@params$combineBSF = optstr

    return(comb.bds)
}

test_that("Test that score annotation function works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # prepare objects
    obj0 = bds
    obj1 = makeBindingSites(object = obj0, bsSize = 9, quiet = TRUE)

    # test with OrganismDB as annotation source
    expect_error(annotateWithScore(object = obj1))
    expect_error(annotateWithScore(object = obj1, match.ranges = obj0))
    expect_silent(annotateWithScore(object = obj1, match.ranges = getRanges(obj0)))
    expect_equal(annotateWithScore(object = obj1, match.ranges = getRanges(obj0)),
                 annotateWithScore(object = obj1, match.ranges = getRanges(obj0), match.option = "max"))

    expect_silent(annotateWithScore(object = obj1, match.ranges = getRanges(obj0), match.option = "sum"))
    expect_silent(annotateWithScore(object = obj1, match.ranges = getRanges(obj0), match.option = "mean"))

    expect_error(annotateWithScore(object = obj1, match.ranges = getRanges(obj0), match.option = "test"))

})

test_that("Binding site merging works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # testing input type errors
    expect_warning(expect_error(supportRatio(bds, bsWidths = "1")))
    expect_warning(expect_error(supportRatio(bds, bsWidths = c(1.1))))
    expect_warning(expect_error(supportRatio(bds, bsWidths = c(3,5), bsFlank = c(5,5,5))))
    expect_warning(expect_error(supportRatio(bds, bsWidths = c(3,5), bsFlank = c(5))))
    expect_warning(expect_error(supportRatio(bds, bsWidths = c(3,5), bsFlank = c(5,5.1))))

    # testing output
    (expect_warning(supportRatio(bds, bsWidths = c(3,5))))
    (expect_warning(supportRatio(bds, bsWidths = c(5,7))))
})

test_that("Reproducibility filter function works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    expect_error(reproducibilityFilter(bds, cutoff = c(0.1,0.1)))
    expect_error(reproducibilityFilter(bds, nReps = c(2,2)))

    expect_is(reproducibilityFilter(bds, cutoff = c(0.1), nReps = c(2), returnType = "BSFDataSet"), "BSFDataSet")
    expect_is(reproducibilityFilter(bds, cutoff = c(0.1), nReps = c(2), returnType = "data.frame"), "data.frame")

    expect_error(reproducibilityFilter(bds, cutoff = c(0.1), nReps = c(1,1)))
    expect_error(reproducibilityFilter(bds, cutoff = c(0.1, 0.2), nReps = 1))

    expect_silent(reproducibilityFilter(bds, cutoff = c(0.1), nReps = 1))
    expect_error(expect_warning(reproducibilityFilter(bds, cutoff = c(0.1))))
    expect_error(expect_warning(reproducibilityFilter(bds, nReps = c(1))))

    expect_message(reproducibilityFilter(bds))

})

test_that("Combine function works", {
    # load clip data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # make binding sites
    bds = makeBindingSites(bds, bsSize = 7)

    # split ranges in two groups
    allRanges = getRanges(bds)
    set.seed(1234)
    idx = sample(1:length(allRanges), size = length(allRanges)/2, replace = FALSE)
    r1 = allRanges[idx]
    r2 = allRanges[-idx]

    # splite meta data
    allMeta = getMeta(bds)
    m1 = allMeta[1:2,]
    m2 = allMeta[3:4,]

    # create new objects
    bds1 = setRanges(bds, r1)
    bds2 = setRanges(bds, r2)
    bds1 = setMeta(bds1, m1)
    bds2 = setMeta(bds2, m2)
    bds1 = setName(bds1, "test1")
    bds2 = setName(bds2, "test2")

    # merge two objects from list
    list = list(bds1, bds2)

    c1 = combineBSF(list = list, overlaps.fix = TRUE, combine.bsSize = NULL, combine.name = NULL, quiet = TRUE)
    r = unique(unique(width(getRanges(list[[1]])), width(getRanges(list[[2]]))))
    expect_identical(unique(width(getRanges(c1))), r)

    inSize = 9
    c2 = combineBSF(list = list, overlaps.fix = TRUE, combine.bsSize = inSize, combine.name = NULL, quiet = TRUE)
    expect_equal(unique(width(getRanges(c2))), inSize)

    c3 = combineBSF(list = list, overlaps.fix = FALSE, combine.bsSize = NULL, combine.name = NULL, quiet = TRUE)
    nr = length(c(getRanges(list[[1]]), getRanges(list[[2]])))
    expect_identical(length(getRanges(c3)), nr)

    inName = "test"
    c4 = combineBSF(list = list, overlaps.fix = FALSE, combine.bsSize = NULL, combine.name = inName, quiet = TRUE)
    expect_identical(getName(c4), inName)


    # tests with different width
    changeRanges = getRanges(bds1)
    bds1C = setRanges(bds1, changeRanges + 2)
    list2 = list(bds1C, bds2)

    expect_error(c1 = combineBSF(list = list2, overlaps.fix = TRUE, combine.bsSize = NULL, combine.name = NULL, quiet = TRUE))

    inSize = 9
    c2 = combineBSF(list = list2, overlaps.fix = TRUE, combine.bsSize = inSize, combine.name = NULL, quiet = TRUE)
    expect_equal(unique(width(getRanges(c2))), inSize)

    c3 = combineBSF(list = list2, overlaps.fix = FALSE, combine.bsSize = NULL, combine.name = NULL, quiet = TRUE)
    nr = length(c(getRanges(list2[[1]]), getRanges(list2[[2]])))
    expect_identical(length(getRanges(c3)), nr)
})

#' Diagnostic plots for the differential binding background
#'
#' To perform differential binding analysis between two conditions the
#' \code{\link{calculateBsBackground}} function groups crosslinks per gene
#' into those from binding sites and those from background regions.
#' The \code{\link{filterBsBackground}} function perfroms certain
#' filtering operations on that background to ensure that it's suitable
#' for differential testing. This function visually displays the effect
#' of these filtering operations.
#'
#'
#' @param object a \code{\link{BSFDataSet}} object with background counts filtered
#' by \code{\link{filterBsBackground}}
#' @param filter character; which filter to display in the plot (one of:
#' 'minCounts', 'balanceBackground', 'balanceCondition')
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{calculateBsBackground}}
#' @seealso \code{\link{filterBsBackground}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#'
#' # make testset
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = imputeBsDifferencesForTestdata(bds)
#' bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)
#'
#' # use all filters and remove binding sites that fail (default settings)
#' bds = filterBsBackground(bds)
#'
#' # display minCount filter
#' plotBsBackgroundFilter(bds, filter = "minCounts")
#'
#' # display balance background filter
#' plotBsBackgroundFilter(bds, filter = "balanceBackground")
#'
#' # display balance condition filter
#' plotBsBackgroundFilter(bds, filter = "balanceCondition")
#'
#' @export
plotBsBackgroundFilter <- function(object, filter = c("minCounts", "balanceBackground" ,"balanceCondition")) {

    # bind locally used variables
    s <- r <- name <- value <- ratio.ref <- ratio.comp <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    if (is.null(object@params$filterBsBackground)) {
        msg0 = paste0("Global filter was not applied yet. Run filterBsBackground() to compute background first. \n")
        stop(msg0)
    }

    # handle display plot options
    filter = match.arg(filter, choices = c("minCounts", "balanceBackground", "balanceCondition"))

    # check for plot specific data
    if (is.null(object@plotData$filterBsBackground$data.minCounts) & filter == "minCounts") {
        msg0 = paste0("Plot selection: '", filter, "' is not possible.")
        msg1 = paste0("Make sure to run filterBsBackground with option '", filter, "' first.")
        stop(c(msg0,msg1))
    }
    if (is.null(object@plotData$filterBsBackground$data.balanceBackground) & filter == "balanceBackground") {
        msg0 = paste0("Plot selection: '", filter, "' is not possible.")
        msg1 = paste0("Make sure to run filterBsBackground with option '", filter, "' first.")
        stop(c(msg0,msg1))
    }
    if (is.null(object@plotData$filterBsBackground$data.balanceCondition) & filter == "balanceCondition") {
        msg0 = paste0("Plot selection: '", filter, "' is not possible.")
        msg1 = paste0("Make sure to run filterBsBackground with option '", filter, "' first.")
        stop(c(msg0,msg1))
    }


    # MAIN
    # --------------------------------------------------------------------------

    # make plot for minCounts filter
    if (filter == "minCounts") {
        # get options used
        optstr = object@params$filterBsBackground
        optstrNice = paste0("Cutoff=", optstr$minCounts.cutoff)

        # get data
        df = object@plotData$filterBsBackground$data.minCounts %>%
            mutate(r = rank(s))

        p = ggplot(df, aes(x = r, y = s)) +
            ggrastr::rasterise(ggpointdensity::geom_pointdensity(size = 2), dpi = 300) +
            theme_bw() +
            geom_hline(yintercept = optstr$minCounts.cutoff, linetype = "dashed") +
            scale_y_log10() +
            viridis::scale_color_viridis(option = "rocket") +
            theme(legend.position = "top") +
            guides(color = guide_colorbar(title.position = 'top', title.hjust = 0.5,
                                          barwidth = unit(20, 'lines'), barheight = unit(.5, 'lines'))) +
            annotate("text", x = max(df$r)/2, y = optstr$minCounts.cutoff,
                     label = paste0("Cutoff=", optstr$minCounts.cutoff),
                     vjust = -1.1) +
            labs(
                title = "filterBsBackground() - minCounts",
                subtitle = optstrNice,
                x = "Rank",
                y = "#Crosslinks per gene",
                color = "#Genes"
            )
    }

    # make plot for balanceBackground filter
    if (filter == "balanceBackground") {
        # get options used
        optstr = object@params$filterBsBackground
        optstrNice = paste0("Cutoff.bs=", optstr$balanceBackground.cutoff.bs,
                            " cutoff.bg=", optstr$balanceBackground.cutoff.bg)

        # get data
        df = object@plotData$filterBsBackground$data.balanceBackground %>%
            mutate(name = ifelse(name == "ratio.bg", "Background", "Binding site"))

        df = df[!is.na(df$value),]

        p = ggplot(df, aes(x = name, y = value)) +
            ggdist::stat_halfeye(aes(fill = name), adjust = 0.5, width = .6, .width = 0, justification = -.2, point_colour = NA) +
            geom_boxplot(aes(color = name), width = .12, outlier.size = 0.5) +
            geom_point(data = subset(df, name == "ratio.bg" & value < optstr$balanceBackground.cutoff.bg), aes(fill = name), shape = 21, alpha = 0.3) +
            geom_point(data = subset(df, name == "ratio.bs" & value > optstr$balanceBackground.cutoff.bs), aes(fill = name), shape = 21, alpha = 0.3) +
            annotate("rect", xmin = 1.75, xmax = 2.25, ymin = optstr$balanceBackground.cutoff.bs, ymax = 1, alpha = .3, fill = "#435B66") +
            annotate("rect", xmin = 0.75, xmax = 1.25, ymin = 0, ymax = optstr$balanceBackground.cutoff.bg, alpha = .3, fill = "#A76F6F") +
            annotate("text", x = 2.25, y = 0.5, vjust = -1, label = paste0("Remove genes with bs ratio above: ", optstr$balanceBackground.cutoff.bs)) +
            annotate("text", x = 1.25, y = 0.5, vjust = -1, label = paste0("Remove genes with bg ratio below: ", optstr$balanceBackground.cutoff.bg)) +
            coord_flip() +
            theme_bw() +
            scale_fill_manual(values = c("#A76F6F", "#435B66")) +
            scale_color_manual(values = c("#A76F6F", "#435B66")) +
            theme(legend.position = "none") +
            labs(
                title = "filterBsBackground() - balanceBackground",
                subtitle = optstrNice,
                x = NULL,
                y = "Background to binding site count ratio"
            )
    }

    # make plot for balanceCondition filter
    if (filter == "balanceCondition") {
        # get options used
        optstr = object@params$filterBsBackground
        optstrNice = paste0("Cutoff=", optstr$balanceCondition.cutoff)

        # get reference condition
        this.meta = getMeta(object)
        this.condition.reference = levels(this.meta$condition)[1]
        this.condition.comp = levels(this.meta$condition)[2]

        # get data
        df = object@plotData$filterBsBackground$data.balanceCondition %>% as.data.frame()
        df$ratio.ref = df[,3] / df[,2]
        df$ratio.comp = df[,4] / df[,2]
        df = df %>% select(ratio.ref, ratio.comp) %>% pivot_longer(everything())

        p = ggplot(df, aes(x = name, y = value)) +
            ggdist::stat_halfeye(aes(fill = name), adjust = 0.5, width = .6, .width = 0, justification = -.2, point_colour = NA) +
            geom_boxplot(aes(color = name), width = .12, outlier.size = 0.5) +
            geom_point(data = subset(df, name == "ratio.ref" & value < optstr$balanceCondition.cutoff), aes(fill = name), shape = 21, alpha = 0.3) +
            geom_point(data = subset(df, name == "ratio.comp" & value < optstr$balanceCondition.cutoff), aes(fill = name), shape = 21, alpha = 0.3) +
            annotate("rect", xmin = 0.05, xmax = 2.95, ymin = 0, ymax = optstr$balanceCondition.cutoff, alpha = .3, fill = "darkgrey") +
            coord_flip() +
            theme_bw() +
            scale_fill_manual(values = c("#A76F6F", "#435B66")) +
            scale_color_manual(values = c("#A76F6F", "#435B66")) +
            theme(legend.position = "none") +
            labs(
                title = "filterBsBackground() - balanceCondition",
                subtitle = optstrNice,
                x = NULL,
                y = "Condition count ratio"
            ) +
            scale_x_discrete(labels = c(paste0(this.condition.comp, "\n (comparison)"),
                                        paste0(this.condition.reference, "\n (reference)")))


    }

    return(p)
}


#' MA style plot
#'
#' Wrapper that plots differential binding results as MA plot. For each binding
#' site the estimated baseMean (log2) is shown on X and the fold-change (log2)
#' is shown on Y.
#'
#' @param object a \code{\link{BSFDataSet}} object with results calculated by
#' \code{\link{calculateBsFoldChange}}
#' @param what character; whether to show results for binding sites or the
#' background (one of: 'bs', 'bg')
#' @param sig.threshold numeric; what P value significance level to use
#' (default = 0.05)
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{calculateBsFoldChange}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#'
#' # make testset
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = imputeBsDifferencesForTestdata(bds)
#' bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)
#'
#' # use all filters and remove binding sites that fail (default settings)
#' bds = filterBsBackground(bds)
#'
#' # calculate fold-changes
#' bds = calculateBsFoldChange(bds)
#'
#' # make MA plot
#' plotBsMA(bds)
#'
#' @export
plotBsMA <- function(object,
                     what = c("bs", "bg"),
                     sig.threshold = 0.05) {

    # bind locally used variables
    bs.padj <- bs.log2FoldChange <- sig <- bs.baseMean <- bg.padj <- NULL
    bg.log2FoldChange <- bg.baseMean <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    if (is.null(object@params$calculateBsFoldChange)) {
        msg0 = paste0("Fold-changes were not calculated yet. Run calculateBsFoldChange() first. \n")
        stop(msg0)
    }

    # handle display plot options
    what = match.arg(what, choices = c("bs", "bg"))

    this.ranges = getRanges(object)
    expected.cols = c("bs.padj", "bs.log2FoldChange", "bs.baseMean",
                      "bg.padj", "bg.log2FoldChange", "bg.baseMean")
    if (!all(expected.cols %in% colnames(as.data.frame(mcols(this.ranges))))) {
        msg0 = paste0("MA plot not possible, results columns: ", paste(expected.cols, collapse = ','), " not present.\n")
        msg1 = paste0("Make sure to run calculateBsFoldChange() first.\n")
        stop(c(msg0, msg1))
    }

    # MAIN
    # --------------------------------------------------------------------------

    optstr = object@params$calculateBsFoldChange
    optstrNice = paste0("alpha=", optstr$alpha)

    bright_up_down_not = c("#999999", "#68b1a5", "#874C62")
    dark_up_down_not = c("#4d4d4d", "#2b544d", "#623747")

    if (what == "bs") {
        df = as.data.frame(mcols(this.ranges)) %>%
            mutate(sig = ifelse(bs.padj < sig.threshold & bs.log2FoldChange > 0, "Up",
                                ifelse(bs.padj < sig.threshold & bs.log2FoldChange < 0, "Down", "Not"))) %>%
            mutate(sig = factor(sig, levels = c("Not", "Up", "Down"))) %>%
            arrange(sig)

        p = ggplot(df, aes(x = log2(bs.baseMean), y = bs.log2FoldChange, color = sig, fill = sig)) +
            ggrastr::rasterise(geom_point(shape = 21, stroke = 0.5, size = 1.5), dpi = 300) +
            geom_hline(yintercept = 0, color = "black", alpha = .5) +
            theme_bw() +
            scale_fill_manual(values = bright_up_down_not) +
            scale_color_manual(values = dark_up_down_not) +
            guides(color = guide_legend(override.aes = list(size = 4))) +
            theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
            labs(
                title = "plotBsMA() - binding sites",
                subtitle = optstrNice,
                x = "Base mean",
                y = "Fold-change (log2)",
                color = "Regulation",
                fill = "Regulation")
    }
    if (what == "bg") {
        df = as.data.frame(mcols(this.ranges)) %>%
            mutate(sig = ifelse(bg.padj < sig.threshold & bg.log2FoldChange > 0, "Up",
                                ifelse(bg.padj < sig.threshold & bg.log2FoldChange < 0, "Down", "Not"))) %>%
            mutate(sig = factor(sig, levels = c("Not", "Up", "Down"))) %>%
            arrange(sig)

        p = ggplot(df, aes(x = log2(bg.baseMean), y = bg.log2FoldChange, color = sig, fill = sig)) +
            ggrastr::rasterise(geom_point(shape = 21, stroke = 0.5, size = 1.5), dpi = 300) +
            geom_hline(yintercept = 0, color = "black", alpha = .5) +
            theme_bw() +
            scale_fill_manual(values = bright_up_down_not) +
            scale_color_manual(values = dark_up_down_not) +
            guides(color = guide_legend(override.aes = list(size = 4))) +
            theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
            labs(
                title = "plotBsMA() - background",
                subtitle = optstrNice,
                x = "Base mean",
                y = "Fold-change (log2)",
                color = "Regulation",
                fill = "Regulation")
    }


    return(p)
}

#' Volcano style plot
#'
#' Wrapper that plots differential binding results as volcano plot. For each binding
#' site the estimated fold-change (log2) is shown on X and the adjusted
#' P value (-log10) is shown on Y.
#'
#' @param object a \code{\link{BSFDataSet}} object with results calculated by
#' \code{\link{calculateBsFoldChange}}
#' @param what character; whether to show results for binding sites or the
#' background (one of: 'bs', 'bg')
#' @param sig.threshold numeric; what P value significance level to use
#' (default = 0.05)
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{calculateBsFoldChange}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#'
#' # make testset
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = imputeBsDifferencesForTestdata(bds)
#' bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)
#'
#' # use all filters and remove binding sites that fail (default settings)
#' bds = filterBsBackground(bds)
#'
#' # calculate fold-changes
#' bds = calculateBsFoldChange(bds)
#'
#' # make volcano plot
#' plotBsVolcano(bds)
#'
#' @export
plotBsVolcano <- function(object,
                          what = c("bs", "bg"),
                          sig.threshold = 0.05) {

    # bind locally used variables
    bs.padj <- bs.log2FoldChange <- sig <- bg.padj <- bg.log2FoldChange <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    if (is.null(object@params$calculateBsFoldChange)) {
        msg0 = paste0("Fold-changes were not calculated yet. Run calculateBsFoldChange() first. \n")
        stop(msg0)
    }

    # handle display plot options
    what = match.arg(what, choices = c("bs", "bg"))

    this.ranges = getRanges(object)
    expected.cols = c("bs.padj", "bs.log2FoldChange", "bs.baseMean",
                      "bg.padj", "bg.log2FoldChange", "bg.baseMean")
    if (!all(expected.cols %in% colnames(as.data.frame(mcols(this.ranges))))) {
        msg0 = paste0("MA plot not possible, results columns: ", paste(expected.cols, collapse = ','), " not present.\n")
        msg1 = paste0("Make sure to run calculateBsFoldChange() first.\n")
        stop(c(msg0, msg1))
    }

    # MAIN
    # --------------------------------------------------------------------------

    optstr = object@params$calculateBsFoldChange
    optstrNice = paste0("alpha=", optstr$alpha)

    bright_up_down_not = c("#999999", "#68b1a5", "#874C62")
    dark_up_down_not = c("#4d4d4d", "#2b544d", "#623747")

    if (what == "bs") {
        df = as.data.frame(mcols(this.ranges)) %>%
            mutate(sig = ifelse(bs.padj < sig.threshold & bs.log2FoldChange > 0, "Up",
                                ifelse(bs.padj < sig.threshold & bs.log2FoldChange < 0, "Down", "Not"))) %>%
            mutate(sig = factor(sig, levels = c("Not", "Up", "Down"))) %>%
            arrange(sig)

        p = ggplot(df, aes(x = bs.log2FoldChange, y = -log10(bs.padj), color = sig, fill = sig)) +
            ggrastr::rasterise(geom_point(shape = 21, stroke = 0.5, size = 1.5), dpi = 300) +
            geom_vline(xintercept = 0, color = "black", alpha = .5) +
            theme_bw() +
            scale_fill_manual(values = bright_up_down_not) +
            scale_color_manual(values = dark_up_down_not) +
            guides(color = guide_legend(override.aes = list(size = 4))) +
            theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
            labs(
                title = "plotBsVolcano() - binding sites",
                subtitle = optstrNice,
                x = "Fold-change (log2)",
                y = "Adjusted P value (-log10)",
                color = "Regulation",
                fill = "Regulation")
    }
    if (what == "bg") {
        df = as.data.frame(mcols(this.ranges)) %>%
            mutate(sig = ifelse(bg.padj < sig.threshold & bg.log2FoldChange > 0, "Up",
                                ifelse(bg.padj < sig.threshold & bg.log2FoldChange < 0, "Down", "Not"))) %>%
            mutate(sig = factor(sig, levels = c("Not", "Up", "Down"))) %>%
            arrange(sig)

        p = ggplot(df, aes(x = bg.log2FoldChange, y = -log10(bg.padj), color = sig, fill = sig)) +
            ggrastr::rasterise(geom_point(shape = 21, stroke = 0.5, size = 1.5), dpi = 300) +
            geom_vline(xintercept = 0, color = "black", alpha = .5) +
            theme_bw() +
            scale_fill_manual(values = bright_up_down_not) +
            scale_color_manual(values = dark_up_down_not) +
            guides(color = guide_legend(override.aes = list(size = 4))) +
            theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
            labs(
                title = "plotBsVolcano() - background",
                subtitle = optstrNice,
                x = "Fold-change (log2)",
                y = "Adjusted P value (-log10)",
                color = "Regulation",
                fill = "Regulation")
    }

    return(p)
}


#' Gene Regulation Plot
#'
#' Display the fold-change of all binding sites from a given gene on a relative
#' per-nucleotide scale. Binding sites are displayed as dots and with increasing
#' log2 fold-change, they deviate stronger from the center line.
#'
#' For this function to work, binding sites must be assigned to hosting genes
#' using \code{\link{assignToGenes}}. It is also recommended to assing
#' binding sites to transcript regions with \code{\link{assignToTranscriptRegions}}.
#'
#' It is also necessary to calculate the log2 fold-change of binding sites between
#' two conditions using the differential binding workflow \code{\link{calculateBsFoldChange}}.
#'
#' If in addition the transcript regions of the binding sites are given, then
#' shapes are changed accordingly. An edge case can arise from the merging of two
#' \code{\link{BSFDataSet}} objects. If binding sites are overlapping and
#' slightly offset close to the end of a particular transcript region annotation,
#' they might be assigned to different regions in both objects. This results in
#' some ambiguity after the merge, where for instance a binding site can be
#' assigned to CDS and 3'UTR. To handle how such edge cases are displayed, the
#' \code{transcript.regions.outlier.handle} exists. As default, simply the
#' region of the object that was merged first is shown. If one is interested in
#' showing all regions, then the options \code{both} displays both annotations
#' at the same time and labels them accordingly.
#'
#' @param object object; a \code{\link{BSFDataSet}} object
#' @param plot.geneID character;  the gene id of the gene to display. The id must
#' match with the gene ids given in the annotation object.
#' @param anno.annoDB object; an object of class \code{OrganismDbi} that contains
#' the gene annotation (!!! Experimental !!!).
#' @param anno.genes object; an object of class \code{\link{GenomicRanges}}
#' that represents the gene ranges directly.
#' @param match.geneID character; meta column name of the gene ID
#' @param match.geneName character; meta column name of the gene name
#' @param plot.gene.n.tiles numeric; number of tiles the gene should be split in
#' @param alpha numeric; the alpha value to show significantly regulated
#' binding sites. This should match the alpha value used in \code{\link{calculateBsFoldChange}}.
#' @param lfc.cutoff numeric; log2 fold-change cutoff to show significantly
#' regulated binding sites. This should match the lfc.cutoff value used in
#' \code{\link{calculateBsFoldChange}}.
#' @param transcript.regions.outlier.handle character; the option how to handle
#' multiple transcript region annotations being present for the same binding
#' site.
#' @param quiet logical; whether to print messages
#'
#' @return an object of class \code{ggplot2}
#'
#' @seealso \code{\link{BSFind}},
#' \code{\link{calculateBsFoldChange}}
#' \code{\link{assignToGenes}}
#' \code{\link{assignToTranscriptRegions}}
#'
#' @import ggplot2
#'
#' @examples
#'
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#'
#' # make testset
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)
#' bds = imputeBsDifferencesForTestdata(bds)
#' bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)
#'
#' # use all filters and remove binding sites that fail (default settings)
#' bds = filterBsBackground(bds)
#'
#' # calculate fold-changes
#' bds = calculateBsFoldChange(bds)
#'
#' # make example plot
#' exampleGeneId = "ENSG00000253352.10"
#' geneRegulationPlot(bds, plot.geneID = exampleGeneId, anno.genes = gns)
#'
#' @export
geneRegulationPlot <- function(object,
                               plot.geneID = NULL,
                               anno.annoDB = NULL,
                               anno.genes = NULL,
                               match.geneID = "gene_id",
                               match.geneName = "gene_name",
                               plot.gene.n.tiles = 100,
                               alpha = 0.05,
                               lfc.cutoff = 2,
                               transcript.regions.outlier.handle = c("first", "second", "both", "remove"),
                               quiet = FALSE) {

    # bind locally used variables
    bs.padj <- bs.log2FoldChange <- sig <- sigDir <- plot.position <- transcriptRegion <-  NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    # type checks
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    # check presence of fold-changes -> must be present
    if (is.null(object@params$calculateBsFoldChange)) {
        msg0 = paste0("Fold-changes were not calculated yet. Run calculateBsFoldChange() first. \n")
        stop(msg0)
    }
    # check presence of gene assignment -> must be present
    if (is.null(getRanges(object)$geneID)) {
        msg0 = paste0("Gene assignment was not calculated. Run assignToGenes() first. \n")
        stop(msg0)
        # TODO  ideally this should test for the param present in object@params$assignToGenes,
        #       but this gets overwritten by the mergeing of two bds objects. Change this in the future.
    }
    # check presence of transcript regions -> can be missing
    if (is.null(getRanges(object)$transcriptRegion)) {
        msg0 = paste0("Transcript region assignment was not calculated. Run assignToTranscriptRegions() first to link binding sites to transcript regions. \n")
        warning(msg0)
        # TODO  ideally this should test for the param present in object@params$assignToTranscriptRegions,
        #       but this gets overwritten by the mergeing of two bds objects. Change this in the future.
        # TODO make a version of the plot that does not encode the transcript regions as shape information
    }

    # check presence of correct annotation resource -> one of both must be present
    # Check if none is specified
    if (is.null(anno.annoDB) & is.null(anno.genes)) {
        msg = paste0("None of the required annotation sources anno.annoDB or anno.genes was specified. ")
        stop(msg)
    }
    # Check if both are specified
    if (!is.null(anno.annoDB) & !is.null(anno.genes)) {
        msg = paste0("Both of the required annotation sources anno.annoDB or anno.genes are specified. Please provide only one of the two. ")
        stop(msg)
    }
    # Checks if anno.annoDB should be used
    if (!is.null(anno.annoDB) & is.null(anno.genes)) {
        stopifnot(is(anno.annoDB, "OrganismDb"))
        if (!is.null(anno.genes)) {
            msg = paste0("Parameter anno.annoDB and anno.genes are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            datasource = "anno.annoDB"
            # extract relevant annotation
            anno.genes = genes(anno.annoDB, columns=c("ENSEMBL", "SYMBOL", "GENEID"))
            # Create matching vectors for columns from input annotation
            # --------------------------------------------------------------------------
            selectID = as.character(anno.genes$GENEID)
            selectName = as.character(anno.genes$SYMBOL)
        }
    }
    # Checks if anno.genes should be used
    if (is.null(anno.annoDB) & !is.null(anno.genes)) {
        stopifnot(is(anno.genes, "GenomicRanges"))
        if (!is.null(anno.annoDB)) {
            msg = paste0("Parameter anno.annoDB and anno.genes are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            datasource = "anno.genes"
            # extract relevant annotation
            # check for duplicated annotation columns
            if (sum(colnames(mcols(anno.genes)) == match.geneID) > 1) {
                msg = paste0("The names of multiple columns of the annotation match the match.geneID parameter. Please use a unique column name for match.geneID in anno.genes.")
                stop(msg)
            }
            if (sum(colnames(mcols(anno.genes)) == match.geneName) > 1) {
                msg = paste0("The names of multiple columns of the annotation match the match.geneName parameter. Please use a unique column name for match.geneName in anno.genes.")
                stop(msg)
            }
            # check correct annotation columns
            inNames = c(match.geneID, match.geneName)
            annoColNames = colnames(mcols(anno.genes))
            presentNames = inNames[inNames %in% annoColNames]

            # extract columns from annotation based on available names
            present.cols = lapply(presentNames, function(x){
                match.col = mcols(anno.genes)[match(x, colnames(mcols(anno.genes)))][[1]]
                return(match.col)
            })
            names(present.cols) = presentNames

            # extract values from annotation for all present columns
            if (match.geneID %in% names(present.cols)) {
                selectID = present.cols[[match(match.geneID, names(present.cols))]]
            } else {
                msg = paste0("No meta column for ", match.geneID, " present. Creating custom ID.\n")
                if (!quiet) message(msg)
                selectID = paste0("CUSTOM", seq_along(anno.genes))
            }
            if (match.geneName %in% names(present.cols)) {
                selectName = present.cols[[match(match.geneName, names(present.cols))]]
            } else {
                msg = paste0("No meta column for ", match.geneName, " present.\n")
                if (!quiet) message(msg)
            }
        }
    }

    # check if gene to plot is in bds object -> must be present
    if (!(plot.geneID %in% getRanges(object)$geneID)) {
        msg0 = stop(paste0("Selected geneID: ", plot.geneID, " is not present in the meta columns of the given bds object. \n"))
        stop(msg0)
    }
    # check if gene to plot is in annotation object -> must be present
    if (!(plot.geneID %in% selectID)) {
        msg0 = stop(paste0("Selected geneID: ", plot.geneID, " is not present in the meta columns of the given gene annotation. \n"))
        stop(msg0)
    }

    transcript.regions.outlier.handle = match.arg(transcript.regions.outlier.handle, choices = c("first", "second", "both", "remove"))

    # --------------------------------------------------------------------------
    # MAIN
    # --------------------------------------------------------------------------

    # get ranges of binding sites for selected gene
    plot.bs = getRanges(object)
    plot.bs = plot.bs[plot.bs$geneID == plot.geneID]
    plot.bs = sort(plot.bs)
    plot.bs = resize(plot.bs, width = 1, fix = "center")

    # get gene ranges for selected gene
    plot.gene = anno.genes[anno.genes$gene_id == plot.geneID]
    plot.gene.name = plot.gene$gene_name

    # claculate relative position of binding sites in gene
    plot.gene.tiles = unlist(tile(plot.gene, n = plot.gene.n.tiles))
    plot.gene.tiles$position = 1:length(plot.gene.tiles)

    ols = findOverlaps(plot.bs, plot.gene.tiles)

    gene.strand = unique(as.character(strand(plot.gene)))
    if(gene.strand == "-") {
        position = plot.gene.n.tiles-(subjectHits(ols))
    } else {
        position = subjectHits(ols)
    }
    plot.bs$plot.position = position

    # --------------------------------------------------------------------------
    # PLOTS
    # --------------------------------------------------------------------------

    # make plots without transcript region annotation
    if (is.null(getRanges(object)$transcriptRegion)) {
        df.plot = mcols(plot.bs) %>% as.data.frame() %>%
            mutate(sig = (ifelse(bs.padj < alpha & abs(bs.log2FoldChange) > log2(lfc.cutoff), TRUE, FALSE))) %>%
            mutate(dir = factor(ifelse(bs.log2FoldChange > 0, "Up", "Down"), levels = c("Up", "Down"))) %>%
            mutate(sigDir = ifelse(sig == TRUE & dir == "Up", "Up", ifelse(sig == TRUE & dir == "Down", "Down", "Not"))) %>%
            mutate(sigDir = factor(sigDir, levels = c("Not", "Up", "Down")))

        p = ggplot(df.plot, aes(x = plot.position, y = bs.log2FoldChange, color = sigDir, fill = sigDir)) +
            geom_hline(yintercept = 0, color = "#4d4d4d") +
            geom_segment(aes(x=plot.position, xend=plot.position, y=0, yend=bs.log2FoldChange), size = 1, color = "#4d4d4d") +
            geom_point(size = 4, stroke = 1, shape = 21) +
            scale_fill_manual(values = c("Not" = "#999999", "Up" = "#68b1a5", "Down" = "#874C62")) +
            scale_color_manual(values = c("Not" = "#4d4d4d", "Up" = "#2b544d", "Down" = "#623747")) +
            theme_minimal() +
            guides(color = guide_legend(override.aes = list(size = 4))) +
            theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
            ylim(-max(abs(df.plot$bs.log2FoldChange)), max(abs(df.plot$bs.log2FoldChange))) +
            xlim(0,plot.gene.n.tiles) +
            labs(
                title = paste0(plot.gene.name, " (", plot.geneID, ")"),
                x = "Relative position on the gene",
                y = "Fold-change (log2)",
                fill = "Regulation",
                color = "Regulation"
            )
        return(p)
    }

    # make plots with transcript region annotation
    if (!is.null(getRanges(object)$transcriptRegion)) {

        expected.regions = c("INTRON", "CDS", "UTR3", "UTR5", "OTHER")
        all.regions.present = unique(plot.bs$transcriptRegion)

        # plot cases where no region duplication is present
        if (all(all.regions.present %in% expected.regions)) {
            df.plot = mcols(plot.bs) %>% as.data.frame() %>%
                mutate(sig = (ifelse(bs.padj < alpha & abs(bs.log2FoldChange) > log2(lfc.cutoff), TRUE, FALSE))) %>%
                mutate(dir = factor(ifelse(bs.log2FoldChange > 0, "Up", "Down"), levels = c("Up", "Down"))) %>%
                mutate(sigDir = ifelse(sig == TRUE & dir == "Up", "Up", ifelse(sig == TRUE & dir == "Down", "Down", "Not"))) %>%
                mutate(sigDir = factor(sigDir, levels = c("Not", "Up", "Down"))) %>%
                mutate(duplicated = "no")

            p = ggplot(df.plot, aes(x = plot.position, y = bs.log2FoldChange, color = sigDir, fill = sigDir, shape = transcriptRegion)) +
                geom_hline(yintercept = 0, color = "#4d4d4d") +
                geom_segment(aes(x=plot.position, xend=plot.position, y=0, yend=bs.log2FoldChange), size = 1, color = "#4d4d4d") +
                geom_point(size = 4, stroke = 1) +
                scale_shape_manual(values=c("INTRON" = 23, "CDS" = 22, "UTR3" = 25, "UTR5" = 24, "OTHER" = 21)) +
                scale_fill_manual(values = c("Not" = "#999999", "Up" = "#68b1a5", "Down" = "#874C62")) +
                scale_color_manual(values = c("Not" = "#4d4d4d", "Up" = "#2b544d", "Down" = "#623747")) +
                theme_minimal() +
                guides(color = guide_legend(override.aes = list(size = 4))) +
                theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
                ylim(-max(abs(df.plot$bs.log2FoldChange)), max(abs(df.plot$bs.log2FoldChange))) +
                xlim(0,plot.gene.n.tiles) +
                labs(
                    title = paste0(plot.gene.name, " (", plot.geneID, ")"),
                    x = "Relative position on the gene",
                    y = "Fold-change (log2)",
                    fill = "Regulation",
                    color = "Regulation",
                    shape = "Transcript Region"
                )
            return(p)
        }

        # plot cases where region duplication is present
        if (any(all.regions.present %in% expected.regions)) {
            msg0 = paste0("At least one additional transcript region type detected compared to the currently supported expected regions: ",
                          paste(expected.regions, collapse = ", "),
                          ". Using transcript.regions.outlier.handle=", transcript.regions.outlier.handle," to solve. \n")
            if (!quiet) {
                warning(msg0)
            }

            if (transcript.regions.outlier.handle == "remove") {
                df.plot = mcols(plot.bs) %>% as.data.frame() %>%
                    filter(transcriptRegion %in% expected.regions) %>%
                    mutate(sig = (ifelse(bs.padj < alpha & abs(bs.log2FoldChange) > log2(lfc.cutoff), TRUE, FALSE))) %>%
                    mutate(dir = factor(ifelse(bs.log2FoldChange > 0, "Up", "Down"), levels = c("Up", "Down"))) %>%
                    mutate(sigDir = ifelse(sig == TRUE & dir == "Up", "Up", ifelse(sig == TRUE & dir == "Down", "Down", "Not"))) %>%
                    mutate(sigDir = factor(sigDir, levels = c("Not", "Up", "Down"))) %>%
                    mutate(duplicated = "no")

                p = ggplot(df.plot, aes(x = plot.position, y = bs.log2FoldChange, color = sigDir, fill = sigDir, shape = transcriptRegion)) +
                    geom_hline(yintercept = 0, color = "#4d4d4d") +
                    geom_segment(aes(x=plot.position, xend=plot.position, y=0, yend=bs.log2FoldChange), size = 1, color = "#4d4d4d") +
                    geom_point(size = 4, stroke = 1) +
                    scale_shape_manual(values=c("INTRON" = 23, "CDS" = 22, "UTR3" = 25, "UTR5" = 24, "OTHER" = 21)) +
                    scale_fill_manual(values = c("Not" = "#999999", "Up" = "#68b1a5", "Down" = "#874C62")) +
                    scale_color_manual(values = c("Not" = "#4d4d4d", "Up" = "#2b544d", "Down" = "#623747")) +
                    theme_minimal() +
                    guides(color = guide_legend(override.aes = list(size = 4))) +
                    theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
                    ylim(-max(abs(df.plot$bs.log2FoldChange)), max(abs(df.plot$bs.log2FoldChange))) +
                    xlim(0,plot.gene.n.tiles) +
                    labs(
                        title = paste0(plot.gene.name, " (", plot.geneID, ")"),
                        x = "Relative position on the gene",
                        y = "Fold-change (log2)",
                        fill = "Regulation",
                        color = "Regulation",
                        shape = "Transcript Region",
                        caption = paste0("Not all binding sites might be shown due to: transcript.regions.outlier.handle=", transcript.regions.outlier.handle)
                    )
                return(p)
            }

            if (transcript.regions.outlier.handle == "first") {
                df.duplicated = mcols(plot.bs) %>% as.data.frame() %>% filter(grepl(",", transcriptRegion)) %>% mutate(duplicated = "yes") %>%
                    mutate(transcriptRegion = sapply(strsplit(transcriptRegion,","), `[`, 1))
                df.single = mcols(plot.bs) %>% as.data.frame() %>% filter(!grepl(",", transcriptRegion)) %>% mutate(duplicated = "no")
                df.plot = rbind.data.frame(df.duplicated, df.single) %>%
                    mutate(sig = (ifelse(bs.padj < alpha & abs(bs.log2FoldChange) > log2(lfc.cutoff), TRUE, FALSE))) %>%
                    mutate(dir = factor(ifelse(bs.log2FoldChange > 0, "Up", "Down"), levels = c("Up", "Down"))) %>%
                    mutate(sigDir = ifelse(sig == TRUE & dir == "Up", "Up", ifelse(sig == TRUE & dir == "Down", "Down", "Not"))) %>%
                    mutate(sigDir = factor(sigDir, levels = c("Not", "Up", "Down"))) %>%
                    mutate(transcriptRegion = stringr::str_trim(transcriptRegion, "left")) %>%
                    mutate(transcriptRegion = stringr::str_trim(transcriptRegion, "right"))

                p = ggplot(df.plot, aes(x = plot.position, y = bs.log2FoldChange, color = sigDir, fill = sigDir, shape = transcriptRegion)) +
                    geom_hline(yintercept = 0, color = "#4d4d4d") +
                    geom_segment(aes(x=plot.position, xend=plot.position, y=0, yend=bs.log2FoldChange), size = 1, color = "#4d4d4d") +
                    geom_point(size = 4, stroke = 1) +
                    scale_shape_manual(values=c("INTRON" = 23, "CDS" = 22, "UTR3" = 25, "UTR5" = 24, "OTHER" = 21)) +
                    scale_fill_manual(values = c("Not" = "#999999", "Up" = "#68b1a5", "Down" = "#874C62")) +
                    scale_color_manual(values = c("Not" = "#4d4d4d", "Up" = "#2b544d", "Down" = "#623747")) +
                    theme_minimal() +
                    guides(color = guide_legend(override.aes = list(size = 4))) +
                    theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
                    ylim(-max(abs(df.plot$bs.log2FoldChange)), max(abs(df.plot$bs.log2FoldChange))) +
                    xlim(0,plot.gene.n.tiles) +
                    labs(
                        title = paste0(plot.gene.name, " (", plot.geneID, ")"),
                        x = "Relative position on the gene",
                        y = "Fold-change (log2)",
                        fill = "Regulation",
                        color = "Regulation",
                        shape = "Transcript Region",
                        caption = paste0("Not all binding sites might be shown due to: transcript.regions.outlier.handle=", transcript.regions.outlier.handle)
                    )
                return(p)
            }

            if (transcript.regions.outlier.handle == "second") {
                df.duplicated = mcols(plot.bs) %>% as.data.frame() %>% filter(grepl(",", transcriptRegion)) %>% mutate(duplicated = "yes") %>%
                    mutate(transcriptRegion = sapply(strsplit(transcriptRegion,","), `[`, 2))
                df.single = mcols(plot.bs) %>% as.data.frame() %>% filter(!grepl(",", transcriptRegion)) %>% mutate(duplicated = "no")
                df.plot = rbind.data.frame(df.duplicated, df.single) %>%
                    mutate(sig = (ifelse(bs.padj < alpha & abs(bs.log2FoldChange) > log2(lfc.cutoff), TRUE, FALSE))) %>%
                    mutate(dir = factor(ifelse(bs.log2FoldChange > 0, "Up", "Down"), levels = c("Up", "Down"))) %>%
                    mutate(sigDir = ifelse(sig == TRUE & dir == "Up", "Up", ifelse(sig == TRUE & dir == "Down", "Down", "Not"))) %>%
                    mutate(sigDir = factor(sigDir, levels = c("Not", "Up", "Down"))) %>%
                    mutate(transcriptRegion = stringr::str_trim(transcriptRegion, "left")) %>%
                    mutate(transcriptRegion = stringr::str_trim(transcriptRegion, "right"))

                p = ggplot(df.plot, aes(x = plot.position, y = bs.log2FoldChange, color = sigDir, fill = sigDir, shape = transcriptRegion)) +
                    geom_hline(yintercept = 0, color = "#4d4d4d") +
                    geom_segment(aes(x=plot.position, xend=plot.position, y=0, yend=bs.log2FoldChange), size = 1, color = "#4d4d4d") +
                    geom_point(size = 4, stroke = 1) +
                    scale_shape_manual(values=c("INTRON" = 23, "CDS" = 22, "UTR3" = 25, "UTR5" = 24, "OTHER" = 21)) +
                    scale_fill_manual(values = c("Not" = "#999999", "Up" = "#68b1a5", "Down" = "#874C62")) +
                    scale_color_manual(values = c("Not" = "#4d4d4d", "Up" = "#2b544d", "Down" = "#623747")) +
                    theme_minimal() +
                    guides(color = guide_legend(override.aes = list(size = 4))) +
                    theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
                    ylim(-max(abs(df.plot$bs.log2FoldChange)), max(abs(df.plot$bs.log2FoldChange))) +
                    xlim(0,plot.gene.n.tiles) +
                    labs(
                        title = paste0(plot.gene.name, " (", plot.geneID, ")"),
                        x = "Relative position on the gene",
                        y = "Fold-change (log2)",
                        fill = "Regulation",
                        color = "Regulation",
                        shape = "Transcript Region",
                        caption = paste0("Not all binding sites might be shown due to: transcript.regions.outlier.handle=", transcript.regions.outlier.handle)
                    )
                return(p)
            }

            if (transcript.regions.outlier.handle == "both") {
                df.duplicated = mcols(plot.bs) %>% as.data.frame() %>% filter(grepl(",", transcriptRegion)) %>% mutate(duplicated = "yes") %>%
                    separate_longer_delim(transcriptRegion, delim = ",")
                df.single = mcols(plot.bs) %>% as.data.frame() %>% filter(!grepl(",", transcriptRegion)) %>% mutate(duplicated = "no")
                df.plot = rbind.data.frame(df.duplicated, df.single) %>%
                    mutate(sig = (ifelse(bs.padj < alpha & abs(bs.log2FoldChange) > log2(lfc.cutoff), TRUE, FALSE))) %>%
                    mutate(dir = factor(ifelse(bs.log2FoldChange > 0, "Up", "Down"), levels = c("Up", "Down"))) %>%
                    mutate(sigDir = ifelse(sig == TRUE & dir == "Up", "Up", ifelse(sig == TRUE & dir == "Down", "Down", "Not"))) %>%
                    mutate(sigDir = factor(sigDir, levels = c("Not", "Up", "Down"))) %>%
                    mutate(transcriptRegion = stringr::str_trim(transcriptRegion, "left")) %>%
                    mutate(transcriptRegion = stringr::str_trim(transcriptRegion, "right"))

                p = ggplot(df.plot, aes(x = plot.position, y = bs.log2FoldChange, color = sigDir, fill = sigDir, shape = transcriptRegion)) +
                    geom_hline(yintercept = 0, color = "#4d4d4d") +
                    geom_segment(aes(x=plot.position, xend=plot.position, y=0, yend=bs.log2FoldChange), size = 1, color = "#4d4d4d") +
                    geom_point(size = 4, stroke = 1) +
                    ggrepel::geom_label_repel(data = df.plot %>% filter(duplicated == "yes"), aes(label = transcriptRegion)) +
                    scale_shape_manual(values=c("INTRON" = 23, "CDS" = 22, "UTR3" = 25, "UTR5" = 24, "OTHER" = 21)) +
                    scale_fill_manual(values = c("Not" = "#999999", "Up" = "#68b1a5", "Down" = "#874C62")) +
                    scale_color_manual(values = c("Not" = "#4d4d4d", "Up" = "#2b544d", "Down" = "#623747")) +
                    theme_minimal() +
                    guides(color = guide_legend(override.aes = list(size = 4))) +
                    theme(legend.key.size = unit(1, 'cm'), legend.position = "top") +
                    ylim(-max(abs(df.plot$bs.log2FoldChange)), max(abs(df.plot$bs.log2FoldChange))) +
                    xlim(0,plot.gene.n.tiles) +
                    labs(
                        title = paste0(plot.gene.name, " (", plot.geneID, ")"),
                        x = "Relative position on the gene",
                        y = "Fold-change (log2)",
                        fill = "Regulation",
                        color = "Regulation",
                        shape = "Transcript Region",
                        size = "Duplicated Region",
                        caption = paste0("Some binding sites might not be visible due to overplotting: transcript.regions.outlier.handle=", transcript.regions.outlier.handle)
                    )
                return(p)
            }

        }
    }
}

#' Function to export sorted RBP target genes
#'
#' Genes with binding sites are target genes of the RBP. They can be exported
#' as 'csv' or 'xlsx' file. Genes can be sorted by the sum of the individual
#' binding sites score, or by the number of binding sites per gene.
#'
#' As output option, one can either output all genes in a single file, or split
#' by either gene-type or transcript-region. This options requires that either
#' \code{\link{BSFind}} or the individual functions \code{\link{assignToGenes}},
#'  and \code{\link{assignToTranscriptRegions}} were run.
#'
#' @param object a \code{BSFDataSet} object with stored ranges
#' @param path A path to where the output should be stored
#' @param format output file format
#' @param sort sorting rule for genes
#' @param split if and how the output file should be split
#'
#' @return a file of the type specified in \code{\link{format}}
#'
#' @importFrom dplyr filter group_by mutate arrange desc ungroup
#' @importFrom utils write.csv
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' \dontrun{
#' # export
#' # exportTargetGenes(bds)
#' }
#'
#' @export
exportTargetGenes <- function(object,
                              path = "./", # where to export
                              format = c("csv", "xslx"), # CSV, Excel
                              sort = c("score", "bs"), # by what to sort (score, number of bs, number of crosslinks),
                              split = c("none", "geneType", "transcriptRegion")
){
    # bind varaibles locally
    geneID <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    # type checks
    stopifnot(is(object, "BSFDataSet"))

    # capture export options
    format = match.arg(format, choices = c("csv", "xslx"))
    sort = match.arg(sort, choices = c("score", "bs"))
    split = match.arg(split, choices = c("none", "geneType", "transcriptRegion"))


    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    expRng = getRanges(object)

    # sort output
    if (sort == "bs") {
        df = getRanges(object) %>% as.data.frame(row.names = NULL) %>% filter(!is.na(geneID)) %>%
            group_by(geneID) %>% mutate(n = n()) %>% arrange(desc(n)) %>% ungroup()
    }
    if (sort == "score") {
        if (! "score" %in% colnames(mcols(getRanges(object)))) {
            stop("Sorting by score not possible, no column found with name 'score'.")
        }
        df = getRanges(object) %>% as.data.frame(row.names = NULL) %>% filter(!is.na(geneID)) %>%
            group_by(geneID) %>% mutate(n = sum(score)) %>% arrange(desc(n)) %>% ungroup()
    }

    # manage format
    if (format == "csv") {
        # manage split
        if (split == "none") {
            write.csv(x = df, file = paste0(path, "results.csv"))
        }
        if (split == "geneType") {
            if (! "geneType" %in% colnames(df)) {
                msg = paste0("Column geneType not found. Make sure to run assignToGenes() or BSFind() to use this option.")
                stop(msg)
            } else {
                sf = split(df, df$geneType)
                for (i in 1:length(sf)) {
                    cName = names(sf[i])
                    write.csv(x = sf[[i]], file = paste0(path = path, cName, ".csv"))
                }
            }
        }
        if (split == "transcriptRegion") {
            if (! "transcriptRegion" %in% colnames(df)) {
                msg = paste0("Column transcriptRegion not found. Make sure to run assignToTranscriptRegion() or BSFind() to use this option.")
                stop(msg)
            } else {
                sf = split(df, df$transcriptRegion)
                for (i in 1:length(sf)) {
                    cName = names(sf[i])
                    write.csv(x = sf[[i]], file = paste0(path = path, cName, ".csv"))
                }
            }
        }
    }

    if (format == "xslx") {
        if (!requireNamespace("xlsx", quietly=TRUE)) {
            stop("format='xlsx' requires installing the package 'xlsx'")
        }
        if (split == "none") {
            xlsx::write.xlsx(x = df, file = paste0(path, "results.xlsx"), sheetName = "All")
        }
        if (split == "geneType") {
            if (! "geneType" %in% colnames(df)) {
                msg = paste0("Column geneType not found. Make sure to run assignToGenes() or BSFind() to use this option.")
                stop(msg)
            }  else {
                sf = split(df, df$geneType)
                for (i in 1:length(sf)) {
                    cName = names(sf[i])
                    xlsx::write.xlsx(x = sf[[i]], file = paste0(path, "results.xlsx"), sheetName = cName, append = TRUE)
                }
            }
        }
        if (split == "transcriptRegion") {
            if (! "transcriptRegion" %in% colnames(df)) {
                msg = paste0("Column transcriptRegion not found. Make sure to run assignToGenes() or BSFind() to use this option.")
                stop(msg)
            }  else {
                sf = split(df, df$transcriptRegion)
                for (i in 1:length(sf)) {
                    cName = names(sf[i])
                    xlsx::write.xlsx(x = sf[[i]], file = paste0(path, "results.xlsx"), sheetName = cName, append = TRUE)
                }
            }
        }

    }

}


#' Wrapper function to export binding sites as BED files
#'
#' Function that serves as a wrapper function for \code{rtracklayer::export}.
#'
#' @param object a \code{\link{BSFDataSet}} object with stored ranges
#' @param con A path or URL
#'
#' @return a .bed file
#'
#' @seealso \code{\link{BSFind}}
#'
#' @import rtracklayer
#' @importFrom utils tail
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' \dontrun{
#' # export
#' # exportToBED(bds, con = "./myfile.bed")
#' }
#'
#' @export
exportToBED <- function(object, con) {
    # initialize locale variables
    funName <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    # range to export
    expRng = getRanges(object)

    # do a basic export if nothing was computed
    if(length(object@params) == 0){
        mcols(expRng)$name = paste0("IDX", ":", seq_along(expRng))
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackline = new("BasicTrackLine", name = "CLS_unprocessed", description = "Unprocessed", useScore = FALSE))
    }

    # check for the highest level function that was computed and export accordingly
    lastFun = tail(object@results, 1) %>% pull(funName)
    lastFun = substr(lastFun, start = 1, stop = nchar(lastFun)-2)
    optionStr = tail(object@results, 1) %>% pull(options)

    allFunctions = c("pureClipGlobalFilter", "estimateBsWidth",
                     "pureClipGeneWiseFilter", "makeBindingSites",
                     "reproducibilityFilter", "assignToGenes",
                     "assignToTranscriptRegions", "annotateWithScore",
                     "calculateSignalToFlankScore")

    if(lastFun == allFunctions[1]) {
        mcols(expRng)$name = paste0("IDX", ":", seq_along(expRng))
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "CLS_pureClipGlobalfilter",
                                            description = paste0("pureClipGlobalFilter; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[2]) {
        mcols(expRng)$name = paste0("IDX", ":", seq_along(expRng))
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "CLS_estimateBsWidth",
                                            description = paste0("estimateBsWidth; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[3]) {
        mcols(expRng)$name = paste0("IDX", ":", seq_along(expRng), "_", expRng$geneID)
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "CLS_pureClipGeneWiseFilter",
                                            description = paste0("pureClipGeneWiseFilter; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[4]){
        mcols(expRng)$name = paste0(expRng$bsID)
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "BS_makeBindingsites",
                                            description = paste0("makeBindingSites; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[5]){
        mcols(expRng)$name = paste0(expRng$bsID)
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "BS_reproducibilityFilter",
                                            description = paste0("reproducibilityFilter; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[6]){
        mcols(expRng)$name = paste0(expRng$bsID, "_", expRng$geneID, "_", expRng$geneType)
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "BS_assignToGenes",
                                            description = paste0("assignToGenes; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[7]){
        mcols(expRng)$name = paste0(expRng$bsID, "_", expRng$geneID, "_", expRng$geneType, "_", expRng$transcriptRegion)
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "BS_assignToTranscriptRegions",
                                            description = paste0("assignToTranscriptRegions; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[8]){
        mcols(expRng)$name = paste0("BsID", ":", expRng$bsID, "_", expRng$geneID, "_", expRng$geneType, "_", expRng$transcriptRegion, "_score:", expRng$score)
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "BS_annotateWithScore",
                                            description = paste0("annotateWithScore; options: ", optionStr), useScore = FALSE))
    }
    if(lastFun == allFunctions[9]){
        mcols(expRng)$name = paste0("BsID", ":", expRng$bsID, "_", expRng$geneID, "_", expRng$geneType, "_", expRng$transcriptRegion, "_score:", expRng$score, "_flankScore:", expRng$signalToFlankRatio)
        mcols(expRng)$score = NULL
        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "BS_calculateSignalToFlankScore",
                                            description = paste0("annotateWithScore; options: ", optionStr), useScore = FALSE))
    }
    if(!lastFun %in% allFunctions) {
        warning("Function names not matching. Please check your data or export manually with `rtracklayer::export(getRanges(myObj))`.\n")

        rtracklayer::export(object = expRng, con = con, format = "BED",
                            trackLine = new("BasicTrackLine", name = "BS", useScore = FALSE))
    }

}


#' Create a table of all workflow steps for reporting
#'
#' Function that creates a printable table with all steps and numbers for each
#' of the workflow steps that were carried out.
#'
#' If \code{\link{option}} is set to `reduced`, only the most necessary information
#' are collected. Option `full` contains a full list of all options and parameters
#' that were set in any of the workflow functions. Option `extended` contains
#' extra information about the binding site merging step.
#'
#' @param object a \code{\link{BSFDataSet}} object with stored ranges
#' @param option character; how detailed the table should be
#'
#' @return a \code{kableExtra} table
#'
#' @seealso \code{\link{BSFind}}
#'
#' @importFrom kableExtra kable
#'
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # apply 5% filter
#' bds = pureClipGlobalFilter(object = bds, cutoff = 0.05)
#' processingStepsTable(bds)
#'
#' @export
processingStepsTable <- function(object, # bindingSiteFinder
                                 option = c("reduced", "full", "extended") # option to which level of detail the results should be shown
) {

    nIn <- nOut <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    # handle options
    option = match.arg(option, choices = c("reduced", "full", "extended"))
    res = object@results

    res = format(res, big.mark = ",", decimal.mark = ".")
    colnames(res) = c("Step", "Type", "#N In", "#N Out", "%", "Options")
    cap = paste0("Processing step overview (", option, " view).")

    if (option == "full") {
        k = kableExtra::kable(res, caption = cap)
    }
    if (option == "reduced") {
        k = kableExtra::kable(res[c(1,3,4,5)], caption = cap)
    }
    if (option == "extended") {
        s = getSummary(object)
        ss = data.frame(step = s[[1]][2:6], nIn = s[[2]][1:5], nOut = s[[2]][2:6]) %>%
            mutate(per = paste0(round(nOut/ nIn, digits = 2)*100, "%")) %>%
            mutate(options = c(paste0("bsSize=",object@params$makeBindingSites$bsSize,
                                      ", minWidth=", object@params$makeBindingSites$minWidth,
                                      ", sub.chr=", object@params$makeBindingSites$sub.chr),
                               paste0("minCrosslinks=", object@params$makeBindingSites$minCrosslinks),
                               paste0("minClSites=", object@params$makeBindingSites$minClSites),
                               paste0("centerIsClSite=", object@params$makeBindingSites$centerIsClSite),
                               paste0("centerIsSummit=", object@params$makeBindingSites$centerIsSummit)))
        resDetail = data.frame(Step = paste0("makeBindingSites()-",ss$step),
                               Type = "binding sites", nIn = ss$nIn, nOut = ss$nOut, per = ss$per,
                               Options = ss$options)
        colnames(resDetail) = c("Step", "Type", "#N In", "#N Out", "%", "Options")

        resNewPre = res[c(1:(which(res$Step == "makeBindingSites()"))),]
        resNewPost = res[c((which(res$Step == "makeBindingSites()")+1):nrow(res)),]
        resNew = rbind.data.frame(resNewPre, resDetail, resNewPost)

        k = kableExtra::kable(resNew, caption = cap)
    }

    return(k)
}


#' Collapse signal from replicates
#'
#' Collapses all replicates merges all samples from a \link{BSFDataSet} object
#' into a single signal stream, only split by minus and plus strand.
#'
#' @param object a \code{BSFDataSet} object
#' @param collapseAll TRUE/FALSE, if all samples should be collapsed (TRUE), or
#' if samples should be kept separate by condition (FALSE)
#' @return object of type \code{\link{BSFDataSet}} with updated signal
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' bdsNew = collapseReplicates(bds)
#'
#' @export
collapseReplicates <- function(object, collapseAll = FALSE) {
    # stop if object is not of class BSF
    stopifnot(is(object, "BSFDataSet"))
    # stop if collapseAll parameter is not logical
    if (!is.logical(collapseAll)) {
        stop("Option collapseAll not logical, set to TRUE/FALSE")
    }

    sgn = getSignal(object)
    mta = getMeta(object)

    # collapse per condition in meta table
    if (!isTRUE(collapseAll)) {
        # get conditions to split by
        cond = levels(mta$condition)
        # handle plus strand
        plus = sgn$signalPlus
        idx = lapply(cond, function(currCond){
            grep(currCond, names(plus))
        })
        plusSum = lapply(seq_along(idx), function(x){
            currSampleIdx = idx[[x]]
            p = sgn$signalPlus[currSampleIdx]
            pSum = 0
            for (i in seq_along(p)) {
                pSum = pSum + p[[i]]
            }
            names(pSum) = names(p[[1]])
            return(pSum)
        })
        names(plusSum) = cond
        # handle minus strand
        minus = sgn$signalMinus
        idx = lapply(cond, function(currCond){
            grep(currCond, names(minus))
        })
        minusSum = lapply(seq_along(idx), function(x){
            currSampleIdx = idx[[x]]
            m = sgn$signalMinus[currSampleIdx]
            mSum = 0
            for (i in seq_along(m)) {
                mSum = mSum + m[[i]]
            }
            names(mSum) = names(m[[1]])
            return(mSum)
        })
        names(minusSum) = cond
        mrgSgn = list(signalPlus = (plusSum), signalMinus = (minusSum))
        newObj = setSignal(object, mrgSgn)
    }

    # collapse all replicates regardless of the conditions
    if (isTRUE(collapseAll)) {

        p = sgn$signalPlus
        m = sgn$signalMinus

        pSum = 0
        for (i in seq_along(p)) {
            pSum = pSum + p[[i]]
        }
        names(pSum) = names(p[[1]])
        mSum = 0
        for (i in seq_along(m)) {
            mSum = mSum + m[[i]]
        }
        names(mSum) = names(m[[1]])

        # set condition name
        lp = list(pSum)
        names(lp) = "All"
        lm = list(mSum)
        names(lm) = "All"

        mrgSgn = list(signalPlus = (lp), signalMinus = (lm))
        newObj = setSignal(object, mrgSgn)
    }

    return(newObj)
}

test_that("Collapse samples", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # Set artificial KD condition
    metaCond = getMeta(bds)
    metaCond$condition = factor(c(rep("WT", 2), rep("KD", 2)), levels = c("WT", "KD"))
    bdsCond = setMeta(bds, metaCond)
    # Fix replicate names in signal
    namesCond = c("1_WT", "2_WT", "3_KD", "4_KD")
    sgn = getSignal(bdsCond)
    names(sgn$signalPlus) = namesCond
    names(sgn$signalMinus) = namesCond
    bdsCond = setSignal(bdsCond, sgn)
    bds = bdsCond

    # test for collapseAll = default
    bdsNew = collapseReplicates(bds)
    expect_identical(getRanges(bds), getRanges(bdsNew))
    expect_identical(getMeta(bds), getMeta(bdsNew))

    cov1 = coverageOverRanges(bds, returnOptions = "merge_positions_keep_replicates")
    cov2 = coverageOverRanges(bdsNew, returnOptions = "merge_positions_keep_replicates")

    cov1Counts = mcols(cov1) %>% as.data.frame() %>%
        dplyr::mutate(WT = rowSums(dplyr::across(contains("WT")))) %>%
        dplyr::mutate(KD = rowSums(dplyr::across(contains("KD")))) %>%
        dplyr::select(WT, KD) %>%
        dplyr::relocate(WT, KD)
    cov2Counts = mcols(cov2) %>% as.data.frame()
    expect_identical(cov1Counts, cov2Counts)

    # test for collapseAll = FALSE
    bdsNew = collapseReplicates(bds, collapseAll = FALSE)
    expect_identical(getRanges(bds), getRanges(bdsNew))
    expect_identical(getMeta(bds), getMeta(bdsNew))

    cov1 = coverageOverRanges(bds, returnOptions = "merge_positions_keep_replicates")
    cov2 = coverageOverRanges(bdsNew, returnOptions = "merge_positions_keep_replicates")

    cov1Counts = mcols(cov1) %>% as.data.frame() %>%
        dplyr::mutate(WT = rowSums(dplyr::across(contains("WT")))) %>%
        dplyr::mutate(KD = rowSums(dplyr::across(contains("KD")))) %>%
        dplyr::select(WT, KD) %>%
        dplyr::relocate(WT, KD)
    cov2Counts = mcols(cov2) %>% as.data.frame()
    expect_identical(cov1Counts, cov2Counts)

    # test for collapseAll = TRUE
    bdsNew = collapseReplicates(bds, collapseAll = TRUE)
    expect_identical(getRanges(bds), getRanges(bdsNew))
    expect_identical(getMeta(bds), getMeta(bdsNew))

    cov1 = coverageOverRanges(bds, returnOptions = "merge_positions_keep_replicates")
    cov2 = coverageOverRanges(bdsNew, returnOptions = "merge_positions_keep_replicates")

    cov1Counts = mcols(cov1) %>% as.data.frame() %>%
        dplyr::mutate(All = rowSums(dplyr::across(everything()))) %>%
        dplyr::select(All)
    cov2Counts = mcols(cov2) %>% as.data.frame()
    expect_identical(cov1Counts, cov2Counts)

})

#' Plot the PureCLIP score distribution with global cutoff indicator
#'
#' A diagnostic function that plots the PureCLIP score distribution on a log2
#' scale. The function \code{\link{pureClipGlobalFilter}} is expected to be
#' executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{pureClipGlobalFilter}}
#'
#' @import ggplot2
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # apply 5% filter
#' bds = pureClipGlobalFilter(object = bds, cutoff = 0.05)
#' pureClipGlobalFilterPlot(bds)
#'
#' @export
pureClipGlobalFilterPlot <- function(object) {

    # bind locally used variables
    x <- y <- quant <- name <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    if (is.null(object@params$pureClipGlobalFilter)) {
        msg0 = paste0("Global filter was not applied yet. Run BSFind() or pureClipGlobalFilter() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$pureClipGlobalFilter$data) |
        is.null(object@plotData$pureClipGlobalFilter$cutoffPoint)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run pureClipGlobalFilter() or BSFind() before plotting. \n")
        stop(msg1)
    }

    optstr = object@params$pureClipGlobalFilter
    optstrNice = paste0("Cutoff=", optstr$cutoff, ", scoreMatchCol=", optstr$match.score)

    # get stored plotting data
    dt = data.frame(s = log2(object@plotData$pureClipGlobalFilter$data))
    cutpoint = log2(object@plotData$pureClipGlobalFilter$cutoffPoint)

    # calc density
    df = density(dt$s)
    df = data.frame(x = df$x, y = df$y)

    # set quantiles to show
    probs = seq(from = 0, to = 1, by = 0.1)
    quants = quantile(dt$s, prob = probs)
    df$quant = factor(findInterval(df$x, quants))

    # set cutoff to show
    cutoffNice = paste0(round(optstr$cutoff, 2) * 100, "%")
    dfCutoff = data.frame(pos = cutpoint, name = paste0(round(optstr$cutoff, 2) * 100, "%"))

    # Setup color pal
    colScaleFun = colorRampPalette(RColorBrewer::brewer.pal(9, "Blues"))

    p = ggplot(df, aes(x = x, y = y)) +
        geom_line() +
        geom_ribbon(aes(ymin = 0, ymax = y, fill = quant)) +
        scale_x_continuous(breaks = quants) +
        scale_fill_manual(values = colScaleFun(length(probs)+1)) +
        theme_bw() +
        theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1, size = 6), legend.position = "none") +
        geom_vline(xintercept = cutpoint, linetype = "dashed", size = 1, color = "#b35900") +
        labs(
            title = "pureClipGlobalFilterPlot()",
            subtitle = optstrNice,
            x = "Score [log2]", y = "Density"
        ) +
        geom_label(data = dfCutoff, aes(x = pos, y = Inf, label = name), vjust = 1.2, color = "#b35900")
    return(p)
}


#' Plot the number of overlaps when assigning crosslink sites to genes
#'
#' A diagnostic function that plots the number of crosslink sites with their
#' respective overlapping rate. The function \code{\link{pureClipGeneWiseFilter}}
#' is expected to be executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{pureClipGeneWiseFilter}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # Load GRanges with genes
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' # apply 5% gene-wise filter
#' bds = pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.5,
#'  overlaps = "keepSingle")
#' duplicatedSitesPlot(bds)
#'
#' @export
duplicatedSitesPlot <- function(object) {

    # bind locally used variables
    `#N overlaps` <- Freq <- FreqNice <- nRes <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    if (is.null(object@params$geneWiseFilter)) {
        msg0 = paste0("Gene-wise filter was not applied yet. Run BSFind() or pureClipGeneWiseFilter() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$geneWiseFilter$data) ) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run pureClipGeneWiseFilter() or BSFind() before plotting. \n")
        stop(msg1)
    }

    optstr = object@params$geneWiseFilter
    optstrNice = paste0("Cutoff=", optstr$cutoff, ", overlaps=", optstr$overlaps)

    df = object@plotData$geneWiseFilter$data
    df = df %>% mutate(FreqNice = format(Freq, big.mark = ",", small.mark = "."))

    # Setup color pal
    cols = c("#576F72", rep("lightgrey", nrow(df)-1))

    # Make Plot
    # --------------------------------------------------------------------------
    # Standard plot
    p = ggplot(df, aes(x = `#N overlaps`, y = Freq, fill = `#N overlaps`, label = FreqNice)) +
        geom_col(color = "black") +
        scale_y_log10() +
        scale_fill_manual(values = cols) +
        theme_bw() +
        theme(legend.position = "none") +
        labs(
            title = "dupliatedSitesPlot()",
            subtitle = optstrNice,
            x = "#N overlaps",
            y = "#N (log10)"
        )

    # Add custom options based on overlappingLociOption
    if (optstr$overlaps == "keepSingle") {
        dfResolved = df %>% mutate(nRes = as.numeric(Freq) / as.numeric(`#N overlaps`))
        p = p +
            geom_col(data = dfResolved, aes(x = `#N overlaps`, y = nRes, fill = `#N overlaps`), fill = "#576F72", color = "black") +
            geom_text(data = df, aes(x = `#N overlaps`, y = Freq, label = FreqNice),
                      color = "#b35900", vjust = 1, hjust = 0.8, fontface = "bold", angle = 45)
    }
    if (optstr$overlaps == "removeAll") {
        dfResolved = df %>% mutate(nRes = ifelse(df$`#N overlaps` == 1, TRUE, FALSE))
        p = p +
            geom_col(data = subset(dfResolved, dfResolved$nRes == FALSE),
                     aes(x = `#N overlaps`, y = Freq, fill = `#N overlaps`),
                     fill = "#ECB390", color = "black") +
            geom_text(data = df, aes(x = `#N overlaps`, y = Freq, label = FreqNice),
                      color = "#b35900", vjust = 1, hjust = 0.5, fontface = "bold", angle = 45)
    }
    if (optstr$overlaps == "keepAll") {
        p = p +
            geom_col(data = df, aes(x = `#N overlaps`, y = Freq, fill = `#N overlaps`),
                     fill = "#576F72", color = "black") +
            geom_text(data = df, aes(x = `#N overlaps`, y = Freq, label = FreqNice),
                      color = "#b35900", vjust = 1, hjust = 0.5, fontface = "bold", angle = 45)
    }

    return(p)
}


#' Plot binding site merging diagnostics
#'
#' A diagnostic function that plots the number of regions to merge over the width
#' of these regions for each merging iteration calculated in
#' \code{\link{makeBindingSites}}. The function \code{\link{makeBindingSites}}
#' is expected to be executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{makeBindingSites}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' mergeCrosslinkDiagnosticsPlot(bds)
#'
#' @export
mergeCrosslinkDiagnosticsPlot <- function(object) {

    # bind locally used variables
    w <- s <- iteration <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$makeBindingSites)) {
        msg0 = paste0("Reproducibility filter was not applied yet. Run BSFind() or makeBindingSites() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$makeBindingSites$mergeCsData) |
        is.null(object@plotData$makeBindingSites$data)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run makeBindingSites() or BSFind() before plotting. \n")
        stop(msg1)
    }

    optstr = object@params$makeBindingSites
    optstrNice = paste0("bsSize=", optstr$bsSize, ", minWidth=", optstr$minWidth, ", minCrosslinks=", optstr$minCrosslinks)

    dfPlot = object@plotData$makeBindingSites$mergeCsData
    dfPlot = dfPlot %>% mutate(iteration = as.factor(iteration), w = as.factor(w))

    # Make Plot
    # --------------------------------------------------------------------------
    # Standard plot
    cols = c("#716F81","#B97A95","#F6AE99","#F2E1C1","#BEAEE2","#F7DBF0","#CDF0EA","#F9F9F9",rep("grey",5))

    p = ggplot(dfPlot, aes(x = w, y = s, color = iteration, group = iteration)) +
        geom_point(size = 3) +
        geom_line(size = 1) +
        scale_color_manual(values = cols) +
        theme_bw() +
        labs(
            title = "mergeCrosslinkDiagnosticsPlot()",
            subtitle = optstrNice,
            x = "Region-to-fit width",
            y = "Region-to-fit #N",
            color = "Fitting iteration"
        ) +
        theme(legend.position = "bottom") +
        guides(colour = guide_legend(nrow = 1))

    return(p)
}

#' Plot binding site filter diagnostics
#'
#' A diagnostic function that plots the number of binding sites retained
#' after each filtering step calculated in \code{\link{makeBindingSites}}.
#' The function \code{\link{makeBindingSites}} is expected to be executed prior
#' to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{makeBindingSites}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' makeBsSummaryPlot(bds)
#'
#' @export
makeBsSummaryPlot <- function(object) {

    # bind locally used variables
    Option <- nRanges <- slice <- nNice <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$makeBindingSites)) {
        msg0 = paste0("Reproducibility filter was not applied yet. Run BSFind() or makeBindingSites() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$makeBindingSites$mergeCsData) |
        is.null(object@plotData$makeBindingSites$data)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run makeBindingSites() or BSFind() before plotting. \n")
        stop(msg1)
    }

    optstr = object@params$makeBindingSites
    optstrNice = paste0("bsSize=", optstr$bsSize, ", minWidth=", optstr$minWidth, ", minCrosslinks=", optstr$minCrosslinks)

    dfPlot = object@plotData$makeBindingSites$data
    dfPlot = dfPlot %>%
        mutate(Option = factor(Option, levels = c(Option))) %>%
        slice(-1) %>%
        mutate(nNice = paste0(format(nRanges, big.mark = ",", decimal.mark = ".")))

    # Make Plot
    # --------------------------------------------------------------------------
    # Standard plot
    cols = c("#262A56","#8D7B68","#A4907C","#C8B6A6","#F1DEC9")

    p = ggplot(dfPlot, aes(x = Option, y = nRanges, fill = Option)) +
        geom_col() +
        scale_fill_manual(values = cols) +
        theme_bw() +
        theme(legend.position = "none") +
        labs(
            title = "makeBsSummaryPlot()",
            subtitle = optstrNice,
            x = "Options",
            y = "Ranges #N"
        ) +
        geom_text(data = dfPlot, aes(x = Option, y = nRanges, label = nNice),
                  color = "#b35900", vjust = 1, hjust = 0.8, fontface = "bold", angle = 45)

    return(p)
}


#' Plot to that shows the crosslink site distribution per replicate
#'
#' A diagnostic function that plots the number of crosslinks sites over the number
#' of crosslink in these sites and highlights the replicate specific
#' reproducibility cutoff that is derived from that distribution.
#' The function \code{\link{reproducibilityFilter}} is expected to be executed prior
#' to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param plotRange numeric; number of crosslinks per sites that should be shown
#' before summing them up
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{reproducibilityFilter}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = reproducibilityFilter(bds)
#' reproducibilityFilterPlot(bds)
#'
#' @export
reproducibilityFilterPlot <- function(object, plotRange = 20) {

    # bind locally used variables
    value <- n <- condtion <- name <- condition <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$reproducibilityFilter)) {
        msg0 = paste0("Reproducibility filter was not applied yet. Run BSFind() or reproducibilityFilter() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$reproducibilityFilterPlot$data) |
        is.null(object@plotData$reproducibilityFilterPlot$cutoffs) |
        is.null(object@plotData$reproducibilitySamplesPlot$data)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run reproducibilityFilter() or BSFind() before plotting. \n")
        stop(msg1)
    }

    optstr = object@params$reproducibilityFilter
    optstrNice = paste0("Cutoff=", optstr$cutoff, ", nReps=", optstr$nReps, ", minCrosslinks=", optstr$minCrosslinks)

    dfPlot = object@plotData$reproducibilityFilterPlot$data
    dfPlotCutoffs = object@plotData$reproducibilityFilterPlot$cutoffs

    # Apply plot range zoom in
    dfPlot$value[dfPlot$value > plotRange] = plotRange

    # Make Plot
    # --------------------------------------------------------------------------
    # Standard plot
    p = ggplot(dfPlot, aes(x = value, y = n)) +
        geom_col(color = "#2d5986", aes(fill = condition)) +
        facet_wrap(~name, scales = "free_x") +
        scale_fill_brewer(palette = "Set2") +
        theme_classic() +
        labs(
            title = "reproducibilityFilterPlot()",
            subtitle = optstrNice,
            x = "Crosslinks per site",
            y = "Number of sites",
            fill = "Condition"
        ) +
        geom_vline(data = dfPlotCutoffs, aes(xintercept = value, group = name), color = "#b35900", size = 1, linetype = "dashed") +
        geom_label(data = dfPlotCutoffs, aes(x = value, y = max(dfPlot$n), label = value), alpha = 0.5)
        # geom_label(data = dfPlotCutoffs, aes(x = value, y = 1, label = value), nudge_y = 2)

    return(p)
}


#' UpSet-plot to that shows how each replicate supports binding sites
#'
#' A diagnostic function that plots the set sizes for each replicate, indicating
#' how many binding site the specific replicate supports given its specific
#' threshold. The function \code{\link{reproducibilityFilter}} is expected to be
#' executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param nIntersections numeric; number of intersection to be shown
#' @param show.title logical; if plot title should be visible
#' @param text.size numeric; fontsize of all numbers on axis
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{reproducibilityFilter}},
#' \code{\link{reproducibilityFilterPlot}}
#'
#' @import ComplexHeatmap
#' @importFrom dplyr desc
#' @importFrom utils head tail
#' @importFrom grid gpar
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = reproducibilityFilter(bds)
#' reproducibilitySamplesPlot(bds)
#'
#' @export
reproducibilitySamplesPlot <- function(object,
                                       nIntersections = 20,
                                       show.title = TRUE,
                                       text.size = NULL) {

    # bind locally used variables
    title <- rowTitle <- Freq <- `.` <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$reproducibilityFilter)) {
        msg0 = paste0("Reproducibility filter was not applied yet. Run BSFind() or reproducibilityFilter() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$reproducibilitySamplesPlot$data)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run reproducibilityFilter() or BSFind() before plotting. \n")
        stop(msg1)
    }

    optstr = object@params$reproducibilityFilter
    optstrNice = paste0("Cutoff=", paste(optstr$cutoff, collapse = ","),
                        ", nReps=", paste(optstr$nReps, collapse = ","),
                        ", minCrosslinks=", paste(optstr$minCrosslinks, collapse = ","))

    # Get full combination matrix
    df = object@plotData$reproducibilitySamplesPlot$data
    # colnames(df) = colnames(df) %>% as.data.frame() %>% separate(col = ".", sep = "_", into = c(NA, "Sample")) %>% pull(Sample)
    m = make_comb_mat(df)

    rowTitle = "All intersections"

    # Reduce combination matrix for plotting
    if (length(comb_size(m)) > nIntersections) {
        plottingCutoff = comb_size(m) %>%
            as.data.frame() %>% rename("Freq" = ".") %>% arrange(desc(Freq)) %>%
            head(.,nIntersections) %>% tail(1) %>% pull(Freq)
        m = m[comb_size(m) >= plottingCutoff]
        rowTitle = paste0(nIntersections, " largest intersections")
    }

    title = paste0("reproducibilitySamplesPlot()\n", optstrNice)

    # set fontsizes
    if (!is.null(text.size)) {
        fontsize.top = text.size
        fontsize.right = text.size
        fontsize.row.title = text.size
        fontsize.column.title = text.size
        fontsize.row.names = text.size
        fontsize.column.names = text.size
    } else {
        fontsize.top = 8
        fontsize.right = 8
        fontsize.row.title = 12
        fontsize.column.title = 12
        fontsize.row.names = 10
        fontsize.column.names = 10
    }

    UpSet(m, column_title = ifelse(isTRUE(show.title), title, ""),
          row_title = ifelse(isTRUE(show.title), rowTitle, ""),
          comb_order = order(comb_size(m), decreasing = TRUE),
          top_annotation = upset_top_annotation(m, add_numbers = TRUE, annotation_name_gp = gpar(fontsize = 0),
                                                numbers_gp = gpar(col = "#4C6793", fontsize = fontsize.top, fontface = "italic")),
          right_annotation = upset_right_annotation(m, add_numbers = TRUE, annotation_name_gp = gpar(fontsize = 0),
                                                    gp = gpar(fill = "#0B2447"),
                                                    numbers_gp = gpar(col = "#0B2447", fontsize = fontsize.right, fontface = "italic"),
                                                    numbers_rot = 25),
          comb_col = "#4C6793", bg_col = "white", pt_size = unit(.5, "cm"),
          border = T, lwd = 2, bg_pt_col = "grey",
          row_title_gp = gpar(fontsize = fontsize.row.title),
          column_title_gp = gpar(fontsize = fontsize.column.title),
          row_names_gp = gpar(fontsize = fontsize.row.names),
          column_names_gp = gpar(fontsize = fontsize.column.names),
          set_order = seq_along(set_size(m)))

}



#' UpSet-plot to that shows the gene type overlaps
#'
#' A diagnostic function that plots the gene types of binding sites on overlapping
#' loci genes. The function \code{\link{assignToGenes}} is expected to be
#' executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param show.title logical; if plot title should be visible
#' @param text.size numeric; fontsize of all numbers on axis
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{assignToGenes}} \code{\link{targetGeneSpectrumPlot}}
#'
#' @import ComplexHeatmap
#' @importFrom grid gpar
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # Load GRanges with genes
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = assignToGenes(bds, anno.genes = gns)
#' geneOverlapsPlot(bds)
#'
#' @export
geneOverlapsPlot <- function(object,
                             text.size = NULL,
                             show.title = TRUE) {

    # bind locally used variables
    title <- rowTitle <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$assignToGenes)) {
        msg0 = paste0("Assignment to anno.genes was not applied yet. Run BSFind() or assignToGenes() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$assignToGenes$dataOverlaps) |
        is.null(object@plotData$assignToGenes$dataSpectrum) ) {
        msg1 = paste0("It seems like you used options `remove` or `keep` in assignToGenes(). The plot you try to make is only present for options `frequency` and `hierarchy`. \n")
        stop(msg1)
    }

    optstr = object@params$assignToGenes
    optstrNice = paste0("Source=", optstr$source, ", Overlaps=", optstr$overlaps,
                        ifelse(!rlang::is_empty(optstr$rule), paste0(", rule=", paste(optstr$rule, collapse = ">")), ""))

    # Get full combination matrix
    df = object@plotData$assignToGenes$dataOverlaps
    m = make_comb_mat(df)

    rowTitle = "All intersections"
    title = paste0("geneOverlapsPlot()\n", optstrNice)


    # set fontsizes
    if (!is.null(text.size)) {
        fontsize.top = text.size
        fontsize.right = text.size
        fontsize.row.title = text.size
        fontsize.column.title = text.size
        fontsize.row.names = text.size
        fontsize.column.names = text.size
    } else {
        fontsize.top = 8
        fontsize.right = 8
        fontsize.row.title = 12
        fontsize.column.title = 12
        fontsize.row.names = 10
        fontsize.column.names = 10
    }

    UpSet(m, column_title = ifelse(isTRUE(show.title), title, ""),
          row_title = ifelse(isTRUE(show.title), rowTitle, ""),
          comb_order = order(comb_size(m), decreasing = TRUE),
          top_annotation = upset_top_annotation(m, add_numbers = TRUE, annotation_name_gp = gpar(fontsize = 0),
                                                numbers_gp = gpar(col = "#4C6793", fontsize = fontsize.top, fontface = "italic")),
          right_annotation = upset_right_annotation(m, add_numbers = TRUE, gp = gpar(fill = "#0B2447"), annotation_name_gp = gpar(fontsize = 0),
                                                    numbers_gp = gpar(col = "#0B2447", fontsize = fontsize.right, fontface = "italic"),
                                                    numbers_rot = 25),
          comb_col = "#4C6793", bg_col = "white", pt_size = unit(.5, "cm"),
          row_title_gp = gpar(fontsize = fontsize.row.title),
          column_title_gp = gpar(fontsize = fontsize.column.title),
          row_names_gp = gpar(fontsize = fontsize.row.names),
          column_names_gp = gpar(fontsize = fontsize.column.names),
          border = T, lwd = 2, bg_pt_col = "grey",
          set_order = seq_along(set_size(m)))
}

#' Bar-chart to show the hosting gene types of binding sites
#'
#' A diagnostic function that plots the gene type of the hosting gene for
#' each binding site. The function \code{\link{assignToGenes}} is expected to be
#' executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param showNGroups numeric; the number of different gene types to show
#' @param text.size numeric; the size of the text elments on the plot
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{assignToGenes}} \code{\link{geneOverlapsPlot}}
#'
#' @import ggplot2
#' @importFrom dplyr lead desc mutate arrange
#' @importFrom utils head
#' @importFrom tibble add_row
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = assignToGenes(bds, anno.genes = gns)
#' targetGeneSpectrumPlot(bds)
#'
#' @export
targetGeneSpectrumPlot <- function(object,
                                   showNGroups = 5,
                                   text.size = 4) {

    # bind locally used variables
    GeneType <- Freq <- FreqNice <- nGenes <- nBs <- geneType <- nice <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$assignToGenes)) {
        msg0 = paste0("Assignment to genes was not applied yet. Run BSFind() or assigneToGenes() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$assignToGenes$dataOverlaps) |
        is.null(object@plotData$assignToGenes$dataSpectrum) ) {
        msg1 = paste0("It seems like you used options `remove` or `keep` in assignToGenes(). The plot you try to make is only present for options `frequency` and `hierarchy`. \n")
        stop(msg1)
    }

    optstr = object@params$assignToGenes
    optstrNice = paste0("Source=", optstr$source, ", Overlaps=", optstr$overlaps)

    df = object@plotData$assignToGenes$dataSpectrum

    # ----
    # Apply topN cutoff
    if (nrow(df) < showNGroups) {
        msg = paste0("Only ", nrow(df), " different gene types, displaying ", nrow(df), " gene types, not ", showNGroups, ". \n")
        showNGroups = nrow(df)
        message(msg)
    }
    if (nrow(df) > showNGroups) {
        otherFreq = df %>% arrange(desc(nGenes), desc(nBs)) %>% lead(showNGroups-1)
        dfOther = data.frame(geneType = "Other",
                             nBs = sum(otherFreq$nBs, na.rm = TRUE),
                             nGenes = sum(otherFreq$nGenes, na.rm = TRUE))
        df = df %>% as.data.frame() %>% arrange(desc(nGenes)) %>% head(showNGroups-1)
        df = rbind.data.frame(df, dfOther)
        df = df %>% mutate(nice = paste0(
            "#G:",
            format(nGenes, big.mark = ",", small.mark = "."),
            " (#BS:",
            format(nBs, big.mark = ",", small.mark = "."),
            ")")) %>%
            mutate(geneType = forcats::fct_reorder(geneType, nGenes))
    }
    if (nrow(df) == showNGroups) {
        df = df %>% arrange(desc(nGenes), desc(nBs)) %>%
            mutate(nice = paste0(
                "#G:",
                format(nGenes, big.mark = ",", small.mark = "."),
                " (#BS:",
                format(nBs, big.mark = ",", small.mark = "."),
                ")")) %>%
            mutate(geneType = forcats::fct_reorder(geneType, nGenes))

    }
    # ----
    # Setup color pal
    cols = RColorBrewer::brewer.pal(n = showNGroups+2, name = "Greys")
    cols = cols[1:showNGroups+1]

    # Make Plot
    # --------------------------------------------------------------------------
    p = ggplot(df, aes(x = geneType, y = nGenes, fill = geneType, label = nice)) +
        geom_col(color = "black") +
        scale_y_log10() +
        coord_flip() +
        scale_fill_manual(values = cols) +
        theme_bw() +
        theme(legend.position = "none") +
        geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
        labs(
            title = "targetGeneSpectrumPlot()",
            subtitle = optstrNice,
            x = "",
            y = "#N (log10)"
        )

    return(p)
}



#' UpSet-plot to that shows the transcript region overlaps
#'
#' A diagnostic function that plots the transcript regions of binding sites
#' on overlapping loci. The function \code{\link{assignToTranscriptRegions}} is expected to be
#' executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param show.title logical; if plot title should be visible
#' @param text.size numeric; fontsize of all numbers on axis
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{assignToTranscriptRegions}} \code{\link{transcriptRegionSpectrumPlot}}
#'
#' @import ComplexHeatmap
#' @importFrom grid gpar
#' @importFrom dplyr desc
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)
#' transcriptRegionOverlapsPlot(bds)
#'
#' @export
transcriptRegionOverlapsPlot <- function(object,
                                         text.size = NULL,
                                         show.title = TRUE) {

    # bind locally used variables
    title <- rowTitle <- Freq <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$assignToTranscriptRegions)) {
        msg0 = paste0("Reproducibility filter was not applied yet. Run BSFind() or assignToTranscriptRegions() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$assignToTranscriptRegions$dataOverlaps) |
        is.null(object@plotData$assignToTranscriptRegions$dataSpectrum)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run assignToTranscriptRegions() BSFind() before plotting. \n")
        stop(msg1)
    }
    optstr = object@params$assignToTranscriptRegions
    optstrNice = paste0("Source=", optstr$source, ", overlaps=", optstr$overlaps,
                        ifelse(!rlang::is_empty(optstr$rule), paste0(", rule=", paste(optstr$rule, collapse = ">")), ""))

    # Get full combination matrix
    df = object@plotData$assignToTranscriptRegions$dataOverlaps
    m = make_comb_mat(df, mode = "distinct")

    rowTitle = "All intersections"
    title = paste0("transcriptRegionOverlapsPlot()\n", optstrNice)

    # set fontsizes
    if (!is.null(text.size)) {
        fontsize.top = text.size
        fontsize.right = text.size
        fontsize.row.title = text.size
        fontsize.column.title = text.size
        fontsize.row.names = text.size
        fontsize.column.names = text.size
    } else {
        fontsize.top = 8
        fontsize.right = 8
        fontsize.row.title = 12
        fontsize.column.title = 12
        fontsize.row.names = 10
        fontsize.column.names = 10
    }

    UpSet(m, column_title = ifelse(isTRUE(show.title), title, ""),
          row_title = ifelse(isTRUE(show.title), rowTitle, ""),
          comb_order = order(comb_size(m), decreasing = TRUE),
          top_annotation = upset_top_annotation(m, add_numbers = TRUE, annotation_name_gp = gpar(fontsize = 0),,
                                                numbers_gp = gpar(col = "#4C6793", fontsize = fontsize.top, fontface = "italic")),
          right_annotation = upset_right_annotation(m, add_numbers = TRUE, gp = gpar(fill = "#0B2447"), annotation_name_gp = gpar(fontsize = 0),,
                                                    numbers_gp = gpar(col = "#0B2447", fontsize = fontsize.right, fontface = "italic"),
                                                    numbers_rot = 25),
          comb_col = "#4C6793", bg_col = "white", pt_size = unit(.5, "cm"),
          row_title_gp = gpar(fontsize = fontsize.row.title),
          column_title_gp = gpar(fontsize = fontsize.column.title),
          row_names_gp = gpar(fontsize = fontsize.row.names),
          column_names_gp = gpar(fontsize = fontsize.column.names),
          border = T, lwd = 2, bg_pt_col = "grey",
          set_order = seq_along(set_size(m)))
}


#' Bar-chart to show the hosting transcript regions of binding sites
#'
#' A diagnostic function that plots the transcript regions of the hosting gene
#' for each binding site. The function \code{\link{assignToTranscriptRegions}}
#' is expected to be executed prior to calling this plot function.
#'
#' Count frequencies can be normalized to the length of the hosting region with
#' option \code{normalize}. The specific factor how the hosting region length is
#' used is given by \code{normalize.factor}. In the case of
#' \code{normalize.factor = "sum"} binding site frequencies are divided by the
#' summed length of all regions that host the specific binding site.
#'
#' Further with option \code{values} once can indicate whether raw or normalized
#' frequencies should be shown 'as-is' or normalized to 'percentages'.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param values character; if values should be presented 'as-is', that means
#' for example as frequencies in case \code{normalization = FALSE}, or as
#' percentages
#' @param normalize logical; whether to normalize values
#' @param normalize.factor character; indicate by what factor values should be
#' normalized to region length by
#' @param show.others logical; whether to show 'others' category. Has to be false
#' if \code{normalize = TRUE}
#' @param text.size numeric; font size of the numbers to be displayed on each bar
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{assignToTranscriptRegions}} \code{\link{transcriptRegionOverlapsPlot}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)
#' transcriptRegionSpectrumPlot(bds)
#'
#' @export
transcriptRegionSpectrumPlot <- function(object,
                                         values = c("asis", "percentage"),
                                         normalize = FALSE,
                                         normalize.factor = c("sum", "median", "mean"),
                                         show.others = FALSE,
                                         text.size = 4) {

    # bind locally used variables
    TranscriptRegion <- FreqNice <- Freq <-  NULL
    per <- perNice <- value <- name <- norm.value <- norm.per <- freq.per <- . <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$assignToTranscriptRegions)) {
        msg0 = paste0("Transcript region assignment was not computed yet. Run BSFind() or assignToTranscriptRegions() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$assignToTranscriptRegions$dataOverlaps) |
        is.null(object@plotData$assignToTranscriptRegions$dataSpectrum)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run assignToTranscriptRegions() or BSFind() before plotting. \n")
        stop(msg1)
    }

    # handle options
    values = match.arg(values, choices = c("asis", "percentage"))
    normalize.factor = match.arg(normalize.factor, choices = c("sum", "median", "mean"))

    # make option string
    optstr = object@params$assignToTranscriptRegions
    optstrNice = paste0("values=", values, ", normalize=", normalize, ifelse(isTRUE(normalize), paste0(", normalize.factor=", normalize.factor), ""),
                        ", overlaps=", optstr$overlaps,
                        ifelse(! rlang::is_empty(optstr$rule), paste0(", rule=", paste(optstr$rule, collapse = ">")), ""))

    # Get full combination matrix
    df = object@plotData$assignToTranscriptRegions$dataSpectrum

    # Nice formatting of standard names
    df = df %>% mutate(TranscriptRegion = ifelse(toupper(TranscriptRegion) == "UTR3", "3'UTR",
                                                 ifelse(toupper(TranscriptRegion) == "UTR5", "5'UTR",
                                                        ifelse(toupper(TranscriptRegion) == "CDS", "CDS",
                                                               ifelse(toupper(TranscriptRegion) == "INTRON", "Intron", TranscriptRegion)))))

    df = df %>% as.data.frame() %>% arrange(desc(Freq)) %>%
        mutate(FreqNice = format(Freq, big.mark = ",", small.mark = ".")) %>%
        mutate(TranscriptRegion = forcats::fct_reorder(TranscriptRegion, Freq))

    # set color scheme
    cols = c("#A7D2CB", "#F2D388", "#C98474", "#874C62", "#576F72", "#B4CDE6", "#7D6E83", "#D0B8A8")
    cols = cols[1:nrow(df)]

    # Make plot variations
    # --------------------------------------------------------------------------
    if (!isTRUE(normalize)) {
        # don't normalize

        if (!isTRUE(show.others)) {
            # remove other category to match raw plot with normalized plot
            df = df %>% drop_na() %>%
                mutate(TranscriptRegion = forcats::fct_reorder(TranscriptRegion, Freq))
        }

        if (values == "asis") {
            # show raw numbers
            p = ggplot(df, aes(x = TranscriptRegion, y = Freq, fill = TranscriptRegion, label = FreqNice)) +
                geom_col(color = "darkgrey", linewidth = 0.5) +
                theme_bw() +
                theme(legend.position = "none") +
                geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                scale_fill_manual(values = cols) +
                coord_flip(clip = "on", expand = TRUE) +
                labs(title = "transcriptRegionSpectrumPlot()",
                     subtitle = optstrNice,
                     x = "Transcript regions",
                     y = "#N Binding sites")
        }
        if (values == "percentage") {
            # show percentages
            df = df %>%
                mutate(per = round(Freq / sum(Freq) * 100, digits = 1)) %>%
                mutate(perNice = paste0(per, "%"))

            p = ggplot(df, aes(x = TranscriptRegion, y = per, fill = TranscriptRegion, label = perNice)) +
                geom_col(color = "darkgrey", linewidth = 0.5) +
                theme_bw() +
                theme(legend.position = "none") +
                geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                scale_fill_manual(values = cols) +
                coord_flip(clip = "on", expand = TRUE) +
                labs(title = "transcriptRegionSpectrumPlot()",
                     subtitle = optstrNice,
                     x = "Transcript regions",
                     y = "% Binding sites")

        }

    }

    if (isTRUE(normalize)) {
        # normalize counts
        df = df %>%
            drop_na() %>%
            pivot_longer(-c(TranscriptRegion, Freq, FreqNice)) %>%
            mutate(norm.value = Freq / value) %>%
            group_by(name) %>%
            mutate(norm.per = round(norm.value / sum(norm.value) * 100, digits = 1),
                   freq.per = round(Freq / sum(Freq) * 100, digits = 1)) %>%
            mutate(diff.per = round(norm.per - freq.per, digits = 1) ) %>%
            mutate(FreqNice = format(Freq, big.mark = ","))

        if (values == "asis") {
            # show raw numbers
            if (normalize.factor == "mean") {
                # normalize by mean length
                p = df %>% filter(name == "norm.hosting.mean") %>%
                    ggplot(., aes(x = TranscriptRegion, y = norm.value, fill = TranscriptRegion, label = round(norm.value, digits = 2))) +
                    geom_col(color = "darkgrey", linewidth = 0.5) +
                    theme_bw() +
                    theme(legend.position = "none") +
                    geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                    scale_fill_manual(values = cols) +
                    coord_flip(clip = "on", expand = TRUE) +
                    labs(title = "transcriptRegionSpectrumPlot()",
                         subtitle = optstrNice,
                         x = "Transcript regions",
                         y = "% Binding sites")
            }
            if (normalize.factor == "median") {
                # normalize by mean length
                p = df %>% filter(name == "norm.hosting.median") %>%
                    ggplot(., aes(x = TranscriptRegion, y = norm.value, fill = TranscriptRegion, label = round(norm.value, digits = 2))) +
                    geom_col(color = "darkgrey", linewidth = 0.5) +
                    theme_bw() +
                    theme(legend.position = "none") +
                    geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                    scale_fill_manual(values = cols) +
                    coord_flip(clip = "on", expand = TRUE) +
                    labs(title = "transcriptRegionSpectrumPlot()",
                         subtitle = optstrNice,
                         x = "Transcript regions",
                         y = "% Binding sites")
            }
            if (normalize.factor == "sum") {
                # normalize by mean length
                p = df %>% filter(name == "norm.hosting.sum") %>%
                    ggplot(., aes(x = TranscriptRegion, y = norm.value, fill = TranscriptRegion, label = round(norm.value, digits = 2))) +
                    geom_col(color = "darkgrey", linewidth = 0.5) +
                    theme_bw() +
                    theme(legend.position = "none") +
                    geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                    scale_fill_manual(values = cols) +
                    coord_flip(clip = "on", expand = TRUE) +
                    labs(title = "transcriptRegionSpectrumPlot()",
                         subtitle = optstrNice,
                         x = "Transcript regions",
                         y = "% Binding sites")
            }

        }
        if (values == "percentage") {
            # show percentages
            if (normalize.factor == "mean") {
                # normalize by mean length
                p = df %>% filter(name == "norm.hosting.mean") %>%
                    ggplot(., aes(x = TranscriptRegion, y = norm.per, fill = TranscriptRegion, label = paste0(round(norm.per, digits = 1), "%"))) +
                    geom_col(color = "darkgrey", linewidth = 0.5) +
                    theme_bw() +
                    theme(legend.position = "none") +
                    geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                    scale_fill_manual(values = cols) +
                    coord_flip(clip = "on", expand = TRUE) +
                    labs(title = "transcriptRegionSpectrumPlot()",
                         subtitle = optstrNice,
                         x = "Transcript regions",
                         y = "% Binding sites")
            }
            if (normalize.factor == "median") {
                # normalize by mean length
                p = df %>% filter(name == "norm.hosting.median") %>%
                    ggplot(., aes(x = TranscriptRegion, y = norm.per, fill = TranscriptRegion, label = paste0(round(norm.per, digits = 1), "%"))) +
                    geom_col(color = "darkgrey", linewidth = 0.5) +
                    theme_bw() +
                    theme(legend.position = "none") +
                    geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                    scale_fill_manual(values = cols) +
                    coord_flip(clip = "on", expand = TRUE) +
                    labs(title = "transcriptRegionSpectrumPlot()",
                         subtitle = optstrNice,
                         x = "Transcript regions",
                         y = "% Binding sites")
            }
            if (normalize.factor == "sum") {
                # normalize by mean length
                p = df %>% filter(name == "norm.hosting.sum") %>%
                    ggplot(., aes(x = TranscriptRegion, y = norm.per, fill = TranscriptRegion, label = paste0(round(norm.per, digits = 1), "%"))) +
                    geom_col(color = "darkgrey", linewidth = 0.5) +
                    theme_bw() +
                    theme(legend.position = "none") +
                    geom_text(aes(y = Inf), color = "#b35900", hjust = 1, fontface = "bold", size = text.size) +
                    scale_fill_manual(values = cols) +
                    coord_flip(clip = "on", expand = TRUE) +
                    labs(title = "transcriptRegionSpectrumPlot()",
                         subtitle = optstrNice,
                         x = "Transcript regions",
                         y = "% Binding sites")
            }
        }

    }

    return(p)
}


#' Plot the PureCLIP score distribution after re-assignment
#'
#' A diagnostic function that plots the PureCLIP score distribution on a log2
#' scale after the re-assignment on binding site level.
#' The function \code{\link{annotateWithScore}} is expected to be executed
#' prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{annotateWithScore}}
#'
#' @import ggplot2
#' @importFrom stats density
#' @importFrom grDevices colorRampPalette
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' bds1 = makeBindingSites(object = bds, bsSize = 9)
#' bds1 = annotateWithScore(bds1, match.ranges = getRanges(bds))
#' globalScorePlot(bds1)
#'
#' @export
globalScorePlot <- function(object) {

    # bind locally used variables
    x <- y <- quant <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    if (is.null(object@params$annotateWithScore)) {
        msg0 = paste0("Score transfer was not applied yet. Run BSFind() or annotateWithScore() to compute values first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$annotateWithScore$data) ) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run annotateWithScore() or BSFind() before plotting. \n")
        stop(msg1)
    }

    optstr = object@params$annotateWithScore
    optstrNice = paste0("match.score=", optstr$match.score, ", match.option=", optstr$match.option)

    # get stored plotting data
    dt = data.frame(s = log2(object@plotData$annotateWithScore$data))

    # calc density
    df = density(dt$s)
    df = data.frame(x = df$x, y = df$y)

    # set quantiles to show
    probs = seq(from = 0, to = 1, by = 0.1)
    quants = quantile(dt$s, prob = probs)
    df$quant = factor(findInterval(df$x, quants))

    # Setup color pal
    colScaleFun = colorRampPalette(RColorBrewer::brewer.pal(9, "Blues"))

    p = ggplot(df, aes(x = x, y = y)) +
        geom_line() +
        geom_ribbon(aes(ymin = 0, ymax = y, fill = quant)) +
        scale_x_continuous(breaks = quants) +
        scale_fill_manual(values = colScaleFun(length(probs)+1)) +
        theme_bw() +
        theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1), legend.position = "none") +
        labs(
            title = "globalScorePlot()",
            subtitle = optstrNice,
            x = "Score [log2]", y = "Density"
        )
    return(p)
}


#' Plot the signal-to-flank score for varying gene-wise filter and binding
#' site width
#'
#' A diagnostic function that plots the the signal-to-flank score as a mean
#' for each binding site width and gene-wise filter as indicated when executing
#' \code{\link{estimateBsWidth}}. Additionally a mean of means visualizes the
#' overall trend and a red line indicates the suggested optimal binding site
#' width. The function \code{\link{estimateBsWidth}} is expected to be
#' executed prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{estimateBsWidth}}
#'
#' @import ggplot2
#' @importFrom stats sd
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' bds = estimateBsWidth(bds, anno.genes = gns, est.maxBsWidth = 19,
#'  geneResolution = "coarse", bsResolution = "coarse", est.subsetChromosome = "chr22")
#' estimateBsWidthPlot(bds)
#' @export
estimateBsWidthPlot <- function(object) {

    # bind locally used variables
    bsSize <- ms <- geneWiseFilter <- signalToFlankRatio <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$estimateBsWidth) | is.null(object@params$bsSize) |
        is.null(object@params$geneFilter)) {
        msg0 = paste0("Binding site width was not estimated. Run estimateBsWidth() or BSFind() first. \n")
        stop(msg0)
    }
    if (is.null(object@plotData$estimateBsWidth$data)) {
        msg1 = paste0("It seems like someting went wrong with your data. Please check the input and make sure to run estimateBsWidth() or BSFind() before plotting. \n")
    }

    optstr = object@params$estimateBsWidth
    optstrNice = paste0("bsResolution=", optstr$bsResolution,
                        ", geneResolution=", optstr$geneResolution,
                        ",\noption=", optstr$option,
                        ", subsetChromosome=", ifelse(!is.null(optstr$est.subsetChromosome),
                                                      ifelse(length(optstr$est.subsetChromosome) == 1,
                                                             optstr$est.subsetChromosome, paste(optstr$est.subsetChromosome, collapse = ",")
                                                             ), "None"
                                                      )
                        )

    est.bsSize = object@params$bsSize
    est.GeneFilter = object@params$geneFilter
    df = object@plotData$estimateBsWidth$data

    dfPlot = df %>%
        mutate(bsSize = as.factor(bsSize)) %>%
        mutate(geneWiseFilter = paste0((geneWiseFilter)*100, "%")) %>%
        mutate(geneWiseFilter = factor(geneWiseFilter, levels = c(paste0(seq(0,95, by = 5), "%"))))

    dfMean = dfPlot %>%
        group_by(bsSize) %>%
        summarise(ms = mean(signalToFlankRatio), sd = sd(signalToFlankRatio)) %>%
        mutate(geneWiseFilter = "mean")


    dfAnno = dfPlot %>%
        mutate(bsSize = as.numeric(as.character(bsSize))) %>%
        group_by(geneWiseFilter) %>%
        filter(bsSize == max(bsSize)) %>%
        mutate(bsSize = as.factor(bsSize))

    colfunc = colorRampPalette(c("#dceaef", "#4C6793"))
    cols = colfunc(n = length(unique(dfPlot$geneWiseFilter)))

    p = ggplot(dfMean, aes(x = bsSize, y = ms, group = geneWiseFilter)) +
        geom_line(data = dfPlot, aes(x = bsSize, y = signalToFlankRatio, group = geneWiseFilter, color = geneWiseFilter), linewidth = 0.7) +
        geom_point(data = dfPlot, aes(x = bsSize, y = signalToFlankRatio, group = geneWiseFilter, color = geneWiseFilter), size = 2) +
        scale_colour_manual(values = cols) +
        geom_errorbar(aes(ymin = ms-sd, ymax = ms+sd), color = "#874C62", width = 0.2, linewidth = 1.1) +
        geom_line(color = "#874C62", linewidth = 1.1) +
        geom_point(size = 4, shape = 21, color = "#874C62", fill = "#874C62", stroke = 1) +
        theme_bw() +
        theme(legend.position = "bottom") +
        guides(color=guide_legend(nrow=2,byrow=TRUE)) +
        labs(
            title = "estimateBsWidthPlot()",
            subtitle = optstrNice,
            y = "Signal-to-flank score",
            x = "Binding site width",
            color = "Gene-wise filter"
        ) +
        geom_vline(xintercept = dfMean$bsSize[dfMean$bsSize == est.bsSize], color = "#b35900", linetype = "dashed", size = 1) +
        geom_label(data = dfMean[dfMean$bsSize == est.bsSize,], aes(x = Inf, y = Inf),
                   label = paste0("Size=", est.bsSize, "/ filter=", est.GeneFilter), vjust = 1, size = 3, hjust = 1, alpha = 0.3)
    return(p)
}


#' Step-wise flowchart plot
#'
#' An overview plot that shows all workflow functions that were executed on the
#' current object, with all input and output binding site numbers and major options
#' that were used. The function can be called at any time in the analysis. Most
#' optimal usage is after a full run of the wrapper function \code{\link{BSFind}}.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param size.all numeric; size of all numbers
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{BSFind}}
#'
#' @import ggplot2
#' @importFrom dplyr n lead slice left_join
#' @importFrom utils tail
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#' bds = BSFind(bds, anno.genes = gns, anno.transcriptRegionList = regions,
#'  est.subsetChromosome = "chr22")
#' processingStepsFlowChart(bds)
#'
#' @export
processingStepsFlowChart <- function(object, size.all = 3) {

    # bind locally used variables
    xmin <- ymin <- xmax <- ymax <- Type <- NULL
    y <- x <- Step <- from <- to <- s_e <- Options <- `#N In` <- `#N Out` <- NULL
    y1 <- y2 <- id <- x1 <- x2 <- opt <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    # get results
    res = object@results
    res = format(res, big.mark = ",", decimal.mark = ".")
    colnames(res) = c("Step", "Type", "#N In", "#N Out", "%", "Options")

    flowNode = res %>%
        mutate(x = 0, y = n():1) %>%
        mutate(y = y *2) %>%
        mutate(xmin = x - 0.45,
               xmax = x + 0.45,
               ymin = y - 0.65,
               ymax = y + 0.65) %>%
        mutate(Type = as.factor(Type)) %>%
        as.data.frame()

    df = res %>%
        mutate(from = Step, to = lead(Step)) %>%
        slice(-n())

    flowEdges = df %>% dplyr::select(from, to) %>%
        mutate(id = row_number()) %>%
        pivot_longer(cols = c("from", "to"),
                     names_to = "s_e",
                     values_to = "Step") %>%
        left_join(flowNode, by = "Step") %>%
        dplyr::select(-c(Step, Type, y, xmin, xmax)) %>%
        mutate(y = ifelse(s_e == "from", ymin, ymax)) %>%
        dplyr::select(-c(ymin, ymax, Options))

    flowArrowIn = flowNode %>%
        mutate(x2 = xmin -0.01, x1 = xmin -0.25, y2 = ymax -0.45, y1 = ymax +0.4) %>%
        mutate(`#N In` = paste0("#N=", `#N In`),`#N Out` = paste0("#N=", `#N Out`))

    flowArrowOut = flowNode %>%
        mutate(x1 = xmin -0.01, x2 = xmin -0.25, y1 = ymax -0.7, y2 = ymax -1.5) %>%
        mutate(`#N In` = paste0("#N=", `#N In`),`#N Out` = paste0("#N=", `#N Out`))

    flowOutLast = flowArrowIn %>% tail(1) %>%
        mutate(y1 = y1-2, y2 = y2-2)

    suppressWarnings({
        flowOption = flowNode %>%
            separate(Options, sep = ",", into = c("opt")) %>% # throws a warning, that additional columns are dropped
            mutate(x2 = xmax+0.25, x1 = xmax +0.25, y2 = ymax -1, y1 = ymax -0.65)

        flowOptionArrow = flowNode %>%
            separate(Options, sep = ",", into = c("opt")) %>% # throws a warning, that additional columns are dropped
            mutate(x2 = xmax+0.01, x1 = xmax +0.25, y2 = ymax -0.85, y1 = ymax -0.65)
    })

    if (!is.null(object@params$estimateBsWidth) |
        !is.null(object@params$calculateSignalToFlankScore)) {
        cols = c("#F2D7D9", "#D3CEDF", "#ECE5C7", "#9CB4CC")
    } else {
        cols = c("#F2D7D9", "#D3CEDF", "#9CB4CC")
    }

    p = ggplot() +
        geom_rect(data = flowNode, aes(xmin = xmin, ymin = ymin, xmax = xmax, ymax = ymax, fill = Type), color = "#4d4d4d") +
        geom_text(data = flowNode, aes(x = x, y = y, label = Step), color = "#585c45", size = size.all) +
        geom_path(data = flowEdges,
                  mapping = aes(x = x, y = y, group = id),
                  colour = "#585c45",
                  arrow = arrow(length = unit(0.3, "cm"), type = "closed")) +
        geom_curve(data = flowArrowIn, aes(x = x1, y = y1, xend = x2, yend = y2),
                   arrow = arrow(length = unit(0.2, "cm"), type = "closed"), size = 0.4,
                   color = "gray20", curvature = -0.2) +
        geom_curve(data = flowArrowOut, aes(x = x1, y = y1, xend = x2, yend = y2),
                   arrow = arrow(length = unit(0.2, "cm"), type = "closed"), size = 0.4,
                   color = "gray20", curvature = -0.2) +
        geom_label(data = flowArrowIn, aes(x = x1, y = y1, label = `#N In`), size = size.all, hjust = 1, color = "#4d4d4d", fill = "#EDEDED") +
        geom_label(data = flowOutLast, aes(x = x1, y = y1, label = `#N Out`), size = size.all, hjust = 1, color = "#4d4d4d", fill = "#EDEDED") +
        xlim(-1.3,1.3) +
        scale_fill_manual(values = cols) +
        theme_void() +
        labs(title = "BindingSiteFinder Flowchart",
             subtitle = "Use the function processingStepsTable(), with option=extended to get a table with the full overview of all options that \nwere used. ",
             fill = "") +
        theme(legend.position = "bottom",
              plot.title = element_text(hjust = 0.5, color = "#748DA6", size = 20, face = "bold"),
              plot.subtitle = element_text(color = "#748DA6", size = 8, face = "italic")) +
        geom_label(data = flowOption, aes(x = x1, y = y1, label = opt), size = size.all, hjust = 0, color = "#4d4d4d", fill = "#EDEDED", fontface = "italic") +
        geom_curve(data = flowOptionArrow, aes(x = x1, y = y1, xend = x2, yend = y2),
                   arrow = arrow(length = unit(0.2, "cm"), type = "closed"), size = 0.4,
                   color = "gray20", curvature = -0.2)

    return(p)
}


#' Binding site definedness plot
#'
#' Binding site definedness is given by the percent of crosslinks that fall
#' diretly inside the binding site compare to those around the binding site.
#' This plotting function shows the distribution of those percentage values
#' grouped by what is indicated in the \code{by} argument.
#'
#' If \code{by} = 'all', then all binding site are grouped into one distribution.
#' For options 'transcript_region' and 'gene_type' binding sites are split into
#' groups according to the respective assignment. This requires that the respective
#' assignment function was executed on the dataset prior to calling this plot function.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param by character; the option by which the plot should be grouped by.
#' Options are: "all", "transcript_region", "gene_type"
#' @param showN.genes numeric; if \code{by} is `gene_type`, then this argument
#' set the maximum number of groups to be shown in the plot
#' @param show.others logical; whether to show 'others' category.
#'
#' @return a plot of type \code{\link{ggplot}}
#'
#' @seealso \code{\link{BSFind}}, \code{\link{calculateSignalToFlankScore}}
#'
#' @import ggplot2
#' @importFrom ggdist stat_halfeye
#' @importFrom dplyr tally
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#' bds = BSFind(bds, anno.genes = gns, anno.transcriptRegionList = regions,
#'  est.subsetChromosome = "chr22")
#' bds = calculateSignalToFlankScore(bds)
#' bindingSiteDefinednessPlot(bds)
#'
#' @export
bindingSiteDefinednessPlot <- function(
        object,
        by = c("all", "transcript_region", "gene_type"),
        showN.genes = 5,
        show.others = FALSE
) {
    # init local vairables
    transcriptRegion <- signalToFlankRatio <- geneType <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    if (is.null(object@params$calculateSignalToFlankScore)) {
        msg0 = paste0("Signal-to-flank ratio was not calculated yet. Run BSFind() or CalculateSignalToFlankScore() to compute values first. \n")
        stop(msg0)
    }

    # handle by options
    by = match.arg(by, choices = c("all", "transcript_region", "gene_type"))


    optstr = object@params$calculateSignalToFlankScore
    optstrNice = paste0("Flank=", optstr$flank)

    # get main plotting dataframe
    df = as.data.frame(mcols(getRanges(object)), row.names = NULL)

    # filter if needed
    if (!isTRUE(show.others)) {
        # remove other category to match raw plot with normalized plot
        df = subset(df, transcriptRegion != "OTHER")
    }

    # Nice formatting of standard names
    df = df %>% mutate(transcriptRegion = ifelse(toupper(transcriptRegion) == "UTR3", "3'UTR",
                                                 ifelse(toupper(transcriptRegion) == "UTR5", "5'UTR",
                                                        ifelse(toupper(transcriptRegion) == "CDS", "CDS",
                                                               ifelse(toupper(transcriptRegion) == "INTRON", "Intron", transcriptRegion)))))

    # set plotting order based on frequency
    df.order = df %>% group_by(transcriptRegion) %>% tally() %>% arrange(desc(n))
    df$transcriptRegion = factor(df$transcriptRegion, levels = rev(df.order$transcriptRegion), ordered = TRUE)

    # check for what plotting options are available
    if (by == "transcript_region") {
        # check if option is really available
        if (is.null(object@params$assignToTranscriptRegions)) {
            msg0 = paste0("Transcript region assignment was not computed yet. Run BSFind() or assignToTranscriptRegions() to compute values first. \n")
            stop(msg0)
        } else {
            # set color scheme
            cols = c("#A7D2CB", "#F2D388", "#C98474", "#874C62", "#576F72", "#B4CDE6", "#7D6E83", "#D0B8A8")
            cols = cols[1:nrow(df)]
            # make plot
            p = ggplot(df, aes(x = transcriptRegion, y = signalToFlankRatio)) +
                geom_boxplot(aes(color = transcriptRegion), width = .15, outlier.shape = NA) +
                ggdist::stat_halfeye(aes(fill = transcriptRegion), adjust = 1.5, width = .6, .width = 0, justification = -.2, point_colour = NA) +
                geom_boxplot(aes(color = transcriptRegion), width = .15, outlier.shape = NA) +
                scale_fill_manual(values = cols) +
                scale_color_manual(values = cols) +
                theme_bw() +
                theme(legend.position = "none") +
                labs(
                    title = "bindingSiteDefinednessPlot()",
                    subtitle = optstrNice,
                    x = "Transcript region",
                    y = "Percent bound"
                )  +
                ylim(0,1) +
                geom_hline(yintercept = mean(df$signalToFlankRatio), linetype = "dashed", color = "#696969") +
                coord_flip()
        }
    }
    if (by == "gene_type") {
        # check if option is really available
        if (is.null(object@params$assignToGenes)) {
            msg0 = paste0("Assignment to genes was not applied yet. Run BSFind() or assigneToGenes() to compute values first. \n")
            stop(msg0)
        } else {
            selGroup = df %>%
                group_by(geneType) %>%
                summarise(n = n()) %>%
                arrange(desc(n)) %>%
                slice_head(n = showN.genes) %>%
                pull(geneType)

            df = df %>% filter(geneType %in% selGroup) %>%
                mutate(geneType  = factor(geneType, levels = rev(selGroup)))

            #  set color scheme
            colfunc = colorRampPalette(c("#cccccc", "#1a1a1a"))
            cols = colfunc(n = showN.genes)

            # make plot
            p = ggplot(df, aes(x = geneType, y = signalToFlankRatio)) +
                ggdist::stat_halfeye(aes(fill = geneType), adjust = 1.5,
                                     width = .6, .width = 0, justification = -.2,
                                     point_colour = NA, slab_color = "#202020",
                                     slab_linewidth = 0.5) +
                geom_boxplot(aes(fill = geneType), width = .15, outlier.shape = NA, color = "#202020") +
                scale_fill_manual(values = cols) +
                scale_color_manual(values = cols) +
                theme_bw() +
                theme(legend.position = "none") +
                labs(
                    title = "bindingSiteDefinednessPlot()",
                    subtitle = optstrNice,
                    x = "Gene type",
                    y = "Percent bound"
                )  +
                ylim(0,1) +
                geom_hline(yintercept = mean(df$signalToFlankRatio), linetype = "dashed", color = "#696969") +
                coord_flip()

        }
    }
    if (by == "all") {
        # make plot
        p = ggplot(df, aes(x = "Total", y = signalToFlankRatio)) +
            ggdist::stat_halfeye(color = "#696969", fill = "#D3D3D3", adjust = 1.5, width = .6, .width = 0, justification = -.2, point_colour = NA) +
            geom_boxplot(color = "#696969", fill = "#D3D3D3", width = .15, outlier.shape = NA) +
            theme_bw() +
            theme(legend.position = "none") +
            labs(
                title = "bindingSiteDefinednessPlot()",
                subtitle = optstrNice,
                x = "All",
                y = "Percent bound"
            ) +
            ylim(0,1) +
            coord_flip()
    }
    return(p)
}



#' Plot crosslink event coverage over binding site range
#'
#' A diagnostic plot function that allows to check the coverage of crosslink
#' events over different merged regions. The coverage is shown as mean over all
#' replicates and conditions, with a standard deviation corridor.
#'
#' If \code{object} is a single BSFDataObject a single coverage plot will be
#' drawn, whereas if it is a list of BSFDataObjects, then faceting is used to
#' make a plot for each list element.
#'
#' @param object a BSFDataSet, or a list of BSFDataSet
#' @param width numeric; set the plotting range to show (in nt)
#' @param show.samples logical; to show individual samples as lines
#' @param subset.chromosome character; subset by a all ranges on the indicated
#' chromosome. Can also be a vector with multiple chromosomes. If NULL then all
#' ranges are being used.
#' @param quiet logical; whether to print messages
#'
#' @return a plot of type \code{ggplot2} displaying the crosslink coverage over
#' the ranges of the given \code{\link{BSFDataSet}}
#'
#' @seealso \code{\link{BSFDataSet}}, \code{\link{makeBindingSites}}
#'
#' @import ggplot2
#' @importFrom matrixStats colSds
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' # plotting a single object
#' bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1)
#' rangeCoveragePlot(bds, subset.chromosome = "chr22")
#'
#' # plotting multiple objects
#' bds1 <- makeBindingSites(object = bds, bsSize = 3, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")
#' bds2 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")
#' l = list(`1. bsSize = 3` = bds1, `2. bsSize = 9` = bds2)
#' rangeCoveragePlot(l, subset.chromosome = "chr22")
#'
#' @export
rangeCoveragePlot <- function(object,
                              width = 20,
                              show.samples = FALSE,
                              subset.chromosome = "chr1",
                              quiet = TRUE
) {

    # bind locally used variables
    position <- sd <- xmin <- xmax <- ymin <- ymax <- NULL
    Sample <- value <- name <- NULL

    # make plot for individual object
    if (!is(object, "list")) {
        stopifnot(is(object, "BSFDataSet"))
        if (!is.numeric(width)) {
            stop("Plotting width has to be numeric")
        }

        # check if subset is part of the seqnames from ranges
        checkRng = getRanges(object)
        if (!all(subset.chromosome %in% levels(seqnames(checkRng)))){
            msg = paste0("Chromosome to estimate on (", subset.chromosome, "), is not included in the ranges: ", paste(levels(seqnames(checkRng)), collapse = ", "))
            stop(msg)
        }

        # limit estimation to a specific chromosome
        if (!is.null(subset.chromosome)) {
            redObj = .subsetByChr(object, chr = subset.chromosome, quiet = quiet)
        } else {
            redObj = object
        }


        nSamples = nrow(getMeta(redObj))
        nBs = length(getRanges(redObj))
        bsSize = unique(width(getRanges(redObj)))
        rng = getRanges(redObj)
        rngResize = resize(rng, width = 1, fix = "center") + width
        newObj = setRanges(redObj, rngResize)

        optstrNice = paste0("#BS=", format(nBs, big.mark = ","), ", bsSize=", bsSize, ", chr=", subset.chromosome)

        # compute coverage
        cov = clipCoverage(newObj, samples.merge = FALSE, ranges.merge = FALSE,
                           positions.merge = TRUE, out.format = "data.frame", quiet = quiet)

        # main data frame to show
        suppressWarnings({
            df = data.frame(
                mean = colMeans(cov),
                sd = colSds(as.matrix(cov)),
                position = c(rev(-seq_len(width)), 0, seq_len(width))
            )
        })

        # make main plot
        p = ggplot() +
            annotate("rect", xmin = (-floor(bsSize / 2)), xmax = (floor(bsSize / 2)),
                     ymin = -Inf, ymax = Inf, alpha = 0.3, color = "grey") +
            geom_ribbon(data = df, aes(x = position, ymin = mean - sd, ymax = mean + sd),
                        fill = "#8cb3d9") +
            geom_line(data = df, aes(x = position, y = mean), size = 1, color = "#2d5986") +
            theme_classic() +
            labs(
                title = "rangeCoveragePlot()",
                subtitle = optstrNice,
                x = "Position relative to binding site center (nt)",
                y = "Number of crosslink events\n(dark blue = mean of replicates)"
            ) +
            theme(legend.position = "none")

        # add replicates to plot
        if (isTRUE(show.samples)) {
            # add sample
            dfReps = cov %>% tibble::rownames_to_column("Sample") %>%
                pivot_longer(-Sample) %>%
                mutate(position = rep(c(rev(-seq_len(width)), 0, seq_len(width)), nSamples))

            # annotate with sample names
            dfNames = dfReps %>% group_by(Sample) %>%
                filter(position == max(position))
            # set color scheme
            colfunc = colorRampPalette(c("#FFC3A1", "#A75D5D"))
            cols = colfunc(n = nSamples)

            p = p +
                scale_color_manual(values = cols) +
                geom_line(data = dfReps, aes(x = position, y = value, color = Sample)) +
                geom_text(data = dfNames, aes(x = position, y = value, label = Sample, color = Sample), nudge_x = 3) +
                theme(legend.position = "none")
        }

    }
    # make plot for a list of BSFDataSet objects
    if (is(object, "list")) {
        stopifnot(vapply(object, function(x) {
            is(x, "BSFDataSet")},
            FUN.VALUE = logical(1))
        )

        objectNames = names(object)
        df = lapply(seq_along(object), function(x) {

            # check if subset is part of the seqnames from ranges
            checkRng = getRanges(object[[x]])
            if (!all(subset.chromosome %in% levels(seqnames(checkRng)))){
                msg = paste0("Chromosome to estimate on (", subset.chromosome, "), is not included in the ranges: ", paste(levels(seqnames(checkRng)), collapse = ", "))
                stop(msg)
            }

            # limit estimation to a specific chromosome
            if (!is.null(subset.chromosome)) {
                redObj = .subsetByChr(object[[x]], chr = subset.chromosome, quiet = quiet)
            } else {
                redObj = object[[x]]
            }

            # set correct ranges
            rngResize = resize(getRanges(redObj), width = 1, fix = "center") + width
            newObj = setRanges(redObj, rngResize)

            # compute coverage
            cov = clipCoverage(newObj, samples.merge = FALSE, ranges.merge = FALSE, positions.merge = TRUE, out.format = "data.frame")

            # main data frame to show
            suppressWarnings({
                df = data.frame(
                    mean = colMeans(cov),
                    sd = colSds(as.matrix(cov)),
                    position = c(rev(-seq_len(width)), 0, seq_len(width)),
                    name = objectNames[[x]]
                )
            })
            # add sample
            dfReps = cov %>% tibble::rownames_to_column("Sample") %>%
                pivot_longer(-Sample) %>%
                select(-name)

            rownames(df) = NULL
            rownames(dfReps) = NULL

            df = cbind(df, dfReps)

            return(df)
        })
        dfPlot = do.call(rbind, df)
        # annotate with sample names
        dfNames = dfPlot %>% group_by(Sample) %>%
            filter(position == max(position))

        bsSize = vapply(object, function(x) {
            unique(width(getRanges(x)))
        }, FUN.VALUE = integer(1))

        dfFrame = data.frame(
            name = names(bsSize),
            xmin = (-floor(bsSize / 2)),
            xmax = (floor(bsSize / 2)),
            ymin = -Inf,
            ymax = Inf
        )

        p = ggplot() +
            geom_rect(data = dfFrame, aes(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax),
                      alpha = 0.2, color = "grey") +
            geom_ribbon(data = dfPlot, aes(x = position, ymin = mean - sd, ymax = mean + sd),
                        fill = "#8cb3d9") +
            geom_line(data = dfPlot, aes(x = position, y = mean), size = 1, color = "#2d5986") +
            theme_classic() +
            labs(
                title = "rangeCoveragePlot()",
                x = "Position relative to binding site center (nt)",
                y = "Number of crosslink events\n(dark blue = mean of replicates)"
            ) +
            facet_wrap( ~name) +
            theme(legend.position = "none")


        if(isTRUE(show.samples)) {
            # set color scheme
            colfunc = colorRampPalette(c("#FFC3A1", "#A75D5D"))
            cols = colfunc(n = length(unique(dfPlot$Sample)))

            p = p +
                geom_line(data = dfPlot, aes(x = position, y = value, color = Sample)) +
                geom_text(data = dfNames, aes(x = position, y = value, label = Sample, color = Sample), nudge_x = 3, size = 3) +
                scale_color_manual(values = cols)+
                theme(legend.position = "none")
        }
    }
    return(p)
}

#' Plot summarized results of the different binding site merging and filtering
#' steps
#'
#' Bar charts produced for the different filter steps in the binding site
#' merging routine. Depending on the selected option (\code{select}) all or
#' only a user defined filter can be shown.
#'
#' If \code{object} is a single BSFDataObject a single coverage plot will be
#' drawn, whereas if it is a list of BSFDataObjects, then faceting is used to
#' make a plot for each list element.
#'
#' @param object a BSFDataObject, with the makeBindingSites function already run
#' @param select one of "all", "filter", "inputRanges",
#' "minCLSites", "mergeCrosslinkSites", "minCrosslinks",
#' "centerIsClSite" or "centerIsSummit".
#' Defines which parameter is selected for plotting.
#' @param ... further arguments passed to ggplot
#'
#' @return a plot of type ggplot after the \code{\link{makeBindingSites}}
#' function was run
#'
#' @import ggplot2
#'
#' @seealso \code{\link{makeBindingSites}}
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' # plotting a single object
#' bds0 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1)
#' mergeSummaryPlot(bds0)
#'
#' # plotting mulitple obejcts
#' bds1 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")
#' bds2 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 3, sub.chr = "chr22")
#' l = list(`1. bsSize = 3` = bds1, `2. bsSize = 9` = bds2)
#' mergeSummaryPlot(l, width = 20)
#'
#' @export
mergeSummaryPlot <- function(object,
                             select = c(
                                 "all",
                                 "filter",
                                 "inputRanges",
                                 "minClSites",
                                 "mergeCrosslinkSites",
                                 "minCrosslinks",
                                 "centerIsClSite",
                                 "centerIsSummit"
                             ),
                             ...) {
    # bind locally used variables
    Option <- nRanges <- name <- NULL

    # make plot if only a single summary is given
    if (!is(object, "list")) {
        stopifnot(is(object, "BSFDataSet"))
        smry = getSummary(object)
        # filter not possible if only a single summary is given
        select = "all"
        df = smry
        df$Option = factor(df$Option, levels = c(df$Option))
        p = ggplot(df, aes(
            x = Option,
            y = nRanges,
            fill = Option
        ), ...) +
            geom_col() +
            scale_fill_brewer(palette = "Dark2") +
            theme_classic() +
            theme(axis.text.x = element_text(angle = 90),
                  legend.position = "none") +
            ggtitle("Processing steps")

    }
    # make plot if a list of summaries is give
    if (is(object, "list")) {
        select = match.arg(
            select,
            choices = c(
                "all",
                "filter",
                "inputRanges",
                "minClSites",
                "mergeCrosslinkSites",
                "minCrosslinks",
                "centerIsClSite",
                "centerIsSummit"
            )
        )

        stopifnot(all(vapply(object, function(o) {
            (is(o, "BSFDataSet"))},
            FUN.VALUE = logical(1))))

        smry = lapply(object, getSummary)
        df = do.call("rbind", smry)
        df$name = rep(names(smry), each = 6)
        df$Option = factor(df$Option, levels = c(smry[[1]][[1]]))
        if (select == "all") {
            p = ggplot(df, aes(
                x = Option,
                y = nRanges,
                fill = Option
            ), ...) +
                geom_col() +
                scale_fill_brewer(palette = "Dark2") +
                theme_classic() +
                theme(axis.text.x = element_text(angle = 90),
                      legend.position = "none") +
                facet_wrap( ~ name) +
                ggtitle("Processing steps")
        }
        if (select == "filter") {
            df = df[df$Option != c("inputRanges", "mergeCrosslinkSites"), ]
            p = ggplot(df, aes(
                x = Option,
                y = nRanges,
                fill = Option
            ), ...) +
                geom_col() +
                scale_fill_brewer(palette = "Dark2") +
                theme_classic() +
                theme(axis.text.x = element_text(angle = 90),
                      legend.position = "none") +
                facet_wrap( ~ name) +
                ggtitle("Filtering options")
        }
        if (select == "inputRanges" |
            select == "mergeCrosslinkSites" |
            select == "minCrosslinks" |
            select == "centerIsClSite" |
            select == "centerIsSummit" | select == "minClSites") {
            df = df[df$Option == c(select), ]
            p = ggplot(df, aes(
                x = name,
                y = nRanges,
                fill = name
            ), ...length()) +
                geom_col() +
                scale_fill_brewer(palette = "Dark2") +
                theme_classic() +
                theme(axis.text.x = element_text(angle = 90),
                      legend.position = "none") +
                ggtitle(select)
        }

    }

    return(p)
}

#' Plot to that shows how many replicates support each binding site
#'
#' Plotting function for settings specified in
#' \code{\link{reproducibilityFilter}}.
#'
#' @param object a BSFDataSet object
#' @param cutoff a vector of length = 1, or of length = levels(meta$conditions)
#' with a single number (between 0-1) indicating the quantile cutoff
#' @param min.crosslinks numeric of length = 1, defines the lower boundary for
#' the minimum number of crosslinks a binding site has to be supported by all
#' replicates, regardless of the replicate specific quantile threshold
#' @param max.range maximum number of crosslinks per sites that should be shown
#' @param ... further arguments passed to ggplot
#'
#' @return a plot of type \code{ggplot2} showing the per replicate
#' reproducibility cutoffs based on a given quantile threshold
#'
#' @seealso \code{\link{reproducibilityFilter}}
#'
#' @import ggplot2
#' @importFrom ggforce geom_mark_rect
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' # merge binding sites
#' bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1)
#'
#' # use the same cutoff for both conditions
#' suppressWarnings(reproducibilityCutoffPlot(bds, max.range = 20, cutoff = c(0.05)))
#'
#' # use different cutoffs for each condition
#' suppressWarnings(reproducibilityCutoffPlot(bds, max.range = 20, cutoff = c(0.1)))
#'
#'
#' @export
reproducibilityCutoffPlot <-
    function(object,
             cutoff = 0.05,
             min.crosslinks = 1,
             max.range = 20,
             ...) {

        # bind locally used variables
        value <- name <- condition <- applyTo <- NULL

        # deprecation notice
        .Deprecated("reproducibilityFilterPlot")

        stopifnot(is(object, "BSFDataSet"))

        cond = getMeta(object)$condition
        df = coverageOverRanges(
            object, returnOptions = "merge_positions_keep_replicates",
            quiet = TRUE)
        df = as.data.frame(mcols(df))

        if (length(cutoff) == 1) {
            if(length(levels(cond)) > 1) {
                stop("Only one cutoff is given for multiple conditions.")
            }
            # calculate sample specific thresholds
            qSel = .selectQuantilesSingleCondtion(
                covDf = df,
                userCond = cond,
                userNreps = 1,
                userCutoff = cutoff
            )
            # apply minimal crosslink threshold
            qSel$value = ifelse(qSel$value < min.crosslinks,
                                qSel$value + min.crosslinks, qSel$value)

            # prepare objects for plotting
            df[df > max.range] = max.range
            df = df %>% pivot_longer(everything())
            df$condition = vapply(strsplit(df$name, "_"), `[`, 2,
                                  FUN.VALUE = character(1))

            qSel$lower = ifelse(qSel$value == min.crosslinks, TRUE, FALSE)

            p = ggplot(df,
                       aes(x = value, group = name)) +
                scale_fill_brewer(palette = "Set1") +
                scale_color_brewer(palette = "Set1") +
                geom_bar(fill = "#2d5986", color = "#2d5986") +
                facet_wrap( ~name, scales = "free_x") +
                geom_vline(data = qSel,
                           aes(xintercept = value, group = name),
                           color = "darkgrey", size = 1.5) +
                theme_classic() +
                geom_label(data = qSel, aes(x = value, y = 1, label = value)) +
                labs(
                    title = paste0(paste(paste0("Cutoff: ", unique(qSel$per), " ",
                                                unique(qSel$sel)), collapse = "; "),
                                   "; min.crosslinks = ", min.crosslinks),
                    x = "Crosslinks per site",
                    y = "Number of sites",
                    fill = "condition"
                )

        }
        if (length(cutoff) > 1) {
            if (length(levels(cond)) == 1) {
                stop("Multiple cutoffs are given but only one condition exists")
            }
            # calculate sample specific thresholds
            qSel = .selectQuantilesMultipleConditions(
                covDf = df,
                userCond = cond,
                userNreps = 1,
                userCutoff = cutoff
            )
            # apply minimal crosslink threshold
            qSel$value = ifelse(qSel$value < min.crosslinks,
                                qSel$value + min.crosslinks, qSel$value)

            # prepare objects for plotting
            df[df > max.range] = max.range
            df = df %>% pivot_longer(everything())
            df$condition = vapply(strsplit(df$name, "_"), `[`, 2,
                                  FUN.VALUE = character(1))
            idx = match(df$condition, unique(cond))
            df$cutoff = cutoff[idx]
            rectDf = data.frame(
                condition = cond,
                name = unique(df$name),
                value = 1
            )
            qSel$lower = ifelse(qSel$value == min.crosslinks, TRUE, FALSE)

            p = ggplot(df,
                       aes(x = value, group = name)) +
                geom_rect(data = rectDf,
                          aes(xmin = -Inf, xmax = Inf, ymin = -Inf, ymax = Inf,
                              fill = condition), alpha = 0.3) +
                scale_fill_brewer(palette = "Set1") +
                # scale_color_brewer(palette = "Set1") +
                geom_bar(fill = "#2d5986", color = "#2d5986") +
                facet_wrap( ~name, scales = "free_x") +
                geom_vline(data = qSel,
                           aes(xintercept = value, group = name,
                               color = applyTo), size = 1.5) +
                theme_classic() +
                geom_label(data = qSel, aes(x = value, y = 1, label = value)) +
                labs(
                    title = paste0(paste(paste0("Cutoff: ", unique(qSel$per), " ",
                                                unique(qSel$sel)), collapse = "; "),
                                   "; min.crosslinks = ", min.crosslinks),
                    x = "Crosslinks per site",
                    y = "Number of sites",
                    fill = "condition"
                ) +
                scale_color_discrete(guide="none")
        }
        return(p)
    }


#' Plot that shows the binding site support ratio
#'
#' Function that shows a ratio to determine how well a given binding
#' site with is supported by the crosslink coverage of the data.
#' Ratios are computed using the \code{\link{supportRatio}} function.
#'
#' The higher the ratio, the more does the given binding site width captures
#' the enrichment of crosslinks compared the the local surrounding. A ratio
#' equal to 1 would mean no enrichemnt at all.
#'
#' @param object a BSFDataSet object
#' @param bsWidths a numeric vector indicating the different binding site
#' width to compute the ratio for
#' @param bsFlank optional; a numeric vector of the same length as
#' \code{bsWidth} used to specify the width of the flanking regions
#' @param ... further arguments passed to \code{makeBindingSites}
#'
#' @return an object of class \code{ggplot2}
#'
#' @import ggplot2
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' suppressWarnings(supportRatioPlot(bds, bsWidths = c(3,7),
#' minWidth = 1, minClSites = 1, minCrosslinks = 2))
#'
#' @export
supportRatioPlot <- function(object, bsWidths, bsFlank = NA, ...){
    # deprecation notice
    .Deprecated("estimateBsWidthPlot")

    df = supportRatio(object, bsWidths, bsFlank, ...)

    p = ggplot(df, aes(x = bsWidths, y = supportRatio, group = 1)) +
        geom_point() +
        geom_line() +
        labs(
            title = "Binding site widths signal ratios",
            x = "Binding site widths",
            y = "Signal to flank ratio"
        ) +
        theme_bw()
    p
}


#' Plot that shows binding site reproducibility as scatter
#'
#' Function compute the number of crosslinks per binding site on a log2 scale
#' for each sample. Samples are pairwise correlated as a scatter and pairwise
#' pearson correlation is shown.
#'
#' Unlike most plotting functions, this function is actively calculating the
#' values to plot.
#'
#' @param object a BSFDataSet object
#' @param quiet logical; whether to print messages
#'
#' @return an object of class \code{ggplot2}
#'
#' @seealso \code{\link{BSFind}},
#' \code{\link{reproducibilityFilter}}
#'
#' @import ggplot2
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 3, sub.chr = "chr22")
#' reproducibilityScatterPlot(bds)
#'
#' @export
reproducibilityScatterPlot <- function(object,
                                       quiet = FALSE

){
    # bind locally used variables
    what <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    if (is.null(object@params$reproducibilityFilter)) {
        what = "BEFORE"
        msg1 = paste0("Reproducibility filter was not applied yet.\n")
        msg2 = paste0("Correlations are shown BEFORE reproducibility filtering.\n")
        msg3 = paste0("Run BSFind() or reproducibilityFilter() to show correlations AFTER reproducibility filtering.\n")
        if (!quiet) message(c(msg1, msg2, msg3))
    } else {
        what = "AFTER"
        msg1 = paste0("Reproducibility filter was applied.\n")
        msg2 = paste0("Correlations are shown AFTER reproducibility filtering.\n")
        if (!quiet) message(c(msg1, msg2))
    }
    cov = coverageOverRanges(object, returnOptions = "merge_positions_keep_replicates")
    df = as.data.frame(mcols(cov))
    df = log2(df+1)
    max.value = max(df)
    p = GGally::ggpairs(df,
                        upper = list(continuous = .pairsCorColor),
                        lower = list(continuous = GGally::wrap(.parisCorrelationPlot, max.value = max.value)),
                        diag = list(continuous = .pairsDensity),
                        progress = quiet) +
        labs(title = "reproducibilityScatterPlot()",
             subtitle = paste0("Correlations are ", what, " reproducibility filter was applied."))

    return(p)
}





#' Quick figures
#'
#' Summarize all results in a set of quick figures. Depending on how the
#' function is called a different set of analytic plots are arranged into
#' either a 'main' or 'supplementary' type multi-panel figure.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param what character; the plotting option. One of: 'main', 'supp'
#' @param save.filename File name to create on the disc
#' @param save.width numeric; plot size width
#' @param save.height numeric; plot size height
#' @param save.device charcter; Device to use. One of: 'pdf', 'png', ...
#' @param quiet whether to print messages
#' @param ... further arguments passed to \code{\link[ggplot2]{ggsave}}
#'
#' @return a plot
#'
#' @seealso \code{\link{BSFind}}
#'
#' @import ggplot2
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#' bds = BSFind(bds, anno.genes = gns, anno.transcriptRegionList = regions,
#'  est.subsetChromosome = "chr22")
#' quickFigure(bds)
#'
#' @export
quickFigure <- function(object,
                        what = c("main", "supp"),
                        save.filename = NULL, # if it is NULL, then plot to normal device, if it is a path, then plot to the path
                        save.width = 10,
                        save.height = 12,
                        save.device = "pdf",
                        quiet = TRUE,
                        ...
){
    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    # handle options
    what = match.arg(what, choices = c("main", "supp"))


    # setting ggplot local theme
    theme_clean <- theme(
        plot.title = element_blank(),
        plot.subtitle = element_blank(),
        axis.text = element_text(size = 6),
        axis.title = element_text(size = 6)
    )


    if (what == "main") {

        # overview
        p1 = processingStepsFlowChart(object, size.all = 2) +
            theme(plot.title = element_blank(),
                  plot.subtitle = element_blank())

        # binding sites
        p2 = estimateBsWidthPlot(object) + theme_clean
        p3 = reproducibilitySamplesPlot(object, text.size = 6, show.title = FALSE)
        p3 = ggplotify::as.grob(p3)
        # targets and assignment
        p4 = targetGeneSpectrumPlot(object, text.size = 2) + theme_clean
        p5 = transcriptRegionSpectrumPlot(object, normalize = TRUE, values = "percentage", normalize.factor = "median", text.size = 2) + theme_clean
        p6 = bindingSiteDefinednessPlot(object, by = "transcript_region") + theme_clean

        # stitch plots into patch

        pl = list(p1, p2, p3, p4, p5, p6)
        patch = patchwork::wrap_plots(pl)

        # set layout
        layout <- "
        AAAABBB#
        AAAABBB#
        AAAABBB#
        AAAACCCC
        AAAACCCC
        AAAACCCC
        AAAACCCC
        AAAACCCC
        #DDDEEFF
        #DDDEEFF
        #DDDEEFF
        #DDDEEFF
        #DDDEEFF
        "

        # annotate patches
        patch = patch + patchwork::plot_layout(guides = "collect", design = layout) &
            theme(legend.position = "bottom",
                  legend.title = element_blank(),
                  legend.text = element_text(size = 6))  &
            theme(legend.key.size = unit(3,"mm")) &
            guides(fill = guide_legend(nrow = 4, byrow = TRUE))
        patch = patch +
            patchwork::plot_annotation(
                title = paste0("Binding spectrum of: ", getName(object)),
                subtitle = "Disclaimer: This is an auto generated figure, handle with care.",
                caption = paste0("Binding spectrum of: ", getName(object), "\n",
                                 "A) processingStepsFlowChart, B) estimateBsWidthPlot,
                                 C) reproducibilitySamplesPlot, D) targetGeneSpectrumPlot,
                                 E) transcriptRegionSpectrumPlot, F) bindingSiteDefinednessPlot"),
                tag_levels = "A"
            )

    }
    if (what == "supp") {
        # binding site definition
        p1 = pureClipGlobalFilterPlot(object) + theme_clean
        p2 = duplicatedSitesPlot(object) + theme_clean
        p3 = mergeCrosslinkDiagnosticsPlot(object) + theme_clean
        p4 = makeBsSummaryPlot(object) + theme_clean
        p5 = globalScorePlot(object) + theme_clean
        # targets and assignment
        p6 = reproducibilityFilterPlot(object) + theme_clean
        p7 = reproducibilityScatterPlot(object, quiet = quiet) + theme_clean

        p8 = geneOverlapsPlot(object, text.size = 6, show.title = FALSE)
        p9 = transcriptRegionOverlapsPlot(object, text.size = 6, show.title = FALSE)

        p = reproducibilitySamplesPlot(object, text.size = 6, show.title = FALSE)

        p7 = patchwork::wrap_elements(GGally::ggmatrix_gtable(p7))
        p8 = ggplotify::as.grob(p8)
        p9 = ggplotify::as.grob(p9)

        # stitch plots into patch
        pl = list(p1, p2, p3, p4, p5, p6, p7, p8, p9)
        patch = patchwork::wrap_plots(pl)

        # patch = p1 + p2 + p3 + p4 + p5 +
        #     p6  + patchwork::wrap_elements(GGally::ggmatrix_gtable(p7)) +
        #     ggplotify::as.grob(p8) + ggplotify::as.grob(p9)

        # set layout
        layout <- "
        AAABBBCCCDDDEEE
        AAABBBCCCDDDEEE
        AAABBBCCCDDDEEE
        FFFFFFGGGGGGGGG
        FFFFFFGGGGGGGGG
        FFFFFFGGGGGGGGG
        FFFFFFGGGGGGGGG
        FFFFFFGGGGGGGGG
        FFFFFFGGGGGGGGG
        FFFFFFGGGGGGGGG
        FFFFFFGGGGGGGGG
        HHHHHHH#IIIIIII
        HHHHHHH#IIIIIII
        HHHHHHH#IIIIIII
        HHHHHHH#IIIIIII
        "

        patch = patch + patchwork::plot_layout(guides = "collect", design = layout) &
            theme(legend.position = "right",
                  legend.title = element_blank(),
                  legend.text = element_text(size = 6))  &
            theme(legend.key.size = unit(3,"mm")) &
            guides(fill = guide_legend(ncol = 2, byrow = FALSE))
        patch = patch +
            patchwork::plot_annotation(
                title = paste0("Binding spectrum of: ", getName(object)),
                subtitle = "Disclaimer: This is an auto generated figure, handle with care.",
                caption = paste0("Binding spectrum of: ", getName(object), "\n",
                                 "A) pureClipGlobalFilterPlot, B) duplicatedSitesPlot, C) mergeCrosslinkDiagnosticsPlot,
                                 D) makeBsSummaryPlot, E) globalScorePlot, F) reproducibilityFilterPlot,
                                 G) reproducibilityScatterPlot, H) geneOverlapsPlot, I) transcriptRegionOverlapsPlot"),
                tag_levels = "A"
            )

    }

    # manage output
    if (is.null(save.filename)) {
        return(patch)
    } else {
        ggsave(filename = save.filename, plot = patch,
               width = save.width, height = save.height,
               device = save.device, ...)
    }
}

#' RBP binding site definition for iCLIP data
#'
#' This is the main function that performs the binding site definition analysis
#' through the following steps:
#' \enumerate{
#' \item Filter PureCLIP sites by their score distribution: \code{\link{pureClipGlobalFilter}}
#' \item Estimate the appropriate binding site width together with the optimal gene-wise filter level: \code{\link{estimateBsWidth}}
#' \item Filter PureCLIP sites by their score distribution per gene: \code{\link{pureClipGeneWiseFilter}}
#' \item Define equally sized binding sites: \code{\link{makeBindingSites}}
#' \item Perform replicate reproducibility filter: \code{\link{reproducibilityFilter}}
#' \item Assign binding sites to their hosting genes: \code{\link{assignToGenes}}
#' \item Assign binding sites to their hosting transcript regions: \code{\link{assignToTranscriptRegions}}
#' \item Re-assign PureCLIP scores to binding sites: \code{\link{annotateWithScore}}
#' \item Calculation of signal-to-flank ratio: \code{\link{calculateSignalToFlankScore}}
#' }
#'
#' If only the annotation is provided (\code{anno.genes} and
#' \code{anno.transcriptRegionList}), then binding sites size (\code{bsSize})
#' and gene-wise cutoff (\code{cutoff.geneWiseFilter}) are estimated using
#' \code{\link{estimateBsWidth}}. To avoid this behavior one has to provide
#' input values for the arguments \code{bsSize} and \code{cutoff.geneWiseFilter}.
#'
#' If no binding site size is provided through \code{bsSize}, then
#' \code{\link{estimateBsWidth}} is called to estimate the optimal size for the
#' given data-set. The result of this estimation can be looked at with
#' \code{\link{estimateBsWidthPlot}} and arguments can be adjusted if needed.
#'
#' Use the \code{\link{processingStepsFlowChart}} function to get an overview
#' of all steps carried out by the function.
#'
#' For complete details on each step, see the manual pages of the respective
#' functions. The \code{BSFind} function returns a \code{\link{BSFDataSet}}
#' with ranges merged into binding sites. A full flowchart for the entire process
#' can be visualized with \code{\link{processingStepsFlowChart}}. For each of the
#' individual steps dedicated diagnostic plots exists. Further information can be
#' found in our Bioconductor vignette:
#' \url{https://www.bioconductor.org/packages/release/bioc/html/BindingSiteFinder.html}
#'
#' @param object a \code{\link{BSFDataSet}} object with stored ranges
#' @param bsSize an odd integer value specifying the size of the output
#' binding sites
#'
#' @param cutoff.geneWiseFilter numeric; defines the cutoff for which sites to
#' remove in in function \code{\link{pureClipGeneWiseFilter}}. The smallest step
#' is 1\% (0.01). A cutoff of 5\% will remove the lowest 5\% sites, given their
#' score, on each gene, thus keeping the strongest 95\%.
#' @param cutoff.globalFilter numeric; defines the cutoff for which sites to
#' keep, the smallest step is 1\% (0.01) in function
#' \code{\link{pureClipGlobalFilter}}
#'
#' @param est.bsResolution character; level of resolution of the binding site
#' width in function \code{\link{estimateBsWidth}}
#' @param est.geneResolution character; level of resolution of the gene-wise
#' filtering in function \code{\link{estimateBsWidth}}
#' @param est.maxBsWidth numeric; the largest binding site width which should
#' considered in the testing
#' @param est.minimumStepGain numeric; the minimum additional gain in the score
#' in percent the next binding site width has to have, to be selected as best option
#' @param est.maxSites numeric; maximum number of PureCLIP sites that are used
#' @param est.subsetChromosome character; define on which chromosome the
#' estimation should be done in function \code{\link{estimateBsWidth}}
#' @param est.minWidth the minimum size of regions that are subjected to the
#' iterative merging routine, after the initial region concatenation.
#' @param est.offset constant added to the flanking count in the signal-to-flank
#' ratio calculation to avoid division by Zero
#' @param est.sensitive logical; whether to enable sensitive pre-filtering before
#' binding site merging or not
#' @param est.sensitive.size numeric; the size (in nucleotides) of the merged
#' sensitive region
#' @param est.sensitive.minWidth numeric; the minimum size (in nucleoties) of the
#' merged sensitive region
#'
#' @param merge.minWidth the minimum size of regions that are subjected to the
#' iterative merging routine, after the initial region concatenation.
#' @param merge.minCrosslinks the minimal number of positions to overlap with at least
#' one crosslink event in the final binding sites
#' @param merge.minClSites the minimal number of crosslink sites that have to
#' overlap a final binding site
#' @param merge.CenterIsClSite logical, whether the center of a final binding
#' site must be covered by an initial crosslink site
#' @param merge.CenterIsSummit logical, whether the center of a final binding
#' site must exhibit the highest number of crosslink events
#'
#' @param repro.cutoff numeric; percentage cutoff to be used for the
#' reproducibility quantile filtering
#' @param repro.nReps numeric; number of replicates that must meet the cutoff
#' defined in \code{repro.cutoff} for a binding site to be called reproducible.
#' Defaults to N-1.
#' @param repro.minCrosslinks numeric; minimal number of crosslinks a binding
#' site needs to have to be called reproducible. Acts as a lower boundary for
#' \code{repro.cutoff}. Defaults to 1.
#'
#' @param overlaps.geneWiseFilter character; how overlaps should be handled in
#'  \code{\link{pureClipGeneWiseFilter}}
#' @param overlaps.geneAssignment character; how overlaps should be handled in
#'  \code{\link{assignToGenes}}
#' @param overlaps.rule.geneAssignment character vector; a vector of gene types
#' that should be used to handle overlaps if option 'hierarchy' is selected
#' for \code{\link{assignToGenes}}. The order of the vector is the order of
#' the hierarchy.
#' @param overlaps.TranscriptRegions character; how overlaps should be handled in
#'  \code{\link{assignToTranscriptRegions}}
#' @param overlaps.rule.TranscriptRegions character vector; a vector of gene types
#' that should be used to handle overlaps if option 'hierarchy' is selected
#' for \code{\link{assignToTranscriptRegions}}. The order of the vector is the order of
#' the hierarchy.
#'
#' @param stf.flank character; how the flanking region shoule be set. Options are
#' 'bs', 'manual'
#' @param stf.flank.size numeric; if flank='manual' provide the desired flanking size
#'
#' @param match.score character; meta column name of the crosslink site
#' @param match.geneID character; meta column name of the genes
#' @param match.geneName character; meta column name of the gene name
#' @param match.geneType character; meta column name of the gene type
#' @param match.ranges.score a GRanges object, with numeric column for the score
#' to match in function \code{\link{annotateWithScore}}
#' @param match.option.score character; meta column name of the crosslink site
#' in function \code{\link{annotateWithScore}}
#'
#' @param anno.annoDB an object of class \code{OrganismDbi} that contains
#' the gene annotation !!! Experimental !!!
#' @param anno.genes an object of class \code{\link{GenomicRanges}} that represents
#' the gene ranges directly
#' @param anno.transcriptRegionList an object of class \code{\link{CompressedGRangesList}}
#' that holds an ranges for each transcript region
#'
#' @param quiet logical; whether to print messages
#' @param veryQuiet logical; whether to suppress all messages
#' @param ... additional arguments passed to \code{\link{estimateBsWidth}},
#' \code{\link{makeBindingSites}} and \code{\link{reproducibilityFilter}}
#'
#' @return an object of class \code{\link{BSFDataSet}} with ranges merged into
#' binding sites given the inputs.
#'
#' @seealso \code{\link{BSFDataSet}}, \code{\link{estimateBsWidth}},
#' \code{\link{pureClipGlobalFilter}}, \code{\link{pureClipGeneWiseFilter}},
#' \code{\link{assignToGenes}}, \code{\link{assignToTranscriptRegions}},
#' \code{\link{annotateWithScore}}, \code{\link{reproducibilityFilter}},
#' \code{\link{calculateSignalToFlankScore}},
#' \code{\link{processingStepsFlowChart}}
#'
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # Load genes
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' # load transcript regions
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#' BSFind(object = bds, bsSize = 9, anno.genes = gns,
#'  anno.transcriptRegionList = regions, est.subsetChromosome = "chr22")
#'
#' @export
BSFind <- function(
        # input object
        object,
        # binding site size
        bsSize = NULL,
        # cutoffs
        cutoff.geneWiseFilter = NULL,
        cutoff.globalFilter = 0.01,
        # binding site size estimation
        est.bsResolution = "medium",
        est.geneResolution = "medium",
        est.maxBsWidth = 13,
        est.minimumStepGain = 0.02,
        est.maxSites = Inf,
        est.subsetChromosome = "chr1",
        est.minWidth = 2,
        est.offset = 1,
        est.sensitive = FALSE,
        est.sensitive.size = 5,
        est.sensitive.minWidth = 2,
        # binding site merging
        merge.minWidth = 2,
        merge.minCrosslinks = 2,
        merge.minClSites = 1,
        merge.CenterIsClSite = TRUE,
        merge.CenterIsSummit = TRUE,
        # reproducibility
        repro.cutoff = NULL,
        repro.nReps = NULL,
        repro.minCrosslinks = 1,
        # overlap arguments
        overlaps.geneWiseFilter = "keepSingle",
        overlaps.geneAssignment = "frequency",
        overlaps.rule.geneAssignment = NULL,
        overlaps.TranscriptRegions = "frequency",
        overlaps.rule.TranscriptRegions = NULL ,
        # signal-to-flank
        stf.flank = "bs",
        stf.flank.size = NULL,
        # matching arguments
        match.score = "score",
        match.geneID = "gene_id",
        match.geneName = "gene_name",
        match.geneType = "gene_type",
        match.ranges.score = NULL,
        match.option.score = "max",
        # data sources
        anno.annoDB = NULL,
        anno.genes = NULL,
        anno.transcriptRegionList = NULL,
        # controls
        quiet = FALSE,
        veryQuiet = FALSE,
        ...
) {
    # INPUT CHECKS
    # --------------------------------------------------------------------------
    # General
    # ---
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))
    stopifnot(is.logical(veryQuiet))

    # check meta data
    this.meta = getMeta(object)
    if (length(levels(this.meta$condition)) > 1) {
        msg0 = paste0("Found ", length(levels(this.meta$condition)), " different conditions in the input object.\n")
        msg1 = paste0("BSFind should only be used on data from a single condition.\n")
        msg2 = paste0("Run BSFind sparately for each condition, then combine both objects with combineBSF.\n ")
        if(!quiet) warning(c(msg0,msg1,msg2))
    }

    # Check annotation source
    # ---
    if (is.null(anno.annoDB) & is.null(anno.genes) & is.null(anno.transcriptRegionList)) {
        msg = paste0("None of the required annotation sources anno.annoDB or anno.genes was specified. ")
        stop(msg)
    }
    if (!is.null(anno.annoDB) & !is.null(anno.genes) | !is.null(anno.annoDB) & !is.null(anno.transcriptRegionList)) {
        msg = paste0("Both of the required annotation sources anno.annoDB or anno.genes/ anno.transcriptRegionList are specified. Please provide only one source type. ")
        stop(msg)
    }
    if (!is.null(anno.annoDB) & is.null(anno.genes) & is.null(anno.transcriptRegionList)) {
        stopifnot(is(anno.annoDB, "OrganismDb"))
        if (!is.null(anno.genes)) {
            msg = paste0("Parameter anno.annoDB and anno.genes are specified at the same time. Use only one of them.")
            stop(msg)
        }
    }
    if (is.null(anno.annoDB) & !is.null(anno.genes) & !is.null(anno.transcriptRegionList)) {
        stopifnot(is(anno.genes, "GenomicRanges"))
        stopifnot(is(anno.transcriptRegionList, "GenomicRangesList"))
    }
    if (is.null(anno.annoDB) & !is.null(anno.genes) & is.null(anno.transcriptRegionList)) {
        msg = paste0("Parameter anno.transcriptRegionList is missing, while anno.annoDB not specified.")
        stop(msg)
    }
    if (is.null(anno.annoDB) & is.null(anno.genes) & !is.null(anno.transcriptRegionList)) {
        msg = paste0("Parameter anno.genes is missing, while anno.annoDB not specified.")
        stop(msg)
    }

    # Check values
    # ---
    obj_colname = colnames(mcols(getRanges(object)))
    if (! match.score %in% obj_colname) {
        msg = paste0("Column specified in `match.score` (", match.score, "), does not match any column in object score.")
        stop()
    }

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    obj = object

    # Begin main workflow
    if(!veryQuiet) message("pureClipGlobalFilter...")
    obj = pureClipGlobalFilter(obj,
                               match.score = match.score,
                               cutoff = cutoff.globalFilter,
                               quiet = quiet)

    # Check if parameter bsSize and cutoff.geneWiseFilter should be estimated
    if (is.null(bsSize) | is.null(cutoff.geneWiseFilter)) {
        if(!veryQuiet) message("estimateBsWidth...")
        obj = estimateBsWidth(obj,
                              bsResolution = est.bsResolution,
                              geneResolution = est.geneResolution,
                              est.maxBsWidth = est.maxBsWidth,
                              est.minimumStepGain = est.minimumStepGain,
                              est.maxSites = est.maxSites,
                              est.subsetChromosome = est.subsetChromosome,
                              est.minWidth = est.minWidth,
                              est.offset = est.offset,
                              sensitive = est.sensitive,
                              sensitive.size = est.sensitive.size,
                              sensitive.minWidth = est.sensitive.minWidth,
                              anno.annoDB = anno.annoDB,
                              anno.genes = anno.genes,
                              quiet = quiet,
                              veryQuiet = veryQuiet)
        if (is.null(bsSize) & !is.null(cutoff.geneWiseFilter)) {
            obj@params$geneFilter = cutoff.geneWiseFilter
        }
        if (!is.null(bsSize) & is.null(cutoff.geneWiseFilter)) {
            obj@params$bsSize = bsSize
        }
    }

    # Continue main workflow
    if(!veryQuiet) message("pureClipGeneWiseFilter...")
    obj = pureClipGeneWiseFilter(obj,
                                 cutoff = cutoff.geneWiseFilter,
                                 overlaps = overlaps.geneWiseFilter,
                                 match.score = match.score,
                                 match.geneID = match.geneID,
                                 anno.annoDB = anno.annoDB,
                                 anno.genes = anno.genes,
                                 quiet = quiet)

    if(!veryQuiet) message("makeBindingSites...")
    obj = makeBindingSites(obj,
                           bsSize = bsSize,
                           minWidth = merge.minWidth,
                           minCrosslinks = merge.minCrosslinks,
                           minClSites = merge.minClSites,
                           centerIsClSite = merge.CenterIsClSite,
                           centerIsSummit = merge.CenterIsSummit,
                           quiet = quiet)

    if(!veryQuiet) message("reproducibilityFilter...")
    obj = reproducibilityFilter(obj,
                                returnType = "BSFDataSet",
                                cutoff = repro.cutoff,
                                nReps = repro.nReps,
                                minCrosslinks = repro.minCrosslinks,
                                quiet = quiet)

    if(!veryQuiet) message("assignToGenes...")
    obj = assignToGenes(obj,
                        overlaps = overlaps.geneAssignment,
                        overlaps.rule = overlaps.rule.geneAssignment,
                        anno.annoDB = anno.annoDB,
                        anno.genes = anno.genes,
                        match.geneID = match.geneID,
                        match.geneName = match.geneName,
                        match.geneType = match.geneType,
                        quiet = quiet)

    if(!veryQuiet) message("assignToTranscriptRegions...")
    obj = assignToTranscriptRegions(obj,
                                    overlaps = overlaps.TranscriptRegions,
                                    overlaps.rule = overlaps.rule.TranscriptRegions,
                                    anno.annoDB = anno.annoDB,
                                    anno.transcriptRegionList = anno.transcriptRegionList,
                                    quiet = quiet)

    if(!veryQuiet) message("annotateWithScore...")
    if (is.null(match.ranges.score)) {
        match.ranges.score = getRanges(object)
        match.score = match.score
    }
    obj = annotateWithScore(obj,
                            match.ranges = match.ranges.score,
                            match.score = match.score,
                            match.option = match.option.score,
                            quiet = quiet)

    if(!veryQuiet) message("calculateSignalToFlankScore...")
    obj = calculateSignalToFlankScore(obj,
                                      flank = stf.flank,
                                      flank.size = stf.flank.size,
                                      quiet = quiet)
    return(obj)
}

test_that("BSFind wrapper works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    # Load genes
    load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
    # load transcript regions
    load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])


    expect_message(BSFind(object = bds, bsSize = 9, anno.genes = gns,
                          anno.transcriptRegionList = regions,
                          est.subsetChromosome = "chr22",
                          quiet = FALSE, veryQuiet = FALSE))

    expect_silent(BSFind(object = bds, bsSize = 9, anno.genes = gns,
                          anno.transcriptRegionList = regions,
                          est.subsetChromosome = "chr22",
                          quiet = TRUE, veryQuiet = TRUE))

    # ------
    my.size = 13
    my.filter = 0.1

    this.bds = BSFind(object = bds, bsSize = my.size, anno.genes = gns,
                     anno.transcriptRegionList = regions,
                     est.subsetChromosome = "chr22",
                     quiet = TRUE, veryQuiet = TRUE)
    expect_identical(my.size, this.bds@params$bsSize)
    expect_false(my.filter == this.bds@params$geneFilter)

    this.bds = BSFind(object = bds, cutoff.geneWiseFilter = my.filter, anno.genes = gns,
                      anno.transcriptRegionList = regions,
                      est.subsetChromosome = "chr22",
                      quiet = TRUE, veryQuiet = TRUE)
    expect_false(my.size == this.bds@params$bsSize)
    expect_identical(my.filter, this.bds@params$geneFilter)

    this.bds = BSFind(object = bds,
                      bsSize = my.size,
                      cutoff.geneWiseFilter = my.filter,
                      anno.genes = gns,
                      anno.transcriptRegionList = regions,
                      est.subsetChromosome = "chr22",
                      quiet = TRUE, veryQuiet = TRUE)
    expect_identical(my.size, this.bds@params$bsSize)

    expect_silent(BSFind(object = bds,
                      bsSize = my.size,
                      cutoff.geneWiseFilter = my.filter,
                      anno.genes = gns,
                      anno.transcriptRegionList = regions,
                      est.subsetChromosome = "chr22",
                      quiet = TRUE, veryQuiet = TRUE))

})

#' Define equally sized binding sites from peak calling results and iCLIP
#' crosslink events.
#'
#' This function performs the merging of single nucleotide crosslink sites into
#' binding sites of a user defined width (\code{bsSize}). Depending on the
#' desired output width crosslink sites with a distance closer than
#' \code{bsSize} -1 are concatenated. Initially all input regions are
#' concatenated and then imperatively merged and extended. Concatenated regions
#' smaller than \code{minWidth} are removed prior to the merge and extension
#' routine. This prevents outlier crosslink pileup, eg. mapping artifacts
#' to be integrated into the final binding sites. All remaining regions are
#' further processed and regions larger than the desired output width are
#' interactively split up by setting always the position with the highest
#' number of crosslinks as center. Regions smaller than the desired width are
#' symmetrically extended. Resulting binding sites are then filtered by the
#' defined constraints.
#'
#' The \code{bsSize} argument defines the final output width of the merged
#' binding sites. It has to be an odd number, to ensure that a binding site
#' has a distinct center.
#'
#' The \code{minWidth} parameter is used to describe the minimum width a ranges
#' has to be after the initial concatenation step. For example:
#' Consider bsSize = 9 and minWidth = 3. Then all initial crosslink sites that
#' are closer to each other than 8 nucleotides (bsSize -1) will be concatenated.
#' Any of these ranges with less than 3 nucleotides of width will be removed,
#' which reflects about 1/3 of the desired binding site width.
#'
#' The argument \code{minCrosslinks} defines how many positions of the binding
#' sites are covered with at least one crosslink event. This threshold has to
#' be defined in conjunction with the binding site width. A default value of 3
#' with a binding site width of 9 means that 1/3 of all positions in the final
#' binding site must be covered by a crosslink event. Setting this filter to 0
#' deactivates it.
#'
#' The \code{minClSites} argument defines how many positions of the binding site
#' must have been covered by the original crosslink site input. If the input was
#' based on the single nucleotide crosslink positions computed by PureCLIP than
#' this filter checks for the number of positions originally identified by
#' PureCLIP in the computed binding sites. The default of \code{minClSites} = 1
#' essentially deactivates this filter. Setting this filter to 0 deactivates it.
#'
#' The options \code{centerIsClSite} and \code{centerIsSummit} ensure that the
#' center of each binding site is covered by an initial crosslink site and
#' represents the summit of crosslink events in the binding site, respectively.
#'
#' The option \code{sub.chr} allows to run the binding site merging on a
#' smaller subset (eg. "chr1") for improoved computational speed when testing
#' the effect of various binding site width and filtering options.
#'
#' @param object a BSFDataSet object (see \code{\link{BSFDataSet}})
#' @param bsSize an odd integer value specifying the size of the output
#' binding sites
#' @param minWidth the minimum size of regions that are subjected to the
#' iterative merging routine, after the initial region concatenation.
#' @param minCrosslinks the minimal number of positions to overlap with at least
#' one crosslink event in the final binding sites
#' @param minClSites the minimal number of crosslink sites that have to
#' overlap a final binding site
#' @param centerIsClSite logical, whether the center of a final binding
#' site must be covered by an initial crosslink site
#' @param centerIsSummit logical, whether the center of a final binding
#' site must exhibit the highest number of crosslink events
#' @param sub.chr chromosome identifier (eg, chr1, chr2) used for subsetting the
#' BSFDataSet object. This option can be used for testing different
#' parameter options
#' @param quiet logical, whether to print info messages
#'
#' @return an object of type BSFDataSet with modified ranges
#'
#' @seealso \code{\link{BSFDataSet}}, \code{\link{BSFind}},
#' \code{\link{mergeCrosslinkDiagnosticsPlot}},
#' \code{\link{makeBsSummaryPlot}}
#'
#' @import GenomicRanges methods
#' @importFrom GenomeInfoDb seqlevels
#' @importFrom dplyr bind_rows
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' # standard options, no subsetting
#' bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1)
#'
#' # standard options, with subsetting
#' bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")
#'
#' @export
makeBindingSites <- function(object,
                             bsSize = NULL,
                             minWidth = 2,
                             minCrosslinks = 2,
                             minClSites = 1,
                             centerIsClSite = TRUE,
                             centerIsSummit = TRUE,
                             sub.chr = NA,
                             quiet = FALSE) {
    stopifnot(is(object, "BSFDataSet"))

    # INPUT CHECKS
    # --------------------------------------------------------------------------

    # check meta data
    this.meta = getMeta(object)
    if (length(levels(this.meta$condition)) > 1) {
        msg0 = paste0("Found ", length(levels(this.meta$condition)), " different conditions in the input object.\n")
        msg1 = paste0("It is recommended to only use data from a single condition.\n")
        msg2 = paste0("Please run makeBindingSite sparately for each condition, then combine both objects with combineBSF.\n ")
        if(!quiet) warning((c(msg0,msg1,msg2)))
    }

    # check if bsSize was estimated before or if input is needed
    if (!is.null(object@params$bsSize) & !is.null(object@params$geneFilter)) {
        bsSize = object@params$bsSize
    } else {
        if (is.null(bsSize)) {
            stop("bsSize not set. Please provide a desired binding site width.")
        }
    }
    # check integrity of input values
    if (c(bsSize %% 2) == 0) {
        stop("bsSize is even. An odd number is required to have a distinct binding site center.")
    }
    if (!is.numeric(bsSize)) {
        stop("bsSize must be of type numeric")
    }
    if (!is.numeric(minWidth)) {
        stop("minWidth must be of type numeric")
    }
    if (!is.numeric(minClSites)) {
        stop("minClSites must be of type numeric")
    }
    if (!is.numeric(minCrosslinks)) {
        stop("minCrosslinks must be of type numeric")
    }
    if (minClSites > bsSize) {
        stop("Number of forced crosslink sites is larger than desired binding site size. ")
    }

    #---------------------------------------------------------------------------
    # logical errors
    if (bsSize < minWidth) {
        warning("Desired binding site size is smaller than the minimum merging width.\n
                Be sure to check the minWidth and bsSize parameters")
    }

    #---------------------------------------------------------------------------
    # subsetting options
    if (!is.na(sub.chr)) {
        if (!is.character(sub.chr)) {
            stop("sub.chr must be of type character")
        }
        # subset the signal
        objectSub = .subsetByChr(object = object, chr = sub.chr)
        sgn = getSignal(objectSub)
        rngS0 = getRanges(objectSub)
    }
    # no subsetting
    if (is.na(sub.chr)) {
        sgn = getSignal(object)
        rngS0 = getRanges(object)
    }
    if (length(rngS0) == 0) {
        stop("0 ranges as input.")
    }

    # ---
    # Store function parameters
    optstr = list(bsSize = bsSize, minWidth = minWidth,
                  minCrosslinks = minCrosslinks, minClSites = minClSites,
                  centerIsClSite = centerIsClSite,
                  centerIsSummit = centerIsSummit, sub.chr = sub.chr)
    object@params$makeBindingSites = optstr

    #---------------------------------------------------------------------------
    # prepare data for merging
    sgnMerge = .collapseSamples(sgn)

    # execute filter and merging routines
    mergeCs = .mergeCrosslinkSites(
        rng = rngS0,
        sgn = sgnMerge,
        bsSize = bsSize,
        minWidth = minWidth,
        computeOption = "full"
    )
    rngS1 = mergeCs$rng

    # ---
    # Store data for diagnostic plot in list
    plotDf = mergeCs$countDf
    object@plotData$makeBindingSites$mergeCsData = plotDf


    if (length(rngS1) <= 0) {
        stop("No ranges left after initial crosslink merging. ")
    }

    rngS2 = .filterMinCrosslinks(rng = rngS1,
                                 sgn = sgnMerge,
                                 minCrosslinks = minCrosslinks)
    if (length(rngS2) <= 0) {
        stop("No ranges left after appying minimum crosslink events filter. ")
    }

    rngS3 = .filterMinClSites(rng = rngS2,
                              rng0 = rngS0,
                              minClSites = minClSites)
    if (length(rngS3) <= 0) {
        stop("No ranges left after appying minimum crosslink sites filter. ")
    }

    if (isTRUE(centerIsClSite)) {
        rngS4 = .filterCenterClSite(rng = rngS3, rng0 = rngS0)
    } else {
        rngS4 = rngS3
    }
    if (length(rngS4) <= 0) {
        stop("No ranges left after appying crosslink site has to be center of binding site filter. ")
    }

    if (isTRUE(centerIsSummit)) {
        rngS5 = .filterCenterSummit(rng = rngS4, sgn = sgnMerge)
    } else {
        rngS5 = rngS4
    }
    if (length(rngS5) <= 0) {
        stop("No ranges left after appying binding site center has to be summit filter. ")
    }
    #---------------------------------------------------------------------------
    # Add meta information
    rngS5$bsSize = bsSize

    #---------------------------------------------------------------------------
    # summarize number of ranges after each step
    reportDf = data.frame(
        Option = c(
            "inputRanges",
            "mergeCrosslinkSites",
            "minCrosslinks",
            "minClSites",
            "centerIsClSite",
            "centerIsSummit"
        ),
        nRanges = c(
            length(rngS0),
            length(rngS1),
            length(rngS2),
            length(rngS3),
            ifelse(isTRUE(centerIsClSite),
                   length(rngS4), NA),
            ifelse(isTRUE(centerIsSummit),
                   length(rngS5), NA)
        )
    )
    #---------------------------------------------------------------------------
    # check output
    if (!all(match(seqlevels(rngS5),unique(seqnames(rngS5)), nomatch = 0) > 0)) {
        msgWarningin = paste0("Current definition does not result in binding sites on all chromosomes where signal was present.\n No binding sites on: ",
                              paste(seqlevels(rngS5)[
                                  !match(seqlevels(rngS5),
                                         unique(seqnames(rngS5)),
                                         nomatch = 0) > 0], collapse = " "), "\n")
        if(!quiet) message(msgWarningin)
    }

    # ---
    # Store results for plotting
    object@plotData$makeBindingSites$data = reportDf
    # ---
    # Store for results
    resultLine = data.frame(
        funName = "makeBindingSites()", class = "transform",
        nIn = length(rngS0), nOut = length(rngS5),
        per = paste0(round(length(rngS5)/ length(rngS0), digits = 2)*100,"%"),
        options = paste0("bsSize=", bsSize, ", minWidth=", minWidth, ", minCrosslinks=", minCrosslinks,
                         ", minClSites=", minClSites, ", centerIsClSite=", centerIsClSite,
                         ", centerIsSummit=", centerIsSummit, ", sub.chr=", sub.chr)
    )
    object@results = rbind(object@results, resultLine)
    object@params$bsSize = bsSize

    #---------------------------------------------------------------------------
    # update BSFDataSet with new ranges information
    objectNew = setRanges(object, rngS5, quiet = TRUE)
    objectNew = setSignal(objectNew, sgn, quiet = TRUE)
    objectNew = setSummary(objectNew, reportDf)
    ClipDS = objectNew
    return(ClipDS)
}

################################################################################
# unexported functions

.mergeCrosslinkSites <- function(rng, # a GRanges object; -> holds PureCLIP sites, is single nt size
                                 sgn, # the crosslink signal merged per replicates
                                 bsSize, # the binding site size to compute
                                 minWidth, # the minimal width sites should be retained after intital merge
                                 computeOption = c("full", "simple")
) {

    # initialize local variables
    w <- n.x <- n.y <- iteration <- s <- NULL

    # handle compute options
    computeOption = match.arg(computeOption, choices = c("full", "simple"))

    # summarize signal over all replicates for mergeing
    sgnMergePlus = sgn$signalPlus
    sgnMergeMinus = sgn$signalMinus

    rngS1 = rng

    ### Merge peaks for given bs size
    rngS2 = reduce(rngS1, min.gapwidth = bsSize - 1)

    ### Keep only regions that are larger or equal to minWidth
    # -> if minWiidth == 3, then the smallest range to consider is 3
    rngS3 = rngS2[width(rngS2) >= minWidth]
    names(rngS3) = seq_along(rngS3)

    ### Center detection and extension
    rngCenterPlus <- GRanges()
    rngCenterMinus <- GRanges()
    rngToProcessPlus <- subset(rngS3, strand == "+")
    rngToProcessMinus <- subset(rngS3, strand == "-")

    countDf = data.frame()
    Counter = 0
    while (TRUE) {
        # quit if no more regions to check
        if (length(rngToProcessMinus) == 0 &
            length(rngToProcessPlus) == 0) {
            break
        } else {
            if (length(rngToProcessPlus) != 0) {
                # get max xlink position of each peak
                peaksMaxPosPlus = as.matrix(sgnMergePlus[rngToProcessPlus])
                peaksMaxPosPlus[is.na(peaksMaxPosPlus)] = -Inf
                peaksMaxPosPlus = max.col(peaksMaxPosPlus,
                                          ties.method = "first")

                # make new peaks centered arround max position
                currentPeaksPlus = rngToProcessPlus
                start(currentPeaksPlus) =
                    start(currentPeaksPlus) + peaksMaxPosPlus -1
                end(currentPeaksPlus) = start(currentPeaksPlus)
                currentPeaksPlus = currentPeaksPlus + ((bsSize - 1) / 2)
                # store peaks
                rngCenterPlus = c(rngCenterPlus, currentPeaksPlus)
                # remove peak regions from rest of possible regions
                currentPeaksPlus = as(currentPeaksPlus + ((bsSize - 1) / 2),
                                      "GRangesList")

                # update peak regions that are left for processing
                rngToProcessPlus = unlist(psetdiff(rngToProcessPlus,
                                                   currentPeaksPlus))
            }
            if (length(rngToProcessMinus) != 0) {
                peaksMaxPosMinus = as.matrix(sgnMergeMinus[rngToProcessMinus])
                peaksMaxPosMinus[is.na(peaksMaxPosMinus)] = -Inf
                peaksMaxPosMinus = max.col(peaksMaxPosMinus,
                                           ties.method = "last")

                currentPeaksMinus = rngToProcessMinus
                start(currentPeaksMinus) =
                    start(currentPeaksMinus) + peaksMaxPosMinus -1
                end(currentPeaksMinus) = start(currentPeaksMinus)
                currentPeaksMinus = currentPeaksMinus + ((bsSize - 1) / 2)

                rngCenterMinus = c(rngCenterMinus, currentPeaksMinus)

                currentPeaksMinus =
                    as(currentPeaksMinus + ((bsSize - 1) /2),
                       "GRangesList")

                rngToProcessMinus = unlist(psetdiff(rngToProcessMinus,
                                                    currentPeaksMinus))
            }
            Counter = Counter + 1
        }
        # compute option simple exits the loop after the first iteration, which is
        # when the first round of merge and extend is done
        if (computeOption == "simple") {
            if (length(rngCenterPlus) > 0 | length(rngCenterMinus) > 0){
                break
            }
        }

        # if not simple option is selected, write down stats
        countPlus = tibble(w = width(rngToProcessPlus)) %>% dplyr::count(w)
        countMinus = tibble(w = width(rngToProcessMinus)) %>% dplyr::count(w)
        currCountDf = dplyr::left_join(countPlus, countMinus, by = c("w")) %>%
            replace_na(replace = list(n.x = 0, n.y = 0)) %>% group_by(w) %>%
            mutate(s = sum(n.x + n.y, na.rm = FALSE)) %>%
            mutate(iteration = Counter) %>%
            dplyr::select(iteration, w, s)
        countDf = bind_rows(countDf, currCountDf)

    }
    rngS4 = c(rngCenterPlus, rngCenterMinus)
    rngS4 = .sortRanges(rngS4)

    mcols(rngS4)$bsID = paste0("BS", seq_along(rngS4))

    # manage return
    ret = list(rng = rngS4, countDf = countDf)
    return(ret)
}

.filterMinCrosslinks <- function(rng,
                                 sgn,
                                 minCrosslinks) {
    # filter option disabled
    if (minCrosslinks == 0) {
        rngCurr = rng
        return(rngCurr)
    }
    # filter option should be used
    if (minCrosslinks != 0) {
        sgnMergePlus = sgn$signalPlus
        sgnMergeMinus = sgn$signalMinus

        # check if both strands exists
        if ("+" %in% unique(strand(rng))) {
            # split by strand plus
            rngCurrPlus = rng[strand(rng) == "+"]
            rngCurrPlusMat = as.matrix(sgnMergePlus[rngCurrPlus])
            rngCurrPlus = rngCurrPlus[apply((rngCurrPlusMat > 0), 1, sum) >
                                          minCrosslinks]
        }
        if (!"+" %in% unique(strand(rng))) {
            rngCurrPlus = NULL
        }

        if ("-" %in% unique(strand(rng))) {
            # split by strand minus
            rngCurrMinus = rng[strand(rng) == "-"]
            rngCurrMinusMat = as.matrix(sgnMergeMinus[rngCurrMinus])
            rngCurrMinus = rngCurrMinus[apply((rngCurrMinusMat > 0), 1, sum) >
                                            minCrosslinks]
        }
        if (!"-" %in% unique(strand(rng))) {
            rngCurrMinus = NULL
        }

        # combine sort return
        rngCurr = c(rngCurrPlus, rngCurrMinus)
        rngCurr = .sortRanges(rngCurr)
        # rngCurr = GenomeInfoDb::sortSeqlevels(rngCurr)
        # rngCurr = sort(rngCurr)
        return(rngCurr)
    }



}

#' @importFrom S4Vectors queryHits
.filterCenterClSite <- function(rng, rng0) {
    rngCurr = rng - (unique(width(rng)) - 1) / 2
    rngCurr = rng[queryHits(findOverlaps(rngCurr, rng0))]
    # combine sort return
    rngCurr = .sortRanges(rngCurr)
    return(rngCurr)
}

.filterCenterSummit <- function(rng, sgn) {
    sgnMergePlus = sgn$signalPlus
    sgnMergeMinus = sgn$signalMinus

    if ("+" %in% unique(strand(rng))) {
        rngPlus = rng[strand(rng) == "+"]
        rngCurrPlusMat = as.matrix(sgnMergePlus[rngPlus])
        rngCurrPlusCount = apply(rngCurrPlusMat, 1, max)
        rngCurrPlus = rngPlus[rngCurrPlusCount == rngCurrPlusMat[, (
            (unique(width(rng)) -1) / 2) + 1]]
    }
    if (!"+" %in% unique(strand(rng))) {
        rngCurrPlus = NULL
    }

    if ("-" %in% unique(strand(rng))) {
        rngMinus = rng[strand(rng) == "-"]
        rngCurrMinusMat = as.matrix(sgnMergeMinus[rngMinus])
        rngCurrMinusCount = apply(rngCurrMinusMat, 1, max)
        rngCurrMinus = rngMinus[rngCurrMinusCount == rngCurrMinusMat[, (
            (unique(width(rng)) - 1) / 2) + 1]]
    }
    if (!"-" %in% unique(strand(rng))) {
        rngCurrMinus = NULL
    }

    # combine sort return
    rngCurr = c(rngCurrPlus, rngCurrMinus)
    rngCurr = .sortRanges(rngCurr)
    return(rngCurr)
}

.filterMinClSites <- function(rng, rng0, minClSites) {
    # filter option disabled
    if (minClSites == 0) {
        rngCurr = rng
        return(rngCurr)
    }
    # filter option is used
    if (minClSites != 0) {
        overlaps = findOverlaps(rng, rng0)
        freq = table(queryHits(overlaps))
        idx = as.numeric(names(freq[freq >= minClSites]))
        rngCurr = rng[idx]
        # combine sort return
        rngCurr = .sortRanges(rngCurr)
        return(rngCurr)
    }
}

test_that("Coverage function works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    bds = makeBindingSites(bds, bsSize = 7)

    n.samples = nrow(getMeta(bds))
    n.conditions = length(levels(getMeta(bds)$condition))
    n.bs = length(getRanges(bds))
    bs.width = bds@params$bsSize

    # all identical
    t0 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE, ranges.merge = TRUE, samples.merge = TRUE, samples.group = "all")
    expect_length(t0, 1)
    t05 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE, ranges.merge = TRUE, samples.merge = TRUE, samples.group = "condition")
    expect_equal(nrow(t05), n.conditions)
    expect_equal(ncol(t05), 1)

    t1 = clipCoverage(bds, out.format = "data.frame", positions.merge = FALSE, ranges.merge = FALSE, samples.merge = FALSE)
    expect_equal(length(t1),n.samples)
    expect_equal(nrow(t1[[1]]),n.bs)
    expect_equal(ncol(t1[[1]]),bs.width)

    # single merge
    t2 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE, ranges.merge = FALSE, samples.merge = FALSE)
    expect_equal(ncol(t2),bs.width)
    expect_equal(nrow(t2),n.samples)

    t3 = clipCoverage(bds, out.format = "data.frame", positions.merge = FALSE, ranges.merge = FALSE, samples.merge = TRUE, samples.group = "all")
    expect_equal(nrow(t3),n.bs)
    expect_equal(ncol(t3),bs.width)
    t35 = clipCoverage(bds, out.format = "data.frame", positions.merge = FALSE, ranges.merge = FALSE, samples.merge = TRUE, samples.group = "condition")
    expect_equal(length(t35), n.conditions)
    expect_equal(nrow(t35[[1]]), n.bs)
    expect_equal(ncol(t35[[1]]), bs.width)

    t4 = clipCoverage(bds, out.format = "data.frame", positions.merge = FALSE, ranges.merge = TRUE, samples.merge = FALSE)
    expect_equal(length(t4), n.samples)
    expect_equal(length(t4[[1]]), n.bs)

    # double merge
    t5 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE, ranges.merge = FALSE, samples.merge = TRUE, samples.group = "all")
    expect_equal(length(t5), bs.width)
    t55 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE, ranges.merge = FALSE, samples.merge = TRUE, samples.group = "condition")
    expect_equal(nrow(t55), n.conditions)
    expect_equal(ncol(t55), bs.width)

    t6 = clipCoverage(bds, out.format = "data.frame", positions.merge = FALSE, ranges.merge = TRUE, samples.merge = TRUE, samples.group = "all")
    expect_equal(length(t6), n.bs)
    t65 = clipCoverage(bds, out.format = "data.frame", positions.merge = FALSE, ranges.merge = TRUE, samples.merge = TRUE, samples.group = "condition")
    expect_equal(length(t65), n.conditions)
    expect_equal(length(t65[[1]]), n.bs)

    t7 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE, ranges.merge = TRUE, samples.merge = FALSE)
    expect_equal(nrow(t7), n.samples)


})

test_that("Coverage function works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # Set artificial KD condition
    metaCond = getMeta(bds)
    metaCond$condition = factor(c(rep("WT", 2), rep("KD", 2)), levels = c("WT", "KD"))
    bdsCond = setMeta(bds, metaCond)
    # Fix replicate names in signal
    namesCond = c("1_WT", "2_WT", "3_KD", "4_KD")
    sgn = getSignal(bdsCond)
    names(sgn$signalPlus) = namesCond
    names(sgn$signalMinus) = namesCond
    bdsCond = setSignal(bdsCond, sgn)
    bds = bdsCond


    bds1 <- makeBindingSites(object = bds, bsSize = 5, minWidth = 2,
                              minCrosslinks = 2, minClSites = 1, quiet = TRUE)

    # 1) Return option results output with same sized input
    # --------------------------------------------------------------------------
    expect_message(coverageOverRanges(
        bds1, returnOptions = "merge_ranges_keep_positions", quiet = FALSE))

    rngNew = getRanges(bds1)
    names(rngNew) = seq_along(rngNew)
    bds1 = setRanges(bds1, rngNew)
    # merge_ranges_keep_positions
    test = coverageOverRanges(
        bds1, returnOptions = "merge_ranges_keep_positions")
    expect_is(test,"matrix")
    expect_identical(nrow(test), nrow(getMeta(bds1)))
    expect_identical(ncol(test), unique(width(getRanges(bds1))))

    # merge_replicates_per_condition
    test = coverageOverRanges(
        bds1, returnOptions = "merge_replicates_per_condition")
    expect_is(test, "list")
    expect_identical(length(test), length(levels(getMeta(bds1)$condition)))


    expect_identical(dim(test[[1]]), dim(test[[2]]))

    # merge_all_replicates
    test = coverageOverRanges(
        bds1, returnOptions = "merge_all_replicates")
    expect_is(test, "matrix")
    expect_identical(nrow(test), length(getRanges(bds1)))
    expect_identical(ncol(test), unique(width(getRanges(bds1))))

    # merge_positions_keep_replicates
    test = coverageOverRanges(
        bds1, returnOptions = "merge_positions_keep_replicates")
    expect_is(test, "GRanges")
    expect_identical(length(test), length(getRanges(bds1)))
    expect_identical(ncol(mcols(test)), nrow(getMeta(bds1)))

    # 2) Return option results output with different sized input
    # --------------------------------------------------------------------------
    rngMod = getRanges(bds1)
    rngMod = resize(x = rngMod, fix = "center",
                    width = sample(1:10, size = length(rngMod),
                                   replace = TRUE))
    bdsMod = setRanges(bds1, rngMod)

    expect_warning(
        coverageOverRanges(
            bdsMod, returnOptions = "merge_ranges_keep_positions"))
    expect_warning(
        coverageOverRanges(
            bdsMod, returnOptions = "merge_replicates_per_condition"))
    expect_warning(
        coverageOverRanges(
            bdsMod, returnOptions = "merge_all_replicates"))


    test = coverageOverRanges(bdsMod,
                              returnOptions = "merge_replicates_per_condition",
                              allowNA = TRUE)
    expect_false(isTRUE(is.na(test$KD)))
    test = coverageOverRanges(bdsMod,
                              returnOptions = "merge_all_replicates",
                              allowNA = TRUE)
    expect_false(isTRUE(is.na(test)))
    test = coverageOverRanges(bdsMod,
                              returnOptions = "merge_ranges_keep_positions",
                              allowNA = TRUE)
    expect_false(isTRUE(is.na(test)))

    # 3) test sum/ mean method
    # --------------------------------------------------------------------------
    test = coverageOverRanges(bds1,
        returnOptions = "merge_positions_keep_replicates")
    test1 = coverageOverRanges(bds1, method = "sum",
                              returnOptions = "merge_positions_keep_replicates")
    expect_identical(test,test1)

    test = coverageOverRanges(bds1, method = "mean",
                              returnOptions = "merge_replicates_per_condition")
    test = coverageOverRanges(bds1, method = "mean",
                             returnOptions = "merge_all_replicates")
    # test = coverageOverRanges(bds1, method = "mean",
    #                          returnOptions = "merge_positions_keep_replicates")
    test = coverageOverRanges(bds1, method = "mean",
                             returnOptions = "merge_ranges_keep_positions")

})

test_that("Binding site merging works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # testing input type errors
    expect_error(makeBindingSites(object = "bds", bsSize = 3, minWidth = 3,
                                  minCrosslinks = 2, minClSites = 1))
    expect_error(makeBindingSites(object = bds, bsSize = "3", minWidth = 3,
                                  minCrosslinks = 2, minClSites = 1))
    expect_error(makeBindingSites(object = bds, bsSize = 3, minWidth = "3",
                                  minCrosslinks = 2, minClSites = 1))
    expect_error(makeBindingSites(object = bds, bsSize = 3, minWidth = 3,
                                  minCrosslinks = "2", minClSites = 1))
    expect_error(makeBindingSites(object = bds, bsSize = 3, minWidth = 3,
                                  minCrosslinks = 2, minClSites = "1"))
    expect_error(makeBindingSites(object = bds, bsSize = 3, minWidth = 3,
                                  minCrosslinks = 2, minClSites = 4))
    expect_error(makeBindingSites(object = bds, bsSize = 3, minWidth = 3,
                                  minCrosslinks = 20, minClSites = 1))
    expect_warning(makeBindingSites(object = bds, bsSize = 3, minWidth = 4,
                                    minCrosslinks = 1, minClSites = 1, quiet = TRUE))

    # testing output
    bds1 <- makeBindingSites(object = bds, bsSize = 3, minWidth = 2,
                             minCrosslinks = 2, minClSites = 1, quiet = TRUE)
    expect_identical(class(bds), class(bds1))
    expect_identical(getMeta(bds), getMeta(bds1))
    expect_equal(length(unique(width(getRanges(bds1)))), 1)
    expect_equal(unique(width(getRanges(bds1))), 3)

    # test defaults
    bds1 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
                            minCrosslinks = 2, minClSites = 1, quiet = TRUE)
    bds2 <- makeBindingSites(object = bds, bsSize = 9, quiet = TRUE)
    expect_identical(bds1,bds2)

    bds1 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
                             minCrosslinks = 2, minClSites = 1, quiet = TRUE)

})

test_that("calculateBsBackground works", {
    # load clip data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

    # make binding sites
    bds = makeBindingSites(bds, bsSize = 7)
    bds = assignToGenes(bds, anno.genes = gns)

    m = getMeta(bds)
    m$condition = factor(c("WT", "WT", "KO", "KO"), levels = c("WT", "KO"))

    s = getSignal(bds)
    names(s$signalPlus) = paste0(m$id, "_", m$condition)
    names(s$signalMinus) = paste0(m$id, "_", m$condition)

    bds = setMeta(bds, m)
    bds = setSignal(bds, s)

    # test normal mode
    expect_warning(calculateBsBackground(bds, anno.genes = gns))
    expect_silent(calculateBsBackground(bds, anno.genes = gns, quiet = TRUE))

    c0 = calculateBsBackground(bds, anno.genes = gns, quiet = TRUE)
    r0 = as.data.frame(mcols(getRanges(c0))) %>% select(contains("counts.bs"))
    r2 = as.data.frame(mcols(getRanges(c0))) %>% select(contains("counts.bg"))
    expect_identical(nrow(m), ncol(r0))
    expect_identical(nrow(m), ncol(r2))

    # test with blacklist
    myBlacklist = getRanges(bds)
    set.seed(1234)
    myBlacklist = sample(myBlacklist, size = 500) + 4

    c1 = calculateBsBackground(bds, anno.genes = gns, quiet = TRUE, blacklist = myBlacklist)
    r0 = as.data.frame(mcols(getRanges(c1))) %>% select(contains("counts.bs"))
    r2 = as.data.frame(mcols(getRanges(c1))) %>% select(contains("counts.bg"))
    expect_identical(nrow(m), ncol(r0))
    expect_identical(nrow(m), ncol(r2))

    # test with assignToGenes
    r = getRanges(bds)
    r$geneID = NULL
    bds = setRanges(bds, r)

    expect_error(calculateBsBackground(bds, anno.genes = gns, quiet = TRUE))

    c2 = calculateBsBackground(bds, anno.genes = gns, quiet = TRUE, generate.geneID.bs = TRUE)
    expect_identical(getRanges(c0)$geneID, getRanges(c2)$geneID)

    expect_warning(calculateBsBackground(bds, anno.genes = gns, quiet = TRUE, generate.geneID.bs = TRUE, overlaps = "remove"))
})

#' Show method to for the BSFDataSet
#'
#' Prints the information for each of the input data slots in the
#' \code{\link{BSFDataSet}} object.
#'
#' @docType methods
#' @name show
#' @rdname show
#' @aliases show show,BSFDataSet-method
#'
#' @param object a \code{BSFDataSet} object
#'
#' @return shows the current object state
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' show(bds)
#'
#' @export
setMethod("show",
          "BSFDataSet",
          function(object) {
              if(length(object@name) == 1) {
                  cat("Dataset: ", unique(object@name))
              }
              cat("\n")
              cat("Object of class BSFDataSet \n")
              cat("#N Ranges: ", format(length(object@ranges), big.mark = ",", decimal.mark = "."), "\n")
              if(length(unique(width(object@ranges))) > 6) {
                  cat("Width ranges: ",
                      unique(width(object@ranges))[seq(1,6)], "...\n")
              } else {
                  cat("Width ranges: ",
                      unique(width(object@ranges)), "\n")
              }
              cat("#N Samples: ", nrow(object@meta),"\n")
              cat("#N Conditions: ", length(levels(object@meta$condition)), "\n")
          })

#' Summary method to for the BSFDataSet
#'
#' Prints the summaryinformation for the \code{\link{BSFDataSet}} object. This
#' includes information on samples, conditions and crosslinks.
#'
#' @docType methods
#' @name summary
#' @rdname summary
#' @aliases summary summary,BSFDataSet-method
#'
#' @param object a \code{BSFDataSet} object
#'
#' @return summary of the current object
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' summary(bds)
#'
#' @export
setMethod("summary",
          "BSFDataSet",
          function(object) {

              funName <- nice <- condition <- samples <- NULL

              if(length(object@name) == 1) {
                  cat("Dataset: ", unique(object@name), "\n")
              } else {
                  cat("Summary of BSFDataSet \n")
              }

              cat("\n")

              cat("Functions executed: \n")
              if(! length(object@results) <= 0){
                  cat(object@results %>% pull(funName))
              } else {
                  cat("None")
              }
              cat("\n")
              cat("\n")

              cat("Ranges and chromosomes: \n")
              cat(object@ranges %>% as.data.frame(row.names = NULL) %>% count(seqnames) %>% mutate(nice = paste0(seqnames, ": ", format(n, big.mark = ","))) %>% pull(nice))
              cat("\n")
              cat("\n")

              cat("Samples and conditions: \n")
              cat(object@meta %>% count(condition, name = "samples") %>% mutate(nice = paste0(condition, ": ", samples)) %>% pull(nice))
              cat("\n")
              cat("\n")

              cat("Signal and samples: \n")
              format(lapply(object@signal$signalPlus, function(x){sum(sum(x))}) %>% unlist() + lapply(object@signal$signalMinus, function(x){sum(sum(x))}) %>% unlist(), big.mark = ",")
          })


#' Accessor method for the name of the BSFDataSet object
#'
#' The name slot holds the name of the dataset
#'
#' @docType methods
#' @name getName
#' @rdname getName
#' @aliases getName getName,BSFDataSet-method
#'
#' @param object a \code{BSFDataSet} object
#'
#' @return returns the name of the dataset
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' getName(bds)
#'
#' @export
setMethod(
    f = "getName",
    signature(object = "BSFDataSet"),
    definition = function(object) {
        validObject(object)
        return(object@name)
    }
)

#' Setter method for the names of the BSFDataSet object
#' The name slot holds the name information of the dataset
#'
#' @docType methods
#' @name setName
#' @rdname setName
#' @aliases setName setName,BSFDataSet-method
#'
#' @param object a \code{BSFDataSet} object
#' @param newName a character that is the name
#' @param ... additional arguments
#'
#' @return object of type \code{\link{BSFDataSet}} with updated name
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' bdsNew = setName(bds, "test01")
#'
#' @export
setMethod(
    f = "setName",
    signature(object = "BSFDataSet"),
    definition = function(object, newName) {
        object@name <- newName
        validObject(object)
        return(object)
    }

)



#' Accessor method for the ranges of the BSFDataSet object
#'
#' The ranges slot holds the genomic ranges information of the sites currently
#' in the object. They are encoded as a GRanges object with each binding site
#' having a single ranges entry.
#'
#' @docType methods
#' @name getRanges
#' @rdname getRanges
#' @aliases getRanges getRanges,BSFDataSet-method
#'
#' @param object a \code{BSFDataSet} object
#'
#' @return returns the genomic ranges (GRanges) of the associated ranges
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' getRanges(bds)
#'
#' @export
setMethod(
    f = "getRanges",
    signature(object = "BSFDataSet"),
    definition = function(object) {
        validObject(object)
        return(object@ranges)
    }
)

#' Setter method for the ranges of the BSFDataSet object
#' The GRanges object that holds the genomic ranges information can be replaced.
#'
#' @docType methods
#' @name setRanges
#' @rdname setRanges
#' @aliases setRanges setRanges,BSFDataSet-method
#'
#' @param object a \code{BSFDataSet} object
#' @param newRanges an object of type \code{GRanges}
#' @param dropSeqlevels enforce seqnames to be the same in ranges and signal,
#' by dropping unused seqlevels which is required for most downstream functions
#' such as \code{coverageOverRanges}
#' @param quiet logical; whether to print messages
#' @param ... additional arguments
#'
#' @return object of type \code{\link{BSFDataSet}} with updated ranges
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' rng = getRanges(bds)
#' rng = rng + 10
#' bdsNew = setRanges(bds, rng)
#'
#' @export
setMethod(
    f = "setRanges",
    signature(object = "BSFDataSet"),
    definition = function(object, newRanges, dropSeqlevels = TRUE, quiet = FALSE) {
        msg = NULL
        # check for ranges and signal integrity
        fixed = .checkForDropSeqlevels(ranges = newRanges, signal = object@signal,
                                    dropSeqlevels = dropSeqlevels)

        object@ranges <- fixed$ranges
        object@signal <- fixed$signal
        if (!is.null(fixed$msg)) {
            if(!quiet) message(fixed$msg)
        }
        validObject(object)
        return(object)
    }

)

#' Accessor method for the meta data of the BSFDataSet object
#'
#' Meta data is stored as a \code{data.frame} and must contain the columns
#' "condition", "clPlus" and "clMinus".
#'
#' @docType methods
#' @name getMeta
#' @rdname getMeta
#' @aliases getMeta getMeta,BSFDataSet-method
#'
#' @param object a BSFDataSet object
#'
#' @return returns the meta data \code{data.frame} with the columns "condition",
#' "clPlus" and "clMinus".
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' getMeta(bds)
#'
#' @export
setMethod(
    f = "getMeta",
    signature(object = "BSFDataSet"),
    definition = function(object) {
        validObject(object)
        return(object@meta)
    }
)


#' Setter method for the meta data of the BSFDataSet object
#'
#' Meta data is stored as a \code{data.frame} and must contain the columns
#' "condition", "clPlus" and "clMinus".
#'
#' @docType methods
#' @name setMeta
#' @rdname setMeta
#' @aliases setMeta setMeta,BSFDataSet-method
#'
#' @param object a BSFDataSet object
#' @param newMeta the replacement meta data table
#' @param ... additional arguments
#'
#' @return an object of type \code{\link{BSFDataSet}} with updated meta data
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' nMeta = getMeta(bds)
#' setMeta(bds, nMeta)
#'
#' @export
setMethod(
    f = "setMeta",
    signature(object = "BSFDataSet"),
    definition = function(object, newMeta) {
        object@meta = newMeta

        validObject(object)
        return(object)
    }
)



#' Accessor method for the signal data of the BSFDataSet object
#'
#' Signal data is loaded from the path specified in \code{\link{getMeta}}
#' columns "clPlus" and "clMinus" and stored as a list of RLE lists.
#'
#' @docType methods
#' @name getSignal
#' @rdname getSignal
#' @aliases getSignal getSignal,BSFDataSet-method
#'
#' @param object a BSFDataSet object
#'
#' @return returns the signal data, as list of RLE list for each strand, named
#' after the meta data columns "clPlus" and "clMinus"
#'
#' @seealso \code{\link{getMeta}} \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' getSignal(bds)
#'
#' @export
setMethod(
    f = "getSignal",
    signature(object = "BSFDataSet"),
    definition = function(object) {
        validObject(object)
        return(object@signal)
    }
)

#' Setter method for the signal data of the BSFDataSet object
#'
#' Signal data is loaded from the path specified in \code{\link{getMeta}}
#' columns "clPlus" and "clMinus" and stored as a list of RLE lists.
#'
#' @docType methods
#' @name setSignal
#' @rdname setSignal
#' @aliases setSignal setSignal,BSFDataSet-method
#'
#' @param object a BSFDataSet object
#' @param newSignal list of RLE lists
#' @param dropSeqlevels enforce seqnames to be the same in ranges and signal,
#' by dropping unused seqlevels which is required for most downstream functions
#' such as \code{coverageOverRanges}
#' @param quiet logical; whether to print messages
#' @param ... additional arguments
#'
#' @return an object of type \code{\link{BSFDataSet}} with updated signal
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' sgn = getSignal(bds)
#' sgn = lapply(sgn, function(selStrand){
#'    lapply(selStrand, function(chrList){
#'        chrList[names(chrList) == "chr22"]
#'    })
#' })
#' bdsNew = setSignal(bds, sgn)
#'
#' @export
setMethod(
    f = "setSignal",
    signature(object = "BSFDataSet"),
    definition = function(object, newSignal, dropSeqlevels = TRUE, quiet = FALSE) {
        msg = NULL
        # check for ranges and signal integrity
        fixed = .checkForDropSeqlevels(ranges = object@ranges, signal = newSignal,
                                       dropSeqlevels = dropSeqlevels)

        object@ranges <- fixed$ranges
        object@signal <- fixed$signal
        if (!is.null(fixed$msg)) {
            if(!quiet) message(fixed$msg)
        }
        validObject(object)
        return(object)
    }
)

#' Accessor method for the summary slot of the BSFDataSet object
#'
#' The summary slot is used to track information of the filtering steps applied
#' in the \code{\link{makeBindingSites}} function
#'
#' @docType methods
#' @name getSummary
#' @rdname getSummary
#' @aliases getSummary getSummary,BSFDataSet-method
#'
#' @param object a \code{BSFDataSet} object
#' @param ... additional arguments
#'
#' @return returns the summary information storted in the summary slot after
#' \code{\link{makeBindingSites}} was run
#'
#' @seealso \code{\link{BSFDataSet}} \code{\link{makeBindingSites}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
#' minCrosslinks = 2, minClSites = 1)
#'
#' getSummary(bds)
#'
#' @export
setMethod(
    f = "getSummary",
    signature(object = "BSFDataSet"),
    definition = function(object) {
        validObject(object)
        if (nrow(object@summary) == 0) {
            warning(
                "Function makeBindingSites() was not run.
                 Run makeBindingSites() to get a summary on the merging steps."
            )
        }
        if (nrow(object@summary) > 0) {
            return(object@summary)
        }
    }
)

#' Setter method for the summary slot of the BSFDataSet object
#'
#' The summary slot is used to track information of the filtering steps applied
#' in the \code{\link{makeBindingSites}} function
#'
#' @docType methods
#' @name setSummary
#' @rdname setSummary
#' @aliases setSummary setSummary,BSFDataSet-method
#'
#' @param object a BSFDataSet object
#' @param summary a data.frame with the summary information to be stored in
#' \code{BSFDataSet}
#' @param ... additional arguments
#'
#' @return an object of type \code{\link{BSFDataSet}} with updated summary info
#'
#' @seealso \code{\link{BSFDataSet}}
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' df = data.frame(processingStep = c(1,2),
#' parameter = c(3,4))
#' bds = setSummary(bds, df)
#'
#' @export
setMethod(
    f = "setSummary",
    signature(object = "BSFDataSet"),
    definition = function(object, summary) {
        object@summary <- summary
        validObject(object)
        return(object)
    }

)


#' Subset a BSFDataSet object
#'
#' You can subset \link{BSFDataSet} by identifier or by position using the
#' \code{`[`} operator. Empty seqlevels are being droppend after the subset.
#'
#' @param x A \link{BSFDataSet} object.
#' @param i Position of the identifier or the name of the identifier itself.
#' @param j Not used.
#' @param ... Additional arguments not used here.
#' @param drop if the signal not covered by the subsetted ranges should be
#' dropped or not
#'
#' @return A \link{BSFDataSet} object.
#'
#' @importFrom GenomeInfoDb seqlevels
#' @importFrom GenomeInfoDb keepSeqlevels
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' bdsNew = bds[1:10]
#'
#' @name subset-BSFDataSet
NULL

#' @rdname subset-BSFDataSet
#' @export
setMethod("[", signature(x = "BSFDataSet", i = "ANY", j = "ANY", drop = "ANY"),
          function(x, i, j, ..., drop=FALSE)
          {
              rng = x@ranges
              sgn = x@signal
              rngSub = rng[i]

              # drop empty seqlevels from ranges
              rngSub = GenomeInfoDb::keepSeqlevels(rngSub,
                                                   value = unique(seqnames(rngSub)))

              # reduce signal to only those seqlevels that are still present in
              # the ranges after subsetting
              sgnSub = lapply(sgn, function(currStrand){
                  lapply(currStrand, function(currSample){
                      currSample = currSample[match(names(currSample), seqnames(rngSub), nomatch = 0) > 0]
                  })
              })

              # drop unused seqlevels
              if (isTRUE(drop)) {
                  # drop signal on all regions not covered by binding sites
                  sgnSub = lapply(sgn, function(currStrand){
                      lapply(currStrand, function(currSample){
                          currSample = currSample[match(names(currSample), seqnames(rngSub), nomatch = 0) > 0]
                          as(lapply(currSample, function(currChr){
                              rngCov = coverage(ranges(rngSub), width = length(currChr))
                              currChr[rngCov == 0] = 0
                              currChr
                          }), "SimpleRleList")
                      })
                  })
              }
              initialize(x, ranges = rngSub, signal = sgnSub)
          }
)



#' Add two \code{\link{BSFDataSet}} objects
#'
#' Use '+' to add two objects of type \code{\link{BSFDataSet}} to each other.
#'
#' Meta data is extended by binding both input tables together. Ranges are
#' added by re-centering partially overlapping ranges according to their combined
#' coverage maximum.
#'
#' Input ranges must be of the same size. Differently size objects cannot be
#' added. For this and other usecases please see \code{\link{combineBSF}}.
#'
#' @param e1 BSFDataSet; the first dataset
#' @param e2 BSFDataSet; the second dataset
#'
#' @return A \link{BSFDataSet} object with ranges, signal and meta data from
#' both inputs.
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' # make binding sites
#' bds = makeBindingSites(bds, bsSize = 7)
#'
#' # split ranges in two groups
#' allRanges = getRanges(bds)
#' set.seed(1234)
#' idx = sample(1:length(allRanges), size = length(allRanges)/2, replace = FALSE)
#' r1 = allRanges[idx]
#' r2 = allRanges[-idx]
#'
#' # splite meta data
#' allMeta = getMeta(bds)
#' m1 = allMeta[1:2,]
#' m2 = allMeta[3:4,]
#'
#' # create new objects
#' bds1 = setRanges(bds, r1)
#' bds2 = setRanges(bds, r2)
#' bds1 = setMeta(bds1, m1)
#' bds2 = setMeta(bds2, m2)
#' bds1 = setName(bds1, "test1")
#' bds2 = setName(bds2, "test2")
#'
#' # merge two objects with '+' operator
#' c1 = bds1 + bds2
#'
#' @name add-BSFDataSet
NULL

#' @rdname add-BSFDataSet
#' @export
setMethod(
    f = "+",
    signature = c("BSFDataSet","BSFDataSet"),
    definition = function(e1, e2) {
        object = combineBSF(list(e1,e2))
        return(object)
    }
)

#' Plot signal coverage of selected ranges
#'
#' Function plots the coverage of the CLIP data in the signal slot and plots it
#' as coverage. The plot is centered around a given binding site, which can be
#' selected by an index.
#'
#' @param object a \link{BSFDataSet} object
#' @param plotIdx integer, index of the range to plot
#' @param flankPos numeric, number of nucleotides by which the plotting frame is
#' symmetrically extended
#' @param shiftPos numeric, nucleotide positions by which the current plotting
#' range should be shifted
#' @param mergeReplicates logical, if replicates should be merge per
#' condition (TRUE) or if every replicates should be shown separately (FALSE)
#' @param autoscale logical, if y-axis should be scaled to the maximum for all
#' replicates (TRUE), or not (FALSE)
#' @param highlight logical, if the central range should be highlighted (TRUE),
#' or not (FALSE)
#' @param showCentralRange logical, if the central range should be shown (TRUE),
#' or not (FALSE)
#' @param customRange \code{GenomicRanges}, a custom range object to be shown
#' underneath the coverage tracks
#' @param customRange.name character, the name of the customRange track
#' @param customAnnotation \code{GenomicRanges} or \code{TxDb}, a custom annotation
#' for eg. gene, exons, etc. to be shown underneath the coverage tracks
#' @param customAnnotation.name character, the name of the customAnnotation track
#' @param title character, set plotting title
#' @param colorPalette vector, hex colors used for the conditions
#'
#' @return an object of class \code{GVIZ}
#'
#' @seealso \code{\link{BSFDataSet}}, \code{\link{BSFind}}
#'
#' @importFrom GenomicRanges findOverlaps
#' @importFrom dplyr everything group_by summarize pull mutate rename
#' @importFrom tibble rownames_to_column
#' @importFrom tidyr pivot_longer
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' bindingSiteCoveragePlot(bds, plotIdx = 3, flankPos = 10)
#'
#' @export
bindingSiteCoveragePlot <- function(object, plotIdx, flankPos, shiftPos = NULL,
                                    mergeReplicates = FALSE, autoscale = FALSE,
                                    highlight = TRUE, showCentralRange = TRUE,
                                    customRange = NULL, customRange.name = "custom",
                                    customAnnotation = NULL, customAnnotation.name = "anno",
                                    title = NULL, colorPalette = NULL
) {

    # test parameter input validity
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is(mergeReplicates, "logical"))
    stopifnot(is(autoscale, "logical"))
    stopifnot(is(highlight, "logical"))
    stopifnot(is(showCentralRange, "logical"))

    if (!all(is.null(customRange) | is(customRange, "GenomicRanges"))) {
        stop("customRange not of the right type. Must be GenomicRanges or NULL")
    }

    if (!all(is.null(customAnnotation) |
             is(customAnnotation, "GenomicRanges") |
             is(customAnnotation, "TxDb"))) {
        stop("customAnnotation not of the right type. Must be GenomicRanges, TxDb or NULL")
    }

    # bind locally used variables
    `.` <- name <- value <- condition <- NULL

    # replicates should be merged
    if(isTRUE(mergeReplicates)) {
        object = collapseReplicates(object, collapseAll = FALSE)
    }

    # get initial objects for plotting
    rng = getRanges(object)
    sgn = getSignal(object)
    mta = getMeta(object)
    rngToPlot = rng[plotIdx]
    rngToPlotExtended = rng[plotIdx] + flankPos

    # shift window if needed
    if(!is.null(shiftPos)) {
        rngToPlotExtended = shift(rngToPlotExtended, shiftPos)
    }

    # find overlapping additional ranges
    idx = findOverlaps(rng, rngToPlotExtended)
    rngSub = rng[queryHits(idx)]

    if (is.null(names(rngSub))) {
        rngSub$symbol = paste0("BS:", seq_along(rngSub))
    } else {
        rngSub$symbol = names(rngSub)
    }

    # setting frame for plot
    currChr = unique(as.character(seqnames(rngToPlot)))
    currStart = start(rngToPlotExtended)
    currEnd = end(rngToPlotExtended)
    currStrand = as.character(strand(rngToPlotExtended))

    centerStart = start(rngToPlot)
    centerWidth = width(rngToPlot)-1

    # determine if window is too large for histogram boarder
    if (width(rngToPlotExtended) < 300) {
        plotRangeSmall = TRUE
    } else {
        plotRangeSmall = FALSE
    }

    # make tracks to plot
    axisTrack = Gviz::GenomeAxisTrack(fontcolor = "black")

    rngTrackOthers = Gviz::GeneRegionTrack(rngSub, chromosome = currChr,
                                     start = currStart, end = currEnd,
                                     name = "BS Other", fontsize = 6,
                                     exonAnnotation = "symbol", fontcolor.item="black",
                                     fill = "salmon", col = "black", col.line = "black",
                                     col.frame = "black", col.axis="black",
                                     col.border.title=NULL, frame = TRUE,
                                     col.title = "black", background.title="white")

    # compute coverage over all replicates
    sgnCov = suppressMessages(suppressWarnings(coverageOverRanges(setRanges(object, rngToPlotExtended),
                                                                  returnOptions = "merge_ranges_keep_positions")))

    # reverse signal if it is on the minus strand
    if (as.character(strand(rngToPlot)) == "-") {
        sgnCov = t(apply(sgnCov, 1, rev))
    }

    sgnRng = GenomicRanges::tile(rngToPlotExtended, width = 1) %>%
        unlist(., use.names = F)
    values(sgnRng) = t(sgnCov)

    if (isTRUE(autoscale)) {
        max = mcols(sgnRng) %>% as.data.frame() %>% max()
        max = rep(max, nrow(mta))
    } else {
        max = mcols(sgnRng) %>% as.data.frame() %>%
            tidyr::pivot_longer(dplyr::everything()) %>% dplyr::group_by(name) %>%
            dplyr::summarize(max = max(value)) %>% dplyr::pull(max)
    }

    # set color palette
    if (is.null(colorPalette)) {
        colorPalette = c("#A7D2CB", "#F2D388", "#C98474", "#874C62")
    }

    if (isTRUE(mergeReplicates)) {
        mta = data.frame(condition = as.factor(levels(mta$condition)))
    }

    if (length(levels(mta$condition)) <= length(colorPalette)) {
        replicatColors = lapply(seq_along(levels(mta$condition)), function(x){
            idx = which(levels(mta$condition)[x] == mta$condition)
            col = rep(colorPalette[x], length(idx))
            return(col)
        })
        names(replicatColors) = levels(mta$condition)
        if (!isTRUE(mergeReplicates)) {
            replicatColors = unlist(replicatColors, use.names = TRUE) %>% as.data.frame() %>%
                tibble::rownames_to_column("condition") %>%
                dplyr::mutate(condition = substring(condition, 1, nchar(condition)-1)) %>%
                dplyr::rename('col' = '.')
        }
        if (isTRUE(mergeReplicates)) {
            replicatColors = unlist(replicatColors, use.names = TRUE) %>% as.data.frame() %>%
                tibble::rownames_to_column("condition") %>%
                dplyr::rename('col' = '.')
        }
        mta$color = replicatColors$col[match(mta$condition, replicatColors$condition)]
    }
    if (length(levels(mta$condition)) == 1) {
        mta$color = "#A7D2CB"
    }
    if (length(levels(mta$condition)) > length(colorPalette)) {
        warning("The color palette supports only up to 4 different conditions at the moment. Setting all colors to grey.")
        mta$color = "#808080"
    }

    sgnTracks = lapply(seq_along(mcols(sgnRng)), function(x){
        currSample = sgnRng[,x]
        histOutlineColor = ifelse(plotRangeSmall, "black", mta$color[x])
        dTrack = Gviz::DataTrack(currSample, type = 'hist', ylim = c(0,max[x]),
                           col.histogram=histOutlineColor, fill.histogram = mta$color[x],
                           name = names(mcols(currSample)),
                           col = "black", col.line = "black",
                           col.frame = "black", col.axis="black",
                           col.border.title=NULL, frame = FALSE,
                           col.title = "black", background.title="white")
        return(dTrack)
    })

    # all main tracks are created at this point
    if (!isTRUE(showCentralRange)) {
        # remove central range from plotting if wanted
        allTracks = c(axisTrack, sgnTracks)
    } else {
        # all main tracks are added -> this is the default look of the plot
        allTracks = c(axisTrack, sgnTracks, rngTrackOthers)
    }

    # add additional custom range to be shown under the binding sites
    if (!is.null(customRange)) {
        # find overlapping custom ranges
        idx = findOverlaps(customRange, rngToPlotExtended)
        customRangeSub = customRange[queryHits(idx)]
        rngCustom = Gviz::GeneRegionTrack(customRangeSub, chromosome = currChr,
                                    start = currStart, end = currEnd,
                                    name = customRange.name, fontsize = 6,
                                    fill = "#B7C4CF", col = "black", col.line = "black",
                                    col.frame = "black", col.axis="black",
                                    col.border.title=NULL, frame = TRUE,
                                    col.title = "black", background.title="white")
        allTracks = c(allTracks, rngCustom)
    }
    # add additional gene annotation to be shown underneath
    if (!is.null(customAnnotation)) {
        rngGene = Gviz::GeneRegionTrack(customAnnotation, chromosome = currChr,
                                  start = currStart, end = currEnd,
                                  name = customAnnotation.name, fontsize = 6,
                                  fill = "#D7C0AE", col = "black", col.line = "black",
                                  col.frame = "black", col.axis="black",
                                  col.border.title=NULL, frame = TRUE,
                                  col.title = "black", background.title="white")
        allTracks = c(allTracks, rngGene)
    }
    # add highlighting option to the central range
    if(isTRUE(highlight)) {
        allTracks = Gviz::HighlightTrack(trackList = allTracks,
                                   start = centerStart, width = centerWidth, chromosome = currChr)
    }

    # manage plot title
    if (is.null(title)) {
        if (is.null(names(rngToPlot))) {
            title = paste0("BS:", as.character(plotIdx), "/ ",
                           currChr, ": ", currStart, "-", currEnd, " ", currStrand)
        } else {
            title = paste0("BS:", as.character(names(rngToPlot)), "/ ",
                           currChr, ": ", currStart, "-", currEnd, " ", currStrand)
        }
    }


    Gviz::plotTracks(trackList = allTracks,
               from = currStart, to = currEnd, main = title, cex.main = 1, col = "black")
}

.sortRanges <- function(rng) {
    rngSort = GenomeInfoDb::sortSeqlevels(rng)
    rngSort = sort(rngSort)
    return(rngSort)
}

.collapseSamples <- function(signal) {
    pSum <- Reduce(`+`, signal$signalPlus)
    names(pSum) <- names(signal$signalPlus[[1]])
    mSum <- Reduce(`+`, signal$signalMinus)
    names(mSum) <- names(signal$signalMinus[[1]])

    mergedSignal <- list(signalPlus = pSum, signalMinus = mSum)
    return(mergedSignal)
}

.subsetByChr <- function(object, chr, quiet = FALSE) {
    # subset ranges
    rng = getRanges(object)
    # rngSub = rng[seqnames(rng) %in% chr]
    rngSub = subset(rng, match(seqnames(rng), chr))

    # subset the signal
    sgn = getSignal(object)
    sgnSub = lapply(sgn, function(selStrand) {
        lapply(selStrand, function(chrList) {
            chrList[names(chrList) %in% chr]
        })
    })

    objectNew = setRanges(object, rngSub, quiet = quiet)
    objectNew = setSignal(objectNew, sgnSub, quiet = quiet)
    return(objectNew)
}

.subsetByChr_old <- function(object, chr, quiet = FALSE) {
    # subset ranges
    rng = getRanges(object)
    rngSub = rng[seqnames(rng) == chr,]

    # subset the signal
    sgn = getSignal(object)
    sgnSub = lapply(sgn, function(selStrand) {
        lapply(selStrand, function(chrList) {
            chrList[names(chrList) == chr]
        })
    })

    objectNew = setRanges(object, rngSub, quiet = quiet)
    objectNew = setSignal(objectNew, sgnSub, quiet = quiet)
    return(objectNew)
}

.checkForDropSeqlevels <- function(ranges, signal, dropSeqlevels) {
    msg = NULL
    # check input signal
    if (isTRUE(dropSeqlevels)) {
        # check which chromosomes do not fit
        rngChrs = sort(as.character(unique(seqnames(ranges))))
        check = lapply(signal, function(currStrand){
            lapply(currStrand, function(currSample){
                currChrs = sort(names(currSample))
                currChrs
            })
        })
        l = append(list(), c(check$signalPlus, check$signalMinus,
                             list('rngChr' = rngChrs)))
        chrsToUse = Reduce(intersect, l)
        # fix ranges
        rngChrsNew = rngChrs[match(chrsToUse, rngChrs)]
        rangesNew = subset(ranges, match(seqnames(ranges), rngChrsNew))
        # fix signal
        signalNew = lapply(signal, function(currStrand) {
            lapply(currStrand, function(currSample) {
                currSample[match(chrsToUse, names(currSample))]
            })
        })
        # logg which chromosome was removed
        removedChr = setdiff(unique(as.character(unlist(l))), chrsToUse)

        # check if signal was modified
        if (!identical(signal, signalNew)) {
            msg = paste0("Fixed signal input, removing chr: ",
                         paste(removedChr, collapse = " "))
        }
        if (!identical(ranges, rangesNew)) {
            msg = paste0("Fixed ranges input, removing chr: ",
                         paste(removedChr, collapse = " "))
        }
    }
    if (!isTRUE(dropSeqlevels)) {
        rngChrs = sort(as.character(unique(seqnames(ranges))))
        check = lapply(signal, function(currStrand){
            lapply(currStrand, function(currSample){
                currChrs = sort(names(currSample))
                identical(currChrs, rngChrs)
            })
        })
        if(!all(unlist(check))) {
            check = lapply(signal, function(currStrand){
                lapply(currStrand, function(currSample){
                    currChrs = sort(names(currSample))
                    currChrs
                })
            })
            l = append(list(), c(check$signalPlus, check$signalMinus,
                                 list('rngChr' = rngChrs)))
            chrsToUse = Reduce(intersect, l)
            # logg which chromosome was removed
            removedChr = setdiff(unique(as.character(unlist(l))), chrsToUse)
            # not all identical
            msg = paste0("dropSeqlevels is FALSE and chromosome found in ",
                           "ranges and singal do not match on chr: ",
                         removedChr)
            warning(msg)
        }
        rangesNew = ranges
        signalNew = signal
    }
    fixed = list(ranges = rangesNew, signal = signalNew, msg = msg)
    return(fixed)
}

.reduceSignalToFrame <- function(object, frame, quiet = FALSE) {
    rng = getRanges(object)
    # extend ranges by desired frame on which signal should be kept
    rng = rng + frame
    newObj = setRanges(object, rng, quiet = quiet)
    # drop signal outside of protective frame
    newObj = newObj[seq_along(rng), drop=TRUE]
    # put original sized ranges back in place
    rng = rng-frame
    newObj = setRanges(newObj, rng, quiet = quiet)

    return(newObj)
}

.capitalize <- function(x) {
    s <- strsplit(x, " ")[[1]]
    paste(toupper(substring(s, 1,1)), substring(s, 2),
          sep="", collapse=" ")
}

if (.Platform$OS.type != "windows") {
    test_that("BSFDataSetFromBigWig() can be build correctly", {
        # load data from which the BSFDataSet is constructed
        csFile <- system.file("extdata", "PureCLIP_crosslink_sites_examples.bed",
                              package="BindingSiteFinder")
        cs = rtracklayer::import(con = csFile, format = "BED",
                                 extraCols=c("additionalScores" = "character"))
        cs$additionalScores = NULL
        clipFiles <- system.file("extdata", package="BindingSiteFinder")
        meta = data.frame(
            id = c(1,2,3,4),
            condition = factor(c("WT", "WT", "KD", "KD"), levels = c("KD", "WT")),
            clPlus = list.files(clipFiles, pattern = "plus.bw$", full.names = TRUE),
            clMinus = list.files(clipFiles, pattern = "minus.bw$", full.names = TRUE)
        )
        expect_warning(BSFDataSetFromBigWig(ranges = cs, meta = meta))

        testCs = cs
        strand(testCs) = c(rep("-", 500), rep("+", 500))

        # check correct range sorting
        bds = expect_message(BSFDataSetFromBigWig(
            ranges = testCs, meta = meta))
        bds = expect_message(BSFDataSetFromBigWig(
            ranges = testCs, meta = meta, silent = FALSE))
        bds = expect_silent(BSFDataSetFromBigWig(
            ranges = testCs, meta = meta, silent = TRUE))
        expect_identical(getRanges(bds), .sortRanges(getRanges(bds)))

        testMeta = meta
        testMeta$id = c(1,2,3,3)
        expect_message(
            expect_error(
                BSFDataSetFromBigWig(
                    ranges = testCs, meta = testMeta))
        )
        expect_error(
            BSFDataSetFromBigWig(
                ranges = testCs, meta = testMeta, silent = TRUE))


        testMeta = meta[,c(1,2)]
        expect_error(BSFDataSetFromBigWig(
            ranges = testCs, meta = testMeta))
        expect_error(BSFDataSetFromBigWig(
            ranges = testCs, meta = testMeta, silent = TRUE))
        expect_error(BSFDataSetFromBigWig(
            ranges = testCs, meta = testMeta, silent = FALSE))

    })
}

test_that("BSFDataSet() can be build correctly", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    rng = getRanges(bds)
    sgn = getSignal(bds)
    mta = getMeta(bds)
    # correct input
    bdsNew = expect_silent(BSFDataSet(ranges = rng, signal = sgn, meta = mta, silent = TRUE))
    bdsNew = expect_silent(BSFDataSet(ranges = rng, signal = sgn, meta = mta, silent = FALSE))
    bdsNew = expect_silent(BSFDataSet(ranges = rng, signal = sgn, meta = mta))

    # meta data table corrupt
    mtaWrong = mta
    mtaWrong$clPlus = rep(mta$clPlus[1],4)
    mtaWrong$clMinus = rep(mta$clMinus[1],4)
    expect_message(BSFDataSet(ranges = rng, signal = sgn, meta = mtaWrong))
    expect_silent(BSFDataSet(ranges = rng, signal = sgn, meta = mtaWrong, silent = TRUE))

    # check correct range sorting
    rng1 = .sortRanges(rng)
    expect_true(identical(rng, rng1))

    rngNew = getRanges(bdsNew)
    expect_silent(BSFDataSet(ranges = rngNew, signal = sgn, meta = mta))

    expect_identical(getRanges(bdsNew), .sortRanges(rngNew))

    # test force equal seqnames
    rng = getRanges(bds)
    sgn = getSignal(bds)

    r = as.data.frame(granges(rng))
    d = data.frame(seqnames = "chr6", start = 1, end = 1, width = 1, strand = "+")
    rng2 = rbind(d, r) %>%
        makeGRangesFromDataFrame(., keep.extra.columns = TRUE)

    expect_message(setRanges(bds, rng2))
    expect_message(setRanges(bds, rng2, dropSeqlevels = TRUE))
    expect_message(expect_warning(setRanges(bds, rng2, dropSeqlevels = FALSE)))

})

#' @rdname BSFDataSet
#' @export
BSFDataSet <- setClass(
    "BSFDataSet",
    representation = list(
        ranges = "GRanges",
        meta = "data.frame",
        signal = "list",
        summary = "data.frame",
        params = "list",
        plotData = "list",
        results = "data.frame",
        name = "character"
    )
)
setValidity("BSFDataSet", function(object) {
    msg <- NULL

    # check input ranges
    if (any(strand(object@ranges) == "*")) {
        msg = c(msg, "Strand can not be of type'*'.")
    }
    # check input meta data
    if (!c("id") %in% colnames(object@meta) |
        !c("condition") %in% colnames(object@meta)) {
        msg = c(msg, "Metadata must contain a column named 'id' and
                a column named 'condtion'.")
    }
    if (length(object@meta$id) != length(unique(object@meta$id))) {
        msg = c(msg, "IDs in meta data are not unique. ")
    }
    # check signal
    if (!c("signalPlus") %in% names(object@signal) &
        !c("signalMinus") %in% names(object@signal)) {
        msg = c(msg, "Signal slot must at least contain one element named
                'signalPlus', or 'signalMinus'. ")
    }
    if (is.null(object@signal$signalPlus) &
        is.null(object@signal$signalMinus)) {
        msg = c(msg, "Signal slot can not be empty. ")
    }
    if (isTRUE(!all(names(object@signal) %in% c("signalPlus") |
                    names(object@signal) %in% c("signalMinus")))) {
        msg = c(msg, "Incorrect name in signal list. ")
    }
    # check signal list structure
    if(length((match(c("signalPlus", "signalMinus"), names(object@signal), nomatch = 0) > 0)) != 2){
        msg = c(msg, "Signal must contain exactly two sublists. ")
    }
    if(!all(match(c("signalPlus", "signalMinus"), names(object@signal), nomatch = 0) > 0)) {
        # check if signal list contains plus/ minus sublists
        msg = c(msg, "Signal must contain lists named 'signalPlus' and 'signalMinus'. ")
    }
    if (!is.null(object@signal$signalPlus)) {
        if (!all(unlist(lapply(object@signal$signalPlus,
                               function(x){is(x, "SimpleRleList")})))) {
            msg = c(msg, "Signal plus elements are not of type RleList")
        }
    }
    if (!is.null(object@signal$signalMinus)) {
        if (!all(unlist(lapply(object@signal$signalMinus,
                               function(x){is(x, "SimpleRleList")})))) {
            msg = c(msg, "Signal minus elements are not of type RleList")
        }
    }
    if (is.null(msg)) {
        TRUE
    } else {
        msg
    }
})

#' BSFDataSet object and constructors
#'
#' \code{BSFDataSet} contains the class \code{GenomicRanges}, which is used to
#' store input ranges. Alongside with the iCLIP signal in \code{list} structure
#' and additional meta data as \code{data.frame}.
#'
#' The ranges are enforced to have to have a "+" or "-" strand
#' annotation,"*" is not allowed. They are expected to be of the same width
#' and a warning is thrown otherwise.
#'
#' The meta information is stored as \code{data.frame} with at least two
#' required columns, 'id' and 'condition'. They are used to build the unique
#' identifier for each replicate split by '_' (eg. id = 1 and condition = WT
#' will result in 1_WT).
#'
#' The meta data needs to have the additional columns 'clPlus' and 'clMinus'
#' to be present if \code{BSFDataSetFromBigWig} is called.
#' It is used to provide the location to the iCLIP coverage
#' files to the import function. On object initialization these files are
#' loaded and internally represented in the signal slot of the object
#' (see \code{\link{BSFDataSet}}).
#'
#' The iCLIP signal is stored in a special list structure.
#' At the lowest level crosslink counts per nucleotide are stored as
#' \code{Rle} per chromosome summarized as a \code{SimpleRleList}. Such a list
#' exits for each replicate and must be named by the replicate identifier
#' (eg. 1_WT). Therefore this list contains always exactly the same number of
#' entries as the number of replicates in the dataset. Since we handle strands
#' initially seperated from each other this list must be given twice, once
#' for each strand. The strand specific entries must be named 'signalPlus' and
#' 'signalMinus'.
#'
#' The option \code{dropSeqlevels} forces the seqnames of the ranges and
#' the signal to be the same. If for a specific chromosome in the ranges
#' no respective entry in the signal list can be found, then entries with
#' that chromosome are dropped This behavior is needed to keep the
#' \link{BSFDataSet} object in sync, which is required for downstream functions
#' such as \code{coverageOverRanges}
#'
#' @param ranges a \code{GenomicRanges} with the desired ranges to process. The
#' strand slot must be either + or -.
#' @param meta a \code{data.frame} with at least two columns. The first column
#' should be a unique numeric id. The second column holds sample type
#' information, such as the condition.
#' @param signal a \code{list} with the two entries 'signalPlus' and
#' 'signalMinus', following a special representation of \code{SimpleRleList}
#' for counts per replicates (see details for more information).
#' @param silent suppress messages but not warnings (TRUE/ FALSE)
#' @param dropSeqlevels enforce seqnames to be the same in ranges and signal,
#' by dropping unused seqlevels which is required for most downstream functions
#' such as \code{coverageOverRanges}
#'
#' @return A BSFDataSet object.
#'
#' @aliases BSFDataSet, BSFDataSet-class, BSFDataSetFromBigWig
#'
#' @docType class
#'
#' @import GenomicRanges
#' @importFrom rtracklayer import
#'
#' @examples
#'
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' rng = getRanges(bds)
#' sgn = getSignal(bds)
#' mta = getMeta(bds)
#' bdsNew = BSFDataSet(ranges = rng, signal = sgn, meta = mta)
#'
#' @rdname BSFDataSet
#' @export
BSFDataSet <- function(ranges, meta, signal, dropSeqlevels = TRUE, silent = FALSE) {
    msg = NULL
    # check input ranges
    if (!all(c(any(strand(ranges) == "-"), any(strand(ranges) == "+")))) {
        warning("Input ranges are only on one strand. \n")
    }
    if (length(unique(width(ranges))) > 1) {
        msg = c(msg, "Ranges are of different width. \n")
        # warning("Ranges are of differnt width. ")
    }
    if (any(width(ranges) > 1)) {
        msg = c(msg, "Input ranges are larger than 1 nt. \n")
        # message("Input ranges are larger than 1 nt. ")
    }
    # check input metadata
    if(!all(c(any(colnames(meta) == "id"),
              any(colnames(meta) == "condition")))){
        msg = c(msg, "Meta data columns must contain 'id' and 'condition'. \n")
        # stop("Meta data columns must contain 'id' and 'condition'. ")
    }
    if (!is.factor(meta$condition)) {
        msg = c(msg, "Condition column is not a factor, converting to factor. \n")
        # message("Condition column is not factor, converting to factor. ")
        meta$condition = factor(meta$condition)
    }
    if (any(duplicated(meta$clPlus)) |
        any(duplicated(meta$clMinus))) {
        msg = c(msg, "Given paths are duplicated. Please check your input. \n")
        # message("Given path are duplicated. Please check your input.")
    }
    if (! identical(.sortRanges(ranges), ranges)){
        msg = c(msg, "Input ranges are not sorted, sorting for you. \n")
        # message('Input ranges are not sorted, sorting for you.')
        ranges = .sortRanges(ranges)
    }

    # check for signal and ranges integrity
    fixed = .checkForDropSeqlevels(ranges = ranges, signal = signal,
                                dropSeqlevels = dropSeqlevels)
    ranges = fixed$ranges
    signal = fixed$signal
    if (!is.null(fixed$msg)) {
        msg = c(msg, fixed$msg)
    }

    if (!isTRUE(silent)) {
        if(!is.null(msg)) {
            message(msg)
        }
    }

    # manage dataset name
    if ("name" %in% colnames(meta)){
        if (length(unique(meta$name)) != 1) {
            # name corrupt
            warning("Dataset name can not be constructed from meta data column named 'name'. ")
        } else {
            # set name from meta
            name = unique(meta$name)
        }
    } else {
        name = character()
    }

    # set placeholder for slots
    summary = data.frame()
    params = list()
    plotData = list()
    results = data.frame()

    # construct final object
    obj = new(
        "BSFDataSet",
        ranges = ranges,
        meta = meta,
        signal = signal,
        summary = summary,
        params = params,
        plotData = plotData,
        results = results,
        name = name
    )
    return(obj)
}

#' @rdname BSFDataSet
#' @export
BSFDataSetFromBigWig <- function(ranges, meta, silent = FALSE,
                                 dropSeqlevels = TRUE) {
    # check the meta data dataframe for additional info where to find
    # the big wig files
    if(!all(c(
        any(colnames(meta) == "id"),
        any(colnames(meta) == "condition"),
        any(colnames(meta) == "clPlus"),
        any(colnames(meta) == "clMinus")))){
        stop("Meta data columns must contain 'id', 'condition' and
             'clPlus', 'clMinus'. ")
    }
    # check if given path point to actual files
    if (all(!file.exists(meta$clPlus) &
            !file.exists(meta$clMinus))) {
        stop("Given path do not point to existing files. ")
    }
    if (any(duplicated(meta$clPlus)) |
        any(duplicated(meta$clMinus))) {
        warning("Given path are duplicated. Please check your input.")
    }
    # build colSignal by importing files as RLE
    signalPlus = unlist(lapply(meta$clPlus, function(x) {
        rtracklayer::import(con = x, as = "RleList")
    }))
    signalMinus = unlist(lapply(meta$clMinus, function(x) {
        abs(rtracklayer::import(con = x, as = "RleList"))
    }))
    names(signalPlus) = paste0(meta$id, "_", meta$condition)
    names(signalMinus) = paste0(meta$id, "_", meta$condition)
    signal = list(signalPlus = signalPlus, signalMinus = signalMinus)

    # construct BindingSiteFinder data set
    object = BSFDataSet(ranges = ranges, meta = meta, signal = signal,
                        dropSeqlevels = dropSeqlevels, silent = silent)
    return(object)
}

.pairsCorColor <- function(data, mapping, color = I("black"), sizeRange = c(1, 5), ...) {

    # set P value cutoff
    pvalCo = 0.05

    # get the x and y data to use the other code
    x <- GGally::eval_data_col(data, mapping$x)
    y <- GGally::eval_data_col(data, mapping$y)

    # calc pairwise pearson correlation
    ct <- stats::cor.test(x,y, method = "pearson")

    # get P value
    pval = ct$p.value

    # get correlation coefficient
    r <- unname(ct$estimate)
    rt <- format(r, digits=3)[1]
    tt <- as.character(rt)

    # plot the cor value
    p <- GGally::ggally_text(
        label = tt,
        mapping = aes(),
        xP = 0.5, yP = 0.5,
        size = 4,
        color = color,
        ...
    ) +
        theme_void() +
        theme(
            panel.background=element_rect(fill="white"),
            panel.grid.minor=element_blank(),
            panel.grid.major=element_blank()
        )

    # set color scheme for background
    colfunc = colorRampPalette(c("#cfd6e2", "#303c50"))
    corColors = colfunc(n = 10)

    # apply color scheme for background
    if(r >= .9){
        corCol = corColors[10]
    }
    else if(r >= .8){
        corCol = corColors[9]
    } else if(r >= .7){
        corCol = corColors[8]
    } else if(r >= .6){
        corCol = corColors[7]
    } else if(r >= .5){
        corCol = corColors[6]
    } else if(r >= .4){
        corCol = corColors[5]
    } else if(r >= .3){
        corCol = corColors[4]
    } else if(r >= .2){
        corCol = corColors[3]
    } else if(r >= .1){
        corCol = corColors[2]
    }
    else {
        corCol = "white"
    }

    # set background color to grey for not significant correlations
    if (pval >= pvalCo) {
        corCol = "#A9A9A9"
    }

    p <- p + theme(
        panel.background = element_rect(fill = corCol,
                                        color = corCol)
    )
    return(p)
}

.pairsDensity <- function(data, mapping, ...){
    # init local variables
    y <- NULL

    data = data %>% drop_na()
    # compute quantiles to display
    x <- GGally::eval_data_col(data, mapping$x)
    x = as.numeric(x)
    dt = data.frame(x = c(1:length(x)), y = x)
    dens <- density(dt$y)
    df <- data.frame(x=dens$x, y=dens$y)
    probs <- c(0, 0.5, 1)
    quantiles <- quantile(dt$y, prob=probs)
    quant <- factor(findInterval(df$x,quantiles))
    df$quant <- factor(findInterval(df$x,quantiles))

    # make plot
    p = ggplot(df, aes(x,y)) +
        geom_line() + geom_ribbon(aes(ymin=0, ymax=y), fill = "#ECE5C7") +
        scale_x_continuous(breaks=quantiles) +
        scale_fill_brewer(guide="none") +
        theme_bw() +
        theme(axis.title.x=element_blank(),
              axis.text.x=element_blank(),
              axis.ticks.x=element_blank(),
              axis.title.y=element_blank(),
              axis.text.y=element_blank(),
              axis.ticks.y=element_blank())
    return(p)
}

.parisCorrelationPlot <- function(data, mapping, max.value) {

    X <- GGally::eval_data_col(data, mapping$x)
    Y <- GGally::eval_data_col(data, mapping$y)

    p = ggplot(data, mapping) +
        geom_bin2d(bins = 100) +
        viridis::scale_fill_viridis() +
        theme_bw() +
        theme(legend.position = "none") +
        scale_x_continuous(limits = c(-1, ceiling(max.value))) +
        scale_y_continuous(limits = c(-1, ceiling(max.value)))

    return(p)
}

#' Compute background coverage for binding sites per gene
#'
#' This function computes the background coverage used for the differential
#' binding analysis to correct for transcript level changes. Essentially,
#' the crosslink signal on each gene is split into crosslinks that can be
#' attributed to the binding sites and all other signal that can be attributed
#' to the background.
#'
#' To avoid that crosslinks from binding sites contaminate the background counts
#' a protective region around each binding sites can be spanned with
#' \code{use.offset} the default width of the offset region is half of the
#' binding site width, but can also be changed with the \code{ranges.offset}
#' parameter.
#'
#' Additional region that one wants to exclude from contributing to the
#' background signal can be incorporated as \code{GRanges} objects through
#' the \code{blacklist} option.
#'
#' It is expected that binding sites are assigned to hosting genes prior to
#' running this funciton (see \code{\link{BSFind}}). This means a unique gene ID
#' is present in the meta columns of each binding site ranges. If this is not the
#' case one can invoce the binding site to gene assignment with
#' \code{generate.geneID.bs}. The same is true for the blacklist regions with
#' option \code{generate.geneID.blacklist}.
#'
#' It is expected that all binding sites are of the same size
#' (See \code{\link{BSFind}} on how to achieve this). If this is however not
#' the case and one wants to keep binding sites of different with then option
#' \code{force.unequalSites} can be used.
#'
#' This function is intended to be used for the generation of the count matrix
#' used for the differential binding analysis. It is usually preceded by
#' \code{\link{combineBSF}} and followed by \code{\link{filterBsBackground}}.
#'
#' @param object a \code{\link{BSFDataSet}} object with two conditions
#' @param anno.annoDB an object of class \code{OrganismDbi} that contains
#' the gene annotation.
#' @param anno.genes an object of class \code{\link{GenomicRanges}} that represents
#' the gene ranges directly
#' @param blacklist GRanges; genomic ranges where the signal should be
#' excluded from the background
#' @param use.offset logical; if an offset region around the binding sites should
#' be used on which the signal is excluded from the background
#' @param ranges.offset numeric; number of nucleotides the offset window around
#' each binding site should be wide (defaults to 1/2 binding site width - NULL)
#' @param match.geneID.gene character; the name of the column with the gene ID
#' in the genes meta columns used for matching binding sites to genes
#' @param match.geneID.bs character; the name of the column with the gene ID
#' in the binding sites meta columns used for matching binding sites to genes
#' @param match.geneID.blacklist character; the name of the column with the gene
#' ID in the blacklist meta columns used for matching the blacklist regions
#' with the genes
#' @param generate.geneID.bs logical; if the binding site to gene matching
#' should be performed if no matching gene ID is provided
#' @param generate.geneID.blacklist logical; if the blacklist to gene matching
#' should be performed if no matching gene ID is provided
#' @param uniqueID.gene character; column name of a unique ID for all genes
#' @param uniqueID.bs character; column name of a unique ID for all binding sites
#' @param uniqueID.blacklist character; column name of a unique ID for all
#' blacklist regions
#' @param force.unequalSites logical; if binding sites of not identical width
#' should be allowed or not
#' @param quiet logical; whether to print messages or not
#' @param veryQuiet logical; whether to print messages or not
#' @param ... additional arguments; passed to \code{\link{assignToGenes}}
#'
#' @seealso \code{\link{combineBSF}},
#' \code{\link{filterBsBackground}}
#'
#' @return an object of class \code{\link{BSFDataSet}} with counts for binding
#' sites, background and total gene added to the meta column of the ranges
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#'
#' # make binding sites
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#'
#' # change meta data
#' m = getMeta(bds)
#' m$condition = factor(c("WT", "WT", "KO", "KO"), levels = c("WT", "KO"))
#' bds = setMeta(bds, m)
#'
#' # change signal
#' s = getSignal(bds)
#' names(s$signalPlus) = paste0(m$id, "_", m$condition)
#' names(s$signalMinus) = paste0(m$id, "_", m$condition)
#' bds = setSignal(bds, s)
#'
#' # make example blacklist region
#' myBlacklist = getRanges(bds)
#'set.seed(1234)
#' myBlacklist = sample(myBlacklist, size = 500) + 4
#'
#' # make background
#' bds.b1 = calculateBsBackground(bds, anno.genes = gns)
#'
#' # make background - no offset
#' bds.b2 = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)
#'
#' # make background - use blacklist
#' bds.b3 = calculateBsBackground(bds, anno.genes = gns, blacklist = myBlacklist)
#'
#' @export
calculateBsBackground <- function(object,
                                  anno.annoDB = NULL,
                                  anno.genes = NULL,
                                  blacklist = NULL,
                                  use.offset = TRUE,
                                  ranges.offset = NULL,
                                  match.geneID.gene = "gene_id",
                                  match.geneID.bs = "geneID",
                                  match.geneID.blacklist = "geneID",
                                  generate.geneID.bs = FALSE,
                                  generate.geneID.blacklist = FALSE,
                                  uniqueID.gene = "gene_id",
                                  uniqueID.bs = "bsID",
                                  uniqueID.blacklist = "bsID",
                                  force.unequalSites = FALSE,
                                  quiet = FALSE,
                                  veryQuiet = TRUE,
                                  ...
){
    # bind local variables
    geneID <- . <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))
    stopifnot(is.logical(veryQuiet))
    stopifnot(is.logical(use.offset))
    stopifnot(is.logical(generate.geneID.bs))
    stopifnot(is.logical(generate.geneID.blacklist))
    stopifnot(is.logical(force.unequalSites))

    # check meta data
    this.meta = getMeta(object)
    if (length(levels(this.meta$condition)) == 1) {
        msg0 = paste0("Found only one conditions in the input object.\n")
        msg1 = paste0("Make sure to combine objects from different conditions in order to compare them.\n")
        stop(c(msg0,msg1))
        # TODO this could also be just a warning
    }

    # Check gene annotation source
    # -> Check if none is specified
    if (is.null(anno.annoDB) & is.null(anno.genes)) {
        msg = paste0("None of the required annotation sources anno.annoDB or anno.genes was specified. ")
        stop(msg)
    }
    # -> Check if both are specified
    if (!is.null(anno.annoDB) & !is.null(anno.genes)) {
        msg = paste0("Both of the required annotation sources anno.annoDB or anno.genes are specified. Please provide only one of the two. ")
        stop(msg)
    }
    # -> Checks if anno.annoDB should be used
    if (!is.null(anno.annoDB) & is.null(anno.genes)) {
        stopifnot(is(anno.annoDB, "OrganismDb"))
        datasource = "anno.annoDB"
        # extract relevant annotation
        anno.genes = genes(anno.annoDB, columns = c("GENEID"))
        # Create matching vectors for columns from input annotation
        selectID = as.character(anno.genes$GENEID)
    }
    # -> Checks if anno.genes should be used
    if (is.null(anno.annoDB) & !is.null(anno.genes)) {
        stopifnot(is(anno.genes, "GenomicRanges"))

        datasource = "anno.genes"
        # extract relevant annotation
        anno.genes = anno.genes
        # check correct annotation columns
        annoColNames = colnames(mcols(anno.genes))
        if (!match.geneID.gene %in% annoColNames) {
            msg0 = paste0("The column '", match.geneID.gene, "', for match.geneID.gene is not present.\n")
            msg1 = paste0("Provide a column for gene id matching. \n")
            stop(c(msg0,msg1))
        }
        # Create matching vectors for columns from input annotation
        selectID.gene = mcols(anno.genes)[match(match.geneID.gene, colnames(mcols(anno.genes)))][[1]]
    }
    # test if unique id is present
    if (!uniqueID.gene %in% colnames(mcols(anno.genes))) {
        # id is not present
        # -> throw error
        msg0 = paste0("No unique ID present for genes with name: ", uniqueID.gene, " for option 'uniqueID.gene'. \n")
        msg1 = paste0("Provide a unique ID for each range. \n")
        stop(c(msg0,msg1))
    } else {
        # id is present
        # -> test if it is unique
        this.uniqueID.gene = (mcols(anno.genes))[colnames(mcols(anno.genes)) == uniqueID.gene][[1]]
        if (any(duplicated(this.uniqueID.gene))) {
            msg0 = paste0("Unique IDs are not duplicated for IDs: ",
                          paste(this.uniqueID.gene[duplicated(this.uniqueID.gene)], collapse = ","))
        }
    }

    # Check if binding site to gene assignment was already done
    # -> if not run gene assignment first
    if (!match.geneID.bs %in% colnames(mcols(getRanges(object)))) {
        # no gene ID found in binding site meta columns
        msg0 = paste0("No matching column for gene ID found with name: '", match.geneID.bs, "' in binding site meta data.\n")
        if (!isTRUE(generate.geneID.bs)){
            # binding site to gene matching should not be done
            # -> stop with error
            msg1 = paste0("Binding site to gene matching not actiavted with: generate.geneID.bs=", generate.geneID.bs)
            stop(c(msg0,msg1))
        } else {
            # -> run binding site to gene matching
            msg1 = paste0("Trying 'assignToGenes()' to assing binding sites to hosting genes. \n")
            if(!quiet) warning(msg0, msg1)
            object = assignToGenes(object, anno.genes = anno.genes, quiet = quiet, ...)
        }
    }
    if (!uniqueID.bs %in% colnames(mcols(getRanges(object)))) {
        # id is not present
        # -> throw error
        msg0 = paste0("No unique ID present for binding sites with name: ", uniqueID.bs, " for option 'uniqueID.bs'. \n")
        msg1 = paste0("Provide a unique ID for each range. \n")
        stop(c(msg0,msg1))
    } else {
        # id is present
        # -> test if it is unique
        this.uniqueID.bs = mcols(getRanges(object))[colnames(mcols(getRanges(object))) == uniqueID.bs][[1]]
        if (any(duplicated(this.uniqueID.bs))) {
            msg0 = paste0("Unique IDs are not duplicated for IDs: ",
                          paste(this.uniqueID.bs[duplicated(this.uniqueID.bs)], collapse = ","))
        }
    }

    # Check if blacklist regions have a gene ID that can be used
    # -> if not run gene assignment first
    blacklist.inUse = FALSE
    if (!is.null(blacklist)) {
        # blacklist regions are present
        blacklist.inUse = TRUE
        # -> check gene ID
        if (!match.geneID.blacklist %in% colnames(mcols(blacklist))) {
            # no geneID found in binding site meta columns
            msg0 = paste0("No matching column for gene ID found with name: '", match.geneID.blacklist, "' in blacklist region meta data.\n")
            if (!isTRUE(generate.geneID.blacklist)){
                # binding site to gene matching should not be done
                # -> stop with error
                msg1 = paste0("Blacklist region to gene matching not actiavted with: generate.geneID.blacklist=", generate.geneID.blacklist)
                stop(c(msg0,msg1))
            } else {
                # -> run binding site to gene matching
                msg1 = paste0("Trying 'assignToGenes()' to assing blacklist regions to hosting genes. \n")
                if(!quiet) warning(msg0, msg1)
                # TODO
                # -> think about which parameters for assignToGenes need to be available to the user
                object.blacklist = setRanges(object, blacklist, quiet = quiet)
                object.blacklist = assignToGenes(object.blacklist, anno.genes = anno.genes, quiet = quiet, ...)
                blacklist = getRanges(object.blacklist)
            }
        }
        if (!uniqueID.blacklist %in% colnames(mcols(blacklist))) {
            # id is not present
            # -> throw error
            msg0 = paste0("No unique ID present for blacklist regions with name: ", uniqueID.blacklist, " for option 'uniqueID.blacklist'. \n")
            msg1 = paste0("Provide a unique ID for each range. \n")
            stop(c(msg0,msg1))
        } else {
            # id is present
            # -> test if it is unique
            this.uniqueID.blacklist = mcols(blacklist)[colnames(mcols(blacklist)) == uniqueID.blacklist][[1]]
            if (any(duplicated(this.uniqueID.blacklist))) {
                msg0 = paste0("Unique IDs are not duplicated for IDs: ",
                              paste(this.uniqueID.blacklist[duplicated(this.uniqueID.blacklist)], collapse = ","))
            }
        }
    }

    # get all ranges
    this.ranges.initial = getRanges(object)
    this.ranges = getRanges(object)

    # check ranges width
    if (length(unique(width(this.ranges))) > 1) {
        # binding sites are not all of the same size
        msg0 = paste0("Binding sites are not all of the same size. \n")
        msg1 = paste0("Found sizes: ", paste(unique(width(this.ranges)), collapse = ","))
        msg2 = paste0("Use 'combineBSF()' to unify binding site sizes. \n")
        # check if user wants to force continue despite binding sites being unequal
        if(!isTRUE(force.unequalSites)){
            stop(c(msg0, msg1, msg2))
        } else {
            if (!quiet) warning(c(msg0, msg1, msg2))
        }
    }

    # check for binding sites with geneID = NA and remove
    all.geneIDs = mcols(this.ranges)[match(match.geneID.bs, colnames(mcols(this.ranges)))][[1]]
    if (any(is.na(all.geneIDs))) {
        msg0 = paste0("Found ", length(which(is.na(all.geneIDs))), " binding sites with 'NA' as geneID. \n")
        msg1 = paste0("Removing those keeps ", length(which(!is.na(all.geneIDs))), " binding sites remaining. \n")
        if(!quiet) warning(c(msg0,msg1))
        this.ranges = this.ranges[which(!is.na(all.geneIDs))]
    }

    # final binding site check
    if (length(this.ranges) == 0) {
        # no more binding site ranges are left
        msg0 = paste0("No more binding sites left to work with.\n")
        msg1 = paste0("Check input data. \n")
        stop(c(msg0,msg1))
    }

    # MAIN COMPUTE
    # --------------------------------------------------------------------------

    # calculate coverage per binding site
    # ------------------------------------
    cov.bs = .coverageForDifferential(object, ranges = this.ranges,
                                      match.rangeID = uniqueID.bs,
                                      prefix = "counts.bs.", quiet = veryQuiet)

    # calculate coverage of hosting genes
    # ----------------------------------------
    this.matchIDs.gene = mcols(anno.genes)[,which(match.geneID.gene == colnames(mcols(anno.genes)))]
    this.matchIDs.bs = mcols(this.ranges)[,which(match.geneID.bs == colnames(mcols(this.ranges)))]
    hosting.genes = anno.genes[this.matchIDs.gene %in% this.matchIDs.bs]
    # calculate gene-wise coverage
    cov.hosting.genes = .coverageForDifferential(object, ranges = hosting.genes,
                                                 match.rangeID = uniqueID.gene,
                                                 prefix = "counts.gene.", quiet = veryQuiet)

    # group counts by gene
    cov.bs.merge = as.data.frame(mcols(cov.bs)) %>%
        select(geneID, contains("counts")) %>%
        group_by(geneID) %>%
        dplyr::summarise_all(sum)
    # match binding site counts per gene
    match.gene = mcols(cov.hosting.genes)[colnames(mcols(cov.hosting.genes)) == match.geneID.gene][[1]]

    match.bs = cov.bs.merge[colnames(cov.bs.merge) == match.geneID.bs][[1]]
    idx.bs = match(match.gene, match.bs)
    mcols(cov.hosting.genes) = cbind(mcols(cov.hosting.genes),
                                     select(cov.bs.merge, starts_with("counts"))[idx.bs,])


    # add the offset ranges per gene
    # ----------------------------------------
    if (isTRUE(use.offset)) {
        # offset should be used
        # -> check if the default or manual value should be used
        if(!is.null(ranges.offset)){
            # ranges.offset is set manually
            stopifnot(is(ranges.offset, "numeric"))
        } else {
            ranges.offset = floor(width(this.ranges)/2)
        }

        # turn offset size into offset ranges
        range.offset = this.ranges + ranges.offset

        #
        # # turn offset size into offset ranges
        # range.before = flank(this.ranges, width = ranges.offset, start = TRUE, both = FALSE)
        # range.after = flank(this.ranges, width = ranges.offset, start = FALSE, both = FALSE)
        #
        # # merge overlapping offsets
        # range.offset = c(range.before, range.after)


        range.offset.merge = reduce(range.offset, with.revmap = TRUE)
        idx = unlist(lapply(range.offset.merge$revmap, `[[`, 1))
        range.offset.merge$geneID = mcols(range.offset)[colnames(mcols(range.offset)) == match.geneID.bs][[1]][idx]
        range.offset.merge$id = seq_along(range.offset.merge)

        # calulate coverage on offset regions
        cov.offset = .coverageForDifferential(object, ranges = range.offset.merge,
                                              match.rangeID = "id",
                                              prefix = "counts.offset.", quiet = veryQuiet)

        # group counts by gene
        cov.offset.gene = as.data.frame(mcols(cov.offset)) %>%
            select(all_of(match.geneID.bs), contains("counts")) %>%
            dplyr::group_by_if(., is.character) %>%
            dplyr::summarise_all(sum)

        # match counts per gene
        match.gene = mcols(cov.hosting.genes)[colnames(mcols(cov.hosting.genes)) == match.geneID.gene][[1]]
        match.offset = cov.offset.gene[colnames(cov.offset.gene) == match.geneID.bs][[1]]
        idx.offset = match(match.gene, match.offset)

        # add counts to gene
        mcols(cov.hosting.genes) = cbind(mcols(cov.hosting.genes),
                                         select(cov.offset.gene, starts_with("counts"))[idx.offset,])
    }
    # add additional blacklist regions per gene
    # ----------------------------------------

    if (!is.null(blacklist)) {
        # blacklist region is present
        # -> calculate coverage
        cov.blacklist = .coverageForDifferential(object, ranges = blacklist,
                                                 match.rangeID = uniqueID.blacklist,
                                                 prefix = "counts.black.", quiet = veryQuiet)

        # group counts by gene
        # cov.blacklist = as.data.frame(mcols(cov.blacklist)) %>%
        #     select(geneID, contains("counts")) %>%
        #     group_by(geneID) %>%
        #     dplyr::summarise_all(sum)

        cov.blacklist = as.data.frame(mcols(cov.blacklist)) %>%
            select(all_of(match.geneID.blacklist), contains("counts")) %>%
            dplyr::group_by_if(., is.character) %>%
            dplyr::summarise_all(sum)

        # match counts per gene
        match.gene = mcols(cov.hosting.genes)[colnames(mcols(cov.hosting.genes)) == match.geneID.gene][[1]]

        match.blacklist = cov.blacklist[colnames(cov.blacklist) == match.geneID.bs][[1]]
        idx.blacklist = match(match.gene, match.blacklist)
        mcols(cov.hosting.genes) = cbind(mcols(cov.hosting.genes),
                                         select(cov.blacklist, starts_with("counts"))[idx.blacklist,])
    }

    # manage returns
    # ----------------------------------------
    countMatr = mcols(cov.hosting.genes) %>% as.data.frame() %>%
        select(starts_with("counts"), ) %>%
        replace(is.na(.), 0)

    if (isTRUE(use.offset) | !is.null(blacklist)) {
        # one additional resource to subtract from gene counts is present
        if (isTRUE(use.offset) & !is.null(blacklist)) {
            # offset and blacklist are used
            # -> subtract both with bs from gene counts
            sel.countsToRemove = countMatr %>% select(contains(c("offset", "black")))

            # combine counts to remove
            sample.ids = paste0(this.meta$id, "_", this.meta$condition)
            sel.countsToRemove = as.matrix(sel.countsToRemove)
            countsToRemove = lapply(sample.ids, function(x) {
                rowSums(sel.countsToRemove[, grepl(x, colnames(sel.countsToRemove))])
            })
            countsToRemove = do.call(cbind, countsToRemove)
        }
        if (isTRUE(use.offset) & is.null(blacklist)) {
            # offset is used and blacklist is not used
            # -> subtract offset with bs from gene counts
            sel.countsToRemove = countMatr %>% select(contains(c("offset")))
            countsToRemove = sel.countsToRemove
        }
        if (!isTRUE(use.offset) & !is.null(blacklist)) {
            # offset is not used and blacklist is used
            # -> subtract blacklist with bs from gene counts
            sel.countsToRemove = countMatr %>% select(contains(c("black")))
            countsToRemove = sel.countsToRemove
        }
    } else {
        # only one type signal is used for background
        if (!isTRUE(use.offset) & is.null(blacklist)) {
            # none of the options are used (no blacklist, not offset)
            # -> only subtract bs count from genes
            sel.countsToRemove = countMatr %>% select(contains(c("bs")))
        }
        countsToRemove = sel.countsToRemove
    }

    # make background counts
    sample.ids = paste0(this.meta$id, "_", this.meta$condition)
    countsBg = countMatr %>% select(contains("gene"))
    countsBg = countsBg - countsToRemove
    colnames(countsBg) = paste0("counts.bg.",sample.ids)

    # select background and gene coverage for each hosting gene and replicate
    # mcols(cov.hosting.genes) = cbind(select(as.data.frame(mcols(cov.hosting.genes)), -(starts_with("counts"))),
    #                                  select(as.data.frame(mcols(cov.hosting.genes)), starts_with("counts.gene")),
    #                                  countsBg)
    # select background coverage for each hosting gene and replicate
    mcols(cov.hosting.genes) = cbind(select(as.data.frame(mcols(cov.hosting.genes)), -(starts_with("counts"))),
                                     countsBg)

    # match hosting gene counts with binding site counts
    this.matchIDs.gene = mcols(cov.hosting.genes)[,which(match.geneID.gene == colnames(mcols(cov.hosting.genes)))]
    idx = match(this.matchIDs.bs, this.matchIDs.gene)
    mcols(cov.bs) = cbind(mcols(cov.bs),
                          select(as.data.frame(mcols(cov.hosting.genes)), starts_with("counts"))[idx,])

    # set object and return
    object = setRanges(object, cov.bs, quiet = veryQuiet)

    # ---
    # Store function parameters in list
    optstr = list(source = datasource,
                  blacklist = blacklist.inUse,
                  use.offset = use.offset,
                  ranges.offset = paste(unique(ranges.offset), sep = ","),
                  match.geneID.gene = match.geneID.gene,
                  match.geneID.bs = match.geneID.bs,
                  match.geneID.blacklist = match.geneID.blacklist,
                  generate.geneID.blacklist = generate.geneID.blacklist,
                  generate.geneID.bs = generate.geneID.bs,
                  uniqueID.blacklist = uniqueID.blacklist,
                  uniqueID.bs = uniqueID.bs,
                  uniqueID.gene = uniqueID.gene,
                  force.unequalSites = force.unequalSites)
    object@params$calculateBsBackground = optstr

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "calculateBsBackground()", class = "transform",
        nIn = length(this.ranges.initial), nOut = length(cov.bs),
        per = paste0(round(length(cov.bs)/ length(this.ranges.initial), digits = 2)*100,"%"),
        options = paste0("Ranges.offset=", optstr$ranges.offset,
                         ", source=", optstr$datasource,
                         ", blacklist=", optstr$blacklist,
                         ", use.offset=", optstr$use.offset,
                         ", match.geneID.gene=", optstr$match.geneID.gene,
                         ", match.geneID.bs=", optstr$match.geneID.bs,
                         ", match.geneID.blacklist=", optstr$match.geneID.blacklist,
                         ", generate.geneID.blacklist=", optstr$generate.geneID.blacklist,
                         ", generate.geneID.bs=", optstr$generate.geneID.bs,
                         ", uniqueID.blacklist=", optstr$uniqueID.blacklist,
                         ", uniqueID.bs=", optstr$uniqueID.bs,
                         ", uniqueID.gene=", optstr$uniqueID.gene,
                         ", force.unequalSites=", optstr$force.unequalSites)
    )

    object@results = rbind(object@results, resultLine)

    return(object)
}


#' Filter for genes not suitable for differential testing
#'
#' This function removes genes where the differential testing protocol can
#' not be applied to, using count coverage information on the binding sites
#' and background regions per gene, through the following steps:
#' \enumerate{
#' \item Remove genes with overall not enough crosslinks: \code{minCounts}
#' \item Remove genes with a disproportion of counts in binding sites vs. the
#' background: \code{balanceBackground}
#' \item Remove genes where the expression between conditions is too much off
#' balance: \code{balanceCondition}
#' }
#'
#' To remove genes with overall not enough crosslinks (\code{minCounts})
#' all counts are summed up per gene across all samples and compared to the
#' minimal count threshold (\code{minCounts.cutoff}).
#'
#' To remove genes with a count disproportion between binding sites and
#' background regions crosslinks are summed up for binding sites and background
#' per gene. These sums are combined in a ratio. Genes where eg. 50\% of all
#' counts are within binding sites would be removed
#' (see \code{balanceBackground.cutoff.bs} and \code{balanceBackground.cutoff.bg}).
#'
#' To remove genes with very large expression differences between conditions,
#' crosslinks are summed up per gene for each condition. If now eg. the total
#' number of crosslinks is for 98\% in one condition and only 2\% of the
#' combined signal is in the second condition, expression levels are too
#' different for a reliable comparisson (see \code{balanceCondition.cutoff}).
#'
#' This function is intended to be used right after a call of \code{\link{calculateBsBackground}}.
#'
#' @param object a \code{\link{BSFDataSet}} object with computed count data
#' for binding sites and background regions
#' @param minCounts logical; whether to use the minimum count filter
#' @param minCounts.cutoff numeric; the minimal number of crosslink
#' per gene over all samples for the gene to be retained (default = 100)
#' @param balanceBackground logical; whether to use the counts balancing filter
#' between binding sites and background
#' @param balanceBackground.cutoff.bs numeric; the maximum fraction of the total signal
#' per gene that can be within binding sites (default = 0.2)
#' @param balanceBackground.cutoff.bg numeric; the minimum fraction of the total signal
#' per gene that can be within the background (default = 0.8)
#' @param balanceCondition logical; whether to use the counts balancing filter
#' between conditions
#' @param balanceCondition.cutoff numeric; the maximum fraction of the total signal
#' that can be attributed to only one condition
#' @param match.geneID character; the name of the column with the gene ID
#' in the binding sites meta columns used for matching binding sites to genes
#' @param flag logical; whether to remove or flag binding sites from genes that do not
#' pass any of the filters
#' @param quiet logical; whether to print messages or not
#' @param veryQuiet logical; whether to print messages or not
#'
#' @seealso \code{\link{calculateBsBackground}}, \code{\link{plotBsBackgroundFilter}}
#'
#' @return an object of class \code{\link{BSFDataSet}} with biniding sites filtered
#' or flagged by the above filter options
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#'
#' # make testset
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = imputeBsDifferencesForTestdata(bds)
#' bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)
#'
#' # use all filters and remove binding sites that fail (default settings)
#' f0 = filterBsBackground(bds)
#'
#' # do not use the condition balancing filter
#' f1 = filterBsBackground(bds, balanceCondition = FALSE)
#'
#' # use only the minimum count filter and flag binding sites instead of
#' # removing them
#' f3 = filterBsBackground(bds, flag = TRUE, balanceCondition = FALSE,
#'  balanceBackground = FALSE)
#'
#' @export
filterBsBackground <- function(object,
                               minCounts = TRUE,
                               minCounts.cutoff = 100,
                               balanceBackground = TRUE,
                               balanceBackground.cutoff.bs = 0.3,
                               balanceBackground.cutoff.bg = 0.7,
                               balanceCondition = TRUE,
                               balanceCondition.cutoff = 0.05,
                               match.geneID = "geneID",
                               flag = FALSE,
                               quiet = FALSE,
                               veryQuiet = FALSE
){
    # Initialize local variables
    . <- s <- name <- value <- type <- bg <- gene.approx <- bs <- ratio.bg <- NULL
    ratio.bs <- f.bg <- f.bs <-condition <- total <- both <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))
    stopifnot(is.logical(veryQuiet))
    stopifnot(is.logical(minCounts))
    stopifnot(is.logical(balanceBackground))
    stopifnot(is.logical(balanceCondition))
    stopifnot(is.logical(flag))

    # get ranges
    this.ranges.initial = getRanges(object)
    this.ranges = getRanges(object)

    # check if ranges have geneID present
    if (!match.geneID %in% colnames(mcols(this.ranges))) {
        msg0 = paste0("No matching column for gene ID found with name: '", match.geneID, "' in binding site meta data.\n")
        msg1 = paste0("Please run calculateBsBackground() first. \n")
        stop(c(msg0, msg1))
    }

    # check if ranges have counts present
    if (!any(grepl("counts", colnames(as.data.frame(mcols(this.ranges)))))) {
        # meta data has no columns named counts
        # -> stop here; point user to calculateBsBackground
        msg0 = paste0("No associated fields with crosslink counts found for present binding sites. \n")
        msg1 = paste0("Make sure to run 'calculateBsBackground()' first. \n")
        stop(msg0,msg1)
    }
    # check if ranges have at least two samples per condition in counts
    this.counts = as.data.frame(mcols(this.ranges)) %>% select(contains("counts"))
    # check binding sites
    this.counts.bs = this.counts %>% select(contains("bs"))
    if (length(this.counts.bs) == 0) {
        msg0 = paste0("No counts for binding sites present. \n")
        msg1 = paste0("Make sure to run 'calculateBsBackground()' first. \n")
        stop(c(msg0, msg1))
    }
    # check background
    this.counts.bg = this.counts %>% select(contains("bg"))
    if (length(this.counts.bg) == 0) {
        msg0 = paste0("No counts for background present. \n")
        msg1 = paste0("Make sure to run 'calculateBsBackground()' first. \n")
        stop(c(msg0, msg1))
    }

    this.meta = getMeta(object)
    this.condition.reference = levels(this.meta$condition)[1]
    this.condition.comp = levels(this.meta$condition)[2]

    # MAIN
    # --------------------------------------------------------------------------

    # filter for genes with less counts than cutoff
    if (isTRUE(minCounts)) {
        # the gene-wise count filter should be used
        # -> group counts by gene and sample and filter
        all.counts = as.data.frame(mcols(this.ranges)) %>%
            select(all_of(match.geneID), contains("counts")) %>%
            dplyr::group_by_at(match.geneID) %>%
            summarise(dplyr::across(everything(), sum)) %>%
            mutate(s = select(., contains("counts")) %>% rowSums(.)) %>%
            select(all_of(match.geneID), s)
        idx = which(all.counts < minCounts.cutoff, arr.ind = TRUE)

        # ---
        # Store for plotting
        object@plotData$filterBsBackground$data.minCounts = all.counts


        if (length(idx) > 0) {
            if (!veryQuiet) message("Count filter ")
            # found positions with counts less than cutoff
            # remove these counts
            msg0 = paste0("Found ", format(nrow(idx), big.mark = ",", decimal.mark = "."),
                          " genes with less total counts than ", minCounts.cutoff, ".\n")
            # remove genes below threshold
            genesToRemove = all.counts[idx[,1],] %>% pull(match.geneID)
            allGenes = mcols(this.ranges)[colnames(mcols(this.ranges)) == match.geneID][[1]]
            idxToRemove = allGenes %in% genesToRemove

            if (isTRUE(flag)) {
                # flag genes with low counts
                # -> add additional meta column
                msg1 = paste0("Flagging a total of ",
                              format(length(genesToRemove), big.mark = ",", decimal.mark = "."),
                              " genes and ",
                              format(sum(idxToRemove), big.mark = ",", decimal.mark = "."),
                              " binding sites.\n")
                mcols(this.ranges)$flag.minCount = idxToRemove
            }
            if (!isTRUE(flag)) {
                # remove genes with low counts
                msg1 = paste0("Removing a total of ",
                              format(length(genesToRemove), big.mark = ",", decimal.mark = "."),
                              " genes and ",
                              format(sum(idxToRemove), big.mark = ",", decimal.mark = "."),
                              " binding sites.\n")
                this.ranges = this.ranges[!idxToRemove]
            }
            # inform user
            if (!quiet) warning(c(msg0, msg1))
        }
    }

    # filter for genes with binding site vs background ratio being off
    if (isTRUE(balanceBackground)) {
        if (!veryQuiet) message("Ratio filter ")
        # binding site to background ratio filter is used
        # -> calculate ratios per gene over all samples
        df.ratio = as.data.frame(mcols(this.ranges)) %>%
            select(all_of(match.geneID), starts_with("counts")) %>%
            pivot_longer(-all_of(match.geneID)) %>%
            mutate(type = sub(".*\\.(.*?)\\..*", "\\1", name)) %>%
            select(-name) %>%
            dplyr::group_by_if(., is.character) %>%
            summarise(sum = sum(value), .groups = "keep") %>%
            pivot_wider(values_from = sum, names_from = type) %>%
            mutate(gene.approx = bg + bs) %>%
            mutate(ratio.bg = bg / (gene.approx + 1)) %>%
            mutate(ratio.bs = bs / (gene.approx + 1)) %>%
            select(all_of(match.geneID), ratio.bg, ratio.bs) %>%
            pivot_longer(-all_of(match.geneID))

        # ---
        # Store for plotting
        object@plotData$filterBsBackground$data.balanceBackground = df.ratio

        # apply filter cutoffs
        df.filter = df.ratio %>% ungroup() %>%
            pivot_wider(names_from = name, values_from = value) %>%
            mutate(f.bg = ifelse(ratio.bg < balanceBackground.cutoff.bs, TRUE, FALSE),
                   f.bs = ifelse(ratio.bs > balanceBackground.cutoff.bs, TRUE, FALSE))

        msg0 = paste0("Found ", format(sum(df.filter$f.bg | df.filter$f.bs), big.mark = ".", decimal.mark = ","),
                      " genes with bs to bg ratio not meeting thresholds.\n")

        genesToRemove = df.filter %>% filter(f.bg | f.bs) %>% pull(match.geneID)
        allGenes = mcols(this.ranges)[colnames(mcols(this.ranges)) == match.geneID][[1]]
        idxToRemove = allGenes %in% genesToRemove

        if (isTRUE(flag)) {
            # flag genes with ratio above thresholds
            # -> add additional meta column
            msg1 = paste0("Flagging a total of ",
                          format(length(genesToRemove), big.mark = ",", decimal.mark = "."),
                          " genes and ",
                          format(sum(idxToRemove), big.mark = ",", decimal.mark = "."),
                          " binding sites.\n")
            mcols(this.ranges)$flag.ratio = idxToRemove
        }
        if (!isTRUE(flag)) {
            # remove genes with low counts
            msg1 = paste0("Removing a total of ",
                          format(length(genesToRemove), big.mark = ",", decimal.mark = "."),
                          " genes and ",
                          format(sum(idxToRemove), big.mark = ",", decimal.mark = "."),
                          " binding sites.\n")
            this.ranges = this.ranges[!idxToRemove]
        }
        # inform user
        if (!quiet) warning(c(msg0, msg1))
    }

    # filter for genes with large count imbalances between both conditions
    if (isTRUE(balanceCondition)) {
        if (!veryQuiet) message("Balance filter ")
        # genes should be filtered for count differences between conditions
        # -> calculate sum of counts per gene and condition
        df.ratio = as.data.frame(mcols(this.ranges)) %>%
            select(all_of(match.geneID), starts_with("counts")) %>%
            pivot_longer(-all_of(match.geneID)) %>%
            mutate(type = sub(".*\\.(.*?)\\..*", "\\1", name)) %>%
            mutate(condition = sub(".*_", "", name)) %>%
            filter(type == "bg") %>%
            select(-c(name, type)) %>%
            dplyr::group_by_if(., is.character) %>%
            summarise(sum = sum(value), .groups = "keep") %>%
            pivot_wider(values_from = sum, names_from = condition) %>%
            mutate(total = rowSums(dplyr::across(dplyr::where(is.numeric)))) %>%
            ungroup() %>%
            dplyr::relocate(all_of(match.geneID), total, all_of(this.condition.reference), all_of(this.condition.comp))

        # ---
        # Store for plotting
        object@plotData$filterBsBackground$data.balanceCondition = df.ratio

        df.filter = data.frame(
            geneID = df.ratio %>% select(all_of(match.geneID)),
            this.condition.reference = ifelse(c(df.ratio[,3][[1]] / df.ratio[,2][[1]]) < balanceCondition.cutoff, TRUE, FALSE),
            his.condition.comp = ifelse(c(df.ratio[,4][[1]] / df.ratio[,2][[1]]) < balanceCondition.cutoff, TRUE, FALSE)
        )
        df.filter$both = df.filter[,2] | df.filter[,3]

        msg0 = paste0("Found ", format(sum(df.filter$both, na.rm = TRUE), big.mark = ".", decimal.mark = ","),
                      " genes with condition balance ratio not meeting thresholds.\n")

        genesToRemove = df.filter %>% filter(both) %>% pull(match.geneID)
        allGenes = mcols(this.ranges)[colnames(mcols(this.ranges)) == match.geneID][[1]]
        idxToRemove = allGenes %in% genesToRemove

        if (isTRUE(flag)) {
            # flag genes with balance ratio above thresholds
            # -> add additional meta column
            msg1 = paste0("Flagging a total of ",
                          format(length(genesToRemove), big.mark = ",", decimal.mark = "."),
                          " genes and ",
                          format(sum(idxToRemove), big.mark = ",", decimal.mark = "."),
                          " binding sites.\n")
            mcols(this.ranges)$flag.balance = idxToRemove
        }
        if (!isTRUE(flag)) {
            # remove genes with low counts
            msg1 = paste0("Removing a total of ",
                          format(length(genesToRemove), big.mark = ",", decimal.mark = "."),
                          " genes and ",
                          format(sum(idxToRemove), big.mark = ",", decimal.mark = "."),
                          " binding sites.\n")
            this.ranges = this.ranges[!idxToRemove]
        }
        # inform user
        if (!quiet) warning(c(msg0, msg1))
    }

    object = setRanges(object, this.ranges, quiet = quiet)

    # ---
    # Store function parameters in list
    optstr = list(
        minCounts = minCounts,
        minCounts.cutoff = minCounts.cutoff,
        balanceBackground = balanceBackground,
        balanceBackground.cutoff.bs = balanceBackground.cutoff.bs,
        balanceBackground.cutoff.bg = balanceBackground.cutoff.bg,
        balanceCondition = balanceCondition,
        balanceCondition.cutoff = balanceCondition.cutoff,
        match.geneID = match.geneID,
        flag = flag)
    object@params$filterBsBackground = optstr

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "filterBsBackground()", class = "transform",
        nIn = length(this.ranges.initial), nOut = length(this.ranges),
        per = paste0(round(length(this.ranges)/ length(this.ranges.initial), digits = 2)*100,"%"),
        options = paste0("minCounts=", optstr$minCounts,
                         ", source=", optstr$datasource,
                         ", minCounts.cutoff=", optstr$minCounts.cutoff,
                         ", balanceBackground=", optstr$balanceBackground,
                         ", balanceBackground.cutoff.bs=", optstr$balanceBackground.cutoff.bs,
                         ", balanceBackground.cutoff.bg=", optstr$balanceBackground.cutoff.bg,
                         ", balanceCondition=", optstr$balanceCondition,
                         ", balanceCondition.cutoff=", optstr$balanceCondition.cutoff,
                         ", match.geneID=", optstr$generate.genmatch.geneIDeID.bs,
                         ", flag=", optstr$flag)
    )
    object@results = rbind(object@results, resultLine)

    return(object)
}






#' Compute fold-changes per binding site
#'
#' Given count data for binding sites and background regions this function will
#' compute fold-changes between two condition for each binding site. Computation
#' is based on \code{\link[DESeq2]{DESeq}} using the Likelihood ratio test to
#' disentangle transcription level changes from binding site level changes.
#'
#' Fold-changes per binding sites are corrected for transcript level changes.
#' Essentially, background counts are used to model transcript level changes,
#' which are then used to compute fold-changes per binding site, which are
#' corrected for the observed transcript level changes. This is done by using a
#' Likelihood ratio test comparing the full model
#' (~condition + type + condition:type) to the reduced model (~condition + type).
#'
#' Fold-changes for the transcript level changes are modeled explicitly in a
#' second round of the \code{\link[DESeq2]{DESeq}} workflow, using the default
#' Wald test to compare changes between the conditions (~condition). Counts
#' attributed to binding sites are removed from the gene level counts.
#'
#' Results from both calculation rounds can be filtered and further manipulated
#' with parameters given in the DESeq2 framework
#' (see \code{\link[DESeq2]{results}}, \code{\link[DESeq2]{lfcShrink}}).
#'
#' This function is intended to be used right after a call of \code{\link{filterBsBackground}}.
#'
#' @param object a \code{\link{BSFDataSet}} object
#' @param fitType either "parametric", "local", "mean", or "glmGamPoi" for the
#' type of fitting of dispersions to the mean intensity.
#' See \code{\link[DESeq2]{DESeq}} for more details.
#' @param sfType either "ratio", "poscounts", or "iterate" for the type of size
#' factor estimation. See \code{\link[DESeq2]{DESeq}} for more details.
#' @param minReplicatesForReplace the minimum number of replicates required in
#' order to use replaceOutliers on a sample. See \code{\link[DESeq2]{DESeq}} for
#' more details.
#' @param independentFiltering logical, whether independent filtering should be
#' applied automatically. See \code{\link[DESeq2]{results}} for more details.
#' @param alpha the significance cutoff used for optimizing the independent
#' filtering. See \code{\link[DESeq2]{results}} for more details.
#' @param pAdjustMethod he method to use for adjusting p-values. See
#' \code{\link[DESeq2]{results}} for more details.
#' @param minmu lower bound on the estimated count. See
#' \code{\link[DESeq2]{results}} for more details.
#' @param filterFun an optional custom function for performing independent
#' filtering and p-value adjustment. See \code{\link[DESeq2]{results}} for more
#' details.
#' @param use.lfc.shrinkage logical; whether to compute shrunken log2 fold
#' changes for the DESeq results. See \code{\link[DESeq2]{lfcShrink}} for more
#' details.
#' @param type if 'ashr', 'apeglml' or 'normal' should be used for fold change
#' shrinkage. See \code{\link[DESeq2]{lfcShrink}} for more details.
#' @param svalue logical, should p-values and adjusted p-values be replaced with
#' s-values when using apeglm or ashr. See \code{\link[DESeq2]{lfcShrink}}
#' for more details.
#' @param apeAdapt logical, should apeglm use the MLE estimates of LFC to adapt
#' the prior. See \code{\link[DESeq2]{lfcShrink}} for more details.
#' @param apeMethod what method to run apeglm, which can differ in terms of
#' speed. See \code{\link[DESeq2]{lfcShrink}} for more details.
#' @param match.geneID character; the name of the column with the gene ID
#' in the binding sites meta columns used for matching binding sites to genes
#' @param quiet logical; whether to print messages or not
#' @param veryQuiet logical; whether to print messages or not
#' @param replaceNegative logical; force negative counts to be replaces by 0.
#' Be careful when using this, having negative counts can point towards problems
#' with the gene annotation in use.
#' @param removeNA logical; force binding sites with any NA value to be removed.
#' Be careful when using this, having negative counts can point towards problems
#' with the gene annotation in use.
#'
#' @return a \code{\link{BSFDataSet}} object, with results from the
#' \code{\link[DESeq2]{DESeq}} analysis added to the meta columns of the
#' binding site ranges.
#'
#' @seealso \code{\link{calculateBsBackground}},
#' \code{\link{filterBsBackground}},
#' \code{\link{plotBsBackgroundFilter}},
#' \code{\link[DESeq2]{DESeq}}
#'
#' @importFrom stats relevel
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#'
#' # make example dataset
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = imputeBsDifferencesForTestdata(bds)
#' bds = calculateBsBackground(bds, anno.genes = gns)
#'
#' # calculate fold changes - no shrinkage
#' bds = calculateBsFoldChange(bds)
#'
#' # calculate fold changes - with shrinkage
#' bds = calculateBsFoldChange(bds, use.lfc.shrinkage = TRUE)
#'
#' @export
calculateBsFoldChange <- function(object,
                                  # for DESeq
                                  fitType = "local",
                                  sfType = "ratio",
                                  minReplicatesForReplace = 10,
                                  # for results
                                  independentFiltering = TRUE,
                                  alpha = 0.05,
                                  pAdjustMethod = "BH",
                                  minmu = 0.5,
                                  filterFun = NULL,
                                  # for lfcShrink
                                  use.lfc.shrinkage = FALSE,
                                  type = c("ashr", "apeglm", "normal"),
                                  svalue = FALSE,
                                  apeAdapt = TRUE,
                                  apeMethod = "nbinomCR",
                                  # general
                                  match.geneID = "geneID",
                                  quiet = TRUE,
                                  veryQuiet = FALSE,
                                  replaceNegative = FALSE,
                                  removeNA = FALSE
                                  # forceThis = FALSE
){
    # Bind locale variables
    geneID <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))
    stopifnot(is.logical(veryQuiet))
    stopifnot(is.logical(use.lfc.shrinkage))

    # check if shrinkage type is set
    if (isTRUE(use.lfc.shrinkage)) {
        type = match.arg(type, choices = c("ashr", "apeglm", "normal"))
    }

    this.meta = getMeta(object)
    # check for correct pairwise setting in meta data
    if (length(levels(this.meta$condition)) != 2) {
        msg0 = paste0("Fold-changes can only be computed for pairwise comparisons.\n")
        if (length(levels(this.meta$condition)) < 2) {
            # only one condition found
            msg1 = paste0("Found that only one condition is present. There is nothing to compare.\n")
            msg2 = paste0("Please provide two conditions that can be compared. \n")
        }
        if (length(levels(this.meta$condition)) > 2) {
            msg1 = paste0("Found more than 2 conditions. Too many comparisons.\n")
            msg2 = paste0("Please split mulitple comparisons into sets of pairwise comparisions.\n")
        }
        stop(c(msg0,msg1,msg2))
    }

    this.ranges = getRanges(object)

    # check if ranges have geneID present
    if (!match.geneID %in% colnames(mcols(this.ranges))) {
        msg0 = paste0("No matching column for gene ID found with name: '", match.geneID, "' in binding site meta data.\n")
        msg1 = paste0("Please run calculateBsBackground() first. \n")
        stop(c(msg0, msg1))
    }

    # check if ranges have counts present
    if (!any(grepl("counts", colnames(as.data.frame(mcols(this.ranges)))))) {
        # meta data has no columns named counts
        # -> stop here; point user to calculateBsBackground
        msg0 = paste0("No associated fields with crosslink counts found for present binding sites. \n")
        msg1 = paste0("Make sure to run 'calculateBsBackground()' first. \n")
        stop(msg0,msg1)
    }

    # check if ranges have at least two samples per condition in counts
    this.counts = as.data.frame(mcols(this.ranges)) %>% select(contains("counts"))
    # check binding sites
    this.counts.bs = this.counts %>% select(contains("bs"))
    if (length(this.counts.bs) == 0) {
        msg0 = paste0("No counts for binding sites present. \n")
        msg1 = paste0("Make sure to run 'calculateBsBackground()' first. \n")
        stop(c(msg0,msg1))
    }
    this.counts.bs.samples = as.data.frame(table(sub(".*_", "", colnames(this.counts.bs))))
    if (!all(this.counts.bs.samples[,2] > 1)) {
        # found less than two replicates per condition for bs
        msg0 = paste0("Less than two replicates are present for each condition. \n")
        msg1 = paste0("Found conditions: ", paste(this.counts.bs.samples[,1], collapse = ", "),
                      " with counts: ", paste(this.counts.bs.samples[,2], collapse = ", "), "\n")
        stop(c(msg0, msg1))
    }
    # check background
    this.counts.bg = this.counts %>% select(contains("bg"))
    if (length(this.counts.bg) == 0) {
        msg0 = paste0("No counts for background present. \n")
        msg1 = paste0("Make sure to run 'calculateBsBackground()' first. \n")
        stop(c(msg0,msg1))
    }
    this.counts.bg.samples = as.data.frame(table(sub(".*_", "", colnames(this.counts.bg))))
    if (!all(this.counts.bg.samples[,2] > 1)) {
        # found less than two replicates per condition for bs
        msg0 = paste0("Less than two replicates are present for each condition. \n")
        msg1 = paste0("Found conditions: ", paste(this.counts.bg.samples[,1], collapse = ", "),
                      " with counts: ", paste(this.counts.bg.samples[,2], collapse = ", "), "\n")
        stop(c(msg0, msg1))
    }

    # TODO
    # ----------------------------------------------------------------
    # check for NA values
    idx = which(is.na(this.counts), arr.ind = TRUE)
    if (nrow(idx) > 0) {
        # found rows with NA values as counts in matrix
        # -> prompt user towards removing or checking
        msg0 = paste0(format(nrow(idx), big.mark = ",", decimal.mark = "."),
                      " ranges found with NA values.\n")
        if (isTRUE(removeNA)) {
            # force remove NA cases
            this.counts = this.counts[-idx[,1],]
            this.ranges = this.ranges[-idx[,1]]
            msg1 = paste0("Removed all cases.\n")
            if (!veryQuiet) warning(c(msg0, msg1))
        } else {
            msg1 = paste0("This can point towards problems with the gene annoation.\n")
            msg2 = paste0("You can either
                      (1) find problematic genes and fix the cause, or
                      (2) set 'removeNA=TRUE' to force remove those cases.\n")
            stop(c(msg0,msg1,msg2))
        }
    }

    # check for negative values
    idx = which(this.counts < 0, arr.ind = TRUE)
    if (nrow(idx) > 0) {
        # found rows with negative values as counts in matrix
        # -> prompt user towards removing or checking
        msg0 = paste0(format(nrow(idx), big.mark = ",", decimal.mark = "."),
                      " ranges found with negative values.\n")
        if (isTRUE(replaceNegative)) {
            # force replace negative counts
            this.counts[idx] = 0
            msg1 = paste0("Replaced all cases with 0.\n")
            if (!veryQuiet) warning(c(msg0, msg1))
        } else {
            msg1 = paste0("This can point towards problems with the gene annoation.\n")
            msg2 = paste0("You can either
                      (1) find problematic genes and fix the cause, or
                      (2) set 'replaceNegative=TRUE' to force them to 0.\n")
            stop(c(msg0,msg1,msg2))
        }
    }

    # if (isTRUE(forceThis)) {
    #     # TODO check if there are negative counts
    #     # -> temporary hack to replace negative counts with 0
    #     # -> this happens in rare instances where offset ranges of neighboring
    #     #    binding sites overlap, which cause crosslinks to be counted twice.
    #     #    On genes with low counts this could result in negative values for the
    #     #    background.
    #     # -> Optimal solution is to use a combination of `reduce + with.revmap` and
    #     #    `disjoin` to split overlapping offset ranges, which avoids the problem
    #     idx = which(this.counts < 0, arr.ind = TRUE)
    #     if (nrow(idx) > 0) {
    #         this.counts[idx] = 0
    #         msg0 = paste0(format(nrow(idx), big.mark = ",", decimal.mark = "."),
    #                       " ranges found with negative counts, replacing them with 0.\n")
    #         if (!veryQuiet) warning(c(msg0))
    #     }
    #
    #     # TODO
    #     # quick fix for NA values
    #     idx = which(is.na(this.counts), arr.ind = TRUE)
    #     if (nrow(idx) > 0) {
    #         this.counts = this.counts[-idx[,1],]
    #         this.ranges = this.ranges[-idx[,1]]
    #
    #         if (!veryQuiet) warning(c(msg0))
    #     }
    # }

    # ----------------------------------------------------------------
    # TODO



    # --------------------------------------------------------------------------
    # MAIN 1) - Testing binding sites
    # --------------------------------------------------------------------------
    if (!veryQuiet) message("1) Testing binding sites ")

    # construct coldata
    # -----------------
    coldata = rbind.data.frame(this.meta, this.meta)
    coldata$clPlus = NULL
    coldata$clMinus = NULL
    coldata$type = as.factor(c(rep("bs", nrow(this.meta)), rep("bg", nrow(this.meta))))

    # make counts matrix and pass on to deseq
    # ---------------------------------------
    count.matr = this.counts %>% select(contains(c("bs", "bg"))) %>% as.matrix()
    se = SummarizedExperiment::SummarizedExperiment(assays = list(counts = count.matr),
                                                    rowRanges = granges(this.ranges),
                                                    colData = coldata)
    mode(SummarizedExperiment::assay(se)) = "integer"
    dds = DESeq2::DESeqDataSet(se, design = ~condition + type + condition:type)

    # manage model parametrisation through factor levels
    # -> reference level is taken from meta data
    # -> is the first level from condition
    reference.level = levels(this.meta$condition)[1] # WT
    dds$condition = stats::relevel(dds$condition, reference.level)
    dds$type = stats::relevel(dds$type, "bg")

    # running the deseq2 test
    # -----------------------
    dds = DESeq2::DESeq(dds, test="LRT", reduced = ~ condition + type,
                        fitType = fitType,
                        sfType = sfType,
                        quiet = quiet,
                        minReplicatesForReplace = minReplicatesForReplace,
                        modelMatrixType = "standard",
                        useT = FALSE)

    # pulling deseq2 results
    # ----------------------
    comp.level = levels(this.meta$condition)[2] # KO
    this.contrast = paste0("condition", comp.level, ".typebs")

    if (!this.contrast %in% DESeq2::resultsNames(dds)) {
        msg0 = paste0("Contrast named: ", this.contrast, "not found in available names: ", paste(DESeq2::resultsNames(dds), collapse = ", "), ".\n")
        msg1 = paste0("Please check the factor levels of the 'condition' column in your meta data.\n")
        stop(c(msg0, msg1))
    }

    if (!is.null(filterFun)) {
        res = DESeq2::results(dds, name = this.contrast,
                              lfcThreshold = 0,
                              altHypothesis = "greaterAbs",
                              independentFiltering = independentFiltering,
                              alpha = alpha, pAdjustMethod = pAdjustMethod,
                              filterFun = filterFun, format = "DataFrame",
                              minmu = minmu)
    } else {
        res = DESeq2::results(dds, name = this.contrast,
                              lfcThreshold = 0,
                              altHypothesis = "greaterAbs",
                              independentFiltering = independentFiltering,
                              alpha = alpha, pAdjustMethod = pAdjustMethod,
                              format = "DataFrame", minmu = minmu)
    }

    # apply fold change shrinkage if needed
    # -> requiers additional packages to be installed
    if (isTRUE(use.lfc.shrinkage)) {
        res = DESeq2::lfcShrink(dds = dds, res = res, coef = this.contrast,
                                type = type, lfcThreshold = 0,
                                svalue = svalue, returnList = FALSE,
                                format = "DataFrame", saveCols = NULL,
                                apeAdapt = apeAdapt, apeMethod = apeMethod,
                                quiet = quiet)
    }

    # match with ranges
    colnames(res) = paste0('bs.', colnames(res))
    mcols(this.ranges) = cbind.data.frame(mcols(this.ranges), res)


    # --------------------------------------------------------------------------
    # MAIN 2) - Testing the background
    # --------------------------------------------------------------------------
    if (!veryQuiet) message("2) Testing genes ")

    # construct coldata
    # -----------------
    coldata.bg = this.meta
    coldata.bg$clPlus = NULL
    coldata.bg$clMinus = NULL

    # make counts matrix and pass on to deseq
    # ---------------------------------------
    count.matr.bg = as.data.frame(count.matr)
    count.matr.bg$geneID = mcols(this.ranges)[colnames(mcols(this.ranges)) == match.geneID][[1]]
    count.matr.bg = count.matr.bg %>% select(geneID, starts_with("counts.bg"))
    count.matr.bg = count.matr.bg[!duplicated(count.matr.bg$geneID),]
    rownames(count.matr.bg) = count.matr.bg$geneID
    count.matr.bg$geneID = NULL

    se.bg = SummarizedExperiment::SummarizedExperiment(assays = list(counts = as.matrix(count.matr.bg)),
                                                       colData = coldata.bg)
    mode(SummarizedExperiment::assay(se.bg)) = "integer"
    dds.bg = DESeq2::DESeqDataSet(se.bg, design = ~ condition)
    dds.bg$condition = stats::relevel(dds.bg$condition, reference.level)

    # running the deseq2 test
    # -----------------------
    dds.bg = DESeq2::DESeq(dds.bg, test = "Wald",
                           fitType = fitType,
                           sfType = sfType,
                           quiet = quiet,
                           minReplicatesForReplace = minReplicatesForReplace,
                           modelMatrixType = "standard",
                           useT = FALSE)


    # pulling deseq2 results
    # ----------------------
    this.contrast.bg = paste0("condition_", comp.level, "_vs_", reference.level)

    if (!is.null(filterFun)) {
        res.bg = DESeq2::results(dds.bg, name = this.contrast.bg,
                                 lfcThreshold = 0,
                                 altHypothesis = "greaterAbs",
                                 independentFiltering = independentFiltering,
                                 alpha = alpha, pAdjustMethod = pAdjustMethod,
                                 filterFun = filterFun, format = "DataFrame",
                                 minmu = minmu)
    } else {
        res.bg = DESeq2::results(dds.bg, name = this.contrast.bg,
                                 lfcThreshold = 0,
                                 altHypothesis = "greaterAbs",
                                 independentFiltering = independentFiltering,
                                 alpha = alpha, pAdjustMethod = pAdjustMethod,
                                 format = "DataFrame", minmu = minmu)
    }


    # apply fold change shrinkage if needed
    # -> requiers additional packages to be installed
    if (isTRUE(use.lfc.shrinkage)) {
        res.bg = DESeq2::lfcShrink(dds = dds.bg, res = res.bg, coef = this.contrast.bg,
                                   type = type, lfcThreshold = 0,
                                   svalue = svalue, returnList = FALSE,
                                   format = "DataFrame", saveCols = NULL,
                                   apeAdapt = apeAdapt, apeMethod = apeMethod,
                                   quiet = quiet)
    }

    # match with binding sites
    match.bs.ranges = mcols(this.ranges)[colnames(mcols(this.ranges)) == match.geneID][[1]]
    colnames(res.bg) = paste0('bg.', colnames(res.bg))
    idx = match(match.bs.ranges, rownames(res.bg))
    mcols(this.ranges) = cbind.data.frame(mcols(this.ranges), res.bg[idx,])

    object = setRanges(object, this.ranges, quiet = quiet)

    # ---
    # Store function parameters in list
    optstr = list(fitType = fitType, sfType = sfType,
                  minReplicatesForReplace = minReplicatesForReplace,
                  lfcThreshold = 0,
                  independentFiltering = independentFiltering,
                  alpha = alpha, pAdjustMethod = pAdjustMethod,
                  minmu = minmu,
                  use.lfc.shrinkage = use.lfc.shrinkage,
                  type = type,
                  svalue = svalue, apeAdapt = apeAdapt,
                  apeMethod = apeMethod, match.geneID = match.geneID)
    object@params$calculateBsFoldChange = optstr

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "calculateBsFoldChange()", class = "transform",
        nIn = length(this.ranges), nOut = length(this.ranges),
        per = paste0(round(length(this.ranges)/ length(this.ranges), digits = 2)*100,"%"),
        options = paste0("Alpha=", optstr$alpha,
                         ", fitType=", optstr$fitType,
                         ", sfType=", optstr$sfType,
                         ", minReplicatesForReplace=", optstr$minReplicatesForReplace,
                         ", lfcThreshold=", optstr$lfcThreshold,
                         ", independentFiltering=", optstr$independentFiltering,
                         ", minmu=", optstr$minmu,
                         ", use.lfc.shrinkage=", optstr$use.lfc.shrinkage,
                         ifelse(length(optstr$type) == 1, paste0(", type=", optstr$type), ""),
                         ", svalue=", optstr$svalue,
                         ", apeAdapt=", optstr$apeAdapt,
                         ", apeMethod=", optstr$apeMethod,
                         ", match.geneID=", optstr$match.geneID)
    )
    object@results = rbind(object@results, resultLine)

    return(object)
}


# ------------------------------------------------------------------------------
# Exported helper
# ------------------------------------------------------------------------------


#' Impute artificial differences in the example data set
#'
#' A function that works only on the test data set provided with the package. It
#' is used for internal testing and the making of examples to showcase the
#' differential binding functions.
#'
#' Differences between samples are artificially introduced by removing the
#' signal on a random set of binding sites of the input.
#'
#' @param object a \code{\link{BSFDataSet}} object; explicitly the test data set
#' from the extdata folder
#' @param size numeric; the number of positions on which signal should be
#' deleted, counting from the start
#' @param change.per numeric; the percentage of ranges that should be effected
#' by the change.
#'
#' @return object a \code{\link{BSFDataSet}} object with the signal slot adapted
#' to reflect changes in binding between two artificial conditions
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#'
#' bds = makeBindingSites(bds, bsSize = 7)
#' bds = assignToGenes(bds, anno.genes = gns)
#'
#' bds = imputeBsDifferencesForTestdata(bds)
#'
#' @export
imputeBsDifferencesForTestdata <- function(object,
                                           size = 5,
                                           change.per = 0.1
){
    # manipulate meta data sample names
    this.meta = getMeta(object)
    this.meta$condition = factor(c("WT", "WT", "KO", "KO"), levels = c("WT", "KO"))

    # manipulate signal names
    s = getSignal(object)
    names(s$signalPlus) = paste0(this.meta$id, "_", this.meta$condition)
    names(s$signalMinus) = paste0(this.meta$id, "_", this.meta$condition)

    # manipulate ranges
    this.ranges = getRanges(object)
    set.seed(1234)
    diff.ranges = sample(this.ranges, size = length(this.ranges)* change.per)
    idx = sample(c(TRUE,FALSE), size = length(diff.ranges), replace = TRUE)
    diff.ranges.up = diff.ranges[idx]
    diff.ranges.down = diff.ranges[!idx]

    # manipulate signal
    vToKill.up = lapply(1:size, function(x){
        start(ranges(diff.ranges.up)) + x
    })
    vToKill.up = do.call(c, vToKill.up)
    s$signalPlus$`1_WT`$chr22[vToKill.up] = 0
    s$signalPlus$`2_WT`$chr22[vToKill.up] = 0
    s$signalMinus$`1_WT`$chr22[vToKill.up] = 0
    s$signalMinus$`2_WT`$chr22[vToKill.up] = 0

    vToKill.down = lapply(1:size, function(x){
        start(ranges(diff.ranges.down)) + x
    })
    vToKill.down = do.call(c, vToKill.down)
    s$signalPlus$`3_KO`$chr22[vToKill.down] = 0
    s$signalPlus$`4_KO`$chr22[vToKill.down] = 0
    s$signalMinus$`3_KO`$chr22[vToKill.down] = 0
    s$signalMinus$`4_KO`$chr22[vToKill.down] = 0

    object = setMeta(object, this.meta)
    object = setSignal(object, s)
    return(object)
}

test_that("calculateBsFoldChange works", {
    # load clip data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

    # make example data
    bds = makeBindingSites(bds, bsSize = 7)
    bds = assignToGenes(bds, anno.genes = gns)
    bds = imputeBsDifferencesForTestdata(bds)
    bds = expect_warning(calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE))

    c0 = expect_silent(calculateBsFoldChange(bds, veryQuiet = TRUE))

    expect_identical(length(getRanges(c0)), length(getRanges(bds)))

    rngMod = getRanges(bds)
    bdsNoCounts = setRanges(bds, granges(rngMod))
    expect_error(calculateBsFoldChange(bdsNoCounts))

    meta = getMeta(bds)
    meta$condition = factor(c(rep("WT",4)), levels = c("WT"))
    bdsChange = setMeta(bds, meta)
    expect_error(calculateBsFoldChange(bdsChange, veryQuiet = TRUE))

    meta = getMeta(bds)
    meta$condition = factor(c(rep("Ctrl",2), rep("Mut",2)), levels = c("Ctrl", "Mut"))
    bdsChange = setMeta(bds, meta)
    c1 = expect_silent(calculateBsFoldChange(bdsChange, veryQuiet = TRUE))
    expect_identical(getRanges(c0), getRanges(c1))

    meta$condition = factor(c(rep("Ctrl",2), rep("Mut",2)), levels = c("Mut", "Ctrl"))
    bdsChange = setMeta(bds, meta)
    c2 = expect_silent(calculateBsFoldChange(bdsChange, veryQuiet = TRUE))
    resNormal = getRanges(c1)
    resChanged = getRanges(c2)
    expect_equal(resNormal$bs.pvalue, resChanged$bs.pvalue)
    # expect_equal(round(resNormal$bs.log2FoldChange,1), round(resChanged$bs.log2FoldChange,1) * -1)
    # expect_equal(round(resNormal$bg.pvalue, 2), round(resChanged$bg.pvalue, 2))
    # expect_equal(round(resNormal$bg.log2FoldChange,1), round(resChanged$bg.log2FoldChange,1) * -1)

    meta$condition = factor(c(rep("Mut",2), rep("Ctrl",2)), levels = c("Mut", "Ctrl"))
    bdsChange = setMeta(bds, meta)
    c3 = expect_silent(calculateBsFoldChange(bdsChange, veryQuiet = TRUE))
    resNormal = getRanges(c1)
    resChanged = getRanges(c3)
    expect_equal(resNormal$bs.pvalue, resChanged$bs.pvalue)
    expect_equal(resNormal$bs.log2FoldChange, resChanged$bs.log2FoldChange)
    expect_equal(resNormal$bg.pvalue, resChanged$bg.pvalue)
    expect_equal(resNormal$bg.log2FoldChange, resChanged$bg.log2FoldChange)

    # negative counts
    rng = getRanges(bds)
    rng$counts.bs.1_WT[1] = -10
    bdsNeg = setRanges(bds, rng)
    expect_error(calculateBsFoldChange(bdsNeg))
    expect_error(calculateBsFoldChange(bdsNeg, removeNA = TRUE))
    expect_warning(calculateBsFoldChange(bdsNeg, replaceNegative = TRUE))

    # NA counts
    rng = getRanges(bds)
    rng$counts.bs.1_WT[1] = NA
    bdsNA = setRanges(bds, rng)
    expect_error(calculateBsFoldChange(bdsNA))
    expect_error(calculateBsFoldChange(bdsNA, replaceNegative = TRUE))
    expect_warning(calculateBsFoldChange(bdsNA, removeNA = TRUE))
})

test_that("filterBsBackground works", {
    # load clip data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

    # make testset
    bds = makeBindingSites(bds, bsSize = 7)
    bds = assignToGenes(bds, anno.genes = gns)
    bds = imputeBsDifferencesForTestdata(bds)
    bds = expect_warning(calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE))

    # tests
    c0 = expect_warning(filterBsBackground(bds, flag = FALSE))
    expect_lt(length(getRanges(c0)), length(getRanges(bds)))

    c1 = expect_warning(filterBsBackground(bds, flag = TRUE))
    expect_equal(length(getRanges(c1)), length(getRanges(bds)))
    expect_equal(ncol(mcols(getRanges(c1))), ncol(mcols(getRanges(bds)))+3)

    c2 = expect_warning(filterBsBackground(bds, flag = TRUE, minCounts = FALSE))
    c3 = expect_warning(filterBsBackground(bds, flag = TRUE, balanceBackground = FALSE))
    c4 = expect_warning(filterBsBackground(bds, flag = TRUE, balanceCondition = FALSE))

    expect_equal(ncol(mcols(getRanges(bds)))+2, ncol(mcols(getRanges(c2))))
    expect_equal(ncol(mcols(getRanges(bds)))+2, ncol(mcols(getRanges(c3))))
    expect_equal(ncol(mcols(getRanges(bds)))+2, ncol(mcols(getRanges(c4))))

    c5 = filterBsBackground(bds, flag = TRUE, balanceCondition = FALSE,
                            balanceBackground = FALSE, minCounts = FALSE)
    expect_equal(ncol(mcols(getRanges(bds))), ncol(mcols(getRanges(c5))))
})

#' @importFrom stats quantile
.selectQuantilesMultipleConditions <-
    function(covDf, userCond, userNreps, userCutoff) {
        # bind locally used variables
        per <- applyTo <- NULL

        # construct general cutoff matrix in 1% steps
        q = as.data.frame(apply(covDf, 2, function(x) {
            quantile(x, probs = seq(0, 1, by = 0.01))
        }))
        q$cut = seq(0, 1, by = 0.01)
        q$per = rownames(q)

        # select that part of q that was chosen by user
        qSel = q[q$cut %in% userCutoff, ]
        applyDf = data.frame(levels(userCond), userCutoff)
        # applyDf = data.frame(levels(userCond), userCutoff, userCond)

        idx = match(qSel$cut, applyDf$userCutoff)
        qSel$applyTo = applyDf$levels.userCond.[idx]
        qSel = qSel %>% pivot_longer(-c(cut, per, applyTo))
        qSel$sel = vapply(strsplit(qSel$name, "_"), `[`, 2,
                          FUN.VALUE = character(1))
        if (length(unique(userCutoff)) > 1) {
            qSel = qSel[qSel$applyTo == qSel$sel, ]
        }
        # add n.reps support to df
        nRepsDf = data.frame(defaultNreps = userNreps, applyTo = levels(userCond))
        idx = match(qSel$sel, nRepsDf$applyTo)
        qSel$defaultNreps = nRepsDf$defaultNreps[idx]
        return(qSel)
    }

#' @importFrom stats quantile
.selectQuantilesSingleCondtion <-
    function(covDf, userCond, userNreps, userCutoff) {
        # bind locally used variables
        per <- NULL

        # construct general cutoff matrix in 1% steps
        q = as.data.frame(apply(covDf, 2, function(x) {
            quantile(x, probs = seq(0, 1, by = 0.01))
        }))
        q$cut = seq(0, 1, by = 0.01)
        q$per = rownames(q)

        # select that part of q that was chosen by user
        qSel = q[q$cut %in% userCutoff, ]
        qSel = qSel %>% pivot_longer(-c(cut, per))
        qSel$sel = vapply(strsplit(qSel$name, "_"), `[`, 2,
                          FUN.VALUE = character(1))

        # add n.reps support to df
        nRepsDf = data.frame(defaultNreps = userNreps, applyTo = levels(userCond))
        idx = match(qSel$sel, nRepsDf$applyTo)
        qSel$defaultNreps = nRepsDf$defaultNreps[idx]
        qSel$value = round(qSel$value, digits = 0)
        return(qSel)
    }

#' @importFrom stats median
.computeSupportRatio <- function(object, flankSize){
    stopifnot(is(object, "BSFDataSet"))
    if(! length(flankSize) == 1) {
        stop("flankSize must be a single integer value. ")
    }
    if (! is.numeric(flankSize)) {
        stop("flankSize needs to be numeric. ")
    }
    if (round(flankSize) != flankSize) {
        stop("flankSize is not an integer. ")
    }

    c0 = rowSums(as.data.frame(mcols(coverageOverRanges(
        object, returnOptions = "merge_positions_keep_replicates",
        quiet = TRUE))))

    objMod = object
    objMod = suppressMessages(setRanges(objMod, getRanges(object) + flankSize))
    c1 = rowSums(as.data.frame(mcols(coverageOverRanges(
        objMod, returnOptions = "merge_positions_keep_replicates",
        quiet = TRUE))))

    # use 0.1 as pseudocount for the score
    score = c0 / (((c1 - c0) + 0.1) / 2)
    # report median over all binding sites
    score = median(score)
    return(score)
}

.resolveGeneOverlapsWithRule <- function(rng, ols, rule, selectID, selectName, selectType) {
    # initialize local variables
    bsIndex <- geneType <- choice <- NULL

    # takes a Hits object from the binding site and gene annotation matching
    # and a rule by which binding sites on multiple genes are matched
    # and the initial ranges
    # returns the initial binding site ranges with matched gene info

    # check if input is complete
    # -> note that selectType cannot be missing since then this option is not
    # available !!!
    if (is.null(selectID)) {
        # the gene_id is missing
        olsInfo = data.frame(geneIndex = queryHits(ols),
                             bsIndex = subjectHits(ols),
                             geneName = selectName[queryHits(ols)],
                             geneType = selectType[queryHits(ols)]) %>%
            group_by(bsIndex) %>%
            arrange(bsIndex) %>%
            mutate(choice = rule$idx[match(geneType, rule$gene_type)]) %>%
            arrange(choice, .by_group = TRUE) %>%
            slice_head(n = 1)
        # matching index
        idx = match(rng$currIdx, olsInfo$bsIndex)
        # handle intergenic cases
        rngIntergenic = rng[which(is.na(idx))]
        rngIntergenic$geneName = NA
        rngIntergenic$geneType = "Intergenic"
        # handle normal cases
        rngRest = rng[countOverlaps(rng, rngIntergenic) == 0]
        rngRest$geneName[idx[!is.na(idx)]] = olsInfo$geneName
        rngRest$geneType[idx[!is.na(idx)]] = olsInfo$geneType
    }
    if (is.null(selectName)) {
        # the gene_name is missing
        olsInfo = data.frame(geneIndex = queryHits(ols),
                             bsIndex = subjectHits(ols),
                             geneID = selectID[queryHits(ols)],
                             geneType = selectType[queryHits(ols)]) %>%
            group_by(bsIndex) %>%
            arrange(bsIndex) %>%
            mutate(choice = rule$idx[match(geneType, rule$gene_type)]) %>%
            arrange(choice, .by_group = TRUE) %>%
            slice_head(n = 1)
        # matching index
        idx = match(rng$currIdx, olsInfo$bsIndex)
        # handle intergenic cases
        rngIntergenic = rng[which(is.na(idx))]
        rngIntergenic$geneID = NA
        rngIntergenic$geneType = "Intergenic"
        # handle normal cases
        rngRest = rng[countOverlaps(rng, rngIntergenic) == 0]
        rngRest$geneID[idx[!is.na(idx)]] = olsInfo$geneID
        rngRest$geneType[idx[!is.na(idx)]] = olsInfo$geneType
    }
    if (! is.null(selectID) & ! is.null(selectName)) {
        # everything is in place
        olsInfo = data.frame(geneIndex = queryHits(ols),
                             bsIndex = subjectHits(ols),
                             geneID = selectID[queryHits(ols)],
                             geneName = selectName[queryHits(ols)],
                             geneType = selectType[queryHits(ols)]) %>%
            group_by(bsIndex) %>%
            arrange(bsIndex) %>%
            mutate(choice = rule$idx[match(geneType, rule$gene_type)]) %>%
            arrange(choice, .by_group = TRUE) %>%
            slice_head(n = 1)
        # matching index
        idx = match(rng$currIdx, olsInfo$bsIndex)
        # handle intergenic cases
        rngIntergenic = rng[which(is.na(idx))]
        rngIntergenic$geneID = NA
        rngIntergenic$geneName = NA
        rngIntergenic$geneType = "Intergenic"
        # handle normal cases
        rngRest = rng[countOverlaps(rng, rngIntergenic) == 0]
        rngRest$geneID[idx[!is.na(idx)]] = olsInfo$geneID
        rngRest$geneName[idx[!is.na(idx)]] = olsInfo$geneName
        rngRest$geneType[idx[!is.na(idx)]] = olsInfo$geneType
    }

    rng = .sortRanges(c(rngIntergenic, rngRest))

    return(rng)
}


# approximate by one round of inital binding site merging
# approximate by center given by highest number of stacked crosslinks
.approximateBindingSites_medium <- function(rng, sgn, bsSize, minWidth) {
    # approximate binding sites by first iteration of merging
    res = .mergeCrosslinkSites(
        rng = rng,
        sgn = sgn,
        bsSize = bsSize,
        minWidth = minWidth,
        computeOption = "simple"
    )
    rng = res$rng
    rng$bsSize = bsSize
    return(rng)
}

# approximate center by highest pureclip score
#' @importFrom IRanges extractList
.approximateBindingSites_coarse <- function(rng, bsSize, minWidth) {
    # initialize variables locally
    i <- group <- NULL

    ### Merge peaks for given bs size
    redRegion = reduce(rng, min.gapwidth = 1, with.revmap = TRUE)

    redRegion = redRegion[width(redRegion) >= minWidth]
    mcols(redRegion)$id = seq_along(redRegion)

    scoresPerRedRegion = extractList(mcols(rng), redRegion$revmap)
    names(scoresPerRedRegion) = seq_along(scoresPerRedRegion)

    scoresPerRedRegion = unlist(scoresPerRedRegion)
    scoresPerRedRegion$group = rownames(scoresPerRedRegion)

    idx = scoresPerRedRegion %>% as.data.frame() %>%
        group_by(group) %>% summarise(i = which.max(score)) %>%
        mutate(group = as.numeric(group)) %>% arrange(group)

    # Convert the GRanges object to a data frame
    redRegionDf <- data.frame(seqnames = seqnames(redRegion),
                              start = start(redRegion),
                              end = end(redRegion),
                              strand = strand(redRegion),
                              group = redRegion$id,
                              i = idx$i) %>%
        mutate(center = start + i -1)

    rngNew = GRanges(seqnames = redRegionDf$seqnames, ranges = IRanges::IRanges(start = redRegionDf$center, width = floor(bsSize/2)),
                     strand = redRegionDf$strand, bsSize = bsSize)
    return(rngNew)
}

.findFirstMaximum <- function(x, est.minimumStepGain){

    est.option <- df.global <- bsFirstDrop <- idxFirstDrop <- bsSize <- NULL
    localEstimates <- est.option <- est.bsSize <- est.geneFilter <- NULL
    signalToFlankRatio <- ms <- growth_per <- increaseOverMin <- geneWiseFilter <- NULL


    df.global = x %>%
        group_by(bsSize) %>%
        summarise(ms = mean(signalToFlankRatio), sd = sd(signalToFlankRatio)) %>%
        mutate(geneWiseFilter = "mean") %>%
        mutate(growth_per = (ms / dplyr::lag(ms)) - 1) %>%
        mutate(increase = growth_per > 0) %>%
        mutate(increaseOverMin = (growth_per) >= est.minimumStepGain) %>%
        slice(-1)

    # check for minimum increase
    if (all(!df.global$increaseOverMin)) {
        # no size leads to score increase larger than est.minimumStepGain
        # -> trigger local estimation
        localEstimates = .findFirstMaximum_local(x, est.minimumStepGain)
        est.option = localEstimates$option
        est.bsSize = localEstimates$est.bsSize
        est.geneFilter = localEstimates$est.geneFilter

    } else {
        # at lease one size leads to score increase larger than est.minimumStepGain
        # -> trigger global estimation

        # finde die zeile in der ein TRUE auf ein FALSE folgt
        bsFirstDrop = df.global %>%
            filter(increaseOverMin == FALSE & dplyr::lag(increaseOverMin, default = TRUE) == TRUE) %>%
            pull(bsSize)

        # check for no possible maximum
        if (length(bsFirstDrop) == 0) {
            est.option = "error"
            res = list(est.bsSize = est.bsSize, est.geneFilter = est.geneFilter, est.option = est.option)
            return(res)
        } else {
            bsFirstDrop = max(bsFirstDrop)
        }


        # firstMax = which(df.global$increaseOverMin == TRUE)[1]
        # df.current = df.global %>% slice(1:firstMax)

        # firstDrop = which(df.global$increaseOverMin == FALSE)[1]

        if (is.na(bsFirstDrop)) {
            localEstimates = .findFirstMaximum_local(x, est.minimumStepGain)
            est.option = localEstimates$option
            est.bsSize = localEstimates$est.bsSize
            est.geneFilter = localEstimates$est.geneFilter

        } else {
            # pull bsSize
            idxFirstDrop = which(df.global$bsSize == bsFirstDrop)
            est.bsSize = df.global %>% slice(idxFirstDrop-1) %>% pull(bsSize)

            est.geneFilter = x %>%
                filter(bsSize == est.bsSize & signalToFlankRatio >= df.global$ms[df.global$bsSize == est.bsSize]) %>%
                arrange(geneWiseFilter) %>%
                slice_head(n = 1) %>%
                pull(geneWiseFilter)
            # set option
            est.option = "global"
        }

    }

    res = list(est.bsSize = est.bsSize, est.geneFilter = est.geneFilter, est.option = est.option)

    return(res)
}

.findFirstMaximum_local <- function(x, est.minimumStepGain) {

    df.local <- option <- est.bsSize <- est.geneFilter <- NULL
    geneWiseFilter <- signalToFlankRatio <- growth_per <- increase <- bsSize <- NULL

    df.local = x %>%
        group_by(geneWiseFilter) %>%
        mutate(growth_per = (signalToFlankRatio / dplyr::lag(signalToFlankRatio)) - 1) %>%
        mutate(increase = growth_per > 0) %>%
        mutate(increaseOverMin = (growth_per) >= est.minimumStepGain) %>%
        filter(!is.na(growth_per))

    if (any(df.local$increase)) {
        # check if at any point an increase of the curve was found
        # -> if so take it as anchor
        df.curr = df.local %>%
            filter(increase == TRUE) %>%
            ungroup() %>%
            filter(growth_per == max(growth_per))
        # pull bsSize
        est.bsSize = df.curr %>% pull(bsSize)
        # pull geneWiseFilter
        est.geneFilter = df.curr %>% pull(geneWiseFilter)
        # set option
        option = "local"
    }
    if (all(!df.local$increase)) {
        # check if there is not a single increase in the curve at any point
        # -> if so use the overall maximum as anchor
        # pull bsSize
        est.bsSize = df.local$bsSize[which.max(df.local$signalToFlankRatio)]
        # pull geneWiseFilter
        est.geneFilter = df.local$geneWiseFilter[which.max(df.local$signalToFlankRatio)]
        # set option
        option = "fallback"
    }
    # return results
    res = list(option = option, est.bsSize = est.bsSize, est.geneFilter = est.geneFilter)
    return(res)
}

.calcNormalizeFactors <- function(object, rng, trl, normalize.exclude.lower,
                                  normalize.exclude.upper){

    # ---
    # Checks
    # ---
    # compute normalization factors for only those regions that actually have binding sites
    this.trl = trl[toupper(names(trl)) %in% toupper(unique(rng$transcriptRegion))]

    # ---
    # Hosting
    # ---
    # -> compute width for all ranges that overlap with at least one binding site
    w.hosting = lapply(seq_along(this.trl), function(x){
        # get all relevant width
        curr.width = width(reduce(IRanges::subsetByOverlaps(this.trl[[x]], rng)))
        # find upper and lower boundary cutoffs
        currIdx = findInterval(curr.width, vec = quantile(x = curr.width, probs = seq(from = 0, to = 1, by = 0.01)), all.inside = TRUE)
        seqs = seq(from = 0.01, to = 1, by = 0.01)
        names(seqs) = seq_along(seqs)
        idxRemoveLower = as.numeric(names(seqs[seqs <= normalize.exclude.lower]))
        idxRemoveUpper = as.numeric(names(seqs[seqs >= 1-normalize.exclude.upper+0.01]))
        # apply cutoffs
        curr.w.reduced = curr.width[which(!currIdx %in% c(idxRemoveLower, idxRemoveUpper))]
        # convert to normalization values
        width.return = c(mean(curr.w.reduced), median(curr.w.reduced), sum(curr.w.reduced))
        return(width.return)
    })
    w.hosting = do.call(rbind, w.hosting)
    colnames(w.hosting) = c("norm.hosting.mean", "norm.hosting.median", "norm.hosting.sum")
    w.hosting = as.data.frame(w.hosting)
    # set names
    w.hosting$TranscriptRegion = names(this.trl)

    w.total = cbind.data.frame(w.hosting)
    return(w.total)
}


.matchMetaData <- function(d){
    # assign local variables
    geneID <- geneName <- geneType <- transcriptRegion <- NULL
    dataset <- signalToFlankRatio <- NULL


    # assign meta data if present
    if ("geneID" %in% colnames(d)) {
        merge.geneID = d %>%
            summarize(geneID = toString(unique(geneID))) %>%
            mutate(geneID = ifelse(geneID == "NA", NA, geneID))
    } else {
        merge.geneID = NA
    }
    if ("geneName" %in% colnames(d)) {
        merge.geneName = d %>%
            summarize(geneName = toString(unique(geneName))) %>%
            mutate(geneName = ifelse(geneName == "NA", NA, geneName))
    } else {
        merge.geneName = NA
    }
    if ("geneType" %in% colnames(d)) {
        merge.geneType = d %>%
            summarize(geneType = toString(unique(geneType))) %>%
            mutate(geneType = ifelse(geneType == "NA", NA, geneType))
    } else {
        merge.geneType = NA
    }
    if ("transcriptRegion" %in% colnames(d)) {
        merge.transcriptRegion = d %>%
            summarize(transcriptRegion = toString(unique(transcriptRegion))) %>%
            mutate(transcriptRegion = ifelse(transcriptRegion == "NA", NA, transcriptRegion))
    } else {
        merge.transcriptRegion = NA
    }
    if ("dataset" %in% colnames(d)) {
        merge.dataset = d %>%
            summarize(dataset = toString(unique(dataset))) %>%
            mutate(dataset = ifelse(dataset == "NA", NA, dataset))
    } else {
        merge.dataset = NA
    }
    if ("score" %in% colnames(d)) {
        merge.score = d %>%
            summarize(score = max(score)) %>%
            mutate(score = ifelse(score == "NA", NA, score))
    } else {
        merge.score = NA
    }
    if ("signalToFlankRatio" %in% colnames(d)) {
        merge.signalToFlankRatio = d %>%
            summarize(signalToFlankRatio = max(signalToFlankRatio)) %>%
            mutate(signalToFlankRatio = ifelse(signalToFlankRatio == "NA", NA, signalToFlankRatio))
    } else {
        merge.signalToFlankRatio = NA
    }

    total.meta = cbind(merge.geneID, merge.geneName, merge.geneType,
                       merge.transcriptRegion, merge.dataset,
                       merge.score, merge.signalToFlankRatio) %>%
        dplyr::distinct_all() %>%
        dplyr::select_if(~sum(!is.na(.)) > 0)
    return(total.meta)
}


.coverageForDifferential <- function(object, ranges, prefix, quiet, match.rangeID){
    this.object = setRanges(object, ranges, quiet = quiet)
    cov = clipCoverage(this.object,
                       match.rangeID = match.rangeID,
                       ranges.merge = TRUE, positions.merge = FALSE,
                       samples.merge = FALSE, out.format = "granges",
                       out.format.overwrite = TRUE, quiet = quiet)
    colnames(mcols(cov)) = paste0(prefix, colnames(mcols(cov)))
    mcols(cov) = cbind(mcols(ranges), mcols(cov))
    return(cov)
}

#' @rdname getRanges
#' @export
setGeneric("getRanges", function(object) {
    standardGeneric("getRanges")
})

#' @rdname getMeta
#' @export
setGeneric("getMeta", function(object) {
    standardGeneric("getMeta")
})

#' @rdname setMeta
#' @export
setGeneric("setMeta", function(object, ...) {
    standardGeneric("setMeta")
})

#' @rdname getSignal
#' @export
setGeneric("getSignal", function(object) {
    standardGeneric("getSignal")
})

#' @rdname setRanges
#' @export
setGeneric("setRanges", function(object, ...) {
    standardGeneric("setRanges")
})

#' @rdname setSignal
#' @export
setGeneric("setSignal", function(object, ...) {
    standardGeneric("setSignal")
})

#' @rdname setSummary
#' @export
setGeneric("setSummary", function(object, ...) {
    standardGeneric("setSummary")
})

#' @rdname getSummary
#' @export
setGeneric("getSummary", function(object, ...) {
    standardGeneric("getSummary")
})

#' @rdname setName
#' @export
setGeneric("setName", function(object, ...) {
    standardGeneric("setName")
})

#' @rdname getName
#' @export
setGeneric("getName", function(object) {
    standardGeneric("getName")
})

#' Coverage function for BSFDataSet objects
#'
#' Function that computes a crosslink coverage with all samples over all ranges
#' given in the \code{\link{BSFDataSet}}. The coverage can be summarized over
#' all combinations of the three dimension (samples, ranges, positions).
#'
#' When summarizing the crosslink coverage over samples (\code{samples.merge=TRUE})
#' one can decide whether to summarize all samples or whether to keep conditions
#' separate (\code{samples.group}). This either reduces the samples dimension
#' to a single matrix, or a list. For a binding site set with 100 binding sites
#' of width=7 and 4 replicates with 2 conditions, the following options are
#' possible. With merging enabled and \code{samples.group='all'} the coverage of
#' all samples is combined. With \code{samples.group='condition'} only samples
#' of the same condition are grouped.
#'
#' When summarizing the crosslink coverage over ranges, all ranges are combined
#' which reduces the ranges dimension to a single vector. This turns eg. a
#' binding site set of 100 binding sites with width=7 into a vector of length
#' 100 with exactly one column. Depending on how the samples were summarized, the
#' result can be a single such vector, or a list.
#'
#' When summarizing the crosslink coverage over positions, all positions are
#' combined which reduces the positions dimension to a single vector. This turns
#' eq. a binding site set of 100 binding sites with width=7 into a vector of
#' length 1 with 7 columns. Depending on how the samples were summarized, the
#' result can be a single such vector, or a list.
#'
#' For all summarizing operations options \code{sum} and \code{mean} exists. This
#' allows for normalization by the eg. the number of binding sites, size of the
#' range, number of sample, etc..
#'
#' If the resulting object does have a dimension that fits to the number of
#' input ranges the result can be directly attached to them. Basically extending
#' the \code{GRanges} object (\code{out.format}).
#'
#' @param object a BSFDataSet object
#' @param ranges.merge logical; whether to merge ranges
#' @param ranges.merge.method character; how to combine ranges ('sum' or 'mean')
#' @param positions.merge logical; whether to merge positions
#' @param positions.merge.method character; how to combine positions ('sum' or 'mean')
#' @param samples.merge logical: whether to merge samples
#' @param samples.group character; how samples should be grouped when combining
#' ('all', 'condition')
#' @param samples.merge.method charater; how to combine positions ('sum' or 'mean')
#' @param out.format character; how the coverage should be returned
#' ('data.frame' or 'granges'). Note that option 'granges' only exists if the
#' output coverage is of the same rows as the input ranges.
#' @param out.format.overwrite logical; if \code{out.format='granges'}, then
#' decide wheter the meta columns should be extended by the coverage information
#' or be overwritten
#' @param match.rangeID character; unique internal identifier.
#' Name of the meta column of the input ranges
#' that should be used as identifier to match the coverage back to the input
#' ranges. Is 'bsID' as default, since that ID exists for all binding sites
#' after \code{makeBindingSites} was called.
#' @param quiet logical; whether to print messages
#'
#' @return an object of class specified in \code{out.format}
#'
#' @seealso \code{\link{BSFDataSet}}, \code{\link{BSFind}}
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' bds = makeBindingSites(bds, bsSize = 7)
#' # sum of each replicate over each binding site position
#' c1 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE,
#'  ranges.merge = FALSE, samples.merge = FALSE)
#' # total signal per binding site from all samples
#' c2 = clipCoverage(bds, out.format = "granges", positions.merge = FALSE,
#'  ranges.merge = TRUE, samples.merge = TRUE, samples.group = "all")
#' # total signal per binding site from all samples - split by condition
#' c3 = clipCoverage(bds, out.format = "granges", positions.merge = FALSE,
#'  ranges.merge = TRUE, samples.merge = TRUE, samples.group = "condition")
#'
#' @export
clipCoverage <- function(
        object,
        ranges.merge = FALSE,
        ranges.merge.method = c("sum", "mean"),
        positions.merge = FALSE,
        positions.merge.method = c("sum", "mean"),
        samples.merge = TRUE,
        samples.group = c("all", "condition"),
        samples.merge.method = c("sum", "mean"),
        out.format = c("granges", "data.frame"),
        out.format.overwrite = FALSE,
        match.rangeID = "bsID",
        quiet = FALSE
) {

    # init local variables
    # out.format <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    # check ranges
    rng = getRanges(object)
    if (length(rng) == 0) {
        msg = paste0("No ranges present. Please provide valid input.")
        stop(msg)
    }
    # initialize identical width check
    rng.width.identical = TRUE
    if (length(unique(width(rng))) > 1) {
        msg0 = paste0("Ranges have ", length(unique(width(rng))), " different length.\n")
        msg1 = paste0("Coverage is computed as a single value per range and sample.\n")
        msg2 = paste0("(merge.range = FALSE, merge.sample = FALSE, merge.positions = TRUE)")
        if (!quiet) warning(c(msg0, msg1, msg2))
        rng.width.identical = FALSE
    }

    # match input for positions
    positions.merge.method = match.arg(positions.merge.method, choices = c("sum", "mean"))
    # match input for ranges
    ranges.merge.method = match.arg(ranges.merge.method, choices = c("sum", "mean"))
    # match input for samples
    samples.merge.method = match.arg(samples.merge.method, choices = c("sum", "mean"))
    samples.group = match.arg(samples.group, choices = c("all", "condition"))
    # match output format
    out.format = match.arg(out.format, choices = c("granges", "data.frame"))

    # manage range names
    metaNames = colnames(mcols(rng))
    if (!match.rangeID %in% metaNames) {
        # match id is not present in meta data
        msg0 = paste0(match.rangeID, " was defined as unique ID for the ranges, but is not present in the meta data of the ranges. \n")
        msg1 = paste0("Specify another column or create an appropriate ID. \n")
        stop(c(msg0, msg1))
    } else {
        # match id is present in meta data
        selectID = mcols(rng)[match(match.rangeID, colnames(mcols(rng)))][[1]]
        names(rng) = selectID
    }

    # get signal
    sgn = getSignal(object)

    # get conditions
    conditions = levels(getMeta(object)$condition)

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    # ranges have all the same size
    if (isTRUE(rng.width.identical)) {
        # compute coverage for all replicates positions and ranges
        # -> this creates list of matrices with the dimensions equal to the
        # --> number of the ranges
        # --> width of the ranges
        cov.final = .makeCoverageMatrix(rng = rng, sgn = sgn)


        # merge output as defined
        cov.final = .mergeSamples(cov.final = cov.final,
                                  samples.merge = samples.merge,
                                  samples.group = samples.group,
                                  samples.merge.method = samples.merge.method,
                                  conditions = conditions)

        cov.final = .mergeRanges(cov.final = cov.final,
                                 ranges.merge = ranges.merge,
                                 ranges.merge.method = ranges.merge.method)

        cov.final = .mergePositions(cov.final = cov.final,
                                    positions.merge = positions.merge,
                                    positions.merge.method = positions.merge.method)
    }
    # ranges have different sizes
    if (!isTRUE(rng.width.identical)) {
        cov.final = .makeCoverageVector(sgn = sgn, rng = rng, method = ranges.merge.method)
    }

    # MANAGE OUTPUT
    # --------------------------------------------------------------------------
    # check output format granges
    if (out.format == "granges") {
        if (is(cov.final, "data.frame")) {
            if (! nrow(cov.final) == length(rng)) {
                msg0 = paste0("Coverage matrix and input ranges do not have the same dimensions. \n")
                msg1 = paste0("Output defaults to 'data.frame' or `vector`.\n")
                if (!quiet) warning(c(msg0, msg1))
                out.format = "data.frame"
            }
        }
        if (is(cov.final, "numeric")) {
            if (!length(cov.final) == length(rng)) {
                msg0 = paste0("Coverage vector and input ranges do not have the same dimensions. \n")
                msg1 = paste0("Output defaults to 'data.frame' or `vector`.\n")
                if (!quiet) warning(c(msg0, msg1))
                out.format = "data.frame"
            }
        }
    }

    # make output
    if (out.format == "granges"){
        rngOut = rng
        if (isTRUE(out.format.overwrite)) {
            mcols(rngOut) = cov.final
        }
        if (!isTRUE(out.format.overwrite)) {
            mcols(rngOut) = cbind(mcols(rngOut), cov.final)
        }
        out = rngOut
    }

    if (out.format == "data.frame") {
        out = cov.final
    }

    return(out)
}

.makeCoverageMatrix <- function(sgn, rng) {

    # split by strand
    rngPlus = rng[strand(rng) == "+"]
    rngMinus = rng[strand(rng) == "-"]

    # Plus strand
    if (length(rngPlus) > 0) {
        # compute initial coverage
        matPlus = lapply(sgn$signalPlus, function(x) {
            y = as.matrix(x[rngPlus])
            rownames(y) = names(rngPlus)
            return(y)
        })
    } else {
        matPlus = NULL
    }

    # Minus strand
    if (length(rngMinus) > 0) {
        # compute initial coverage
        matMinus = lapply(sgn$signalMinus, function(x) {
            y = as.matrix(x[rngMinus])
            rownames(y) = names(rngMinus)
            # reverse minus strand matrices to fit to plus orientation
            y = y[, ncol(y):1]
            return(y)
        })
    } else {
        matMinus = NULL
    }

    # Combine strands
    if (length(rngMinus) > 0 & length(rngPlus) > 0) {
        # if both strands had ranges
        matCov = lapply(seq_along(matPlus), function(x){
            y = dplyr::bind_rows(as.data.frame(matPlus[[x]]), as.data.frame(matMinus[[x]]))
            y = y[match(names(rng), rownames(y)),]
            y
        })
        names(matCov) = names(matPlus)
    }
    if (length(rngMinus) == 0) {
        # only plus strand with ranges
        matCov = lapply(seq_along(matPlus), function(x){
            y = as.data.frame(matPlus[[x]])
            y
        })
        names(matCov) = names(matPlus)
    }
    if (length(rngPlus) == 0) {
        # only minus strand with ranges
        matCov = lapply(seq_along(matMinus), function(x){
            y = as.data.frame(matMinus[[x]])
            y
        })
        names(matCov) = names(matMinus)
    }

    return(matCov)
}

.makeCoverageVector <- function(sgn, rng, method) {

    # split by strand
    rngPlus = rng[strand(rng) == "+"]
    rngMinus = rng[strand(rng) == "-"]

    # Plus strand
    # --------------------------------------------------------------------------
    if (length(rngPlus) > 0) {
        # compute initial coverage
        if (method == "sum") {
            covPlus = as.matrix(
                do.call(cbind, lapply(sgn$signalPlus, function(x) {
                    sum(x[rngPlus])
                })))
            rownames(covPlus) = names(rngPlus)
        }
        if (method == "mean") {
            covPlus = as.matrix(
                do.call(cbind, lapply(sgn$signalPlus, function(x) {
                    mean(x[rngPlus])
                })))
            rownames(covPlus) = names(rngPlus)
        }
    }

    # Minus strand
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0) {
        # compute initial coverage
        if (method == "sum") {
            covMinus = as.matrix(
                do.call(cbind, lapply(sgn$signalMinus, function(x) {
                    sum(x[rngMinus])
                })))
            rownames(covMinus) = names(rngMinus)
        }
        if (method == "mean") {
            covMinus = as.matrix(
                do.call(cbind, lapply(sgn$signalMinus, function(x) {
                    mean(x[rngMinus])
                })))
            rownames(covMinus) = names(rngMinus)
        }
    }

    # Combine strands for return
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0 & length(rngPlus) > 0) {
        cov.final = rbind(covPlus, covMinus)
    }
    if (length(rngMinus) == 0) {
        cov.final = rngPlus
    }
    if (length(rngPlus) == 0) {
        cov.final = rngMinus
    }
    # match results to manage order
    idx = match(rownames(cov.final), names(rng))
    cov.final = cov.final[order(idx),]

    return(cov.final)
}

.mergeRanges <- function(cov.final,
                         ranges.merge,
                         ranges.merge.method
){
    # merge ranges
    if (isTRUE(ranges.merge)) {
        if (is(cov.final, "list")) {
            # merge a list of matrices
            if (ranges.merge.method == "sum") {
                cov.final = lapply(cov.final, rowSums, na.rm = TRUE)
            }
            if (ranges.merge.method == "mean") {
                cov.final = lapply(cov.final, rowMeans, na.rm = TRUE)
            }
        }
        if (is(cov.final, "data.frame")) {
            # merge a single matrix
            if (ranges.merge.method == "sum") {
                cov.final = rowSums(cov.final, na.rm = TRUE)
            }
            if (ranges.merge.method == "mean") {
                cov.final = rowMeans(cov.final, na.rm = TRUE)
            }
        }

    }
    return(cov.final)
}

.mergeSamples <- function(cov.final,
                          samples.merge,
                          samples.group,
                          samples.merge.method,
                          conditions
){
    if (isTRUE(samples.merge)) {
        if (samples.group == "condition") {
            # merge samples by condition
            # -> return list of matrices with length equal to number of conditions
            # -> should work with NAs
            mat.list = lapply(conditions, function(x){cov.final[c(grep(x, names(cov.final)))]})
            names(mat.list) = conditions

            if (samples.merge.method == "sum") {
                # merge by sum
                cov.final = lapply(mat.list, function(x){Reduce('+', x)})
            }
            if (samples.merge.method == "mean") {
                # merge by mean
                cov.final = lapply(mat.list, function(x){Reduce('+', x) / length(x)})
            }
        }
        if (samples.group == "all") {
            # merge all samples
            # -> returns a single matrix
            if (samples.merge.method == "sum") {
                # merge by sum
                cov.final = Reduce("+", cov.final)
            }
            if (samples.merge.method == "mean") {
                # merge by mean
                cov.final = Reduce("+", cov.final) / length(cov.final)
            }
        }

    }
    return(cov.final)
}

.mergePositions <- function(cov.final,
                            positions.merge,
                            positions.merge.method
){
    # merge positions
    if (isTRUE(positions.merge)) {
        if (is(cov.final, "numeric")) {
            # merge a single vector
            # -> from merge ranges
            if (positions.merge.method == "sum") {
                cov.final = sum(cov.final)
            }
            if (positions.merge.method == "mean") {
                cov.final = mean(cov.final)
            }
        }
        if (is(cov.final, "data.frame")) {
            # merge a single matrix
            # -> from samples option 'all'
            if (positions.merge.method == "sum") {
                cov.final = colSums(cov.final)
            }
            if (positions.merge.method == "mean") {
                cov.final = colMeans(cov.final)
            }
        }
        if (is(cov.final, "list")) {
            # merge a list
            if (all(unlist(lapply(cov.final, class)) == "numeric")) {
                # the list is made from vectors
                if (positions.merge.method == "sum") {
                    cov.final = lapply(cov.final, sum)
                    cov.final = as.data.frame(do.call(rbind, cov.final))
                }
                if (positions.merge.method == "mean") {
                    cov.final = lapply(cov.final, mean)
                    cov.final = as.data.frame(do.call(rbind, cov.final))
                }
            }
            if (all(unlist(lapply(cov.final, class)) == "data.frame")) {
                # the list is made from matrices
                if (positions.merge.method == "sum") {
                    cov.final = lapply(cov.final, colSums)
                    cov.final = as.data.frame(do.call(rbind, cov.final))
                }
                if (positions.merge.method == "mean") {
                    cov.final = lapply(cov.final, colMeans)
                    cov.final = as.data.frame(do.call(rbind, cov.final))
                }
            }
        }
    }
    return(cov.final)
}



#-------------------------------------------------------------------------------
# OLD Function
#-------------------------------------------------------------------------------

#' Coverage function for BSFDataSet objects
#'
#' The crosslink coverage is computed for all ranges in the the given
#' \code{BSFDataSet} object (see \code{\link{BSFDataSet}} for details).
#' Depending on the \code{returnOptions} the resulting coverage information is
#' summarized, suitable for diverse computation and plotting tasks. The
#' coverage can only be compute for objects with identical ranges.
#'
#' If \code{returnOptions} is set to merge_ranges_keep_positions:
#' Returns a matrix with ncol being the nucleotides of the ranges (equal to the
#' width of the input ranges) and nrow being the number of replicates in the
#' meta information.
#'
#' If \code{returnOptions} is set to merge_replicates_per_condition:
#' Returns a list of matrices. Each list corresponds to one condition set in
#' the meta information. The matrix in each entry has ncols equal to the
#' ranges width and nrow equal to the number of ranges. Counts per ranges
#' and position are summed.
#'
#' If \code{returnOptions} is set to merge_all_replicates:
#' Returns a matrix with ncols equal to the range width and nrow equal to the
#' number of ranges. Counts per range and position are summed.
#'
#' If \code{returnOptions} is set to merge_positions_keep_replicates:
#' Returns a GRanges object where the counts are summed for each replicate
#' and added to the original granges object.
#'
#' @param object a BSFDataSet object
#' @param returnOptions one of merge_ranges_keep_positions,
#' merge_replicates_per_condition, merge_all_replicates,
#' merge_positions_keep_replicates
#' @param method sum/ mean, select how replicates/ ranges should be summarized
#' @param allowNA TRUE/ FALSE, allow NA values in the output if input ranges
#' are of different width
#' @param quiet logical, whether to print messages
#'
#' @return an object of class specified in \code{returnOptions}
#' @import GenomicRanges
#' @importFrom plyr rbind.fill
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#'
#' rng = coverageOverRanges(bds, returnOptions = "merge_ranges_keep_positions")
#' rng = coverageOverRanges(bds, returnOptions = "merge_replicates_per_condition")
#' rng = coverageOverRanges(bds, returnOptions = "merge_all_replicates")
#' rng = coverageOverRanges(bds, returnOptions = "merge_positions_keep_replicates")
#'
#' @export
coverageOverRanges <- function(
    object,
    returnOptions = c("merge_ranges_keep_positions",
                      "merge_replicates_per_condition",
                      "merge_all_replicates",
                      "merge_positions_keep_replicates"),
    method = "sum",
    allowNA = FALSE,
    quiet = TRUE
    ) {

    # deprecation notice
    # .Deprecated("clipCoverage")
    # first exchange everywhere it is used

    # Check input
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))

    # manage return option
    returnOptions = match.arg(
        returnOptions,
        choices = c("merge_ranges_keep_positions",
                    "merge_replicates_per_condition",
                    "merge_all_replicates",
                    "merge_positions_keep_replicates"))
    method = match.arg(method,choices = c("sum", "mean"))
    # get range
    rng = getRanges(object)

    # check if lenght is all the same or different
    if (length(unique(width(getRanges(object)))) > 1 &
        allowNA == FALSE &
        returnOptions != "merge_positions_keep_replicates") {
        # ranges do not have all the same size
        warning("Width of ranges are not identical. This will cause NAs in the
                output. Option `merge_positions_keep_replicates` is enforced.
                Avoid this behavior by setting allowNA = TRUE.")
        returnOptions = "merge_positions_keep_replicates"
    }

    # check for unique range names
    if(any(duplicated(names(rng)))) {
        msg0 = paste0("duplicate names in ranges, adding preceding ID")
        if(!quiet) message(c(msg0))
        names(rng) = paste0(c(seq_along(rng)), "_", names(rng))
    }

    # prepare signal
    sgn = getSignal(object)
    # extract present conditions
    condition = levels(getMeta(object)$condition)


    # Forward to helper functions
    # --------------------------------------------------------------------------
    if (returnOptions == "merge_replicates_per_condition") {
        covRet = .coverageOverRanges.merge_replicates_per_condition(
            sgn = sgn, rng = rng, condition = condition, method = method, allowNA = allowNA
        )
    }
    if (returnOptions == "merge_all_replicates") {
        covRet = .coverageOverRanges.merge_all_replicates(
            sgn = sgn, rng = rng, method = method, allowNA = allowNA
        )
    }
    if (returnOptions == "merge_positions_keep_replicates") {
        covRet = .coverageOverRanges.merge_positions_keep_replicates(
            sgn = sgn, rng = rng, method = method
        )
    }
    if (returnOptions == "merge_ranges_keep_positions") {
        covRet = .coverageOverRanges.merge_ranges_keep_positions(
            sgn = sgn, rng = rng, method = method
        )
    }
    return(covRet)
}

.coverageOverRanges.merge_positions_keep_replicates <- function(sgn, rng, method) {

    # split by strand
    rngPlus = rng[strand(rng) == "+"]
    rngMinus = rng[strand(rng) == "-"]

    # Plus strand
    # --------------------------------------------------------------------------
    if (length(rngPlus) > 0) {
        # compute initial coverage
        if (method == "sum") {
            mcols(rngPlus) = as.matrix(
                do.call(cbind, lapply(sgn$signalPlus, function(x) {
                    sum(x[rngPlus])
                })))
        }
        if (method == "mean") {
            mcols(rngPlus) = as.matrix(
                do.call(cbind, lapply(sgn$signalPlus, function(x) {
                    mean(x[rngPlus])
                })))
        }

    }

    # Minus strand
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0) {
        # compute initial coverage
        if (method == "sum") {
            mcols(rngMinus) = as.matrix(
                do.call(cbind, lapply(sgn$signalMinus, function(x) {
                    sum(x[rngMinus])
                })))
        }
        if (method == "mean") {
            mcols(rngMinus) = as.matrix(
                do.call(cbind, lapply(sgn$signalMinus, function(x) {
                    mean(x[rngMinus])
                })))
        }
    }

    # Combine strands for return
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0 & length(rngPlus) > 0) {
        retCov = c(rngPlus, rngMinus)
    }
    if (length(rngMinus) == 0) {
        retCov = rngPlus
    }
    if (length(rngPlus) == 0) {
        retCov = rngMinus
    }
    # sort ranges
    rngCov = c(retCov)
    rngCov = .sortRanges(rngCov)

    return(rngCov)
}

.coverageOverRanges.merge_ranges_keep_positions <- function(sgn, rng, method) {
    # split by strand
    rngPlus = rng[strand(rng) == "+"]
    rngMinus = rng[strand(rng) == "-"]

    # Plus strand
    # --------------------------------------------------------------------------
    if (length(rngPlus) > 0) {
        # compute initial coverage
        matPlus = lapply(sgn$signalPlus, function(x) {
            y = as.matrix(x[rngPlus])
            rownames(y) = names(rngPlus)
            return(y)
        })
        # summarize
        if (method == "sum") {
            covPlus = do.call(rbind, lapply(matPlus, colSums))
        }
        if (method == "mean") {
            covPlus = do.call(rbind, lapply(matPlus, colMeans))
        }
    }
    if (length(rngPlus) == 0) {
        # set cov on plus strand to zero if no range is present
        covPlus = 0
    }

    # Minus strand
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0) {
        # compute initial coverage
        matMinus = lapply(sgn$signalMinus, function(x) {
            y = as.matrix(x[rngMinus])
            rownames(y) = names(rngMinus)
            return(y)
        })
        # summarize
        if (method == "sum") {
            covMinus = do.call(rbind, lapply(matMinus, colSums))
        }
        if (method == "mean") {
            covMinus = do.call(rbind, lapply(matMinus, colMeans))
        }
        # reverse orientation
        covMinus = as.matrix(rev(as.data.frame(covMinus)))
    }
    if (length(rngMinus) == 0) {
        # set cov on plus strand to zero if no range is present
        covMinus = 0
    }

    # Combine strands for return
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0 & length(rngPlus) > 0) {
        retCov = covPlus + covMinus
    }
    if (length(rngMinus) == 0) {
        retCov = covPlus
    }
    if (length(rngPlus) == 0) {
        retCov = covMinus
    }
    return(retCov)
}

.coverageOverRanges.merge_all_replicates <- function(sgn, rng, method, allowNA) {
    # set names
    names(rng) = seq_along(rng)
    # split by strand
    rngPlus = rng[strand(rng) == "+"]
    rngMinus = rng[strand(rng) == "-"]

    # Plus strand
    # --------------------------------------------------------------------------
    if (length(rngPlus) > 0) {
        # compute initial coverage
        matPlus = lapply(sgn$signalPlus, function(x) {
            y = as.matrix(x[rngPlus])
            rownames(y) = names(rngPlus)
            return(y)
        })
        # summarize
        if (method == "sum") {
            covPlus = Reduce('+', matPlus)
        }
        if (method == "mean") {
            covPlus = Reduce('+', matPlus) / length(matPlus)
        }
    }
    if (length(rngPlus) == 0) {
        # set cov on plus strand to zero if no range is present
        covPlus = 0
    }

    # Minus strand
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0) {
        # compute initial coverage
        matMinus = lapply(sgn$signalMinus, function(x) {
            y = as.matrix(x[rngMinus])
            rownames(y) = names(rngMinus)
            return(y)
        })
        # summarize
        if (method == "sum") {
            covMinus = Reduce('+', matMinus)
        }
        if (method == "mean") {
            covMinus = Reduce('+', matMinus) / length(matMinus)
        }
        # reverse orientation
        covMinus = as.matrix(rev(as.data.frame(covMinus)))
    }
    if (length(rngMinus) == 0) {
        # set cov on plus strand to zero if no range is present
        covMinus = 0
    }

    # Combine strands for return
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0 & length(rngPlus) > 0) {
        if (allowNA == TRUE) {
            retCov = plyr::rbind.fill(as.data.frame(covPlus), as.data.frame(covMinus))
        }
        if (allowNA == FALSE) {
            retCov = rbind(covPlus, covMinus)
        }
    }
    if (length(rngMinus) == 0) {
        retCov = covPlus
    }
    if (length(rngPlus) == 0) {
        retCov = covMinus
    }
    # match results to manage order
    idx = match(rownames(retCov), names(rng))
    retCov = retCov[order(idx),]
    return(retCov)
}

.coverageOverRanges.merge_replicates_per_condition <- function(sgn, rng, condition, method, allowNA) {
    # set names
    names(rng) = seq_along(rng)
    # split by strand
    rngPlus = rng[strand(rng) == "+"]
    rngMinus = rng[strand(rng) == "-"]

    # Plus strand
    # --------------------------------------------------------------------------
    if (length(rngPlus) > 0) {
        # compute initial coverage
        matPlus = lapply(sgn$signalPlus, function(x) {
            y = as.matrix(x[rngPlus])
            rownames(y) = names(rngPlus)
            return(y)
        })
        # summarize per condition
        matPlusCond = lapply(condition, function(x){
            matPlus[c(grep(x, names(matPlus)))]})
        names(matPlusCond) = condition
        # manage combine method
        if (method == "sum") {
            covPlus = lapply(matPlusCond, function(x){Reduce('+', x)})
        }
        if (method == "mean") {
            covPlus = lapply(matPlusCond, function(x){Reduce('+', x) / length(x)})
        }
    }
    if (length(rngPlus) == 0) {
        # set cov on plus strand to zero if no range is present
        covPlus = 0
    }

    # Minus strand
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0) {
        # compute initial coverage
        matMinus = lapply(sgn$signalMinus, function(x) {
            y = as.matrix(x[rngMinus])
            rownames(y) = names(rngMinus)
            return(y)
        })
        # summarize per condition
        matMinusCond = lapply(condition, function(x){
            matMinus[c(grep(x, names(matMinus)))]})
        names(matMinusCond) = condition
        # mange combine method
        if (method == "sum") {
            covMinus = lapply(matMinusCond, function(x){Reduce('+', x)})
        }
        if (method == "mean") {
            covMinus = lapply(matMinusCond, function(x){Reduce('+', x) / length(x)})
        }
        # reverse orientation
        covMinus = lapply(covMinus, function(x){
            as.matrix(rev(as.data.frame(x)))})
    }
    if (length(rngMinus) == 0) {
        # set cov on plus strand to zero if no range is present
        covMinus = 0
    }

    # Combine strands for return
    # --------------------------------------------------------------------------
    if (length(rngMinus) > 0 & length(rngPlus) > 0) {
        if (allowNA == TRUE) {
            retCov = lapply(seq_along(covPlus), function(x){
                retCov = plyr::rbind.fill(as.data.frame(covPlus[[x]]), as.data.frame(covMinus[[x]]))
            })
        }
        if (allowNA == FALSE) {
            retCov = lapply(seq_along(covPlus), function(x){
                rbind(covPlus[[x]],covMinus[[x]])
            })
        }
        names(retCov) = names(covPlus)
    }
    if (length(rngMinus) == 0) {
        retCov = covPlus
    }
    if (length(rngPlus) == 0) {
        retCov = covMinus
    }
    # match results to manage order
    retCov = lapply(retCov, function(x){
        idx = match(rownames(x), names(rng))
        x = x[order(idx),]
    })
}

################################################################################
# Workflow functions
# - these functions are used in the main bindingSiteFinde workflow
# - each function updates the bindingSiteFinder object
# - functions are either designed to work on single nt wide crosslink sites or
# on merged binding sites
# - for each function a related plotting funciton exists to visualize the results
# - the wrapper function BSFind() executes the functions below in the desired order
################################################################################


#' Filter PureCLIP sites by their score distribution
#'
#' Function that applies a filter on the global crosslink site score distribution.
#' The \code{\link{GenomicRanges}} contained in the \code{\link{BSFDataSet}} need to
#' have a meta-column that holds a numeric score value, which is used for filtering.
#' The name of the column can be set with \code{match.score}.
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a \code{\link{BSFDataSet}} object with stored crosslink ranges
#' of width=1
#' @param cutoff numeric; defines the cutoff for which sites to keep, the
#' smallest step is 1\% (0.01)
#' @param match.score character; meta column name of the crosslink site
#' \code{\link{GenomicRanges}} object that holds the score which is used for
#' sub-setting
#' @param quiet logical; whether to print messages
#'
#' @return an object of class  \code{\link{BSFDataSet}} with its ranges filtered
#' by those that passed the threshold set with \code{cutoff}
#'
#' @seealso \code{\link{BSFind}}, \code{\link{pureClipGlobalFilterPlot}}
#'
#' @examples
#' # load data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # apply 5% filter
#' pureClipGlobalFilter(object = bds, cutoff = 0.05)
#'
#' @export
pureClipGlobalFilter <- function(object, # bindingSiteFinder
                                 cutoff = 0.01, # defines the cutoff for which PureCLIP sites to keep; smallest are 1% (0.01); a cutoff of 0.05 means to remove the lowest 5% PureCLIP sites base on the score
                                 match.score = "score",
                                 quiet = FALSE
) {
    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))
    # check for correct cutoff value
    possibleCutoffs = seq(from = 0, to = 1, by = 0.01)
    if (! cutoff %in% possibleCutoffs) {
        msg = paste0("Specified cutoff (", cutoff,
                     "), does not refer to a percentage value. Try cutoffs of the form 0.01, 0.05, ... instead.")
        stop(msg)
    }
    # Check for correct score column
    metaNames = colnames(mcols(getRanges(object)))
    if (! match.score %in% metaNames) {
        msg = paste0("Specified match.score (", match.score, "), is present in the ranges. Add respective column. ")
        stop(msg)
    }

    # Create matching vectors for score column from input source
    # --------------------------------------------------------------------------
    rngInitial = getRanges(object)
    score = mcols(rngInitial)[match(match.score, colnames(mcols(rngInitial)))][[1]]

    # ---
    # Store function parameters in list
    optstr = list(cutoff = cutoff, match.score = match.score)
    object@params$pureClipGlobalFilter = optstr

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    # compute quantile based on PureCLIP score
    qCut = quantile(score, probs = seq(0,1, by = 0.01))

    # apply user set cutoff
    names(qCut) = as.numeric(sub("%", "", names(qCut))) / 100
    idx = match(cutoff, names(qCut))
    rngFiltered = rngInitial[score >= qCut[idx]]

    # ---
    # Store for plotting
    dfPlot = score
    cutoffPoint = qCut[idx]
    object@plotData$pureClipGlobalFilter$data = dfPlot
    object@plotData$pureClipGlobalFilter$cutoffPoint = cutoffPoint

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "pureClipGlobalFilter()", class = "crosslink sites",
        nIn = length(rngInitial), nOut = length(rngFiltered),
        per = paste0(round(length(rngFiltered)/ length(rngInitial), digits = 2)*100,"%"),
        options = paste0("Cutoff=", optstr$cutoff, ", scoreMatchCol=", optstr$match.score)
    )
    object@results = rbind(object@results, resultLine)

    # return BSF object with only ranges above cutoff
    objectNew = setRanges(object, rngFiltered, quiet = quiet)
    return(objectNew)
}


#' Filter PureCLIP sites by their score distribution per gene
#'
#' Function that applies a filter on the crosslink site score distribution at
#' gene level. This allows to filter for those sites with the strongest signal
#' on each gene. Since scores are tied to the expression level of the hosting
#' transcript this function allows a fair filter for all genes partially
#' independent of the expression level.
#'
#' The \code{\link{GenomicRanges}} contained in the \code{\link{BSFDataSet}} need to
#' have a meta-column that holds a numeric score value, which is used for filtering.
#' The name of the column can be set with \code{scoreCol}.
#'
#' In the case of overlapping gene annotation, a single crosslink site will be
#' attributed to multiple genes. The \code{\link{overlaps}} parameter allows
#' to control these cases. Option `keepSingle` will only keep a single instance
#' of the site; `removeAll` will remove both sites; `keepAll` will keep both
#' sites.
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a \code{\link{BSFDataSet}} object with stored crosslink ranges
#' of width=1
#' @param cutoff numeric; defines the cutoff for which sites to remove, the
#' smallest step is 1\% (0.01). A cutoff of 5\% will remove the lowest 5\% sites,
#' given their score, on each gene, thus keeping the strongest 95\%.
#' @param overlaps character; how overlapping gene loci should be handled.
#' @param anno.annoDB an object of class \code{OrganismDbi} that contains
#' the gene annotation (!!! Experimental !!!).
#' @param anno.genes an object of class \code{\link{GenomicRanges}} that represents
#' the gene ranges directly
#' @param match.score character; meta column name of the crosslink site
#' \code{\link{GenomicRanges}} object that holds the score which is used for
#' sub-setting
#' @param match.geneID character; meta column name of the genes
#' \code{\link{GenomicRanges}} object that holds a unique geneID
#' @param quiet logical; whether to print messages
#'
#' @return an object of class \code{\link{BSFDataSet}} with its ranges filtered
#' by those that passed the gene-wise threshold set with \code{cutoff}
#'
#' @seealso \code{\link{BSFind}}, \code{\link{estimateBsWidthPlot}}
#'
#' @importFrom dplyr filter reframe group_by mutate mutate_all arrange row_number
#' @importFrom GenomicFeatures genes
#' @importFrom stats density
#' @importFrom grDevices colorRampPalette
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # Load GRanges with genes
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' # apply 5% gene-wise filter
#' pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.5, overlaps = "keepSingle")
#'
#' @export
pureClipGeneWiseFilter <- function(object, # bindingSiteFinder
                                   cutoff = 0.05, # defines the cutoff for which PureCLIP sites to keep; smallest steps are in 1% (0.01); cutoff of 0.7 means that the top 70% of PureCLIP sites per gene are retained.
                                   overlaps = c("keepSingle", "removeAll", "keepAll"), # Options to deal with PureCLIP sites on overlapping loci genes; use keepSingle as default; keepAll will inflate the number of binding sites by the number of overlaps, removeAll will kill all duplicated instances
                                   anno.annoDB = NULL, # annotation data base of class OrganismDbi; can be NULL -> but then anno.genes must be provided
                                   anno.genes = NULL, # ranges of the genes of class GenomicRanges; can be NULL -> but then anno.annoDB must be provided
                                   match.score = "score",
                                   match.geneID = "gene_id",
                                   quiet = FALSE) {
    # Local variables
    datasource <- cutoffs <- quantiles <- name <- NULL
    # INPUT CHECKS
    # --------------------------------------------------------------------------
    # type checks
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    # Check if none is specified
    if (is.null(anno.annoDB) & is.null(anno.genes)) {
        msg = paste0("None of the required annotation sources anno.annoDB or anno.genes was specified. ")
        stop(msg)
    }
    # Check if both are specified
    if (!is.null(anno.annoDB) & !is.null(anno.genes)) {
        msg = paste0("Both of the required annotation sources anno.annoDB or anno.genes are specified. Please provide only one of the two. ")
        stop(msg)
    }
    # Checks if anno.annoDB should be used
    if (!is.null(anno.annoDB) & is.null(anno.genes)) {
        stopifnot(is(anno.annoDB, "OrganismDb"))
        datasource = "anno.annoDB"
        # extract relevant annotation
        anno.genes = genes(anno.annoDB, columns = c("GENEID"))
        # Create matching vectors for columns from input annotation
        selectID = as.character(anno.genes$GENEID)
    }
    # Checks if anno.genes should be used
    if (is.null(anno.annoDB) & !is.null(anno.genes)) {
        stopifnot(is(anno.genes, "GenomicRanges"))

        datasource = "anno.genes"
        # extract relevant annotation
        anno.genes = anno.genes
        # check correct annotation columns
        annoColNames = colnames(mcols(anno.genes))
        if (!all(match.geneID %in% annoColNames)) {
            msg = paste0("The provided matching column for the geneID (",
                         match.geneID, ") is not present in the provided annotation. \n")
            stop(msg)
        }
        # Create matching vectors for columns from input annotation
        selectID = mcols(anno.genes)[match(match.geneID, colnames(mcols(anno.genes)))][[1]]

    }
    # check if cutoff was estimated before or if default should be used
    if (!is.null(object@params$bsSize) &
        !is.null(object@params$geneFilter)) {
        cutoff = object@params$geneFilter
    }

    # handle options (remove, keep)
    overlaps = match.arg(overlaps, choices = c("keepSingle", "removeAll", "keepAll"))
    # Check for correct score column
    # get all ranges
    rngInitial = getRanges(object)
    metaNames = colnames(mcols(rngInitial))
    if (! match.score %in% metaNames) {
    # if (! any(metaNames == match.score)) {
        msg = paste0("Specified match.score (", match.score, "), is present in the ranges. Add respective column. ")
        stop(msg)
    }

    # ---
    # Store function parameters in list
    optstr = list(cutoff = cutoff, overlaps = overlaps,
                  source = datasource, match.geneID = match.geneID)
    object@params$geneWiseFilter = optstr

    # MAIN COMPUTE
    # --------------------------------------------------------------------------

    # Create matching vectors for score column from input source
    # --------------------------------------------------------------------------
    score = mcols(rngInitial)[match(match.score, colnames(mcols(rngInitial)))][[1]]
    rngInitial$score = score
    # get all intergenic cases
    rngIntergeneic = rngInitial[countOverlaps(rngInitial, anno.genes) == 0]
    mcols(rngIntergeneic)$geneID = "Intergenic"
    # get all cases to solve
    # rngToSolve = subset(rngInitial, !rngInitial %in% rngIntergeneic)
    # rngToSolve = rngInitial[!rngInitial %in% rngIntergeneic]
    rngToSolve = rngInitial[countOverlaps(rngInitial, rngIntergeneic) == 0]

    # Gene specific cutoffs
    # ---------------------
    # set cutpoints
    potentialCutoffs = seq(0, 1, by = 0.01)
    # group scores by gene range
    ols = findOverlaps(anno.genes, rngToSolve) %>%
        as.data.frame() %>%
        mutate(score = rngToSolve$score[subjectHits]) %>%
        group_by(queryHits)
    # calculate gene range specific quantile cutoff
    quants = ols %>%
        dplyr::reframe(quantiles = quantile(score, probs = potentialCutoffs)) %>%
        group_by(queryHits) %>%
        mutate(cutoffs = potentialCutoffs) %>%
        dplyr::filter(as.character(cutoffs) == as.character(cutoff))
    # apply cutoff to every PureCLIP peak per gene
    peakIdToKeep = ols
    peakIdToKeep$quantiles = quants$quantiles[match(peakIdToKeep$queryHits, quants$queryHits)]
    peakIdToKeep = peakIdToKeep %>%
        dplyr::filter(score >= quantiles)
    # add gene information
    filteredPerRegion = rngToSolve[peakIdToKeep$subjectHits]
    mcols(filteredPerRegion)$geneID = selectID[peakIdToKeep$queryHits]
    # ---
    # Store for plotting
    dfPlot = countOverlaps(filteredPerRegion) %>% table() %>% as.data.frame() %>% rename("#N overlaps" = ".")
    object@plotData$geneWiseFilter$data = dfPlot

    # Deal with multiple loci
    # -----------------------
    # check for overlapping loci peaks
    potentialOverlaps = findOverlaps(filteredPerRegion)
    overlappingLociHits = potentialOverlaps[queryHits(potentialOverlaps) != subjectHits(potentialOverlaps)]
    # reduce to a single pair
    singlePair = overlappingLociHits %>%
        as.data.frame() %>%
        dplyr::filter(queryHits > subjectHits)
    overlappingLociHits = overlappingLociHits[queryHits(overlappingLociHits) %in% singlePair$queryHits]
    # get numbers
    overlappingLociGenePairs = paste0(unique(names(filteredPerRegion[queryHits(overlappingLociHits)])),
                                      ",",
                                      unique(names(filteredPerRegion[subjectHits(overlappingLociHits)])), " | ")
    fractionOfRangesRemoved = paste0(round(length(overlappingLociHits) / length(rngInitial) * 100, digits = 2), "%")
    # Handle cases
    if (length(overlappingLociHits) > 0) {
        msg0 = paste0(fractionOfRangesRemoved, " (", length(overlappingLociHits),"/", length(rngInitial), ")",
                      " peaks overlap with multiple anno.genes in the given gene annotation. \n")
        if (overlaps == "keepSingle") {
            msg1 = "A single instance of each peak is kept. This is recommended. \n "
            filteredPerRegion = filteredPerRegion[!duplicated(filteredPerRegion)]
            # report message
            if (!quiet) message(c(msg0, msg1))
        }
        if (overlaps == "removeAll") {
            msg1 = "All peaks on any duplicated range are removed. This is not recommended, since many peaks are lost. \n"
            filteredPerRegion = filteredPerRegion[countOverlaps(filteredPerRegion) == 1]
            warning(c(msg0, msg1))
        }
        if (overlaps == "keepAll") {
            msg1 = "Duplicated peaks due to overlapping loci are not removed. This is not recommended, since peak numbers are inflated. \n"
            warning(msg0, msg1)
        }
    }

    # add intergenic cases again
    filteredPerRegion = c(filteredPerRegion, rngIntergeneic)

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "pureClipGeneWiseFilter()", class = "crosslink sites",
        nIn = length(rngInitial), nOut = length(filteredPerRegion),
        per = paste0(round(length(filteredPerRegion)/ length(rngInitial), digits = 2)*100,"%"),
        options = paste0("Cutoff=", optstr$cutoff, ", overlapp option=", optstr$overlaps)
    )
    object@results = rbind(object@results, resultLine)

    objectOut = setRanges(object, filteredPerRegion, quiet = quiet)
    return(objectOut)
}


#' Assign binding sites to their hosting genes
#'
#' Function that assigns each binding site in the \code{\link{BSFDataSet}} to its
#' hosting gene given a gene annotation (\code{anno.annoDB}, \code{anno.genes}).
#'
#' Regardless of the annotation source that is being used, the respective meta
#' information about the genes have to be present. They can be set by the
#' \code{match.geneID}, \code{match.geneName} and \code{match.geneType} arguments.
#'
#' In the case of overlapping gene annotation, a single binding site will be
#' associated with multiple genes. The \code{\link{overlaps}} parameter allows
#' to decide in these cases. Option `frequency` will take the most frequently
#' observed gene type, option `hierarchy` works in conjunction with a user defined
#' rule (\code{overlaps.rule}). Options `remove` and `keep` will remove or
#' keep all overlapping cases, respectively.
#'
#' Note that if an overlaps exists, but gene types are identical options
#' `frequency` and `hierarchy` will cause the gene that was seen first to be
#' selected as representative.
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a \code{\link{BSFDataSet}} object with stored binding sites. This
#' means that ranges should be > 1
#' @param overlaps character; how overlapping gene loci should be handled.
#' @param overlaps.rule character vector; a vector of gene type that should
#' be used to handle overlapping cases in a hierarchical manor. The order of the
#' vector is the order of the hierarchy.
#' @param anno.annoDB an object of class \code{OrganismDbi} that contains
#' the gene annotation (!!! Experimental !!!).
#' @param anno.genes an object of class \code{\link{GenomicRanges}} that represents
#' the gene ranges directly
#' @param match.geneID character; meta column name of the gene ID
#' @param match.geneName character; meta column name of the gene name
#' @param match.geneType character; meta column name of the gene type
#' @param quiet logical; whether to print messages
#'
#' @return an object of class \code{\link{BSFDataSet}} with binding sites having
#' hosting gene information added to their meta columns.
#'
#' @seealso \code{\link{BSFind}}, \code{\link{geneOverlapsPlot}},
#' \code{\link{targetGeneSpectrumPlot}}
#'
#' @importFrom dplyr select slice_head group_by count ungroup arrange desc left_join
#' @importFrom GenomicFeatures genes
#' @importFrom dplyr desc
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' # Load GRanges with genes
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = assignToGenes(bds, anno.genes = gns)
#'
#' @export
assignToGenes <- function(object,
                          overlaps = c("frequency", "hierarchy", "remove", "keep"),
                          overlaps.rule = NULL,
                          anno.annoDB = NULL,
                          anno.genes = NULL,
                          match.geneID = "gene_id",
                          match.geneName = "gene_name",
                          match.geneType = "gene_type",
                          quiet = FALSE
) {
    # local variables
    datasource <- geneIndex <- geneID <- geneName <- value <- bsIndex <- Freq <- geneType <- match.missing <- NULL
    selectType <- selectID <- selectName <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    # Check if none is specified
    if (is.null(anno.annoDB) & is.null(anno.genes)) {
        msg = paste0("None of the required annotation sources anno.annoDB or anno.genes was specified. ")
        stop(msg)
    }
    # Check if both are specified
    if (!is.null(anno.annoDB) & !is.null(anno.genes)) {
        msg = paste0("Both of the required annotation sources anno.annoDB or anno.genes are specified. Please provide only one of the two. ")
        stop(msg)
    }
    # Checks if anno.annoDB should be used
    if (!is.null(anno.annoDB) & is.null(anno.genes)) {
        stopifnot(is(anno.annoDB, "OrganismDb"))
        if (!is.null(anno.genes)) {
            msg = paste0("Parameter anno.annoDB and anno.genes are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            datasource = "anno.annoDB"
            # extract relevant annotation
            anno.genes = genes(anno.annoDB, columns=c("ENSEMBL", "GENETYPE", "SYMBOL", "GENEID"))
            # Create matching vectors for columns from input annotation
            # --------------------------------------------------------------------------
            selectID = as.character(anno.genes$GENEID)
            selectName = as.character(anno.genes$SYMBOL)
            selectType = as.character(anno.genes$GENETYPE)
        }
    }
    # Checks if anno.genes should be used
    if (is.null(anno.annoDB) & !is.null(anno.genes)) {
        stopifnot(is(anno.genes, "GenomicRanges"))
        if (!is.null(anno.annoDB)) {
            msg = paste0("Parameter anno.annoDB and anno.genes are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            datasource = "anno.genes"
            # extract relevant annotation
            # check for duplicated annotation columns
            if (sum(colnames(mcols(anno.genes)) == match.geneID) > 1) {
                msg = paste0("The names of multiple columns of the annotation match the match.geneID parameter. Please use a unique column name for match.geneID in anno.genes.")
                stop(msg)
            }
            if (sum(colnames(mcols(anno.genes)) == match.geneName) > 1) {
                msg = paste0("The names of multiple columns of the annotation match the match.geneName parameter. Please use a unique column name for match.geneName in anno.genes.")
                stop(msg)
            }
            if (sum(colnames(mcols(anno.genes)) == match.geneType) > 1) {
                msg = paste0("The names of multiple columns of the annotation match the match.geneType parameter. Please use a unique column name for match.geneType in anno.genes.")
                stop(msg)
            }
            # check correct annotation columns
            inNames = c(match.geneID, match.geneName, match.geneType)
            annoColNames = colnames(mcols(anno.genes))
            presentNames = inNames[inNames %in% annoColNames]

            # extract columns from annotaiton based on available names
            present.cols = lapply(presentNames, function(x){
                match.col = mcols(anno.genes)[match(x, colnames(mcols(anno.genes)))][[1]]
                return(match.col)
            })
            names(present.cols) = presentNames

            # extract values from annotation for all present columns
            if (match.geneID %in% names(present.cols)) {
                selectID = present.cols[[match(match.geneID, names(present.cols))]]
            } else {
                msg = paste0("No meta column for ", match.geneID, " present. Creating custom ID.\n")
                if (!quiet) message(msg)
                selectID = paste0("CUSTOM", seq_along(anno.genes))
            }
            if (match.geneName %in% names(present.cols)) {
                selectName = present.cols[[match(match.geneName, names(present.cols))]]
            } else {
                msg = paste0("No meta column for ", match.geneName, " present.\n")
                if (!quiet) message(msg)
            }
            if (match.geneType %in% names(present.cols)) {
                selectType = present.cols[[match(match.geneType, names(present.cols))]]
            } else {
                msg = paste0("No meta column for ", match.geneType, " present.\n")
                if (!quiet) message(msg)
            }
        }
    }

    # handle options (hierarchy, frequency, remove, keep)
    overlaps = match.arg(overlaps, choices = c("frequency", "hierarchy", "remove", "keep"))
    if (is.null(selectType)) {
        # gene type is missing
        # -> check for correct overlap option
        if (overlaps == "frequency" | overlaps == "hierarchy") {
            msg0 = paste0(match.geneType, " is missing in the annotation. Options 'frequency' and 'hierarchy' are not available.\n")
            msg1 = paste0("Select one of: 'remove' or 'keep'.\n")
            stop(c(msg0, msg1))
        }
    }
    # Check multiple loci options
    if (overlaps == "hierarchy" & is.null(overlaps.rule)) {
        msg1 = paste0("Binding sites on anno.genes with overlapping annotaitons is set to be handled by 'hierarchy', but no rule is provided. \n")
        msg2 = paste0("Change the 'overlaps' option or provide a valid rule. \n")
        stop(c(msg1, msg2))
    }
    # --------------------------------------------------------------------------
    # MAIN
    # --------------------------------------------------------------------------

    # ---
    # Store function parameters in list
    optstr = list(source = datasource, match.geneID = match.geneID,
                  match.geneName = match.geneName, match.geneType = match.geneType,
                  overlaps = overlaps, overlaps.rule = overlaps.rule)
    object@params$assignToGenes = optstr

    # Match Binding sites and anno.genes
    # --------------------------------------------------------------------------
    # split ranges in duplicated event and non duplicated events
    rngInitial = getRanges(object)
    rngInitial$currIdx = seq_along(rngInitial)

    # remove possibility of partial overlaps by setting `minoverlap` parameter
    # to the size of the binding sites
    # -> use the minimum of the size for uneven binding sites sizes
    # -> this is not optimal since partial overlaps can be re-introduced
    # -> does not affect ranges that are true binding sites
    bs.set.range = min(unique(width(rngInitial)))
    ols = findOverlaps(anno.genes, rngInitial, minoverlap = bs.set.range)
    # ols = findOverlaps(anno.genes, rngInitial)

    # Deal with multiple loci
    # --------------------------------------------------------------------------
    # Count out how many cases there are
    countOlsSize = as.data.frame(table(duplicated(subjectHits(ols))))
    if (nrow(countOlsSize) == 1){
        # no duplications are present
        # -> no overlaps have to be resolved
        totalBS = countOlsSize$Freq[1]
        duplicatedBS = 0
    } else {
        totalBS = countOlsSize$Freq[1]
        duplicatedBS = countOlsSize$Freq[2]
    }
    duplicatedFraction = paste0(round(duplicatedBS / totalBS * 100, digits = 2), "%")

    # ----
    # Store results for plotting
    # -> select data
    if (all(!is.null(selectID) & !is.null(selectType) & !is.null(selectName))) {
        # nothing is missing
        dfPlot = data.frame(geneIndex = queryHits(ols), bsIndex = subjectHits(ols),
                            geneType = selectType[queryHits(ols)]) %>% #TODO type
            dplyr::mutate(value = 1) %>%
            dplyr::select(-geneIndex) %>%
            tidyr::pivot_wider(names_from = geneType, values_from = value, values_fn = length, values_fill = 0) %>%
            dplyr::select(-bsIndex) %>%
            dplyr::mutate_all(~ ifelse(. > 1, 1, .))
    } else {
        # check what is missing
        if (!is.null(selectType)) {
            # anything other than gene type is missing
            dfPlot = data.frame(geneIndex = queryHits(ols), bsIndex = subjectHits(ols),
                                geneType = selectType[queryHits(ols)]) %>% #TODO type
                dplyr::mutate(value = 1) %>%
                dplyr::select(-geneIndex) %>%
                tidyr::pivot_wider(names_from = geneType, values_from = value, values_fn = length, values_fill = 0) %>%
                dplyr::select(-bsIndex) %>%
                dplyr::mutate_all(~ ifelse(. > 1, 1, .))
        } else {
            # gene type is missing
            dfPlot = NULL
        }

    }
    object@plotData$assignToGenes$dataOverlaps = dfPlot

    # Case where there are no overlaps
    if (duplicatedBS == 0) {
        msg = paste0("No binding sites (",
                     format(totalBS, big.mark = ',', decimal.mark = "."),
                     ") on overlapping gene loci. \n")
        msg2 = paste0("Parameter 'overlaps' set to '",
                      overlaps, "' not in use. \n")
        if(!quiet) message(c(msg, msg2))
    }
    # Cases where there are multiple overlaps
    msg0 = paste0(duplicatedFraction, " (", duplicatedBS,"/", totalBS, ")",
                  " of binding sites overlap with multiple anno.genes in the given gene annotation. \n")
    if (overlaps == "hierarchy") {
        # apply hierarchy solution
        msg1 = paste0("Apply 'hierarchy' solution")
        if(!quiet) message(c(msg0, msg1))
        ruleMod = data.frame(gene_type = overlaps.rule, idx = seq_along(overlaps.rule))
        rngResolved = .resolveGeneOverlapsWithRule(rng = rngInitial, ols = ols, rule = ruleMod,
                                                   selectID = selectID, selectName = selectName,
                                                   selectType = selectType)
        if (length(rngInitial) != length(rngResolved)) {
            if(!quiet) message(paste0(format(length(rngInitial)-length(rngResolved), big.mark = ",", decimal.mark = "."),
                                      " binding sites could not be assigned to a gene range.\n"))
        }

    }
    if (overlaps == "frequency") {
        # apply frequency solution
        msg1 = paste0("Apply 'frequency' solution")
        if (!quiet) message(c(msg0, msg1))
        ruleFreq = selectType %>% table() %>% as.data.frame() %>%
            rename(gene_type = 1) %>% arrange(desc(Freq)) %>%
            mutate(idx = row_number())  %>% rename(n = 2)
        rngResolved = .resolveGeneOverlapsWithRule(rng = rngInitial, ols = ols, rule = ruleFreq,
                                                   selectID = selectID, selectName = selectName, selectType = selectType)
        if (length(rngInitial) != length(rngResolved)) {
            if (!quiet) message(paste0(format(length(rngInitial)-length(rngResolved), big.mark = ",", decimal.mark = "."),
                                       " binding sites could not be assigned to a gene range.\n"))
        }

    }
    if (overlaps == "remove") {
        # apply remove option
        msg1 = "Binding sites from overlapping loci are removed. This is not recommended. Please see options 'heirarchy' and 'frequency'. \n"
        warning(c(msg0, msg1))
        rngResolved = rngInitial[subjectHits(ols)]
        # manage meta columns
        if (!is.null(selectID)) {
            mcols(rngResolved) = cbind(mcols(rngResolved), geneID = selectID[queryHits(ols)])
        }
        if (!is.null(selectType)) {
            mcols(rngResolved) = cbind(mcols(rngResolved), geneType = selectType[queryHits(ols)])
        }
        if (!is.null(selectName)) {
            mcols(rngResolved) = cbind(mcols(rngResolved), geneName = selectName[queryHits(ols)])
        }
        # remove duplicates
        rngResolved = rngResolved[countOverlaps(rngResolved) == 1]
    }
    if (overlaps == "keep") {
        msg1 = "Binding sites from overlapping loci are kept. This is not recommended. Please see options 'heirarchy' and 'frequency'. \n"
        warning(c(msg0, msg1))
        rngResolved = rngInitial[subjectHits(ols)]
        # manage meta columns
        if (!is.null(selectID)) {
            mcols(rngResolved) = cbind(mcols(rngResolved), geneID = selectID[queryHits(ols)])
        }
        if (!is.null(selectType)) {
            mcols(rngResolved) = cbind(mcols(rngResolved), geneType = selectType[queryHits(ols)])
        }
        if (!is.null(selectName)) {
            mcols(rngResolved) = cbind(mcols(rngResolved), geneName = selectName[queryHits(ols)])
        }
    }

    # ---
    # Store results for plotting
    if (all(!is.null(selectID) & !is.null(selectType) & !is.null(selectName))) {
        # nothing is missing
        nCountGenes = rngResolved %>% mcols() %>% as.data.frame() %>%
            group_by(geneID) %>% dplyr::count(geneType) %>% ungroup() %>%
            dplyr::count(geneType) %>% arrange(desc(n))
        nCountBs = rngResolved %>% mcols() %>% as.data.frame() %>%
            dplyr::count(geneType) %>% arrange(desc(n))
        nCountAll = left_join(nCountBs, nCountGenes, by = "geneType") %>%
            rename("nBs" = "n.x", "nGenes" = "n.y")
    } else {
        # check what is missing
        if (!is.null(selectType)) {
            # anything other than gene type is missing
            nCountGenes = rngResolved %>% mcols() %>% as.data.frame() %>%
                group_by(geneID) %>% dplyr::count(geneType) %>% ungroup() %>%
                dplyr::count(geneType) %>% arrange(desc(n))
            nCountBs = rngResolved %>% mcols() %>% as.data.frame() %>%
                dplyr::count(geneType) %>% arrange(desc(n))
            nCountAll = left_join(nCountBs, nCountGenes, by = "geneType") %>%
                rename("nBs" = "n.x", "nGenes" = "n.y")
        }
        else {
            # gene type is missing
            nCountAll = NULL
        }

    }
    object@plotData$assignToGenes$dataSpectrum = nCountAll

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "assignToGenes()", class = "binding sites",
        nIn = length(rngInitial), nOut = length(rngResolved),
        per = paste0(round(length(rngResolved)/ length(rngInitial), digits = 2)*100,"%"),
        options = paste0("Overlaps=", optstr$overlaps, ", Source=", optstr$source,
                         ifelse(!rlang::is_empty(optstr$overlaps.rule),
                                paste0(", overlaps.rule=",
                                       paste(optstr$overlaps.rule, collapse = ">")), ""))
    )
    object@results = rbind(object@results, resultLine)

    objectOut = setRanges(object, rngResolved, quiet = quiet)
    return(objectOut)
}


#' Assign binding sites to their hosting transcript regions
#'
#' Function that assigns each binding site in the \code{\link{BSFDataSet}} to its
#' hosting transcript region given an annotation database (\code{anno.annoDB}), or
#' a GRanges list / \code{\link{CompressedGRangesList}} (\code{anno.transcriptRegionList})
#' that holds the ranges for the transcript regions of interest.
#'
#' Since the assignment is based on the overlaps of annotated transcript ranges
#' and binding sites, no matching meta data is needed.
#'
#' In the case of transcript regions overlaps are very frequent. To resolve these
#' cases the \code{\link{overlaps}} argument can be used. Option `frequency`
#' will take the most frequently observed transcript region, option `hierarchy`
#' works in conjunction with a user defined rule (\code{overlaps.rule}).
#' Options `flag` and `remove` will label binding sites with an ambiguous tag or
#' remove all overlapping cases, respectively.
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a \code{\link{BSFDataSet}} object with stored binding sites. This
#' means that ranges should be > 1
#' @param overlaps character; how overlapping transcript regions should be handled.
#' @param overlaps.rule character vector; a vector of transcript region
#' names that should be used to handle overlapping cases in a hierarchical manor.
#' The order of the vector is the order of the hierarchy.
#' @param anno.annoDB an object of class \code{OrganismDbi} that contains
#' the transcript region annotation (!!! Experimental !!!).
#' @param anno.transcriptRegionList an object of class \code{\link{CompressedGRangesList}}
#' that holds an ranges for each transcript region
#'
#' @param normalize.exclude.upper numeric; percentage value that indicates the
#' upper boundary for transcript region length to be considered when calculating
#' normalization factors for regions.
#' @param normalize.exclude.lower numeric; percentage value that indicates the
#' lower boundary for transcript region length to be considered when calculating
#' normalization factors for regions.
#'
#' @param quiet logical; whether to print messages
#'
#' @return an object of class \code{\link{BSFDataSet}} with binding sites having
#' hosting transcript region information added to their meta columns.
#'
#' @seealso \code{\link{BSFind}},
#' \code{\link{transcriptRegionOverlapsPlot}},
#' \code{\link{transcriptRegionSpectrumPlot}}
#'
#' @importFrom dplyr rename_with desc
#' @importFrom GenomicFeatures cds intronsByTranscript threeUTRsByTranscript fiveUTRsByTranscript
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#'
#' bds = makeBindingSites(object = bds, bsSize = 9)
#' bds = assignToGenes(bds, anno.genes = gns)
#' bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)
#'
#' @export
assignToTranscriptRegions <- function(object, # bindingSiteFinder
                                      overlaps = c("frequency", "hierarchy", "flag", "remove"), # overlapping loci solution options
                                      overlaps.rule = NULL, # rule to apply when option hierarchy is selected
                                      anno.annoDB = NULL, # annotation data base of class OrganismDbi; can be NULL -> but then anno.transcriptRegionList must be provided
                                      anno.transcriptRegionList = NULL, # GRangesList object with the transcript features to be used; can be NULL -> but then anno.annoDB must be provided
                                      normalize.exclude.upper = 0.02,
                                      normalize.exclude.lower = 0.02,
                                      quiet = FALSE
) {
    # local variables
    datasource <- `.` <- Freq <- transcript_type <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    # Check if none is specified
    if (is.null(anno.annoDB) & is.null(anno.transcriptRegionList)) {
        msg = paste0("None of the required annotation sources anno.annoDB or anno.transcriptRegionList was specified. ")
        stop(msg)
    }
    # Check if both are specified
    if (!is.null(anno.annoDB) & !is.null(anno.transcriptRegionList)) {
        msg = paste0("Both of the required annotation sources anno.annoDB or anno.transcriptRegionList are specified. Please provide only one of the two. ")
        stop(msg)
    }
    # anno.annoDB should be used
    if (!is.null(anno.annoDB) & is.null(anno.transcriptRegionList)) {
        stopifnot(is(anno.annoDB, "OrganismDb"))
        if (!is.null(anno.transcriptRegionList)) {
            msg = paste0("Parameter anno.annoDB and anno.transcriptRegionList are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            datasource = "anno.annoDB"
            # extract relevant annotation - transcripts
            cdseq = cds(anno.annoDB)
            intrns = unlist(intronsByTranscript(anno.annoDB))
            utrs3 = unlist(threeUTRsByTranscript(anno.annoDB))
            utrs5 = unlist(fiveUTRsByTranscript(anno.annoDB))
            anno.transcriptRegionList = GRangesList(CDS = cdseq, Intron = intrns, UTR3 = utrs3, UTR5 = utrs5)
            names(anno.transcriptRegionList) = toupper(names(anno.transcriptRegionList))
        }
    }
    # anno.transcriptRegionList should be used
    if (is.null(anno.annoDB) & !is.null(anno.transcriptRegionList) ) {
        stopifnot(is(anno.transcriptRegionList, "GenomicRangesList"))
        if (!is.null(anno.annoDB)) {
            msg = paste0("Parameter anno.annoDB and anno.transcriptRegionList are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            datasource = "anno.transcriptRegionList"
            # extract relevant annotation - transcripts
            anno.transcriptRegionList = anno.transcriptRegionList
            names(anno.transcriptRegionList) = toupper(names(anno.transcriptRegionList))
        }
    }
    # handle options (hierarchy, frequency, remove, keep)
    overlaps = match.arg(overlaps, choices = c("frequency", "hierarchy", "flag", "remove"))
    # Check multiple loci options
    if (overlaps == "hierarchy" & is.null(overlaps.rule)) {
        msg1 = paste0("Multiple region problem is set to be handled by 'hierarchy', but no overlaps.rule is provided. \n")
        msg2 = paste0("Change the 'overlaps' option or provide a valid overlaps.rule. \n")
        stop(c(msg1, msg2))
    }
    overlaps.rule = toupper(overlaps.rule)
    if (! all(overlaps.rule %in% names(anno.transcriptRegionList))){
        msg = paste0("Regions defined in overlaps.rule (", paste(overlaps.rule, collapse = " > "),
                     ") does not match the input region names (", paste(names(anno.transcriptRegionList), collapse = ", "), ")")
        stop(msg)
    }
    if (overlaps == "hierarchy" & is.null(overlaps.rule)) {
        msg1 = paste0("Overlap are set to be solved by 'hierarchy', but no overlaps.rule is provided. \n")
        msg2 = paste0("Change the 'overlaps' option or provide a valid overlaps.rule. \n")
        stop(c(msg1, msg2))
    }

    # ---
    # Store function parameters in list
    optstr = list(source = datasource, overlaps = overlaps, overlaps.rule = overlaps.rule)
    object@params$assignToTranscriptRegions = optstr

    # MAIN COMPUTE
    # --------------------------------------------------------------------------
    # Match transcript annotation with binding sites
    # --------------------------------------------------------------------------
    rngInitial = getRanges(object)
    rng = getRanges(object)

    cRange = resize(rng, width = 1, fix = "center")
    countDf = lapply(anno.transcriptRegionList, function(x) {
        countOverlaps(cRange, x)
    }) %>% as.data.frame() %>% rename_with(toupper)
    # Add intergenic cases
    countDf = countDf %>% mutate(OTHER = ifelse(rowSums(.) == 0, 1, 0))

    # ---
    # Store results for plotting
    plotDf = countDf
    plotDf[plotDf >= 1] = 1
    object@plotData$assignToTranscriptRegions$dataOverlaps = plotDf

    # Deal with multiple loci
    # --------------------------------------------------------------------------
    # Count out how many cases there are
    totalBs = nrow(countDf)
    numberOfLociHit = rowSums(countDf != 0)
    ambigousBs = nrow(countDf[numberOfLociHit != 1,])
    ambigousBsFraction = paste0(round(ambigousBs / totalBs * 100, digits = 2), "%")

    # Case where there are no overlaps
    if (ambigousBs == 0) {
        msg0 = paste0("No binding sites (", format(totalBs, big.mark = ',', decimal.mark = "."), ") on overlapping transcript loci. \n")
        msg1 = paste0("Parameter 'overlaps' set to '", overlaps, "' not in use. \n")
        if(!quiet) message(c(msg0, msg1))
        # assign regions
        countDfSorted = countDf
        countDfSorted[countDfSorted > 0] = 1
        col_names = colnames(countDfSorted)
        transcriptRegion = ifelse(rowSums(countDfSorted) == 1, col_names[max.col(countDfSorted, ties.method = "first")], "Ambiguous")
        mcols(rng) = cbind.data.frame(mcols(rng), transcriptRegion = transcriptRegion)
    }
    # Cases where we have to resolve overlapping loci
    if (ambigousBs > 0) {
        msg0 = paste0(ambigousBsFraction, " (", ambigousBs,"/", totalBs, ")",
                      " of binding sites overlap with multiple different transcript regions in the given annotation. \n")
        if (overlaps == "hierarchy") {
            # apply hierarchy solution
            msg1 = paste0("Apply 'hierarchy' solution. \n ")
            if(!quiet) message(c(msg0, msg1))
            countDfSorted = countDf[, overlaps.rule]
            countDfSorted[countDfSorted > 0] = 1
            transcriptRegion = names(countDfSorted)[max.col(countDfSorted, ties.method = "first")]
            mcols(rng) = cbind.data.frame(mcols(rng), transcriptRegion = transcriptRegion)
        }
        if (overlaps == "frequency") {
            msg1 = paste0("Apply 'frequency' solution. \n ")
            if(!quiet) message(c(msg0, msg1))
            countDfSorted = countDf
            # countDfSorted[countDfSorted > 0] = 1
            # get overlaps.rule by frequency
            overlaps.rule = names(countDfSorted)[max.col(countDfSorted, ties.method = "first")] %>%
                table() %>% as.data.frame() %>% rename(transcript_type = 1) %>%
                arrange(desc(Freq)) %>% pull(transcript_type) %>% as.character()
            # apply frequency solution
            countDfSorted = countDfSorted[, overlaps.rule]
            transcriptRegion = names(countDfSorted)[max.col(countDfSorted, ties.method = "first")]
            mcols(rng) = cbind.data.frame(mcols(rng), transcriptRegion = transcriptRegion)
        }
        if (overlaps == "flag") {
            msg1 = paste0("Binding sites with multiple different overlapping transcript regions are being flaged with the tag 'Ambiguous'. \n")
            if(!quiet) message(c(msg0, msg1))
            countDfSorted = countDf
            countDfSorted[countDfSorted > 0] = 1
            # flag multiple loci binding sites
            col_names = colnames(countDfSorted)
            transcriptRegion = ifelse(rowSums(countDfSorted) == 1, col_names[max.col(countDfSorted, ties.method = "first")], "Ambiguous")
            mcols(rng) = cbind.data.frame(mcols(rng), transcriptRegion = transcriptRegion)
            # remove mulitple loci binding sitges
            mcols(rng) = cbind.data.frame(mcols(rng), transcriptRegion = transcriptRegion)
        }
        if (overlaps == "remove") {
            msg1 = paste0("Binding sites with multiple different overlapping transcript regions are being removed. This is not recommended! \n")
            if(!quiet) message(c(msg0, msg1))
            countDfSorted = countDf
            countDfSorted[countDfSorted > 0] = 1
            # flag multiple loci binding sites
            col_names = colnames(countDfSorted)
            transcriptRegion = ifelse(rowSums(countDfSorted) == 1, col_names[max.col(countDfSorted, ties.method = "first")], "Ambiguous")
            mcols(rng) = cbind.data.frame(mcols(rng), transcriptRegion = transcriptRegion)
            # remove mulitple loci binding sitges
            mcols(rng) = cbind.data.frame(mcols(rng), transcriptRegion = transcriptRegion)
            rng = rng[rng$transcriptRegion != "Ambiguous"]
        }
    }

    # ---
    # Calculate width for length based normalization
    ## normalize by all expressed regions
    debugList = list(object = object, rng = rng)
    # saveRDS(debugList, file = "./debugList.rds")
    norm.factors = .calcNormalizeFactors(object, rng, anno.transcriptRegionList,
                                         normalize.exclude.lower = normalize.exclude.lower,
                                         normalize.exclude.upper = normalize.exclude.upper)

    # ---
    # Store results for plotting
    dfPlot = rng$transcriptRegion %>%
        table() %>%
        as.data.frame() %>%
        rename("TranscriptRegion" = ".") %>%
        arrange(desc(Freq)) %>%
        left_join(., y = norm.factors, by = "TranscriptRegion")
    object@plotData$assignToTranscriptRegions$dataSpectrum = dfPlot

    # ---
    # Store for results
    resultLine = data.frame(
        funName = "assignToTranscriptRegions()", class = "binding sites",
        nIn = length(rngInitial), nOut = length(rng),
        per = paste0(round(length(rng)/ length(rngInitial), digits = 2)*100,"%"),
        options = paste0("Overlaps=", optstr$overlaps, ", Source=", datasource,
                         ifelse(!rlang::is_empty(optstr$overlaps.rule), paste0(", overlaps.rule=", paste(optstr$overlaps.rule, collapse = ">")), ""))
    )
    object@results = rbind(object@results, resultLine)

    outObject = setRanges(object, rng, quiet = quiet)
    return(outObject)
}


#' Function to estimate the appropriate binding site width together with the
#' optimal gene-wise filter level.
#'
#' This function tests different width of binding sites for different gene-wise
#' filtering steps. For each test the signal-to-score ratio is calculated. The
#' mean over all gene-wise filterings at each binding site width is used to
#' extract the optimal width, which serves as anchor to select the optimal
#' gene-wise filter.
#'
#' Parameter estimation is done on a subset of all crosslink sites
#' (\code{est.subsetChromosome}).
#'
#' Gene-level filter can be tested with varying
#' levels of accuracy ranging from `finest` to `coarse`, representing 1% and
#' 20% steps, respectively.
#'
#' Binding site computation at each step can be done on three different accuracy
#' level (\code{bsResolution}). Option `fine` is equal to a normal run
#' of the \code{\link{makeBindingSites}} function. `medium` will perform
#' a shorter version of the binding site computation, skipping some of the
#' refinement steps. Option `coarse` will approximate binding sites by merged
#' crosslinks regions, aligning the center at the site with the highest score.
#'
#' For each binding site in each set given the defined resolutions a signal-to-
#' flank score ratio is calculated and the mean of this score per set is returned.
#' Next a mean of means is created which results in a single score for each
#' binding site width that was tested. The width that yielded the highest score
#' is selected as optimal. In addtion the \code{minimumStepGain} option
#' allows control over the minimum additional gain in the score that a tested
#' width has to have to be selected as the best option.
#'
#' To enhance the sensitivity of the binding site estimation, the sensitivity
#' mode exists. In this mode crosslink sites undergo a pre-filtering and merging
#' step, to exclude potential artifical peaks (experimental-, mapping-biases).
#' If sensitivity mode is activated the \code{est.minWidth} option should be set
#' to 1.
#'
#' The optimal geneFilter is selected as the first one that passes the merged
#' mean of the selected optimal binding site width.
#'
#' The function is part of the standard workflow performed by \code{\link{BSFind}}.
#'
#' @param object a \code{\link{BSFDataSet}} object with stored crosslink sites.
#' This means that ranges should have a width = 1.
#' @param bsResolution character; level of resolution at which different binding
#' site width should be tested
#' @param geneResolution character; level of resolution at which gene-wise filtering
#' steps should be tested
#' @param est.maxBsWidth numeric; the largest binding site width which should
#' considered in the testing
#' @param est.minimumStepGain numeric; the minimum additional gain in the score
#' in percent the next binding site width has to have, to be selected as best option
#' @param est.maxSites numeric; maximum number of PureCLIP sites that are used;
#' @param est.subsetChromosome character; define on which chromosome the
#' estimation should be done in function \code{\link{estimateBsWidth}}
#' @param est.minWidth the minimum size of regions that are subjected to the
#' iterative merging routine, after the initial region concatenation.
#' @param est.offset constant added to the flanking count in the signal-to-flank
#' ratio calculation to avoid division by Zero
#'
#' @param sensitive logical; whether to enable sensitive pre-filtering before
#' binding site merging or not
#' @param sensitive.size numeric; the size (in nucleotides) of the merged
#' sensitive region
#' @param sensitive.minWidth numeric; the minimum size (in nucleoties) of the
#' merged sensitive region
#'
#' @param anno.annoDB an object of class \code{OrganismDbi} that contains
#' the gene annotation (!!! Experimental !!!).
#' @param anno.genes an object of class \code{\link{GenomicRanges}} that represents
#' the gene ranges directly
#' @param bsResolution.steps numeric vector; option to use a user defined threshold
#' for binding site width directly. Overwrites \code{bsResolution}
#' @param geneResolution.steps numeric vector; option to use a user defined threshold
#' vector for gene-wise filtering resolution. Overwrites \code{geneResolution}
#' @param quiet logical; whether to print messages
#' @param veryQuiet logical; whether to suppress all messages
#' @param reportScoresPerBindingSite report the ratio score for each binding site
#' separately. Warning! This is for debugging and testing only. Downstream
#' functions can be impaired.
#' @param ... additional arguments passed to \code{\link{pureClipGeneWiseFilter}}
#'
#' @return an object of class \code{\link{BSFDataSet}} with binding sites with
#' the `params` slots `bsSize` and `geneFilter` being filled
#'
#' @seealso \code{\link{BSFind}},
#' \code{\link{estimateBsWidthPlot}}
#'
#' @importFrom dplyr summarise slice_head slice_tail
#' @importFrom utils head
#' @importFrom GenomicFeatures genes
#' @importFrom stats sd
#'
#' @examples
#' # load clip data
#' files <- system.file("extdata", package="BindingSiteFinder")
#' load(list.files(files, pattern = ".rda$", full.names = TRUE))
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
#' load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
#' estimateBsWidth(bds, anno.genes = gns, est.maxBsWidth = 19,
#'  geneResolution = "coarse", bsResolution = "coarse", est.subsetChromosome = "chr22")
#'
#' @export
estimateBsWidth <- function(object, # BindingSiteFinder object
                            bsResolution = c("medium", "fine", "coarse"),
                            geneResolution = c("medium", "coarse", "fine", "finest"),
                            est.maxBsWidth = 13,
                            est.minimumStepGain = 0.02,
                            est.maxSites = Inf,
                            est.subsetChromosome = "chr1", #
                            est.minWidth = 2,
                            est.offset = 1,
                            sensitive = FALSE,
                            sensitive.size = 5,
                            sensitive.minWidth = 2,
                            anno.annoDB = NULL,
                            anno.genes = NULL,
                            bsResolution.steps = NULL,
                            geneResolution.steps = NULL,
                            quiet = TRUE,
                            veryQuiet = FALSE,
                            reportScoresPerBindingSite = FALSE,
                            ...
) {

    # initialize local variables
    bsSize <- ms <- growth_per <- sel <- geneWiseFilter <- est.option <- signalToFlankRatio <- NULL

    # INPUT CHECKS
    # --------------------------------------------------------------------------
    stopifnot(is(object, "BSFDataSet"))
    stopifnot(is.logical(quiet))

    if (c(est.maxBsWidth %% 2) == 0) {
        stop("est.maxBsWidth is even. An odd number is required to have a distinct binding site center.")
    }

    # Check if none is specified
    if (is.null(anno.annoDB) & is.null(anno.genes)) {
        msg = paste0("None of the required annotation sources anno.annoDB or anno.genes was specified. ")
        stop(msg)
    }
    # Check if both are specified
    if (!is.null(anno.annoDB) & !is.null(anno.genes)) {
        msg = paste0("Both of the required annotation sources anno.annoDB or anno.genes are specified. Please provide only one of the two. ")
        stop(msg)
    }
    # Checks if anno.annoDB should be used
    if (!is.null(anno.annoDB) & is.null(anno.genes)) {
        stopifnot(is(anno.annoDB, "OrganismDb"))
        if (!is.null(anno.genes)) {
            msg = paste0("Parameter anno.annoDB and anno.genes are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            # extract relevant annotation
            anno.genes = genes(anno.annoDB)
        }
    }
    # Checks if anno.genes should be used
    if (is.null(anno.annoDB) & !is.null(anno.genes)) {
        stopifnot(is(anno.genes, "GenomicRanges"))
        if (!is.null(anno.annoDB)) {
            msg = paste0("Parameter anno.annoDB and anno.genes are specified at the same time. Use only one of them.")
            stop(msg)
        } else {
            # extract relevant annotation
            anno.genes = anno.genes
        }
    }

    # handle gene filter resolution
    geneResolution = match.arg(geneResolution, choices = c("medium", "coarse", "fine", "finest"))
    if (!is.null(geneResolution.steps)) {
        geneResolution.steps = geneResolution.steps
    } else {
        if (geneResolution == "coarse") {
            geneResolution.steps = seq(from = 0, to = 1, by = 0.2)
            geneResolution.steps = geneResolution.steps[1:length(geneResolution.steps)-1]
        }
        if (geneResolution == "medium") {
            geneResolution.steps = seq(from = 0, to = 1, by = 0.1)
            geneResolution.steps = geneResolution.steps[1:length(geneResolution.steps)-1]
        }
        if (geneResolution == "fine") {
            geneResolution.steps = seq(from = 0, to = 1, by = 0.05)
            geneResolution.steps = geneResolution.steps[1:length(geneResolution.steps)-1]
        }
        if (geneResolution == "finest") {
            geneResolution.steps = seq(from = 0, to = 1, by = 0.01)
            geneResolution.steps = geneResolution.steps[1:length(geneResolution.steps)-1]
        }
    }

    # handle binding site compute resolution
    bsResolution = match.arg(bsResolution, choices = c("medium", "fine", "coarse"))

    # handle est.maxBsWidth
    if (!is.null(bsResolution.steps)) {
        bsResolution.steps = bsResolution.steps
    } else {
        bsResolution.steps = seq(from = 3, to = est.maxBsWidth, by = 2)
    }

    # ---
    # Store function parameters in list
    optstr = list(bsResolution = bsResolution,
                  geneResolution = geneResolution,
                  est.maxBsWidth = est.maxBsWidth,
                  est.maxSites = est.maxSites,
                  est.minimumStepGain = est.minimumStepGain,
                  est.subsetChromosome = est.subsetChromosome,
                  est.minWidth = est.minWidth,
                  est.offset = est.offset,
                  sensitive = sensitive,
                  sensitive.size = sensitive.size,
                  sensitive.minWidth = sensitive.minWidth)
    object@params$estimateBsWidth = optstr

    # PREPARE TEST RANGES + SIGNAL
    # --------------------------------------------------------------------------
    checkRng = getRanges(object)
    # check if subset is part of the seqnames from ranges
    if (!all(est.subsetChromosome %in% levels(seqnames(checkRng)))){
        msg = paste0("Chromosome to estimate on (", est.subsetChromosome, "), is not included in the ranges: ", paste(levels(seqnames(checkRng)), collapse = ", "))
        stop(msg)
    }

    # limit estimation to a specific chromosome
    if (!is.null(est.subsetChromosome)) {
        redObj = .subsetByChr(object, chr = est.subsetChromosome, quiet = quiet)
    } else {
        redObj = object
    }

    # limit estimation to a maximum number of sites
    estRng = getRanges(redObj)
    if (length(estRng) > est.maxSites) {
        estRng = head(estRng, est.maxSites)
        redObj = setRanges(redObj, estRng, quiet = quiet)
    }

    # limit the clip signal to the ranges used for estimation (plus some extra offset)
    if (!is.null(est.subsetChromosome)) {
       # reduce signal to frame
        maxFrame = ceiling((max(bsResolution.steps) *3))
        redObj = .reduceSignalToFrame(redObj, frame = maxFrame, quiet = quiet)
    } else {
        # don't reduce signal
        redObj = redObj
    }

    # collapse signal from replicates
    sgnMerge = .collapseSamples(getSignal(redObj))


    # MAIN COMPUTE
    # --------------------------------------------------------------------------

    # a counter that shows to how many percent the computation is done
    counterTotalIterations = length(geneResolution.steps) * length(bsResolution.steps)
    counterCurrentIterations = 0

    # print start message
    msg = paste0("Estimation at: ", round((counterCurrentIterations/counterTotalIterations)*100 ), "%"  )
    if(!veryQuiet) message(msg)

    # calculate binding sites for each filter step and width
    scoreAllDf = lapply(geneResolution.steps, function(bsFilterStep){
        # bsFilterStep = 0.2

        # apply current gene-wise filter (has two modes)
        if (isTRUE(sensitive)) {
            # -> sensitive mode
            # define regions in which sites should be kept
            cRng = getRanges(redObj)
            cRngMerge = reduce(cRng, min.gapwidth = sensitive.size)
            cRngMerge = cRngMerge[width(cRngMerge) >= sensitive.minWidth]

            # currFilterObj = pureClipGeneWiseFilter(object = redObj, anno.genes = anno.genes, cutoff = bsFilterStep, quiet = quiet, ...)
            currFilterObj = pureClipGeneWiseFilter(object = redObj, anno.genes = anno.genes, cutoff = bsFilterStep, quiet = quiet)
            cRngFilter = getRanges(currFilterObj)
            currRng = IRanges::subsetByOverlaps(cRngFilter, cRngMerge)

        } else {
            # -> normal mode
            # currFilterObj = pureClipGeneWiseFilter(object = redObj, anno.genes = anno.genes, cutoff = bsFilterStep, quiet = quiet, ...)
            currFilterObj = pureClipGeneWiseFilter(object = redObj, anno.genes = anno.genes, cutoff = bsFilterStep, quiet = quiet)
            currRng = getRanges(currFilterObj)
        }

        # calculate binding sites for current bsWidth
        rngPerWidth = lapply(bsResolution.steps, function(bsWidthStep){
            if (bsResolution == "fine") {
                # calculate with full binding sites
                currBsObj = makeBindingSites(object = currFilterObj,
                                             bsSize = bsWidthStep,
                                             minWidth = est.minWidth,
                                             quiet = quiet)
                currBs = getRanges(currBsObj)
            }
            if (bsResolution == "medium") {
                # approximate binding sites by a single merge and extend round
                currBs = .approximateBindingSites_medium(rng = currRng,
                                                         sgn = sgnMerge,
                                                         bsSize = bsWidthStep,
                                                         minWidth = est.minWidth)
            }
            if (bsResolution == "coarse") {
                # approximate binding sites by reduced pureclip sites center
                currBs = .approximateBindingSites_coarse(rng = currRng,
                                                         bsSize = bsWidthStep,
                                                         minWidth = est.minWidth)
            }
            return(currBs)
        })
        rngPerWidth = unlist(GRangesList(rngPerWidth))

        # handle all plus ranges
        cRangePlus = subset(rngPerWidth, strand == "+")
        if (length(cRangePlus) > 0) {
            bsSumPlus = sum(sgnMerge$signalPlus[cRangePlus])
            extendedRangePlus = cRangePlus + cRangePlus$bsSize
            exSumPlus = sum(sgnMerge$signalPlus[extendedRangePlus])
            # extendedRangePlus$signalToFlankRatio = (bsSumPlus) / (((exSumPlus - bsSumPlus) / 2) + est.offset)
            extendedRangePlus$signalToFlankRatio = (bsSumPlus / exSumPlus)
        } else {
            extendedRangePlus = cRangePlus
        }

        # handle all minus ranges
        cRangeMinus = subset(rngPerWidth, strand == "-")
        if (length(cRangeMinus) > 0) {
            bsSumMinus = sum(sgnMerge$signalMinus[cRangeMinus])
            extendedRangeMinus = cRangeMinus + cRangeMinus$bsSize
            exSumMinus = sum(sgnMerge$signalMinus[extendedRangeMinus])
            # extendedRangeMinus$signalToFlankRatio = (bsSumMinus) / (((exSumMinus - bsSumMinus) / 2) + est.offset)
            extendedRangeMinus$signalToFlankRatio = (bsSumMinus / exSumMinus)
        } else {
            extendedRangeMinus = cRangeMinus
        }

        if (isTRUE(reportScoresPerBindingSite)) {
            # report the scores for each binding site
            # -> does not allow the usage of the standard plot
            df = rbind(as.data.frame(mcols(extendedRangeMinus)), as.data.frame(mcols(extendedRangePlus))) %>%
                group_by(bsSize) %>%
                mutate(geneWiseFilter = bsFilterStep)

        } else {
            # combine both and make results dataframe
            df = rbind(as.data.frame(mcols(extendedRangeMinus)), as.data.frame(mcols(extendedRangePlus))) %>%
                group_by(bsSize) %>%
                summarize(signalToFlankRatio = median(signalToFlankRatio), .groups = "keep") %>%
                mutate(geneWiseFilter = bsFilterStep)
        }

        # update chunk counter
        counterCurrentIterations <<- counterCurrentIterations + length(bsResolution.steps)
        msg = paste0("Estimation at: ", round((counterCurrentIterations/counterTotalIterations)*100 ), "%"  )
        if(!veryQuiet) message(msg)
        return(df)
    })

    df = do.call("rbind", scoreAllDf)

    # ---
    # Store results for plotting
    if (isTRUE(reportScoresPerBindingSite)){
        object@plotData$estimateBsWidth$dataEnhanced = df
    } else {
        object@plotData$estimateBsWidth$data = df
    }

    resEstimate = .findFirstMaximum(df, est.minimumStepGain = est.minimumStepGain)
    est.bsSize = resEstimate$est.bsSize
    est.geneFilter = resEstimate$est.geneFilter
    est.option = resEstimate$est.option

    if (est.option == "local") {
        msg = paste0("No global maximum found with est.maxBsWidth=", est.maxBsWidth, ". Using local maximum instead.\n")
        if(!veryQuiet) warning(msg)
    }
    if (est.option == "fallback") {
        msg = paste0("No global or local maximum found with est.maxBsWidth=", est.maxBsWidth, ". Falling back to simple maximum instead.\n")
        if(!veryQuiet) warning(msg)
    }
    if (est.option == "error") {
        msg = paste0("No maximum found at all. Raise 'est.minimumStepGain' argument/, change bsResolution and/or geneResolution arguments/ raise est.maxBsWidth and check visually. \n")
        stop(msg)
    }

    # Store estimate type in options (global/ local)
    optstr$option = est.option
    object@params$estimateBsWidth = optstr

    # Store estimated parameters
    object@params$bsSize = est.bsSize
    object@params$geneFilter = est.geneFilter

    # ---
    # Store for results
    rng = getRanges(object)
    resultLine = data.frame(
        funName = "estimateBsWidth()", class = "estimate",
        nIn = length(rng), nOut = length(rng),
        per = paste0(round(length(rng)/ length(rng), digits = 2)*100,"%"),
        options = paste0("bsResolution=", bsResolution, ", geneResolution=", geneResolution,
                         ", est.maxBsWidth=", est.maxBsWidth, ", est.maxSites=", est.maxSites, ", est.subsetChromosome=", paste(est.subsetChromosome, collapse = ","))
    )
    object@results = rbind(object@results, resultLine)

    return(object)
}

test_that("Test for estimateBsWidth()", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    # load genes
    load(list.files(files, pattern = "gns.rds$", full.names = TRUE))
    # load transcript region list
    load(list.files(files, pattern = "regions.rds$", full.names = TRUE))

    # Test for correct input
    expect_error(estimateBsWidth(object = bds, anno.genes = gns, est.maxBsWidth = 7, geneResolution = "medium", bsResolution = "medium"))
    expect_error(estimateBsWidth(object = bds, anno.genes = gns, est.maxBsWidth = 7, geneResolution = "medium", bsResolution = "medium",
                                   est.subsetChromosome = "chr21", quiet = TRUE))

    # Test for correct execution
    expect_error(estimateBsWidth(object = bds, anno.genes = gns, est.maxBsWidth = 7, geneResolution = "medium", bsResolution = "coarse",
                                   est.subsetChromosome = "chr22", quiet = TRUE, veryQuiet = TRUE))

    expect_silent(estimateBsWidth(object = bds, anno.genes = gns, est.maxBsWidth = 13, geneResolution = "medium", bsResolution = "medium",
                                  est.subsetChromosome = "chr22", quiet = TRUE, veryQuiet = TRUE))

})

test_that("Test for pureClipGlobalFilter()", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # tests with genes as annotion resources
    expect_silent(pureClipGlobalFilter(object = bds, match.score = "score"))
    expect_silent(pureClipGlobalFilter(object = bds, cutoff = 0.1, match.score = "score"))
    expect_error(pureClipGlobalFilter(object = bds, cutoff = 0.1, match.score = "score2"))

    rng = getRanges(bds)
    mcols(rng)$newScore = rng$score
    mcols(rng)$score = NULL
    objNew = setRanges(bds, rng)

    expect_error(pureClipGlobalFilter(object = objNew, cutoff = 0.1, match.score = "score"))
    expect_silent(pureClipGlobalFilter(object = objNew, cutoff = 0.1, match.score = "newScore"))
})

test_that("Test for assignToGenes()", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    # load genes
    load(list.files(files, pattern = "gns.rds$", full.names = TRUE))

    # Prepare objects for test
    obj1 = pureClipGlobalFilter(object = bds, cutoff = 0.05)
    obj2_manual = pureClipGeneWiseFilter(object = obj1, anno.genes = gns, quiet = TRUE)
    # obj2_db = pureClipGeneWiseFilter(object = obj1, anno.annoDB = Homo.sapiens38)
    obj3_manual = makeBindingSites(object = obj2_manual, bsSize = 9)
    # obj3_db = makeBindingSites(object = obj2_db, bsSize = 9)

    # Prepare rule
    selectTerms = c("protein_coding", "miRNA", "tRNA", "lincRNA", "snRNA")
    GENE_RULE = unique(gns$gene_type)
    GENE_RULE = GENE_RULE[!GENE_RULE %in% selectTerms]
    GENE_RULE = c(selectTerms, GENE_RULE)
    rule = GENE_RULE

    # test with OrganismDB as annotation source
    # expect_message(assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38))
    # expect_message(assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "frequency"))
    # expect_message(expect_error(assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "hierarchy")))
    # expect_message(assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "hierarchy", rule = rule))
    # expect_message(expect_warning(assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "remove")))
    # expect_message(expect_warning(assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "keep")))
    # expect_message(expect_error(assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "test")))

    # test with manual as annotation source
    expect_message(assignToGenes(object = obj3_manual, anno.genes = gns))
    expect_message(assignToGenes(object = obj3_manual, anno.genes = gns, overlaps = "frequency"))
    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns, overlaps = "hierarchy"))
    expect_message(assignToGenes(object = obj3_manual, anno.genes = gns, overlaps = "hierarchy", overlaps.rule = rule))
    expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns, overlaps = "remove"))
    expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns, overlaps = "keep"))
    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns, overlaps = "test"))

    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns, geneMatchID = "GeneID"))
    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns, geneMatchName = "GeneName"))
    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns, geneMatchType = "GeneType"))

    # test quiet
    expect_silent(assignToGenes(object = obj3_manual, anno.genes = gns, quiet = TRUE))

    # test with missing gene annotation meta columns
    gns1 = gns
    gns1$gene_type = NULL
    expect_message(expect_error(assignToGenes(object = obj3_manual, anno.genes = gns1)))
    expect_message(expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns1, overlaps = "remove")))
    expect_message(expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns1, overlaps = "keep")))

    gns1 = gns
    gns1$gene_name = NULL
    expect_message(assignToGenes(object = obj3_manual, anno.genes = gns1))
    expect_message(expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns1, overlaps = "remove")))
    expect_message(expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns1, overlaps = "keep")))

    gns1 = gns
    gns1$gene_id = NULL
    expect_message(assignToGenes(object = obj3_manual, anno.genes = gns1))
    expect_message(expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns1, overlaps = "remove")))
    expect_message(expect_warning(assignToGenes(object = obj3_manual, anno.genes = gns1, overlaps = "keep")))
    
    # test with duplicated gene annotation meta columns
    gns1 = gns
    meta = elementMetadata(gns1)
    meta = cbind(meta, gene_id = meta$gene_id)
    elementMetadata(gns1) = meta
    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns1))
    
    gns1 = gns
    meta = elementMetadata(gns1)
    meta = cbind(meta, gene_type = meta$gene_type)
    elementMetadata(gns1) = meta
    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns1))
    
    gns1 = gns
    meta = elementMetadata(gns1)
    meta = cbind(meta, gene_name = meta$gene_name)
    elementMetadata(gns1) = meta
    expect_error(assignToGenes(object = obj3_manual, anno.genes = gns1))
    

})

test_that("Test for assignToTranscriptRegions()", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    # load genes
    load(list.files(files, pattern = "gns.rds$", full.names = TRUE))
    # load transcript region list
    load(list.files(files, pattern = "regions.rds$", full.names = TRUE))


    # Prepare objects for test
    obj1 = pureClipGlobalFilter(object = bds, cutoff = 0.05)
    obj2_manual = pureClipGeneWiseFilter(object = obj1, anno.genes = gns, quiet = TRUE)
    # obj2_db = pureClipGeneWiseFilter(object = obj1, anno.annoDB = Homo.sapiens38)
    obj3_manual = makeBindingSites(object = obj2_manual, bsSize = 9, quiet = TRUE)
    # obj3_db = makeBindingSites(object = obj2_db, bsSize = 9)
    obj4_manual = assignToGenes(object = obj3_manual, anno.genes = gns, quiet = TRUE)
    # obj4_db = assignToGenes(object = obj3_db, anno.annoDB = Homo.sapiens38)

    # test with OrganismDB as annotation source
    # expect_message(assignToTranscriptRegions(object = obj4_db, anno.annoDB = Homo.sapiens38))
    # expect_message(assignToTranscriptRegions(object = obj4_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "frequency"))
    # expect_message(assignToTranscriptRegions(object = obj4_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "flag"))
    # expect_message(assignToTranscriptRegions(object = obj4_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "remove"))
    #
    # expect_message(assignToTranscriptRegions(object = obj4_db, anno.annoDB = Homo.sapiens38,
    #                                          overlappingLoci = "hierarchy", rule = c("utr3", "cds", "utr5", "intron")))
    # expect_error(assignToTranscriptRegions(object = obj4_db, anno.annoDB = Homo.sapiens38, overlappingLoci = "hierarchy"))
    # expect_error(assignToTranscriptRegions(object = obj4_db, anno.annoDB = Homo.sapiens38,
    #                                        overlappingLoci = "hierarchy", rule = c("hamster", "maus")))

    # test with manual as annotation source
    expect_message(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions))
    expect_message(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions, overlaps = "frequency"))
    expect_message(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions, overlaps = "flag"))
    expect_message(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions, overlaps = "remove"))

    expect_message(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions,
                                             overlaps = "hierarchy", overlaps.rule = c("utr3", "cds", "utr5", "intron")))
    expect_error(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions, overlaps = "hierarchy"))
    expect_error(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions,
                                           overlaps = "hierarchy", overlaps.rule = c("hamster", "maus")))

    x = assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions, overlaps = "frequency")
    expect_equal("frequency", x@params$assignToTranscriptRegions$overlaps)
    x = assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions, overlaps = "flag")
    expect_equal("flag", x@params$assignToTranscriptRegions$overlaps)

    # test quiet
    expect_silent(assignToTranscriptRegions(object = obj4_manual, anno.transcriptRegionList = regions, overlaps = "frequency", quiet = TRUE))
})

test_that("Test for pureClipGeneWiseFilter()", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    # load genes
    load(list.files(files, pattern = "gns.rds$", full.names = TRUE))

    # tests with anno.genes as annotion resources
    expect_message(pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.2, overlaps = "keepSingle"))
    expect_warning(pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.2, overlaps = "removeAll"))
    expect_warning(pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.2, overlaps = "keepAll"))

    expect_message(pureClipGeneWiseFilter(object = bds, anno.genes = gns, overlaps = "keepSingle"))
    expect_warning(pureClipGeneWiseFilter(object = bds, anno.genes = gns, overlaps = "removeAll"))
    expect_warning(pureClipGeneWiseFilter(object = bds, anno.genes = gns, overlaps = "keepAll"))

    expect_message(pureClipGeneWiseFilter(object = bds, anno.genes = gns))

    expect_error(pureClipGeneWiseFilter(object = bds))

    # tests with anno.annoDB as annotaiton resource
    # expect_message(pureClipGeneWiseFilter(object = bds, anno.annoDB = Homo.sapiens38, cutoff = 0.2, overlaps = "keepSingle"))
    # expect_message(expect_warning(pureClipGeneWiseFilter(object = bds, anno.annoDB = Homo.sapiens38, cutoff = 0.2, overlaps = "removeAll")))
    # expect_message(expect_warning(pureClipGeneWiseFilter(object = bds, anno.annoDB = Homo.sapiens38, cutoff = 0.2, overlaps = "keepAll")))
    #
    # expect_message(pureClipGeneWiseFilter(object = bds, anno.annoDB = Homo.sapiens38, overlaps = "keepSingle"))
    # expect_message(expect_warning(pureClipGeneWiseFilter(object = bds, anno.annoDB = Homo.sapiens38, overlaps = "removeAll")))
    # expect_message(expect_warning(pureClipGeneWiseFilter(object = bds, anno.annoDB = Homo.sapiens38, overlaps = "keepAll")))
    #
    # expect_message(pureClipGeneWiseFilter(object = bds, anno.annoDB = Homo.sapiens38))
    #
    # expect_error(pureClipGeneWiseFilter(object = bds, anno.genes = gns, anno.annoDB = Homo.sapiens38))

    # check cutoff 0
    testObj1 = expect_message(pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0, overlaps = "keepSingle"))
    expect_identical(length(getRanges(testObj1)), length(getRanges(bds)))

    # check quiet mode
    expect_silent(pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.1, quiet = TRUE))

})

test_that("Binding site merging works", {
    # load data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))

    # test range of subset 1
    rng1 = getRanges(bds)[1:100]
    rng2 = getRanges(bds[1:100])
    expect_identical(rng1, rng2)

    # test range of subset 2
    set.seed(1234)
    samp = sample(c(1:length(getRanges(bds))), 10)
    rng1 = getRanges(bds)[samp]
    rng2 = getRanges(bds[samp])
    expect_identical(rng1, rng2)

    # check signal for correct subsetting - no drop
    set.seed(1234)
    samp = sample(c(1:length(getRanges(bds))), 10)
    rng1 = getRanges(bds)[samp]
    rng1 = keepSeqlevels(rng1, value = unique(seqnames(rng1)))
    rng1Seqlevels = seqlevels(rng1)
    sgn2 = getSignal(bds[samp])
    sgn2Seqlevels = unique(unlist(lapply(sgn2, function(x){lapply(x, function(y){names(y)})})))
    expect_identical(sort(rng1Seqlevels), sort(sgn2Seqlevels))
    # check signal for correct subsetting - with drop 1
    set.seed(1234)
    samp = sample(c(1:length(getRanges(bds))), 10)
    rng1 = getRanges(bds)[samp, drop = TRUE]
    rng1 = keepSeqlevels(rng1, value = unique(seqnames(rng1)))
    rng1Seqlevels = seqlevels(rng1)
    sgn2 = getSignal(bds[samp, drop = TRUE])
    sgn2Seqlevels = unique(unlist(lapply(sgn2, function(x){lapply(x, function(y){names(y)})})))
    expect_identical(sort(rng1Seqlevels), sort(sgn2Seqlevels))
    # check signal for correct subsetting - with drop 2
    sgn1 = getSignal(bds[samp])
    sgn2 = getSignal(bds[samp, drop = TRUE])
    expect_false(identical(sgn1, sgn2))
    sgn1 = getSignal(bds[samp])
    sgn2 = getSignal(bds[samp, drop = FALSE])
    expect_true(identical(sgn1, sgn2))
    sgn1 = getSignal(bds[samp])
    sgn2 = getSignal(bds[samp])
    expect_true(identical(sgn1, sgn2))
})

---
title: "Definition of binding sites from iCLIP signal"
author:
- name: Mirko Brüggemann
  affiliation: Buchmann Institute for Molecular Life Sciences (BMLS), Frankfurt Germany
- name: Kathi Zarnack
  affiliation: Buchmann Institute for Molecular Life Sciences (BMLS), Frankfurt Germany
output:
  BiocStyle::html_document:
    toc_float: true
package: BindingSiteFinder
abstract: |
    Precise knowledge on the binding sites of an RNA-binding protein (RBP) is key to understand (post-) transcriptional regulatory processes. The package BindingSiteFinder provides all functionalities to define exact binding sites defined from iCLIP data. The following vignette describes the complete workflow. 
vignette: |
    %\VignetteIndexEntry{Definition of binding sites from iCLIP signal}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r setup, echo=FALSE, results="hide"}
knitr::knit_hooks$set(optipng = knitr::hook_optipng)
knitr::opts_chunk$set(tidy = FALSE,
                      cache = FALSE,
                      dev = "png",
                      message = FALSE, error = FALSE, warning = TRUE,
                      optipng = '-o7')
this.dpi = 100
```	

```{r, echo=FALSE, results="hide", warning=FALSE}
suppressPackageStartupMessages({
    library(GenomicRanges)
    library(GenomicAlignments)
    library(rtracklayer)
    library(ggplot2)
    library(tidyr)
    library(ComplexHeatmap)
    library(BindingSiteFinder)
    library(forcats)
    library(dplyr)
})
```

# Preface

## Motivation

Most cellular processes are regulated by RNA-binding proteins (RBPs). Knowledge on their exact positioning can be obtained from individual-nucleotide resolution UV crosslinking and immunoprecipitation (iCLIP) experiments. In a recent publication we described a complete analysis workflow to detect RBP binding sites from iCLIP data. The workflow covers all essential steps from quality control of sequencing reads, different peak calling options, to the downstream analysis and definition of binding sites. The pre-processing and peak calling steps rely on publicly available software, whereas the definition of the final binding sites follows a custom procedure implemented in BindingSiteFinder. This vignette explains how equally sized binding sites can be defined from a genome-wide iCLIP coverage.

## Prerequisites

The workflow described herein is based on our recently published complete iCLIP analysis pipeline [@busch2020]. Thus, we expect the user to have preprocessed their iCLIP sequencing reads up to the point of the peak calling step. In brief, this includes basic processing of the sequencing reads, such as quality filtering, barcode handling, mapping and the generation of a single nucleotide crosslink file for all replicates under consideration. As we describe in our manuscript, replicate .bam files may or may not be merged prior to peak calling, for which we suggest PureCLIP [@Krakau2017]. For simplicity, we address only the case in which peak calling was performed on the merge of all replicates.

![**Overview of the preprocessing workflow**. ](images/preprocessing.png){#preprocessing}

## Installation

The `r Biocpkg("BindingSiteFinder")` package is available at [https://bioconductor.org](https://bioconductor.org) and can be installed via `BiocManager::install`:

```{r BiocManager, eval=FALSE}
if (!require("BiocManager"))
    install.packages("BiocManager")
BiocManager::install("BindingSiteFinder")
```

**Note:** If you use BindingSiteFinder in published research, please cite:

> Busch, A., Brüggemann, M., Ebersberger, S., & Zarnack, K. (2020) iCLIP data analysis: A complete pipeline from sequencing reads to RBP binding sites. *Methods*, 178, 49-62. <https://doi.org/10.1016/j.ymeth.2019.11.008>




# Quick start

This part is for the impatient user that wants to start as fast as possible. For that reason this section contains only the most basic steps and function calls to get to the final binding site set. For more details on each step see the [Standard workflow](# Standard workflow) section.

## A quick look at the input data

It all starts with your input iCLIP data, which has to be in form of a `BSFDataSet`. This data-structure combines crosslink sites (the output from *PureCLIP*) with the individual crosslinks per replicate (iCLIP coverage). Further instructions on the data-set constructions are given in the chapter [Construction of the BindingSiteFinder dataset](## Construction of the BindingSiteFinder dataset).

```{r}
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds
```

To intersect binding sites with genomic features, such as genes and transcript regions the respective regions have to be presented as `GenomicRanges` objects. Further instructions on how to import annotations from various sources are given in chapter [Construct the annotation objects](## Construct the annotation objects).

```{r}
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
gns
```

```{r}
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
regions
```


## A quick way to define binding sites

The fastest way to accurately define binding site with very little manual intervention is running the wrapper function `BSFind()`. It will perform all necessary computations and will store all results and diagnostics internally. Various plotting functions can be used to make all sorts of diagnostics. Export functions can be used to extract resulting binding sites for further processing. Each step is covered in more detail in the [Standard workflow](# Standard workflow) section.

```{r, echo=TRUE, eval=FALSE}
# run BSFind to compute binding sites
bdsQuick = BSFind(object = bds, anno.genes = gns, anno.transcriptRegionList = regions, est.subsetChromosome = "chr22")
# export output as .bed
exportToBED(bdsQuick, con = "./myBindingSites.bed")
```

## The quick picture, two figures for the lazy one

To get a fast glance at the data-set with minimal user intervention that is also visually appealing, the `quickFigure()` function exists. 

```{r, echo=TRUE, eval=FALSE}
# not run
quickFigure(bdsQuick, what = "main", save.filename = "./my_fig_main.pdf", save.device = "pdf")
```

![**Quick Figure - Main**. ](../vignettes/my_fig_main.png){#main}

```{r, echo=TRUE, eval=FALSE}
# not run
quickFigure(bdsQuick, what = "supp", save.filename = "./my_fig_supp.pdf", save.device = "pdf")
```

![**Quick Figure - Supp**. ](../vignettes/my_fig_supp.png){#supp}

# Standard workflow

We showcase the standard workflow on the example of the RNA-binding protein U2AF2. The data-set consists of four replicates. Sequencing reads from all replicates were pre-processed up to the point of .bam files. These were then merged and subjected to peak calling with *PureCLIP*. Resulting peaks as well as the processed coverage files (.bigwigs) are what is needed as input for the following sections.

## Manage the input data

### Build the BSFDataSet

To build the `BSFDataSet` one needs two types of information. First, the result of the peak calling step (crosslink sites) and second the iCLIP coverage in the form of .bigwig files for each replicate. Here we first import the crosslink sites to be represented as a `GenomicRanges` object. We also expect that all ranges are a single nucleotide wide^[Note: We remove the column `additionalScore` after the initial load, to clean our input].

```{r}
# load crosslink sites
csFile <- system.file("extdata", "PureCLIP_crosslink_sites_examples.bed", 
                      package="BindingSiteFinder")
cs = rtracklayer::import(con = csFile, format = "BED", extraCols=c("additionalScores" = "character"))
cs$additionalScores = NULL
cs
```

The .bigwig files with the coverage information are located in a single folder. The path is given in the `files` variable. We then use the path to list all .bigwig files from each strand separately. 

```{r}
# Load clip signal files and define meta data object
files <- system.file("extdata", package="BindingSiteFinder")
clipFilesP <- list.files(files, pattern = "plus.bw$", full.names = TRUE)
clipFilesM <- list.files(files, pattern = "minus.bw$", full.names = TRUE)
```

The strand specific vectors with the `path` information that point to each replicate .bigwig file are combined with additional meta information into a meta table `data.frame`. This `data.frame` has to have at minimum three columns, which must be named `condition`, `clPlus` and `clMinus`. The `clPlus` and `clMinus` columns point towards the strand-specific coverage for each replicate. This information will be imported as `Rle objects` upon object initialization. 

```{r}
# make the meta table
meta = data.frame(
  id = c(1:4),
  condition = factor(rep("WT", 4)), 
  clPlus = clipFilesP, clMinus = clipFilesM)
meta
```

With the crosslink sites and the meta table ready we are good to construct the `BSFDataSet` object using the `BSFDataSetFromBigWig()` function. We can get information of the constructed object using it's `show` method, which show the number of ranges and crosslinks present in the object^[Note: We load a previously compiled `BSFDataSet` to save disc space.].


```{r, eval=FALSE}
library(BindingSiteFinder)
bds = BSFDataSetFromBigWig(ranges = cs, meta = meta, silent = TRUE)
```

```{r, eval=TRUE}
exampleFile <- list.files(files, pattern = ".rda$", full.names = TRUE)
load(exampleFile)
bds
```


```{r, eval=TRUE, echo=FALSE}
# exampleFile <- list.files(files, pattern = ".rda$", full.names = TRUE)
# load(exampleFile)
# meta = data.frame(
#   id = c(1:4),
#   condition = factor(rep("WT", 4)), 
#   clPlus = clipFilesP, clMinus = clipFilesM)
# bds@meta = meta
# names(bds@signal$signalPlus) = c("1_WT", "2_WT", "3_WT", "4_WT")
# names(bds@signal$signalMinus) = c("1_WT", "2_WT", "3_WT", "4_WT")
```

```{r, eval=FALSE}
# exampleFile <- list.files(files, pattern = ".rda$", full.names = TRUE)
# load(exampleFile)
```



## Construct the annotation objects

To take advantage of the full functionality of the package one needs to provide gene annotation information. These will be used to map computed binding sites directly on target genes and and transcript regions of protein-coding genes. Here, we use [GENCODE](https://www.gencodegenes.org/) hg38 gene annotations.

### Create annotation for genes

To present gene-level information we import the annotation as a .gff3 file into a `GenomicRanges` object. To combine the ranges with the available meta information for each gene (such as a `gene_id`, `gene_name` or `gene_type`), we have to read our annotation twice. First, we import the ranges into a `TxDB` database using functionalists from [GenomicFeatures](https://bioconductor.org/packages/release/bioc/html/GenomicFeatures.html). This database can be queried using the `genes()` function to retrieve the positions of all genes from our resource as `GRanges` objects^[Note: It is of particular importance that one uses a valid format for the *geneID* (ENSEMBL, UCSC, ...) at this point.]. Second we use `rtracklayer` to import all ranges with the corresponding meta information. We then match both imports to create a single `GRanges` object which contains all genes from our annotation with relevant meta information added^[Note: We again only load a previously compiled object at this point to save disc space.].

```{r, echo=TRUE, eval=FALSE}
library(txdbmaker)
# Make annotation database from gff3 file
annoFile = "./gencode_v37_annotation.gff3"

annoDb = txdbmaker::makeTxDbFromGFF(file = annoFile, format = "gff3")
annoInfo = rtracklayer::import(annoFile, format = "gff3")
# Get genes as GRanges
gns = genes(annoDb)
idx = match(gns$gene_id, annoInfo$gene_id)
meta = cbind(elementMetadata(gns),
             elementMetadata(annoInfo)[idx,])
meta = meta[,!duplicated(colnames(meta))]
elementMetadata(gns) = meta
```

```{r}
# Load GRanges with genes
geneFile <- list.files(files, pattern = "gns.rds$", full.names = TRUE)
load(geneFile)
gns
```


### Create annotation for transcript regions

To represent all possible transcript regions, we re-use the previously created `TxDB` annotation database. The [GenomicFeatures](https://bioconductor.org/packages/release/bioc/html/GenomicFeatures.html) package provides dedicated information to retrieve specific sets of ranges from such a database object, like *CDS*, *UTRs*, etc. Here we store these ranges into a `GRangestList` and name the elements accordingly^[Note: Just like for the genes, we load a pre-compiled list to save disc space.].

```{r, eval=FALSE}
# Count the overlaps of each binding site for each region of the transcript. 
cdseq = cds(annoDb) 
intrns = unlist(intronsByTranscript(annoDb)) 
utrs3 = unlist(threeUTRsByTranscript(annoDb)) 
utrs5 = unlist(fiveUTRsByTranscript(annoDb)) 
regions = GRangesList(CDS = cdseq, Intron = intrns, UTR3 = utrs3, UTR5 = utrs5)
```

```{r}
# Load list with transcript regions
regionFile <- list.files(files, pattern = "regions.rds$", full.names = TRUE)
load(regionFile)
```


## Performing the main analysis

Now that we have prepared our input data, we are good to use the main workhorse, the `BSFind()` function. This function is essentially a high-level wrapper around all core utility functions and provides an easy entry point into the analysis. At it's core each analysis step consists of a utility function that computes the respective result. Each utility function further comes along with a set of diagnostic/ result plotting functions. These can be used to assess the quality of every computational step. We recommend to once run the entire analysis with `BSFind()` and then follow up on each step in detail using the plot functionalities^[Note: Here we set `est.subsetChromosom='chr22'` because we have only signal on chromosome 22 to save disc space]. 

```{r}
bdsOut = BSFind(object = bds, anno.genes = gns, anno.transcriptRegionList = regions,
              est.subsetChromosome = "chr22", veryQuiet = TRUE, est.maxBsWidth = 29)
```

It is good practice to once visualize the entire binding site definition process after `BSFind()` was executed. This helps to keep track what steps were performed and how the flow into each other^[Note: Not all binding sites can be assigned to genes or transcript regions, in these cases NA is introduced, so no binding site is lost.].

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="processingStepsFlowChart"}
processingStepsFlowChart(bdsOut)
```

As a default all of the shown steps are executed sequentially in the given order. To change that order, each utility function can be called individually to set up a customized workflow (for details see [Construct your own pipeline](## Construct your own pipeline)). 

### Deciding on the binding site width

The first and main step of any binding site analysis is the actual binding site definition step. In the following we showcase how you can apply the main workflow functions to accurately define binding sites and judge the results. 

#### Pre-filtering of crosslink sites

An optional step prior to the actual merging of crosslink sites into binding sites is pre-filtering. Depending on the experiment type or sequencing depth, it might be useful to retain only the most informative crosslink sites. In the case of *PureCLIP*, the positions called as significantly enriched (hereafter called *PureCLIP sites*) are associated with a binding affinity strength score (hereafter called *PureCLIP score*) which can be used as a metric for the pre-filtering step. As default *PureCLIP* sites with the lowest 1% score will be removed. 

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="pureClipGlobalFilterPlot. The diagnostic shows the distribution of PureCLIP scores on a log2 scale with a 10% step increased coloring for easier visualization and comparability. The selected cutoff is indicated by the dashed line. ", fig.small=TRUE}
pureClipGlobalFilterPlot(bdsOut)
```

#### Selection of the binding site width

In case the option `bsSize` is not set, the optimal binding site width is estimated from the data directly (using the `estimateBsWidth` function). Since the final binding site width depends on the quality of the input crosslink sites, `bsSize` is estimated together with the gene-wise filter level (`cutoff.geneWiseFilter`). That is filtering by the top X% crosslink sites per gene. Filtering on gene level has been shown to be really effective in diluting the most informative signal from the data^[Note: Since we did not specify `bsSize` in `BSFind()` the binding site size is estimated automatically.].

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="estimateBsWidthPlot. The diagnostic show the ratio between the crosslink events within binding sites and the crosslink events in adjacent windows to the binding sites. This is effectively a 'percent-bound' type ratio, with the higher the better the associated binding site width captures the distribution of the underlying crosslink events. The plot shows how that ratio behaves for different binding site width under varying levels of the gene-wise filter.", fig.small=TRUE}
estimateBsWidthPlot(bdsOut)
```

*Signal-to-flank* score values are estimated on all binding sites on a given chromosome (`est.bsResolution`) for varying levels of accuracy. For `est.geneResolution` options "coarse" results in 20% steps, "medium" in 10%, "fine" in 5% and "finest" in 1% steps. For `est.bsResolution` option "coarse" and "medium" approximate binding sites by extending the center of the each merged crosslink region. In "coarse" the center is found by the maximum PureCLIP score, in "medium" the center is found by the position with the most stacked crosslinks. Option "fine" employs a full binding site definition cycle. Binding sites are tested from the minimum size, which is 3, up to `maxBsWidth`. 

After iterating over all sets given the above resolutions a mean score per binding site width is calculated and the binding site width that yielded the highest score is selected as optimal. The `est.minimumStepGain` option allows control over the minimum additional gain in the score that a tested width has to have to be selected as the best option (the default is 2% increase). This essentially lets binding site sizes to be computed more conservatively, opting for smaller binding sites rather than larger ones. This is particularly useful when the estimated mean score flattens out at a certain point showing only a very small increase in the score per binding site width increase. Samples with broad binding RBPs or samples with poor coverage tend to show this behavior. 

In case you already know your binding site width and/or gene-wise filtering threshold you can invoke `BSFind()` with these parameters, to avoid the execution of the utility function `estimateBsWidth()`. For details see section [If you already know your binding site size](## If you already know your binding site size).

In case you want to manually investigate which binding site size to use see the section [How to manually test different binding site sizes](## How to manually test different binding site sizes).

#### Apply the gene-wise filter

The gene-wise filter cutoff estimated in the previous step (`cutoff.geneWiseFilter`) is applied to keep only the top percent of crosslink sites, ranked by their *PureCLIP* score. In case the value for `cutoff.geneWiseFilter` was given directly without estimation, that value is used directly.

Such a filter essentially helps to improve the binding site quality by using only the defined subset per gene of all *PureCLIP* sites for the binding site definition. This is because the strength of a binding site is tied to the expression level of the hosting gene. Thus a gene-wise percentage filter will increase the quality on all hosting genes in the same order.

Since a single *PureCLIP* site might overlap with multiple different genes given in the annotation, a diagnostic function helps to visualize the magnitude of that problem. 

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="duplicatedSitesPlot. Diagnostic plot that show how many crosslink sites overlap how many different genes in the annotation. The grey area for 2 and 3 overlaps indicates what proportion the option overlaps='keepSingle' removes in order to keep crosslink site numbers stable. ", fig.small=TRUE}
duplicatedSitesPlot(bdsOut)
```

To handle such overlapping cases the `overlaps` parameter can be used with options `"keepSingle"`, `"removeAll"` and `"keepAll"`. We recommend to stick with the default (`overlaps="keepSingle"`), which reduces the number of crosslink sites on overlapping loci to a single instance. This will conserve the number of crosslink sites. The additional options `overlaps="removeAll"` and `overlaps="keepAll"` will either increase or decrease the number of crosslink sites in the set and should be handled with care. 


#### Merge crosslink sites into binding sites

In this step we turn the single-nucleotide wide PureCLIP sites into binding sites of a desired size (`bsSize`). This most essential parameter can either be estimated automatically, or be set directly. For the inexperienced user we recommend to use the estimated size in almost all cases. An exception would be very sparse or incomplete data, which should result in warning messages during the estimation step. If you do want to use a manual size, because you have eg. prior-knowledge of your RBP please have a look at the [If you already know your binding site size](## If you already know your binding site size) section first. 

Binding site merging happens by concatenation of neighboring crosslink sites into larger regions. These regions are then split up again using the crosslink coverage to guide the position of the final binding sites. This process happens iteratively, where each merge and resize round resolves more regions unitl all regions are converted into binding sites. The following diagnostic plots the number of regions that are still left to be resolved after each round and compares it to the size of these regions^[Note: Since we use a very small data-set for demonstration purpose, this plot might look different with a real-sized data-set.].

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="mergeCrosslinkDiagnosticsPlot. Diagnostic plot with number and region size at each fitting itteration loop. ", fig.small=TRUE}
mergeCrosslinkDiagnosticsPlot(bdsOut)
```

A variety of different filtering options for placing a final binding site inside a merged region exist. Most importantly `merge.minClSites` gives the minimum number of initial crosslink sites that have to overlap a final binding site. How this and all other options affect the final binding site set size is displayed by the following diagnostic.

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="makeBsSummaryPlot. Diagnostic bar chart indicating the number of binding site after each of the indicated filter options is applied given their input values.", fig.small=TRUE}
makeBsSummaryPlot(bdsOut)
```


### How to ensure reproducibility among replicates

Since the initially used crosslink sites are computed from the merged signal of all replicates, binding sites resulting from the previous merge might not be reproducible among all replicates. For that reason, we specifically ask in this section which of the computed binding sites are reproduced by the individual replicates. 

#### The replicate-specific threshold

Since replicates might differ in library size, a replicate-specific threshold is computed base on the binding site support distribution. That is how many crosslinks fall into the computed binding site for each replicate. In other words, we compute the number of crosslinks per binding site and replicate. This results in the following histogram.

When calling `BSFind()` we did not specify any further cutoff for this particular step, so the 5% quantile (`cutoff`) with a lower boundary of 1 crosslink (`minCrosslinks`) is used as default. Here for example a binding site would be called reproducible by replicate No.3 if at least 4 crosslinks from that replicate fall into the range of the binding site[^8].

[^8] Note: These settings can be adapted to for example produce high and low confidence binding site sets.

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilityFilterPlot. Reproducibility crosslink histogram, with the number of crosslink sites over the number of crosslinks per binding site for each replicate. The dashed line indicates the replicate-specific cutoff."}
reproducibilityFilterPlot(bdsOut)
```


#### How many replicates are enough? 

After computing the replicate-specific thresholds as indicated above, one must decide on the number of replicates that must meet the thresholds for a binding site to be seen as reproducible. 

A binding site that meets the threshold for all replicates in the set would be highly reproducible, whereas a binding site that meets the threshold for a single replicate only would be not reproducible at all. To find a balance in this question we suggest to stay with the default of N-1 for the `nReps` parameter. However if your replicates show a very high variability among each other changing this parameter could be beneficial to increase the specificity of the reproducible set. The following diagnostic assists in that decision, by visualizing the degree of which reproducible sets overlap among each other.

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilitySamplesPlot. UpSet plot indicating the degree of reproducibility overlap between each replicate. "}
reproducibilitySamplesPlot(bdsOut)
```

#### Do my replicate corrlate well?

A common question for reproducibility is the visualization of the between replicate pairwise reproducibility. The following scatter plot does exactly that. Depending on when the correlations are calculated, the binding site set is either already corrected for reproducibility or not. To compare pairwise correlations before and after the application of the reproducibility filtering see [Reproducibility scatter before and after correction](## Reproducibility scatter before and after correction).

```{r, fig.retina = 1, dpi = this.dpi, message=FALSE, fig.cap="reproducibilityScatterPlot. Bottom left) Pairwise correlations between all replicates as scatter. Upper right) Pairwise pearson correlation coefficient. Diagnoal) Coverage distribution as density."}
reproducibilityScatterPlot(bdsOut)
```


### Genomic target identification

A major question the follow from binding site definition is the assessment of the genomic targets that the RBP under consideration binds to. In the following section we showcase how target genes and transcript regions can be identified and what one has to consider doing so. In case you don't have a gene annotation at hand but still want to compute binding sites, have a look at the section [Work without a gene annotation](## Work without a gene annotation).

#### Target gene identification

Here we essentially overlap the annotation given in `anno.genes` with the reproducible set of binding site computed above. Depending on the source and organism, gene annotations overlap each other to some degree.
We implemented several strategies to resolve these overlaps with `overlaps="frequency"` being the recommended default. Here the most frequently present *gene_type* is used to resolve overlaps. This however requires that the annotated genes (given in `anno.genes`) carry gene type information as a meta column. The exact column name can be specified with `match.geneType`. The same requirement holds true for the option `overlaps="hierarchy"`, where a manual hierarchical rule on the gene type is used (`overlaps.rule` is required)^[Note: If an overlaps exists, but gene types are identical options "frequency" and "hierarchy" will cause the gene that was seen first to be selected as representative. ].
The following plot visualizes among which gene types the most overlaps exist^[Note: Strategies that work without gene type information are `overlaps="remove"` and `overlaps="keep"` that respectively remove or keep all overlapping cases.].

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="geneOverlapsPlot. UpSet plot with the number of binding sites overlapping a specific gene type from the annotation before overlaps were resolved. "}
geneOverlapsPlot(bdsOut)
```

The final target gene spectrum can now be visualized with the following plot. Depending on which of the overlap strategies are being used, numbers in this plot will change slightly. 

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="targetGeneSpectrumPlot. Bar chart with the number of gene and binding sites of a given gene type after overlaps were resolved. ", fig.small=TRUE}
targetGeneSpectrumPlot(bdsOut)
```

The assigned gene with it's ID, name and gene type (if given) is also attached to the binding site meta data, which can be retrieved as `GRanges` object using the `getRanges()` function.

#### Transcript region identification

Similar to the target gene, each binding site can be assigned to a hosting transcript region. This is done using the `anno.transcriptRegionList` resource that we defined earlier. Essentially each binding site is overlapped with each of the ranges given in the region list. As one would expect, the problem of overlapping annotations is increased manifold on the level of transcript regions compared to entire genes. 

Options to resolve these are similar to those on gene level. Option `overlaps="frequency"` will resolve based on the most frequently observed transcript region, `overlaps="hierarchy"` uses a manual hierarchic assignment based on `overlaps.rule`^[Note: This time the transcript type is pulled from the names of the `GRangesList` defined in `anno.transcriptRegionList`. ]. Additionally, option `overlaps="flag"` will just flag all overlaps with the tag *ambiguous* and option `overlaps="remove"` will remove all those cases. The following plot illustrates the degree of overlap for our example data-set. 

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionOverlapsPlot. UpSet plot with the number of binding sites overlapping a specific transcript region from the annotation before overlaps were resolved."}
transcriptRegionOverlapsPlot(bdsOut)
```

We observe that most overlaps are generated by binding sites that are either within an intron or a 3'UTR. As indicated by the plot we use the default strategy `overlaps="frequency"` is used to resolve these cases. Since introns were most frequently observed in the data, the transcript region *intron* was selected in these cases.

The final binding spectrum on transcript region level can be visualized with the following plot. Depending on which of the overlap strategies are being used, numbers in this plot will change slightly. 

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionSpectrumPlot. Bar chart with the number of binding sites for each transcript region after overlaps were resolved. ", fig.small=TRUE}
transcriptRegionSpectrumPlot(bdsOut)
```

The assigned transcript region is also attached to the binding site meta data, which can be retrieved as `GRanges` object using the `getRanges()` function. 

If you want to learn about how to normalize these profiles by the length of each region see section [Normalize transcript regions](## Normalize transcript regions).

The described matching of binding sites and transcript regions is entirely based on overlaps with the regions given in `anno.transcriptRegionList`. This means one can customize these regions by using different sets of ranges to extend and modify this list. For further details on how to do this see section [Use a different set of transcript regions](## Use a different set of transcript regions).


### Assessing further binding site properties

Another common task is to assess the strength and shape of a binding site. We provide two additional metrics as part of our core workflow that allow to assess exactly that. 

#### Binding site strength

The affinity of a binding site to the RNA is already given to some degree by *PureCLIP*. The *PureCLIP* model computes a strength score that captures affinity for each crosslink site. Here we re-map these scores from the initial crosslink sites to the final binding sites. This is done through the `match.score` argument that point to the meta column name with the score from the initial crosslink sites^[Note: Essentially any numerical score can be transferred from the initial crosslinks to the final binding site using this option.]. As for the default we assign the highest score from all overlapping crosslink site to the binding site (`match.option="max"`). The following plot shows the distribution of scores as density for all binding sites after re-assignment. 

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="globalScorePlot. The diagnostic shows the distribution of PureCLIP scores on a log2 scale with a 10% step increased coloring for easier visualization and comparability.", fig.small=TRUE}
globalScorePlot(bdsOut)
```

#### Binding site definedness

How well the defined binding sites fit to the observed crosslink coverage can be assess through binding site definedness. This also gives first insights into 
the overall binding site shape and the binding mode of the RBP. Here we assess this property through a *percent bound* score, that divides the number of crosslinks inside the binding sites by the total number of crosslinks in a region around the binding site. As default, this background region is two times the size of the binding site. In other words, we compare the signal within a binding site to a same sized window directly flanking the binding site on both sides.
A splicing factor that recognizes very clear motifs will for example form very sharp peaks, which will lead to a high *percent bound* degree. A broad binding RBP on the other hand, that scans more along the RNA, will lead to a lower definedness score.

Depending on which steps were executed in the [Genomic target identification](### Genomic target identification) section, the plot can be grouped by the assignment (`by="all"`, `by="transcript_region"` or `by="gene_type"`).

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="bindingSiteDefinednessPlot. Density distribution of binding site definedness as percent bound grouped by transcript region.", fig.small=TRUE}
bindingSiteDefinednessPlot(bdsOut, by = "transcript")
```


## Exporting your results

For proper integration with any other downstream analysis and application we provide a set of export functionalities that allow to quickly and easily retrieve any type of result. 

### For additional analysis in R/Bioconductor

If one stays inside the R/ Bioconductor environment one can simply extract the final binding sites as `GenomicRanges`, which enables full functionality that comes with this powerful data class. As indicated by the print all information from each processing step is kept in the meta data of the granges object. This makes it easy to combine your analysis results with orthogonal data and projects.

```{r}
getRanges(bdsOut)
```

### Export as UCSC BED

For visual inspection and integration with further data-sets results can be easily exported in the widely used .bed format. This offers a quick way for manual inspection of the resulting binding sites in tools like IGV or UCSC genome browser.

```{r, eval=FALSE}
exportToBED(bdsOut, con = "./myResults.bed")
```

### Export genes and feature lists

Gene targets from the target gene assignment can be exported to either `csv` or `xslx` grouped by `geneType` or `transcriptRegion`. Genes can additionally be sorted by the number of overlapping binding sites or by binding site strength score. 

```{r, eval=FALSE}
exportTargetGenes(bdsOut, format = "xslx", split = "transcriptRegion")
```

### Export iCLIP signal

To export iCLIP coverage directly we do not provide a dedicated function, but rather we use the `export` function from the [rtracklayer](https://www.bioconductor.org/packages/release/bioc/html/rtracklayer.html) package. For details see the section [Exporting iCLIP signal](## Exporting iCLIP signal).


# Diagnostic coverage polots

The analysis workflow described in the [Standard workflow](# Standard workflow) section contains all steps for a comprehensive transcriptome-wide binding site analysis. However in some instances it might be useful to drill these analysis down to specific binding sites. In the following chapter we provide some useful ideas and functions when this could be useful and what type of questions could be addressed. 

## Visualize the iCLIP coverage

To visualize the iCLIP signal (aka coverage) on a specific binding site, we provide the `bindingSiteCoveragePlot()` function. It takes a reference index for the binding site and then plots the coverage as bars in a defined range around a selected binding site. The function is based on the [Gviz](http://bioconductor.org/packages/release/bioc/html/Gviz.html) package. The `plotIdx` indicates which range should be used as center, the `flankPos` parameter allows to zoom in and out. In addition to the selected range, also all other ranges which fall into the selected window will be shown.

To specifically assess for example the reproducibility on a particular binding site, one could plot the coverage from each replicate on that site. The plot below shows the coverage from each replicate on an exemplaric binding site.

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="bindingSiteCoveragePlot. Coverage plot on replicate level."}
bindingSiteCoveragePlot(bdsOut, plotIdx = 8, flankPos = 100, autoscale = TRUE)
```

Sometimes it can also be useful to merge all replicates into a single combined coverage. This can be done by setting `mergeReplicates=TRUE`. 


## Trace back a binding site

A common case that makes use of this function is when one wants to see why a particular binding site is lost after a certain filtering step. Here, we look at a binding site that was filtered out from the final object by the reproducibility filter. Doing so, we can visually confirm that binding sites 5 and 6 were correctly removed by the reproducibility filter function.

```{r, fig.retina = 1, dpi = this.dpi}
rangesBeforeRepFilter = getRanges(bds)
rangesAfterRepFilter = getRanges(bdsOut)
idx = which(!match(rangesBeforeRepFilter, rangesAfterRepFilter, nomatch = 0) > 0)
rangesRemovedByFilter = rangesBeforeRepFilter[idx]
bdsRemovedRanges = setRanges(bds, rangesRemovedByFilter)

bindingSiteCoveragePlot(bdsRemovedRanges, plotIdx = 2, flankPos = 50)
```

We could also look at the final binding site definition and see how they were derived from the initial *PureCLIP sites*. To achieve this, we make use of the `customRange` slot to add these sites at the bottom of the coverage plot.

```{r, fig.retina = 1, dpi = this.dpi, eval=FALSE}
pSites = getRanges(bds)
bindingSiteCoveragePlot(bdsOut, plotIdx = 8, flankPos = 20, autoscale = TRUE,
                        customRange = pSites, customRange.name = "pSites", shiftPos = -10)
```


## Gene wise coverage

Another use case would be to check the coverage on some example genes, or as we do here on the binding site with the highest number of crosslinks: 

```{r, fig.retina = 1, dpi = this.dpi, eval=FALSE}
bindingSiteCoverage = coverageOverRanges(bdsOut, returnOptions = "merge_positions_keep_replicates")
idxMaxCountsBs = which.max(rowSums(as.data.frame(mcols(bindingSiteCoverage))))
bindingSiteCoveragePlot(bdsOut, plotIdx = idxMaxCountsBs, flankPos = 100, mergeReplicates = FALSE, shiftPos = 50)
```

It is also possible to anchor the plot on any other `GenomicRange`. Here, we take annotated CDS regions and ask for the one with the most overlapping binding sites. We then use this ranges as center for the plot and further zoom in to a particular range. We then make use of the `customRange` slot to re-include the binding site ranges as additional annotation shown underneath the signal. Additionally, one could also add a custom annotation track in the form of a `GenomicRanges` or `TxDB` object. 

```{r, fig.retina = 1, dpi = this.dpi, eval=FALSE}
bdsCDS = setRanges(bdsOut, regions$CDS)
cdsWithMostBs = which.max(countOverlaps(regions$CDS, getRanges(bdsOut)))

bindingSiteCoveragePlot(bdsCDS, plotIdx = cdsWithMostBs, showCentralRange = FALSE,
                       flankPos = 250, shiftPos = 50, mergeReplicates = TRUE,
                       highlight = FALSE, customRange = getRanges(bdsOut),
                       customAnnotation = regions$CDS)
```


# Variations of the standard workflow

## Normalize transcript regions

When assessing the number of binding sites per transcript region, it is common sense that more binding sites will map to *intron* as appose to for example *5'UTRs* due to the size difference of these regions. Here we describe how one can be used to normalize for this effect. 

The normalization scheme we provide is based on the hosting region length of each binding site. That means the number of binding sites per region can be divided by either the sum, mean or median of the hosting region length from all binding sites (see `normalize.factor` when using `transcriptRegionSpectrumPlot()`). To avoid biases by very short or long transcript regions an upper and lower boundary can be set (see `normalize.exclude.upper` and `normalize.exclude.lower` when using `assignToTranscriptRegions()`). 

Here we show the transcript region spectrum on a percentage scale once normalized and once un-normalized.

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionSpectrumPlot. Without region size normalization.", fig.small=TRUE, eval=FALSE}
transcriptRegionSpectrumPlot(bdsOut, normalize = FALSE, values = "percentage")
```

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="transcriptRegionSpectrumPlot. With region size normalization.", fig.small=TRUE, eval=FALSE}
transcriptRegionSpectrumPlot(bdsOut, normalize = TRUE, values = "percentage", normalize.factor = "median")
```


## How to manually test different binding site sizes

In some instances the `estimateBsWidth()` function used in the `BSFind()` wrapper might not lead to optimal results. This can be the case if the starting data is of below average quality with very few crosslink sites to begin with. Here manual binding site profiling can be used to achieve the best possible result given the input data.

To guide the choice of a good width, we recommend to explicitly compute binding sites for a variety of width. Here we selected 3, 7, 19 and 29nt and manually execute `makeBindingSites()`^[Note: Notice that we do not subset on a specific chromosome, but rather include all ranges to avoid any local pitfalls. ].

```{r}
# compute binding sites
bds1 <- makeBindingSites(object = bds, bsSize = 3)
bds2 <- makeBindingSites(object = bds, bsSize = 9)
bds3 <- makeBindingSites(object = bds, bsSize = 19)
bds4 <- makeBindingSites(object = bds, bsSize = 29)
# summarize in list
l = list(`1. bsSize = 3` = bds1, `2. bsSize = 9` = bds2, 
         `3. bsSize = 19` = bds3, `4. bsSize = 29` = bds4)
```

The combined coverage over these different binding sites can help to quickly judge if the selected size does fit the data appropriately (`rangeCoveragePlot()`). Here, each plot is centered around the binding site’s midpoint and the computed width is indicated by the gray frame.

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="rangeCoveragePlot. Crosslink coverage summarized at potential binding sites, with current size given by the grey area."}
rangeCoveragePlot(l, width = 20, show.samples = TRUE, subset.chromosome = "chr22") 
```

In our example, size = 3 appears too small, since not all of the relevant peak signal seems to be captured. On the contrary, size = 29 appears extremely large. Here, we decided for size = 9 because it seems to capture the central coverage peak best. 


## If you already know your binding site size

In case you already know what binding site size you want to use `BSFind()` can be called directly with that input size. This is usually the case if one has prior knowledge from orthogonal experiments or if the present data-set should be fix to the same size as another data-set for comparison. With a fixed binding site at hand we can call `BSFind()` with that explicit size to avoid estimation and use our manual input instead^[Note: We also set `cutoff.geneWiseFilter=0` to turn off gene-wise filtering.].

```{r}
bdsManual = BSFind(object = bds, anno.genes = gns, anno.transcriptRegionList = regions, bsSize = 9, cutoff.geneWiseFilter = 0)
```

If you are not sure how to choose an appropriate binding site size for your data-set, see [Standard workflow](# Standard workflow) which makes use of the  `estimateBsWidth()` function, or see the chapter [How to manually test different binding site sizes](## How to manually test different binding site sizes), to manually select a size value.


## Call utility functions seperately

In the case on does not want to work with `BSFind()` as main wrapper function, all core utility function can be called separately. In our example from above where we manually decided on a binding site width of 9nt we could also call `makeBindingSites()` directly. This avoids turning off certain default options and will result in the most basic result possible, a simple set of merged binding site.

```{r}
bdsSimple = makeBindingSites(object = bds, bsSize = 9)
```


## Construct your own pipeline

If one is using more than one utility function with some pre-defined input options it is recommended to construct a small custom pipeline. This greatly increase readability and reproducibility of the code. Here we extend our binding site merging to 9nt from above by a pre-filter and add another layer of post-processing to it.

```{r}
customBSFind <- function(object) {
    this.obj = pureClipGlobalFilter(object)
    this.obj = makeBindingSites(this.obj, bsSize = 9)
    this.obj = calculateSignalToFlankScore(this.obj)
    return(this.obj)
}
bdsSimple = customBSFind(bds)
```

Processing steps in custom pipelines can also be visualized with the flow chart function.

## Work without a gene annotation

Setting up a custom pipeline can also be needed when one does not have any supporting gene annotation at hand. In those cases a simplified workflow pipeline can be defined that still allows binding site definition. To achieve this we simply combine all processing functions that do not require an explicit annotation into a new reduced wrapper^[Note: We highly recommend using an adequate gene annotation since it improves the quality of the analysis.].

```{r}
noAnnotationBSFind <- function(object) {
    this.obj = pureClipGlobalFilter(object)
    this.obj = makeBindingSites(this.obj, bsSize = 9)
    this.obj = reproducibilityFilter(this.obj)
    this.obj = annotateWithScore(this.obj, getRanges(object))
    this.obj = calculateSignalToFlankScore(this.obj)
    return(this.obj)
}
bdsSimple = noAnnotationBSFind(bds)
```

Since no gene or transcript level assignment was performed (`assignToGenes()` and `assignToTranscriptRegions()` was not run), the related diagnostic plots are not available. Yet the diagnostics for the functions we have called in `noAnnotationBSFind()` can be used to assess the performance of our analysis. 


## Reproducibility scatter before and after correction

In order to visualize how the reproducibility filtering function improves pairwise replicate reproducibility, the `reproducibilityScatterPlot()` function can be called before and after reproducibility filtering. 

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilityScatterPlot. Before reproducibility filter.", eval=FALSE}
bds.before = makeBindingSites(bds, bsSize = 9)
reproducibilityScatterPlot(bds.before)
```

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="reproducibilityScatterPlot. After reproducibility filter.", eval=FALSE}
bds.after = reproducibilityFilter(bds.before, minCrosslinks = 2)
reproducibilityScatterPlot(bds.after)
```


## Use a different set of transcript regions

When assigning binding sites to transcript regions, the `assignToTranscriptRegions()` utility function takes the annotation ranges as a `GRangesList` input. Then binding sites are simply assigned to each list entry using the respective list name. This fact can be exploited to essentially use any kind of range set as input to assign binding site to any custom set of genomic ranges, while handling overlaps as it is described in `assignToTranscriptRegions()`. 


# Additional functions

Besides the standard workflow and main processing functions a variety of further useful functions that come in handy when working with iCLIP data are provided to the user. In the following sections we demonstrate some these functionalities.

## Subset data for faster iterations 

Sub-setting a `BSFinderData` object can be useful in a variety of cases, e.g. for reducing the object size for faster parameter testing, limiting the analysis to some candidate genes etc. Here, we subset the object by a random set of 100 binding sites and plot their count distribution.

```{r, fig.retina = 1, dpi = this.dpi, fig.small=TRUE, eval=FALSE}
set.seed(1234)
bdsSub = bds[sample(seq_along(getRanges(bds)), 100, replace = FALSE)]

cov = coverageOverRanges(bdsSub, returnOptions = "merge_positions_keep_replicates")
df = mcols(cov) %>%
    as.data.frame() %>%
    pivot_longer(everything())

ggplot(df, aes(x = name, y = log2(value+1), fill = name)) +
    geom_violin() +
    geom_boxplot(width = 0.1, fill = "white") +
    scale_fill_brewer(palette = "Greys") +
    theme_bw() +
    theme(legend.position = "none") +
    labs(x = "Samples", y = "#Crosslinks (log2)")
```

## Merge replicate signal 

Depending on the task at hand, one either wants to keep the iCLIP signal separated by replicates or merge the signal over the replicates (e.g. of the same condition). Merging signal can be done using the `collapseReplicates()` function. Doing so allows for example to identify the proportion of crosslink events that each sample contributes to the total of a binding site. Here, we did this for the first 100 binding sites. We sort all binding sites by their fraction and color the plot based on the replicate. 

```{r, fig.retina = 1, dpi = this.dpi, fig.small=TRUE, eval=FALSE}
bdsMerge = collapseReplicates(bds)[1:100]
covTotal = coverageOverRanges(bdsMerge, returnOptions = "merge_positions_keep_replicates")

covRep = coverageOverRanges(bds[1:100], returnOptions = "merge_positions_keep_replicates")

df = cbind.data.frame(mcols(covTotal), mcols(covRep)) %>%
    mutate(rep1 = round(`1_WT`/ WT, digits = 2) * 100,
           rep2 = round(`2_WT`/ WT, digits = 2) * 100,
           rep3 = round(`3_WT`/ WT, digits = 2) * 100,
           rep4 = round(`4_WT`/ WT, digits = 2) * 100) %>%
    tibble::rowid_to_column("BsID") %>%
    dplyr::select(BsID, rep1, rep2, rep3, rep4) %>%
    pivot_longer(-BsID) %>%
    group_by(BsID) %>%
    arrange(desc(value), .by_group = TRUE) %>%
    mutate(name = factor(name, levels = name)) %>%
    group_by(name) %>%
    arrange(desc(value), .by_group = TRUE) %>%
    mutate(BsID = factor(BsID, levels = BsID))


ggplot(df, aes(x = BsID, y = value, fill = name)) +
    geom_col(position = "fill", width = 1) +
    theme_bw() +
    scale_fill_brewer(palette = "Set3") +
    theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1, size = 7)) +
    labs(x = "Binding site ID",
         y = "Percentage",
         fill = "Replicate"
         ) +
    scale_y_continuous(labels = scales::percent)
```

## Exporting iCLIP signal

Similar to exporting the binding sites as *BED* file for external visualization, one can do the same with the iCLIP signal itself. Here we use the `export` function from the [rtracklayer](https://www.bioconductor.org/packages/release/bioc/html/rtracklayer.html) package. We could now first merge the signal from all replicates into a single combined signal. This signal can then be exported to *.bigwig* format for easy visualization in IGV or th UCSC genome browser.

```{r, eval=FALSE}
sgn = getSignal(bds)
export(sgn$signalPlus$`1_WT`, con = "./WT_1_plus.bw", format = "bigwig")
export(sgn$signalMinus$`1_WT`, con = "./WT_1_minus.bw", format = "bigwig")
```

```{r, eval=FALSE}
bdsMerge = collapseReplicates(bds)
sgn = getSignal(bdsMerge)
export(sgn$signalPlus$WT, con = "./sgn_plus.bw", format = "bigwig")
export(sgn$signalPlus$WT, con = "./sgn_minus.bw", format = "bigwig")
```

## Calculate coverage

Each `BSFDataSet` contains the binding sites as ranges as well as the individual crosslinks per sample. With the `clipCoverage()` function one can easily compute all sorts of coverage over these ranges. 

In the purest form this would result in a matrix with the number of columns being equal to the number of nucleotides in the binding sites (`bsSize`) and the number of rows being equal to the number of binding sites. If this matrix is computed for all samples the result is a list of matrices with the described dimensions. 

To simplify this output three parameter options exist to merge either of the three dimensions (`samples.merge`, `positions.merge`, `ranges.merge`). Merging samples will combine the coverage signal from individual samples^[Notes: Samples can also be combined by condition if multiple conditions are present.]. Merging positions combine all single nucleotide positions of the binding sites into a single value. Merging ranges will combine all ranges. For each merging operation one can choose between using the `sum` or `mean`. Depending on the shape of the result it can either be attached to the ranges present in the `BSFDataSet` which results in a `granges` type of output, or it can be returned as a `data.frame`. 

```{r}
cov = clipCoverage(bdsOut, ranges.merge = TRUE, samples.merge = FALSE, positions.merge = FALSE)
cov
```

Here we added the number of crosslinks per binding site and replicate to the `granges` object of the binding sites. 



# Differential binding analysis

In this section we go beyond basic binding site definition and demonstrate how binding sites of a given RBP can be compared between conditions. For example one might be interested in how the binding pattern on of an RBP changes upon a specific mutation or condition. 

Comparing two sets of binding sites from different biological conditions to each other is not straight forward, since a change observed between any two binding sites can also be caused by expression level changes of the hosting transcript. For that reason the following computational approach will correct for this effect resulting in a log2 fold-change for each binding site that is corrected for a potential transcript expression level change.

## Concept and idea

The main idea of the approach is to divide given iCLIP crosslinks into two bins for each gene. The first bin holds all crosslinks that are overlapping defined binding sites and thus can be attributed to the binding affinity of the RBP. We call this bin the *binding-bin*. The second bin holds all other crosslinks, which can be attributed to the expression level of the hosting RNA. We call this bin the *background-bin*. 

We now make use of the condition specific changes in the *background-bin* to correct all fold-changes for each binding site in the *binding-bin*. This is done using the `DESeq2` likelihood ratio test, by comparing a "full" model to a reduced "null" model. 

![**Overview of differential binding model and workflow.**](images/model.png){#model}

## Preparation of the input data

In the most basic approach one wants to compute fold-changes for each binding site for the same RBP found in of two conditions. For the following example we name our conditions *WT* and *KO*. 

### Define binding sites for each condition

At first binding sites have to be defined for each condition separately. This ensures that regardless of sequencing depth or expression levels not binding site can be missed or masked by either condition ^[Note: This requires that both data sets need to be handled strict separately, from pre-processing to peak-calling and binding site definition.]. 

Here we load two pre-computed data sets, each having two replicates for *WT* and two for the *KO* ^[Note: Both data-sets were artificially created from the same condition for demonstration purposes.]. To learn how these sets can be generated from two individual data set, see the [Starting from scratch](### Starting from scratch) section.

```{r, eval=TRUE, echo=FALSE}
makeExampleDataSets <- function() {
    # load clip data
    files <- system.file("extdata", package="BindingSiteFinder")
    load(list.files(files, pattern = ".rda$", full.names = TRUE))
    # make binding sites
    bds = makeBindingSites(bds, bsSize = 5, quiet = TRUE)
    bds = assignToGenes(bds, anno.genes = gns, quiet = TRUE)
    # create artificial changes between binding sites
    bds = imputeBsDifferencesForTestdata(bds)
    
    # handle ranges
    bs.all = getRanges(bds)
    set.seed(1234)
    idx = sample(c(TRUE,FALSE), size = length(bs.all), replace = TRUE)       
    bs.wt = bs.all[idx]
    bs.ko = bs.all[!idx]
    
    # handle meta
    meta.all = getMeta(bds)
    meta.wt = meta.all[c(1:2),]
    meta.ko = meta.all[c(3:4),]
    meta.ko$id = 1:2
    
    # handle signal
    signal.all = getSignal(bds)
    signal.wt.p = signal.all$signalPlus[1:2]
    signal.wt.m = signal.all$signalMinus[1:2]
    signal.wt = list(signalPlus = signal.wt.p, signalMinus = signal.wt.m)
    
    signal.all = getSignal(bds)
    signal.ko.p = signal.all$signalPlus[3:4]
    signal.ko.m = signal.all$signalMinus[3:4]
    signal.ko = list(signalPlus = signal.ko.p, signalMinus = signal.ko.m)
    
    # set objects
    bds.wt = setRanges(bds, bs.wt)
    bds.wt = setMeta(bds.wt, meta.wt)
    bds.wt = setSignal(bds.wt, signal.wt)
    bds.wt = setName(bds.wt, "Example.WT")
    
    bds.ko = setRanges(bds, bs.ko)
    bds.ko = setMeta(bds.ko, meta.ko)
    bds.ko = setSignal(bds.ko, signal.ko)
    bds.ko = setName(bds.ko, "Example.KO")
    
    # return
    l = list(wt = bds.wt, ko = bds.ko)
    return(l)
}
```

```{r}
# make artifical data sets
myDataSets = makeExampleDataSets()

bds.wt = myDataSets$wt
bds.ko = myDataSets$ko

bds.wt
bds.ko
```

### Merge binding sites into a single object for testing

We follow up on the separate binding site definition by combining both binding site sets into a single set on which we then perform the differential binding testing. Merging two `BSFDataSets` can be done using the `+` operator as shown below or by using the `combineBSF()` function.

```{r}
bds.comb = bds.wt + bds.ko
getRanges(bds.comb)
```

## Count matrix generation

To perform the differential testing we first need to extract crosslink counts for each binding site and *background-bin* per gene. It is all crosslinks that are attributed for the *background-bin* that are used for the expression change correction. This makes the differentiation of the iCLIP signal in *binding-bin* and *background-bin* one of the must critical steps. 

To fine-tune this step a variety of options exist. First each binding site can be extend by an *offset* region. All crosslinks overlapping the *offset* will be removed. This means they are not counted towards the *background-bin* or the *binding-bin*. The same holds true for the *blacklist-region* which is an additional container which can be used to remove counts from certain ranges from the *background-bin*^[Note: The background counts per replicate is identical for binding sites on the same hosting gene.].

![**Background-bin generation options.**](images/background.png){#background}

### Calculate binding sites and background counts

Calculating the background counts requires a gene annotation information to be present in the form of a `GRanges` object. For more details on how to generate this object from a .gtf/.gff3 file please see the section [Create annotation for genes](### Create annotation for genes). The final counts are added to the meta columns of the binding site `GRanges` objects ^[Note: Binding sites that do not overlap a distinct gene may end with NA values as background counts and can be removed in the next step.].

```{r, warning=FALSE, message=FALSE}
bds.diff = calculateBsBackground(bds.comb, anno.genes = gns)
getRanges(bds.diff)
```

### Filter binding site and background counts

Depending on the crosslink signal at hand some genes might not be suitable for differential testing. This is for example when there is overall not enough signal per gene, if the distribution between conditions is heavily tailored towards one side or if there is a strong imbalance between the *background-bin* and *binding-bin* per gene. The following function can be used to remove such cases to create a more reliable test set^[Note: The `plotBsBackgroundFilter()` function can be used to visualize each filtering option.].

```{r}
bds.diff = filterBsBackground(bds.diff)
```

## Perform differential testing

Finally differential testing can be carried out using the `calculateBsFoldChange()` function. This function computes fold-changes for each binding site using the *background-bin* for correction and the likelihood ratio test for significance testing (for details on these test see the `DESeq2` user manual). Additionally, gene-level fold-changes are computed for each hosting gene using the standard Wald test (again see the `DESeq2` user manual for further details)^[Note: Depending on the data one might have to use the `removeNA` argument to remove edge cases where no `background-bin` counts could be computed. It is worth to manually look at such cases.]. 

```{r}
bds.diff = calculateBsFoldChange(bds.diff)
getRanges(bds.diff)
```

Resulting fold-changes and P values are again added to the meta columns of the binding sites (`GRanges` format). As a quick visualization of the results the functions `plotBsMA()` and `plotBsVolcano()` can be used, each of which can produce a plot either for the binding sites (`what = "bs"`) or for the background (`what = "bg"`).

```{r, fig.retina = 1, dpi = this.dpi, fig.cap="plotBsMA. MA plot for binding site changes.", fig.small=TRUE}
plotBsMA(bds.diff, what = "bs")
```

### Display differential results on gene level

To investigate differential binding results for a certain gene, the `geneRegulationPlot` can be used. The function displays the regulation (log2 fold-changes) of your binding sites on a given gene relative to their position with in the gene and optional adds the transcript region to it as well. 

The differential expression workflow must be followed to obtain log2 fold-changes (see section [Differential binding analysis](# Differential binding analysis)). Binding sites must also be assigned to their host genes with `assignToGenes` prior to this plot. 

For maximum information also `assignToTranscriptRegions` should be executed as well^[Note: The present plot is made from a small example data set due to size limitations, a plot made from real data will look more informative.]. If transcript region assignments are missing, one can re-calculate them for the merged object even after the differential binding workflow. The function allows a quick and easy way to visualize regulatory patterns due to binding site positioning and transcript region location. 

```{r}
bds.diff = assignToTranscriptRegions(bds.diff, anno.transcriptRegionList = regions)
geneRegulationPlot(bds.diff, plot.geneID = "ENSG00000180957.18", anno.genes = gns)
```



## Differential binding variations

### Starting from scratch

The following code chunks show how one can start fully from scratch with the binding site definition for the purpose of a downstream differential binding analysis. In the following example we will demonstrate the binding site definition for the conditions *WT* and *KO*.

At first one has to construct a `BSFDataSet` for each condition with the crosslinks form the peak-calling step (PureCLIP output). This serves as input for the binding sites definition with `BSFind()` (for further details see the [Performing the main analysis](## Performing the main analysis) section). As a result one gets a set of binding sites. This has to be done for each condition (in our example these are *WT* and *KO*)^[Note: It is desired that the same binding site size is used for both conditions, which can require setting the `bsSize` argument explicitly when calling `BSFind()`.].

```{r, eval=FALSE}
############################
# Construct WT binding sites
############################
# Locate peak calling result
peaks.wt = "./myFolder/wt/pureclip.bed"
peaks.wt = import(con = peaks.wt, format = "BED", extraCols=c("additionalScores" = "character"))
peaks.wt = keepStandardChromosomes(peaks.wt, pruning.mode = "coarse")
# Locate bigwig files
clipFiles = "./myFolder/wt/bigwig/"
clipFiles = list.files(clipFiles, pattern = ".bw$", full.names = TRUE)
clipFiles = clipFiles[grepl("WT", clipFiles)]
clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# Organize clip data in dataframe
colData = data.frame(
    name = "Example.WT", id = c(1:2),
    condition = factor(c("WT", "WT"), levels = c("WT")),
    clPlus = clipFilesP, clMinus = clipFilesM)
# Construct BSFDataSet object
bdsWT.initial = BSFDataSetFromBigWig(ranges = peaksInitial, meta = colData)
# Define binding sites
bdsWT = BSFind(bdsWT.initial, anno.genes = gns, anno.transcriptRegionList = trl)
```

```{r, eval=FALSE}
############################
# Construct KO binding sites
############################
# Locate peak calling result
peaks.ko = "./myFolder/ko/pureclip.bed"
peaks.ko = import(con = peaks.ko, format = "BED", extraCols=c("additionalScores" = "character"))
peaks.ko = keepStandardChromosomes(peaks.ko, pruning.mode = "coarse")
# Locate bigwig files
clipFiles = "./myFolder/ko/bigwig/"
clipFiles = list.files(clipFiles, pattern = ".bw$", full.names = TRUE)
clipFiles = clipFiles[grepl("KO", clipFiles)]
clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# Organize clip data in dataframe
colData = data.frame(
    name = "Example.KO", id = c(1:2),
    condition = factor(c("KO", "KO"), levels = c("KO")),
    clPlus = clipFilesP, clMinus = clipFilesM)
# Construct BSFDataSet object
bdsKO.initial = BSFDataSetFromBigWig(ranges = peaksInitial, meta = colData)
# Define binding sites
bdsKO = BSFind(bdsKO.initial, anno.genes = gns, anno.transcriptRegionList = trl)
```

### Starting from existing binding sites

Sometimes one already has a set of binding sites computed and just wants to perform the differential binding analysis part for the specific set at hand. In this case the pre-computed binding sites can be imported alongside the bigwig files from both conditions into a single `BSFDataSet`. This set can then directly be used for background generation (see section [Count matrix generation](## Count matrix generation)).

```{r, eval=FALSE}
# Locate bigwig files - KO
clipFilesKO = "./myFolder/ko/bigwig/"
clipFilesKO = list.files(clipFilesKO, pattern = ".bw$", full.names = TRUE)
clipFilesKO = clipFilesKO[grepl("KO", clipFilesKO)]
# Locate bigwig files - WT
clipFilesWT = "./myFolder/wt/bigwig/"
clipFilesWT = list.files(clipFilesWT, pattern = ".bw$", full.names = TRUE)
clipFilesWT = clipFilesWT[grepl("WT", clipFilesWT)]
# Merge both conditions
clipFiles = c(clipFilesWT, clipFilesKO)
# Split by strand
clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# Organize clip data in dataframe
colData = data.frame(
    name = "Example.KO", id = c(1:4),
    condition = factor(c("WT", "WT", "KO", "KO"), levels = c("WT", "KO")),
    clPlus = clipFilesP, clMinus = clipFilesM)
# Construct BSFDataSet object from pre-defined binding sites
# -> myBindingSites must be a GRanges object
bds.comb = BSFDataSetFromBigWig(ranges = myBindingSites, meta = colData)
```

### How to deal with multiple comparisons? 

In the case of a more complex experimental design that requires the comparison of more than two conditions, we recommend to compute as many pair-wise comparison as necessary. If one for example wants to incorporate an additional control condition (named *CTRL*) to the above setup, we would split the comparisons into three pair-wise designs (*CTRL-WT*, *CTRL-KO* and *WT-KO*). To ensure that all sets contain all binding sites found in either of the three conditions a merge over all three conditions has to be done prior to the testing. 

Assuming that binding sites were defined for each condition, one would merge all binding sites doing the following:^[Note: This also merges the meta data which we have to revert for the specific comparison that we want to compute.].

```{r, eval=FALSE}
bds.complex = combineBSF(list(ctrl = bds.ctrl, wt = bds.wt, ko = bds.ko))
```

Next all binding site ranges from this set can be extracted and used to create a new `BSFDataSet` object for each comparison that should be computed (here we compare WT-vs-KO). This is very similar to starting a differential binding analysis from an existing set of binding sites (see chapter [Starting from existing binding sites](### Starting from existing binding sites)).

```{r, eval=FALSE}
# extract binding site ranges from complex merge
bindingSites.complex = getRanges(bds.complex)
# Locate bigwig files - KO
clipFilesKO = "./myFolder/ko/bigwig/"
clipFilesKO = list.files(clipFilesKO, pattern = ".bw$", full.names = TRUE)
clipFilesKO = clipFilesKO[grepl("KO", clipFilesKO)]
# Locate bigwig files - WT
clipFilesWT = "./myFolder/wt/bigwig/"
clipFilesWT = list.files(clipFilesWT, pattern = ".bw$", full.names = TRUE)
clipFilesWT = clipFilesWT[grepl("WT", clipFilesWT)]
# Merge both conditions
clipFiles = c(clipFilesWT, clipFilesKO)
# Split by strand
clipFilesP = clipFiles[grep(clipFiles, pattern = "plus")]
clipFilesM = clipFiles[grep(clipFiles, pattern = "minus")]
# Organize clip data in dataframe
colData = data.frame(
    name = "Example.KO", id = c(1:2),
    condition = factor(c("KO", "KO"), levels = c("KO")),
    clPlus = clipFilesP, clMinus = clipFilesM)
# Construct BSFDataSet object from pre-defined binding sites
bds.comb = BSFDataSetFromBigWig(ranges = bindingSites.complex, meta = colData)
```

### Using the blacklist regions

In the case that the binding site definition was performed using a particularly high gene-wise level filter or if differential testing should be performed only on a subset of all defined binding sites, we recommend to use this information for the background generation by making use of the `blacklist` argument. This ensures that counts from all ranges that were at some point in the analysis identified as either a peak or a binding site will be removed from the *background-bin*.

# Session info

```{r}
sessionInfo()
```

# Bibliography 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflow.R
\name{estimateBsWidth}
\alias{estimateBsWidth}
\title{Function to estimate the appropriate binding site width together with the
optimal gene-wise filter level.}
\usage{
estimateBsWidth(
  object,
  bsResolution = c("medium", "fine", "coarse"),
  geneResolution = c("medium", "coarse", "fine", "finest"),
  est.maxBsWidth = 13,
  est.minimumStepGain = 0.02,
  est.maxSites = Inf,
  est.subsetChromosome = "chr1",
  est.minWidth = 2,
  est.offset = 1,
  sensitive = FALSE,
  sensitive.size = 5,
  sensitive.minWidth = 2,
  anno.annoDB = NULL,
  anno.genes = NULL,
  bsResolution.steps = NULL,
  geneResolution.steps = NULL,
  quiet = TRUE,
  veryQuiet = FALSE,
  reportScoresPerBindingSite = FALSE,
  ...
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored crosslink sites.
This means that ranges should have a width = 1.}

\item{bsResolution}{character; level of resolution at which different binding
site width should be tested}

\item{geneResolution}{character; level of resolution at which gene-wise filtering
steps should be tested}

\item{est.maxBsWidth}{numeric; the largest binding site width which should
considered in the testing}

\item{est.minimumStepGain}{numeric; the minimum additional gain in the score
in percent the next binding site width has to have, to be selected as best option}

\item{est.maxSites}{numeric; maximum number of PureCLIP sites that are used;}

\item{est.subsetChromosome}{character; define on which chromosome the
estimation should be done in function \code{\link{estimateBsWidth}}}

\item{est.minWidth}{the minimum size of regions that are subjected to the
iterative merging routine, after the initial region concatenation.}

\item{est.offset}{constant added to the flanking count in the signal-to-flank
ratio calculation to avoid division by Zero}

\item{sensitive}{logical; whether to enable sensitive pre-filtering before
binding site merging or not}

\item{sensitive.size}{numeric; the size (in nucleotides) of the merged
sensitive region}

\item{sensitive.minWidth}{numeric; the minimum size (in nucleoties) of the
merged sensitive region}

\item{anno.annoDB}{an object of class \code{OrganismDbi} that contains
the gene annotation (!!! Experimental !!!).}

\item{anno.genes}{an object of class \code{\link{GenomicRanges}} that represents
the gene ranges directly}

\item{bsResolution.steps}{numeric vector; option to use a user defined threshold
for binding site width directly. Overwrites \code{bsResolution}}

\item{geneResolution.steps}{numeric vector; option to use a user defined threshold
vector for gene-wise filtering resolution. Overwrites \code{geneResolution}}

\item{quiet}{logical; whether to print messages}

\item{veryQuiet}{logical; whether to suppress all messages}

\item{reportScoresPerBindingSite}{report the ratio score for each binding site
separately. Warning! This is for debugging and testing only. Downstream
functions can be impaired.}

\item{...}{additional arguments passed to \code{\link{pureClipGeneWiseFilter}}}
}
\value{
an object of class \code{\link{BSFDataSet}} with binding sites with
the `params` slots `bsSize` and `geneFilter` being filled
}
\description{
This function tests different width of binding sites for different gene-wise
filtering steps. For each test the signal-to-score ratio is calculated. The
mean over all gene-wise filterings at each binding site width is used to
extract the optimal width, which serves as anchor to select the optimal
gene-wise filter.
}
\details{
Parameter estimation is done on a subset of all crosslink sites
(\code{est.subsetChromosome}).

Gene-level filter can be tested with varying
levels of accuracy ranging from `finest` to `coarse`, representing 1% and
20% steps, respectively.

Binding site computation at each step can be done on three different accuracy
level (\code{bsResolution}). Option `fine` is equal to a normal run
of the \code{\link{makeBindingSites}} function. `medium` will perform
a shorter version of the binding site computation, skipping some of the
refinement steps. Option `coarse` will approximate binding sites by merged
crosslinks regions, aligning the center at the site with the highest score.

For each binding site in each set given the defined resolutions a signal-to-
flank score ratio is calculated and the mean of this score per set is returned.
Next a mean of means is created which results in a single score for each
binding site width that was tested. The width that yielded the highest score
is selected as optimal. In addtion the \code{minimumStepGain} option
allows control over the minimum additional gain in the score that a tested
width has to have to be selected as the best option.

To enhance the sensitivity of the binding site estimation, the sensitivity
mode exists. In this mode crosslink sites undergo a pre-filtering and merging
step, to exclude potential artifical peaks (experimental-, mapping-biases).
If sensitivity mode is activated the \code{est.minWidth} option should be set
to 1.

The optimal geneFilter is selected as the first one that passes the merged
mean of the selected optimal binding site width.

The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
estimateBsWidth(bds, anno.genes = gns, est.maxBsWidth = 19,
 geneResolution = "coarse", bsResolution = "coarse", est.subsetChromosome = "chr22")

}
\seealso{
\code{\link{BSFind}},
\code{\link{estimateBsWidthPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{transcriptRegionOverlapsPlot}
\alias{transcriptRegionOverlapsPlot}
\title{UpSet-plot to that shows the transcript region overlaps}
\usage{
transcriptRegionOverlapsPlot(object, text.size = NULL, show.title = TRUE)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{text.size}{numeric; fontsize of all numbers on axis}

\item{show.title}{logical; if plot title should be visible}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the transcript regions of binding sites
on overlapping loci. The function \code{\link{assignToTranscriptRegions}} is expected to be
executed prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
bds = makeBindingSites(object = bds, bsSize = 9)
bds = assignToGenes(bds, anno.genes = gns)
bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)
transcriptRegionOverlapsPlot(bds)

}
\seealso{
\code{\link{assignToTranscriptRegions}} \code{\link{transcriptRegionSpectrumPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{transcriptRegionSpectrumPlot}
\alias{transcriptRegionSpectrumPlot}
\title{Bar-chart to show the hosting transcript regions of binding sites}
\usage{
transcriptRegionSpectrumPlot(
  object,
  values = c("asis", "percentage"),
  normalize = FALSE,
  normalize.factor = c("sum", "median", "mean"),
  show.others = FALSE,
  text.size = 4
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{values}{character; if values should be presented 'as-is', that means
for example as frequencies in case \code{normalization = FALSE}, or as
percentages}

\item{normalize}{logical; whether to normalize values}

\item{normalize.factor}{character; indicate by what factor values should be
normalized to region length by}

\item{show.others}{logical; whether to show 'others' category. Has to be false
if \code{normalize = TRUE}}

\item{text.size}{numeric; font size of the numbers to be displayed on each bar}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the transcript regions of the hosting gene
for each binding site. The function \code{\link{assignToTranscriptRegions}}
is expected to be executed prior to calling this plot function.
}
\details{
Count frequencies can be normalized to the length of the hosting region with
option \code{normalize}. The specific factor how the hosting region length is
used is given by \code{normalize.factor}. In the case of
\code{normalize.factor = "sum"} binding site frequencies are divided by the
summed length of all regions that host the specific binding site.

Further with option \code{values} once can indicate whether raw or normalized
frequencies should be shown 'as-is' or normalized to 'percentages'.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
bds = makeBindingSites(object = bds, bsSize = 9)
bds = assignToGenes(bds, anno.genes = gns)
bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)
transcriptRegionSpectrumPlot(bds)

}
\seealso{
\code{\link{assignToTranscriptRegions}} \code{\link{transcriptRegionOverlapsPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{reproducibilityCutoffPlot}
\alias{reproducibilityCutoffPlot}
\title{Plot to that shows how many replicates support each binding site}
\usage{
reproducibilityCutoffPlot(
  object,
  cutoff = 0.05,
  min.crosslinks = 1,
  max.range = 20,
  ...
)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{cutoff}{a vector of length = 1, or of length = levels(meta$conditions)
with a single number (between 0-1) indicating the quantile cutoff}

\item{min.crosslinks}{numeric of length = 1, defines the lower boundary for
the minimum number of crosslinks a binding site has to be supported by all
replicates, regardless of the replicate specific quantile threshold}

\item{max.range}{maximum number of crosslinks per sites that should be shown}

\item{...}{further arguments passed to ggplot}
}
\value{
a plot of type \code{ggplot2} showing the per replicate
reproducibility cutoffs based on a given quantile threshold
}
\description{
Plotting function for settings specified in
\code{\link{reproducibilityFilter}}.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

# merge binding sites
bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1)

# use the same cutoff for both conditions
suppressWarnings(reproducibilityCutoffPlot(bds, max.range = 20, cutoff = c(0.05)))

# use different cutoffs for each condition
suppressWarnings(reproducibilityCutoffPlot(bds, max.range = 20, cutoff = c(0.1)))


}
\seealso{
\code{\link{reproducibilityFilter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflow.R
\name{pureClipGlobalFilter}
\alias{pureClipGlobalFilter}
\title{Filter PureCLIP sites by their score distribution}
\usage{
pureClipGlobalFilter(
  object,
  cutoff = 0.01,
  match.score = "score",
  quiet = FALSE
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored crosslink ranges
of width=1}

\item{cutoff}{numeric; defines the cutoff for which sites to keep, the
smallest step is 1\% (0.01)}

\item{match.score}{character; meta column name of the crosslink site
\code{\link{GenomicRanges}} object that holds the score which is used for
sub-setting}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class  \code{\link{BSFDataSet}} with its ranges filtered
by those that passed the threshold set with \code{cutoff}
}
\description{
Function that applies a filter on the global crosslink site score distribution.
The \code{\link{GenomicRanges}} contained in the \code{\link{BSFDataSet}} need to
have a meta-column that holds a numeric score value, which is used for filtering.
The name of the column can be set with \code{match.score}.
}
\details{
The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
# apply 5\% filter
pureClipGlobalFilter(object = bds, cutoff = 0.05)

}
\seealso{
\code{\link{BSFind}}, \code{\link{pureClipGlobalFilterPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{setRanges}
\alias{setRanges}
\alias{setRanges,BSFDataSet-method}
\title{Setter method for the ranges of the BSFDataSet object
The GRanges object that holds the genomic ranges information can be replaced.}
\usage{
setRanges(object, ...)

\S4method{setRanges}{BSFDataSet}(object, newRanges, dropSeqlevels = TRUE, quiet = FALSE)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}

\item{...}{additional arguments}

\item{newRanges}{an object of type \code{GRanges}}

\item{dropSeqlevels}{enforce seqnames to be the same in ranges and signal,
by dropping unused seqlevels which is required for most downstream functions
such as \code{coverageOverRanges}}

\item{quiet}{logical; whether to print messages}
}
\value{
object of type \code{\link{BSFDataSet}} with updated ranges
}
\description{
Setter method for the ranges of the BSFDataSet object
The GRanges object that holds the genomic ranges information can be replaced.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

rng = getRanges(bds)
rng = rng + 10
bdsNew = setRanges(bds, rng)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{getSignal}
\alias{getSignal}
\alias{getSignal,BSFDataSet-method}
\title{Accessor method for the signal data of the BSFDataSet object}
\usage{
getSignal(object)

\S4method{getSignal}{BSFDataSet}(object)
}
\arguments{
\item{object}{a BSFDataSet object}
}
\value{
returns the signal data, as list of RLE list for each strand, named
after the meta data columns "clPlus" and "clMinus"
}
\description{
Signal data is loaded from the path specified in \code{\link{getMeta}}
columns "clPlus" and "clMinus" and stored as a list of RLE lists.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

getSignal(bds)

}
\seealso{
\code{\link{getMeta}} \code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{getSummary}
\alias{getSummary}
\alias{getSummary,BSFDataSet-method}
\title{Accessor method for the summary slot of the BSFDataSet object}
\usage{
getSummary(object, ...)

\S4method{getSummary}{BSFDataSet}(object)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}

\item{...}{additional arguments}
}
\value{
returns the summary information storted in the summary slot after
\code{\link{makeBindingSites}} was run
}
\description{
The summary slot is used to track information of the filtering steps applied
in the \code{\link{makeBindingSites}} function
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1)

getSummary(bds)

}
\seealso{
\code{\link{BSFDataSet}} \code{\link{makeBindingSites}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{subset-BSFDataSet}
\alias{subset-BSFDataSet}
\alias{[,BSFDataSet,ANY,ANY,ANY-method}
\title{Subset a BSFDataSet object}
\usage{
\S4method{[}{BSFDataSet,ANY,ANY,ANY}(x, i, j, ..., drop = FALSE)
}
\arguments{
\item{x}{A \link{BSFDataSet} object.}

\item{i}{Position of the identifier or the name of the identifier itself.}

\item{j}{Not used.}

\item{...}{Additional arguments not used here.}

\item{drop}{if the signal not covered by the subsetted ranges should be
dropped or not}
}
\value{
A \link{BSFDataSet} object.
}
\description{
You can subset \link{BSFDataSet} by identifier or by position using the
\code{`[`} operator. Empty seqlevels are being droppend after the subset.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

bdsNew = bds[1:10]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{reproducibilitySamplesPlot}
\alias{reproducibilitySamplesPlot}
\title{UpSet-plot to that shows how each replicate supports binding sites}
\usage{
reproducibilitySamplesPlot(
  object,
  nIntersections = 20,
  show.title = TRUE,
  text.size = NULL
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{nIntersections}{numeric; number of intersection to be shown}

\item{show.title}{logical; if plot title should be visible}

\item{text.size}{numeric; fontsize of all numbers on axis}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the set sizes for each replicate, indicating
how many binding site the specific replicate supports given its specific
threshold. The function \code{\link{reproducibilityFilter}} is expected to be
executed prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds = makeBindingSites(object = bds, bsSize = 9)
bds = reproducibilityFilter(bds)
reproducibilitySamplesPlot(bds)

}
\seealso{
\code{\link{reproducibilityFilter}},
\code{\link{reproducibilityFilterPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Functions.R
\name{annotateWithScore}
\alias{annotateWithScore}
\title{Annotation function for BSFDataSet object}
\usage{
annotateWithScore(
  object,
  match.ranges = NULL,
  match.score = "score",
  match.option = c("max", "sum", "mean"),
  scoreRanges = lifecycle::deprecated(),
  MatchColScore = lifecycle::deprecated(),
  quiet = FALSE
)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{match.ranges}{a GRanges object, with numeric column for the score to match}

\item{match.score}{character; meta column name of the crosslink site
\code{\link{GenomicRanges}} object that holds the score to match}

\item{match.option}{character; option how score should be matched}

\item{scoreRanges}{deprecated -> use match.ranges instead}

\item{MatchColScore}{deprecated -> use match.score instead}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class BSFDataSet with updated meta columns of the ranges
}
\description{
This function can be used to annotate a \code{BSFDataSet} object with
merged binding sites with scores from the initial ranges
(eg. PureCLIP scores).
}
\details{
The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
if (.Platform$OS.type != "windows") {
    # load data
    csFile <- system.file("extdata", "PureCLIP_crosslink_sites_examples.bed",
                        package="BindingSiteFinder")
    cs = rtracklayer::import(con = csFile, format = "BED",
    extraCols=c("additionalScores" = "character"))
    cs$additionalScores = NULL
    clipFiles <- system.file("extdata", package="BindingSiteFinder")
    # two experimental conditions
    meta = data.frame(
    id = c(1,2,3,4),
    condition = factor(c("WT", "WT", "KD", "KD"),
    levels = c("KD", "WT")),
    clPlus = list.files(clipFiles, pattern = "plus.bw$", full.names = TRUE),
    clMinus = list.files(clipFiles, pattern = "minus.bw$",
     full.names = TRUE))
    bds = BSFDataSetFromBigWig(ranges = cs, meta = meta, silent = TRUE)

    # merge binding sites
    bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
    minCrosslinks = 2, minClSites = 1)

    # annotate with original pureCLIP score
    bdsRe = annotateWithScore(bds, cs)
}
}
\seealso{
\code{\link{BSFind}}, \code{\link{globalScorePlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exports.R
\name{exportToBED}
\alias{exportToBED}
\title{Wrapper function to export binding sites as BED files}
\usage{
exportToBED(object, con)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored ranges}

\item{con}{A path or URL}
}
\value{
a .bed file
}
\description{
Function that serves as a wrapper function for \code{rtracklayer::export}.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
\dontrun{
# export
# exportToBED(bds, con = "./myfile.bed")
}

}
\seealso{
\code{\link{BSFind}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/differentialPlots.R
\name{plotBsMA}
\alias{plotBsMA}
\title{MA style plot}
\usage{
plotBsMA(object, what = c("bs", "bg"), sig.threshold = 0.05)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with results calculated by
\code{\link{calculateBsFoldChange}}}

\item{what}{character; whether to show results for binding sites or the
background (one of: 'bs', 'bg')}

\item{sig.threshold}{numeric; what P value significance level to use
(default = 0.05)}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
Wrapper that plots differential binding results as MA plot. For each binding
site the estimated baseMean (log2) is shown on X and the fold-change (log2)
is shown on Y.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

# make testset
bds = makeBindingSites(bds, bsSize = 7)
bds = assignToGenes(bds, anno.genes = gns)
bds = imputeBsDifferencesForTestdata(bds)
bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)

# use all filters and remove binding sites that fail (default settings)
bds = filterBsBackground(bds)

# calculate fold-changes
bds = calculateBsFoldChange(bds)

# make MA plot
plotBsMA(bds)

}
\seealso{
\code{\link{calculateBsFoldChange}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{setName}
\alias{setName}
\alias{setName,BSFDataSet-method}
\title{Setter method for the names of the BSFDataSet object
The name slot holds the name information of the dataset}
\usage{
setName(object, ...)

\S4method{setName}{BSFDataSet}(object, newName)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}

\item{...}{additional arguments}

\item{newName}{a character that is the name}
}
\value{
object of type \code{\link{BSFDataSet}} with updated name
}
\description{
Setter method for the names of the BSFDataSet object
The name slot holds the name information of the dataset
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

bdsNew = setName(bds, "test01")

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{quickFigure}
\alias{quickFigure}
\title{Quick figures}
\usage{
quickFigure(
  object,
  what = c("main", "supp"),
  save.filename = NULL,
  save.width = 10,
  save.height = 12,
  save.device = "pdf",
  quiet = TRUE,
  ...
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{what}{character; the plotting option. One of: 'main', 'supp'}

\item{save.filename}{File name to create on the disc}

\item{save.width}{numeric; plot size width}

\item{save.height}{numeric; plot size height}

\item{save.device}{charcter; Device to use. One of: 'pdf', 'png', ...}

\item{quiet}{whether to print messages}

\item{...}{further arguments passed to \code{\link[ggplot2]{ggsave}}}
}
\value{
a plot
}
\description{
Summarize all results in a set of quick figures. Depending on how the
function is called a different set of analytic plots are arranged into
either a 'main' or 'supplementary' type multi-panel figure.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
bds = BSFind(bds, anno.genes = gns, anno.transcriptRegionList = regions,
 est.subsetChromosome = "chr22")
quickFigure(bds)

}
\seealso{
\code{\link{BSFind}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflow.R
\name{assignToGenes}
\alias{assignToGenes}
\title{Assign binding sites to their hosting genes}
\usage{
assignToGenes(
  object,
  overlaps = c("frequency", "hierarchy", "remove", "keep"),
  overlaps.rule = NULL,
  anno.annoDB = NULL,
  anno.genes = NULL,
  match.geneID = "gene_id",
  match.geneName = "gene_name",
  match.geneType = "gene_type",
  quiet = FALSE
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored binding sites. This
means that ranges should be > 1}

\item{overlaps}{character; how overlapping gene loci should be handled.}

\item{overlaps.rule}{character vector; a vector of gene type that should
be used to handle overlapping cases in a hierarchical manor. The order of the
vector is the order of the hierarchy.}

\item{anno.annoDB}{an object of class \code{OrganismDbi} that contains
the gene annotation (!!! Experimental !!!).}

\item{anno.genes}{an object of class \code{\link{GenomicRanges}} that represents
the gene ranges directly}

\item{match.geneID}{character; meta column name of the gene ID}

\item{match.geneName}{character; meta column name of the gene name}

\item{match.geneType}{character; meta column name of the gene type}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class \code{\link{BSFDataSet}} with binding sites having
hosting gene information added to their meta columns.
}
\description{
Function that assigns each binding site in the \code{\link{BSFDataSet}} to its
hosting gene given a gene annotation (\code{anno.annoDB}, \code{anno.genes}).
}
\details{
Regardless of the annotation source that is being used, the respective meta
information about the genes have to be present. They can be set by the
\code{match.geneID}, \code{match.geneName} and \code{match.geneType} arguments.

In the case of overlapping gene annotation, a single binding site will be
associated with multiple genes. The \code{\link{overlaps}} parameter allows
to decide in these cases. Option `frequency` will take the most frequently
observed gene type, option `hierarchy` works in conjunction with a user defined
rule (\code{overlaps.rule}). Options `remove` and `keep` will remove or
keep all overlapping cases, respectively.

Note that if an overlaps exists, but gene types are identical options
`frequency` and `hierarchy` will cause the gene that was seen first to be
selected as representative.

The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
# Load GRanges with genes
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
bds = makeBindingSites(object = bds, bsSize = 9)
bds = assignToGenes(bds, anno.genes = gns)

}
\seealso{
\code{\link{BSFind}}, \code{\link{geneOverlapsPlot}},
\code{\link{targetGeneSpectrumPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{getMeta}
\alias{getMeta}
\alias{getMeta,BSFDataSet-method}
\title{Accessor method for the meta data of the BSFDataSet object}
\usage{
getMeta(object)

\S4method{getMeta}{BSFDataSet}(object)
}
\arguments{
\item{object}{a BSFDataSet object}
}
\value{
returns the meta data \code{data.frame} with the columns "condition",
"clPlus" and "clMinus".
}
\description{
Meta data is stored as a \code{data.frame} and must contain the columns
"condition", "clPlus" and "clMinus".
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

getMeta(bds)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{duplicatedSitesPlot}
\alias{duplicatedSitesPlot}
\title{Plot the number of overlaps when assigning crosslink sites to genes}
\usage{
duplicatedSitesPlot(object)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the number of crosslink sites with their
respective overlapping rate. The function \code{\link{pureClipGeneWiseFilter}}
is expected to be executed prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
# Load GRanges with genes
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
# apply 5\% gene-wise filter
bds = pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.5,
 overlaps = "keepSingle")
duplicatedSitesPlot(bds)

}
\seealso{
\code{\link{pureClipGeneWiseFilter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Functions.R
\name{combineBSF}
\alias{combineBSF}
\title{Combine multiple \code{\link{BSFDataSet}} objects}
\usage{
combineBSF(
  list,
  overlaps.fix = TRUE,
  combine.bsSize = NULL,
  combine.name = NULL,
  force.reload = FALSE,
  quiet = TRUE,
  veryQuiet = FALSE
)
}
\arguments{
\item{list}{list; a list of objects from class \code{\link{BSFDataSet}} that
should be combined}

\item{overlaps.fix}{logical; if partially overlapping binding sites should be
re-centered}

\item{combine.bsSize}{numeric; the binding site size that the merged sites
should have. Default=NULL, then bsSize is taken from the input objects
in \code{list}.}

\item{combine.name}{character; meta table name of the combined object.
Default=NULL; then name is set to 'combined'}

\item{force.reload}{logical; whether the signal should be derived from the
merge of the input objects given in \code{list} or if the signal should be
re-loaded from the path given in the meta data.}

\item{quiet}{logical; whether to print messages or not}

\item{veryQuiet}{logical; whether to print status messages or not}
}
\value{
an object of class \code{\link{BSFDataSet}} with ranges, signal and
meta data resulting from the merge of the input objects.
}
\description{
This function combines all \code{\link{BSFDataSet}} objects from the input
list into a single \code{\link{BSFDataSet}} object.
}
\details{
Meta-data tables are added to each other by performing a row-wise bind,
basically adding all meta data tables underneath each other.

The default way of signal combination is merging all signal lists on the level
of the indivdual samples. One can also force a re-load of the signal list
component by using \code{force.reload=TRUE}.
The signal can be combined by

The ranges are combined by adding both granges objects together. With option
\code{overlaps.fix} one can decide if partially overlapping ranges should be
combined into a single range or not. If this option is FALSE one is likely to
have overlapping binding sites after the merge. If this option is TRUE, then
the combined coverage is used to guide the new center point for these cases.

The \code{combine.bsSize} option allows one to set a unique bsSize for all
objects that should be combined. Although it is recommended to combine only
objects with the same bsSize this option can be used to ensure that the
merged result has the same bsSize for all ranges.

This function is usually used to combine two datasets in the context of a
differntial testing analysis.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

# make binding sites
bds = makeBindingSites(bds, bsSize = 7)

# split ranges in two groups
allRanges = getRanges(bds)
set.seed(1234)
idx = sample(1:length(allRanges), size = length(allRanges)/2, replace = FALSE)
r1 = allRanges[idx]
r2 = allRanges[-idx]

# splite meta data
allMeta = getMeta(bds)
m1 = allMeta[1:2,]
m2 = allMeta[3:4,]

# create new objects
bds1 = setRanges(bds, r1)
bds2 = setRanges(bds, r2)
bds1 = setMeta(bds1, m1)
bds2 = setMeta(bds2, m2)
bds1 = setName(bds1, "test1")
bds2 = setName(bds2, "test2")

# merge two objects with '+' operator
c1 = bds1 + bds2

# merge two objects from list
list = list(bds1, bds2)
c1 = combineBSF(list = list, overlaps.fix = TRUE,
 combine.bsSize = NULL, combine.name = NULL, quiet = TRUE)

}
\seealso{
\code{\link{processingStepsFlowChart}}, \code{\link{calculateBsBackground}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{targetGeneSpectrumPlot}
\alias{targetGeneSpectrumPlot}
\title{Bar-chart to show the hosting gene types of binding sites}
\usage{
targetGeneSpectrumPlot(object, showNGroups = 5, text.size = 4)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{showNGroups}{numeric; the number of different gene types to show}

\item{text.size}{numeric; the size of the text elments on the plot}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the gene type of the hosting gene for
each binding site. The function \code{\link{assignToGenes}} is expected to be
executed prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
bds = makeBindingSites(object = bds, bsSize = 9)
bds = assignToGenes(bds, anno.genes = gns)
targetGeneSpectrumPlot(bds)

}
\seealso{
\code{\link{assignToGenes}} \code{\link{geneOverlapsPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/differentialPlots.R
\name{geneRegulationPlot}
\alias{geneRegulationPlot}
\title{Gene Regulation Plot}
\usage{
geneRegulationPlot(
  object,
  plot.geneID = NULL,
  anno.annoDB = NULL,
  anno.genes = NULL,
  match.geneID = "gene_id",
  match.geneName = "gene_name",
  plot.gene.n.tiles = 100,
  alpha = 0.05,
  lfc.cutoff = 2,
  transcript.regions.outlier.handle = c("first", "second", "both", "remove"),
  quiet = FALSE
)
}
\arguments{
\item{object}{object; a \code{\link{BSFDataSet}} object}

\item{plot.geneID}{character;  the gene id of the gene to display. The id must
match with the gene ids given in the annotation object.}

\item{anno.annoDB}{object; an object of class \code{OrganismDbi} that contains
the gene annotation (!!! Experimental !!!).}

\item{anno.genes}{object; an object of class \code{\link{GenomicRanges}}
that represents the gene ranges directly.}

\item{match.geneID}{character; meta column name of the gene ID}

\item{match.geneName}{character; meta column name of the gene name}

\item{plot.gene.n.tiles}{numeric; number of tiles the gene should be split in}

\item{alpha}{numeric; the alpha value to show significantly regulated
binding sites. This should match the alpha value used in \code{\link{calculateBsFoldChange}}.}

\item{lfc.cutoff}{numeric; log2 fold-change cutoff to show significantly
regulated binding sites. This should match the lfc.cutoff value used in
\code{\link{calculateBsFoldChange}}.}

\item{transcript.regions.outlier.handle}{character; the option how to handle
multiple transcript region annotations being present for the same binding
site.}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class \code{ggplot2}
}
\description{
Display the fold-change of all binding sites from a given gene on a relative
per-nucleotide scale. Binding sites are displayed as dots and with increasing
log2 fold-change, they deviate stronger from the center line.
}
\details{
For this function to work, binding sites must be assigned to hosting genes
using \code{\link{assignToGenes}}. It is also recommended to assing
binding sites to transcript regions with \code{\link{assignToTranscriptRegions}}.

It is also necessary to calculate the log2 fold-change of binding sites between
two conditions using the differential binding workflow \code{\link{calculateBsFoldChange}}.

If in addition the transcript regions of the binding sites are given, then
shapes are changed accordingly. An edge case can arise from the merging of two
\code{\link{BSFDataSet}} objects. If binding sites are overlapping and
slightly offset close to the end of a particular transcript region annotation,
they might be assigned to different regions in both objects. This results in
some ambiguity after the merge, where for instance a binding site can be
assigned to CDS and 3'UTR. To handle how such edge cases are displayed, the
\code{transcript.regions.outlier.handle} exists. As default, simply the
region of the object that was merged first is shown. If one is interested in
showing all regions, then the options \code{both} displays both annotations
at the same time and labels them accordingly.
}
\examples{

# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])

# make testset
bds = makeBindingSites(bds, bsSize = 7)
bds = assignToGenes(bds, anno.genes = gns)
bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)
bds = imputeBsDifferencesForTestdata(bds)
bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)

# use all filters and remove binding sites that fail (default settings)
bds = filterBsBackground(bds)

# calculate fold-changes
bds = calculateBsFoldChange(bds)

# make example plot
exampleGeneId = "ENSG00000253352.10"
geneRegulationPlot(bds, plot.geneID = exampleGeneId, anno.genes = gns)

}
\seealso{
\code{\link{BSFind}},
\code{\link{calculateBsFoldChange}}
\code{\link{assignToGenes}}
\code{\link{assignToTranscriptRegions}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflow.R
\name{assignToTranscriptRegions}
\alias{assignToTranscriptRegions}
\title{Assign binding sites to their hosting transcript regions}
\usage{
assignToTranscriptRegions(
  object,
  overlaps = c("frequency", "hierarchy", "flag", "remove"),
  overlaps.rule = NULL,
  anno.annoDB = NULL,
  anno.transcriptRegionList = NULL,
  normalize.exclude.upper = 0.02,
  normalize.exclude.lower = 0.02,
  quiet = FALSE
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored binding sites. This
means that ranges should be > 1}

\item{overlaps}{character; how overlapping transcript regions should be handled.}

\item{overlaps.rule}{character vector; a vector of transcript region
names that should be used to handle overlapping cases in a hierarchical manor.
The order of the vector is the order of the hierarchy.}

\item{anno.annoDB}{an object of class \code{OrganismDbi} that contains
the transcript region annotation (!!! Experimental !!!).}

\item{anno.transcriptRegionList}{an object of class \code{\link{CompressedGRangesList}}
that holds an ranges for each transcript region}

\item{normalize.exclude.upper}{numeric; percentage value that indicates the
upper boundary for transcript region length to be considered when calculating
normalization factors for regions.}

\item{normalize.exclude.lower}{numeric; percentage value that indicates the
lower boundary for transcript region length to be considered when calculating
normalization factors for regions.}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class \code{\link{BSFDataSet}} with binding sites having
hosting transcript region information added to their meta columns.
}
\description{
Function that assigns each binding site in the \code{\link{BSFDataSet}} to its
hosting transcript region given an annotation database (\code{anno.annoDB}), or
a GRanges list / \code{\link{CompressedGRangesList}} (\code{anno.transcriptRegionList})
that holds the ranges for the transcript regions of interest.
}
\details{
Since the assignment is based on the overlaps of annotated transcript ranges
and binding sites, no matching meta data is needed.

In the case of transcript regions overlaps are very frequent. To resolve these
cases the \code{\link{overlaps}} argument can be used. Option `frequency`
will take the most frequently observed transcript region, option `hierarchy`
works in conjunction with a user defined rule (\code{overlaps.rule}).
Options `flag` and `remove` will label binding sites with an ambiguous tag or
remove all overlapping cases, respectively.

The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])

bds = makeBindingSites(object = bds, bsSize = 9)
bds = assignToGenes(bds, anno.genes = gns)
bds = assignToTranscriptRegions(object = bds, anno.transcriptRegionList = regions)

}
\seealso{
\code{\link{BSFind}},
\code{\link{transcriptRegionOverlapsPlot}},
\code{\link{transcriptRegionSpectrumPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{mergeCrosslinkDiagnosticsPlot}
\alias{mergeCrosslinkDiagnosticsPlot}
\title{Plot binding site merging diagnostics}
\usage{
mergeCrosslinkDiagnosticsPlot(object)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the number of regions to merge over the width
of these regions for each merging iteration calculated in
\code{\link{makeBindingSites}}. The function \code{\link{makeBindingSites}}
is expected to be executed prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds = makeBindingSites(object = bds, bsSize = 9)
mergeCrosslinkDiagnosticsPlot(bds)

}
\seealso{
\code{\link{makeBindingSites}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{setSignal}
\alias{setSignal}
\alias{setSignal,BSFDataSet-method}
\title{Setter method for the signal data of the BSFDataSet object}
\usage{
setSignal(object, ...)

\S4method{setSignal}{BSFDataSet}(object, newSignal, dropSeqlevels = TRUE, quiet = FALSE)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{...}{additional arguments}

\item{newSignal}{list of RLE lists}

\item{dropSeqlevels}{enforce seqnames to be the same in ranges and signal,
by dropping unused seqlevels which is required for most downstream functions
such as \code{coverageOverRanges}}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of type \code{\link{BSFDataSet}} with updated signal
}
\description{
Signal data is loaded from the path specified in \code{\link{getMeta}}
columns "clPlus" and "clMinus" and stored as a list of RLE lists.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

sgn = getSignal(bds)
sgn = lapply(sgn, function(selStrand){
   lapply(selStrand, function(chrList){
       chrList[names(chrList) == "chr22"]
   })
})
bdsNew = setSignal(bds, sgn)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{setSummary}
\alias{setSummary}
\alias{setSummary,BSFDataSet-method}
\title{Setter method for the summary slot of the BSFDataSet object}
\usage{
setSummary(object, ...)

\S4method{setSummary}{BSFDataSet}(object, summary)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{...}{additional arguments}

\item{summary}{a data.frame with the summary information to be stored in
\code{BSFDataSet}}
}
\value{
an object of type \code{\link{BSFDataSet}} with updated summary info
}
\description{
The summary slot is used to track information of the filtering steps applied
in the \code{\link{makeBindingSites}} function
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

df = data.frame(processingStep = c(1,2),
parameter = c(3,4))
bds = setSummary(bds, df)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\docType{methods}
\name{summary}
\alias{summary}
\alias{summary,BSFDataSet-method}
\title{Summary method to for the BSFDataSet}
\usage{
\S4method{summary}{BSFDataSet}(object)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}
}
\value{
summary of the current object
}
\description{
Prints the summaryinformation for the \code{\link{BSFDataSet}} object. This
includes information on samples, conditions and crosslinks.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

summary(bds)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\docType{methods}
\name{show}
\alias{show}
\alias{show,BSFDataSet-method}
\title{Show method to for the BSFDataSet}
\usage{
\S4method{show}{BSFDataSet}(object)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}
}
\value{
shows the current object state
}
\description{
Prints the information for each of the input data slots in the
\code{\link{BSFDataSet}} object.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

show(bds)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoverageFunctions.R
\name{clipCoverage}
\alias{clipCoverage}
\title{Coverage function for BSFDataSet objects}
\usage{
clipCoverage(
  object,
  ranges.merge = FALSE,
  ranges.merge.method = c("sum", "mean"),
  positions.merge = FALSE,
  positions.merge.method = c("sum", "mean"),
  samples.merge = TRUE,
  samples.group = c("all", "condition"),
  samples.merge.method = c("sum", "mean"),
  out.format = c("granges", "data.frame"),
  out.format.overwrite = FALSE,
  match.rangeID = "bsID",
  quiet = FALSE
)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{ranges.merge}{logical; whether to merge ranges}

\item{ranges.merge.method}{character; how to combine ranges ('sum' or 'mean')}

\item{positions.merge}{logical; whether to merge positions}

\item{positions.merge.method}{character; how to combine positions ('sum' or 'mean')}

\item{samples.merge}{logical: whether to merge samples}

\item{samples.group}{character; how samples should be grouped when combining
('all', 'condition')}

\item{samples.merge.method}{charater; how to combine positions ('sum' or 'mean')}

\item{out.format}{character; how the coverage should be returned
('data.frame' or 'granges'). Note that option 'granges' only exists if the
output coverage is of the same rows as the input ranges.}

\item{out.format.overwrite}{logical; if \code{out.format='granges'}, then
decide wheter the meta columns should be extended by the coverage information
or be overwritten}

\item{match.rangeID}{character; unique internal identifier.
Name of the meta column of the input ranges
that should be used as identifier to match the coverage back to the input
ranges. Is 'bsID' as default, since that ID exists for all binding sites
after \code{makeBindingSites} was called.}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class specified in \code{out.format}
}
\description{
Function that computes a crosslink coverage with all samples over all ranges
given in the \code{\link{BSFDataSet}}. The coverage can be summarized over
all combinations of the three dimension (samples, ranges, positions).
}
\details{
When summarizing the crosslink coverage over samples (\code{samples.merge=TRUE})
one can decide whether to summarize all samples or whether to keep conditions
separate (\code{samples.group}). This either reduces the samples dimension
to a single matrix, or a list. For a binding site set with 100 binding sites
of width=7 and 4 replicates with 2 conditions, the following options are
possible. With merging enabled and \code{samples.group='all'} the coverage of
all samples is combined. With \code{samples.group='condition'} only samples
of the same condition are grouped.

When summarizing the crosslink coverage over ranges, all ranges are combined
which reduces the ranges dimension to a single vector. This turns eg. a
binding site set of 100 binding sites with width=7 into a vector of length
100 with exactly one column. Depending on how the samples were summarized, the
result can be a single such vector, or a list.

When summarizing the crosslink coverage over positions, all positions are
combined which reduces the positions dimension to a single vector. This turns
eq. a binding site set of 100 binding sites with width=7 into a vector of
length 1 with 7 columns. Depending on how the samples were summarized, the
result can be a single such vector, or a list.

For all summarizing operations options \code{sum} and \code{mean} exists. This
allows for normalization by the eg. the number of binding sites, size of the
range, number of sample, etc..

If the resulting object does have a dimension that fits to the number of
input ranges the result can be directly attached to them. Basically extending
the \code{GRanges} object (\code{out.format}).
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

bds = makeBindingSites(bds, bsSize = 7)
# sum of each replicate over each binding site position
c1 = clipCoverage(bds, out.format = "data.frame", positions.merge = TRUE,
 ranges.merge = FALSE, samples.merge = FALSE)
# total signal per binding site from all samples
c2 = clipCoverage(bds, out.format = "granges", positions.merge = FALSE,
 ranges.merge = TRUE, samples.merge = TRUE, samples.group = "all")
# total signal per binding site from all samples - split by condition
c3 = clipCoverage(bds, out.format = "granges", positions.merge = FALSE,
 ranges.merge = TRUE, samples.merge = TRUE, samples.group = "condition")

}
\seealso{
\code{\link{BSFDataSet}}, \code{\link{BSFind}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/differentialFunctions.R
\name{imputeBsDifferencesForTestdata}
\alias{imputeBsDifferencesForTestdata}
\title{Impute artificial differences in the example data set}
\usage{
imputeBsDifferencesForTestdata(object, size = 5, change.per = 0.1)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object; explicitly the test data set
from the extdata folder}

\item{size}{numeric; the number of positions on which signal should be
deleted, counting from the start}

\item{change.per}{numeric; the percentage of ranges that should be effected
by the change.}
}
\value{
object a \code{\link{BSFDataSet}} object with the signal slot adapted
to reflect changes in binding between two artificial conditions
}
\description{
A function that works only on the test data set provided with the package. It
is used for internal testing and the making of examples to showcase the
differential binding functions.
}
\details{
Differences between samples are artificially introduced by removing the
signal on a random set of binding sites of the input.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

bds = makeBindingSites(bds, bsSize = 7)
bds = assignToGenes(bds, anno.genes = gns)

bds = imputeBsDifferencesForTestdata(bds)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Functions.R
\name{supportRatio}
\alias{supportRatio}
\title{Support ratio function for BSFDataSet objects}
\usage{
supportRatio(object, bsWidths, bsFlank = NA, sub.chr = NA, ...)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{bsWidths}{a numeric vector indicating the different binding site
width to compute the ratio for}

\item{bsFlank}{optional; a numeric vector of the same length as
\code{bsWidth} used to specify the width of the flanking regions}

\item{sub.chr}{chromosome identifier (eg, chr1, chr2) used for subsetting the
BSFDataSet object.}

\item{...}{further arguments passed to \code{makeBindingSites}}
}
\value{
an object of class \code{data.frame}
}
\description{
Functions that computes a ratio to determine how well a given binding site
with is supported by the crosslink coverage of the data. For a given
\code{BSFDataSet} object binding sites are computed for each width indicated
in the \code{bsWidths} vector (using the \code{\link{coverageOverRanges}}
function). These coverages are compared to the coverage of regions flanking
the binding sites. If not indicated in \code{bsFlank} these regions are of
the same width as the binding sites.
}
\details{
Testing the width of 3nt for example, would result in a coverage within all
3nt wide binding sites (c1) and a coverage computed on the adjacent 3nt
flanking the binding sites up- and downstream (f1, f2). Based on these
 numbers the ratio is computed by: c1/(1/2(f1+f2)).

The median over all ratios is reported as representative value.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

suppressWarnings(supportRatio(bds, bsWidths = c(3,7)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{estimateBsWidthPlot}
\alias{estimateBsWidthPlot}
\title{Plot the signal-to-flank score for varying gene-wise filter and binding
site width}
\usage{
estimateBsWidthPlot(object)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the the signal-to-flank score as a mean
for each binding site width and gene-wise filter as indicated when executing
\code{\link{estimateBsWidth}}. Additionally a mean of means visualizes the
overall trend and a red line indicates the suggested optimal binding site
width. The function \code{\link{estimateBsWidth}} is expected to be
executed prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
bds = estimateBsWidth(bds, anno.genes = gns, est.maxBsWidth = 19,
 geneResolution = "coarse", bsResolution = "coarse", est.subsetChromosome = "chr22")
estimateBsWidthPlot(bds)
}
\seealso{
\code{\link{estimateBsWidth}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exports.R
\name{processingStepsTable}
\alias{processingStepsTable}
\title{Create a table of all workflow steps for reporting}
\usage{
processingStepsTable(object, option = c("reduced", "full", "extended"))
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored ranges}

\item{option}{character; how detailed the table should be}
}
\value{
a \code{kableExtra} table
}
\description{
Function that creates a printable table with all steps and numbers for each
of the workflow steps that were carried out.
}
\details{
If \code{\link{option}} is set to `reduced`, only the most necessary information
are collected. Option `full` contains a full list of all options and parameters
that were set in any of the workflow functions. Option `extended` contains
extra information about the binding site merging step.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
# apply 5\% filter
bds = pureClipGlobalFilter(object = bds, cutoff = 0.05)
processingStepsTable(bds)

}
\seealso{
\code{\link{BSFind}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exports.R
\name{exportTargetGenes}
\alias{exportTargetGenes}
\title{Function to export sorted RBP target genes}
\usage{
exportTargetGenes(
  object,
  path = "./",
  format = c("csv", "xslx"),
  sort = c("score", "bs"),
  split = c("none", "geneType", "transcriptRegion")
)
}
\arguments{
\item{object}{a \code{BSFDataSet} object with stored ranges}

\item{path}{A path to where the output should be stored}

\item{format}{output file format}

\item{sort}{sorting rule for genes}

\item{split}{if and how the output file should be split}
}
\value{
a file of the type specified in \code{\link{format}}
}
\description{
Genes with binding sites are target genes of the RBP. They can be exported
as 'csv' or 'xlsx' file. Genes can be sorted by the sum of the individual
binding sites score, or by the number of binding sites per gene.
}
\details{
As output option, one can either output all genes in a single file, or split
by either gene-type or transcript-region. This options requires that either
\code{\link{BSFind}} or the individual functions \code{\link{assignToGenes}},
 and \code{\link{assignToTranscriptRegions}} were run.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
\dontrun{
# export
# exportTargetGenes(bds)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflow.R
\name{pureClipGeneWiseFilter}
\alias{pureClipGeneWiseFilter}
\title{Filter PureCLIP sites by their score distribution per gene}
\usage{
pureClipGeneWiseFilter(
  object,
  cutoff = 0.05,
  overlaps = c("keepSingle", "removeAll", "keepAll"),
  anno.annoDB = NULL,
  anno.genes = NULL,
  match.score = "score",
  match.geneID = "gene_id",
  quiet = FALSE
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored crosslink ranges
of width=1}

\item{cutoff}{numeric; defines the cutoff for which sites to remove, the
smallest step is 1\% (0.01). A cutoff of 5\% will remove the lowest 5\% sites,
given their score, on each gene, thus keeping the strongest 95\%.}

\item{overlaps}{character; how overlapping gene loci should be handled.}

\item{anno.annoDB}{an object of class \code{OrganismDbi} that contains
the gene annotation (!!! Experimental !!!).}

\item{anno.genes}{an object of class \code{\link{GenomicRanges}} that represents
the gene ranges directly}

\item{match.score}{character; meta column name of the crosslink site
\code{\link{GenomicRanges}} object that holds the score which is used for
sub-setting}

\item{match.geneID}{character; meta column name of the genes
\code{\link{GenomicRanges}} object that holds a unique geneID}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class \code{\link{BSFDataSet}} with its ranges filtered
by those that passed the gene-wise threshold set with \code{cutoff}
}
\description{
Function that applies a filter on the crosslink site score distribution at
gene level. This allows to filter for those sites with the strongest signal
on each gene. Since scores are tied to the expression level of the hosting
transcript this function allows a fair filter for all genes partially
independent of the expression level.
}
\details{
The \code{\link{GenomicRanges}} contained in the \code{\link{BSFDataSet}} need to
have a meta-column that holds a numeric score value, which is used for filtering.
The name of the column can be set with \code{scoreCol}.

In the case of overlapping gene annotation, a single crosslink site will be
attributed to multiple genes. The \code{\link{overlaps}} parameter allows
to control these cases. Option `keepSingle` will only keep a single instance
of the site; `removeAll` will remove both sites; `keepAll` will keep both
sites.

The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
# Load GRanges with genes
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
# apply 5\% gene-wise filter
pureClipGeneWiseFilter(object = bds, anno.genes = gns, cutoff = 0.5, overlaps = "keepSingle")

}
\seealso{
\code{\link{BSFind}}, \code{\link{estimateBsWidthPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bsfind.R
\name{BSFind}
\alias{BSFind}
\title{RBP binding site definition for iCLIP data}
\usage{
BSFind(
  object,
  bsSize = NULL,
  cutoff.geneWiseFilter = NULL,
  cutoff.globalFilter = 0.01,
  est.bsResolution = "medium",
  est.geneResolution = "medium",
  est.maxBsWidth = 13,
  est.minimumStepGain = 0.02,
  est.maxSites = Inf,
  est.subsetChromosome = "chr1",
  est.minWidth = 2,
  est.offset = 1,
  est.sensitive = FALSE,
  est.sensitive.size = 5,
  est.sensitive.minWidth = 2,
  merge.minWidth = 2,
  merge.minCrosslinks = 2,
  merge.minClSites = 1,
  merge.CenterIsClSite = TRUE,
  merge.CenterIsSummit = TRUE,
  repro.cutoff = NULL,
  repro.nReps = NULL,
  repro.minCrosslinks = 1,
  overlaps.geneWiseFilter = "keepSingle",
  overlaps.geneAssignment = "frequency",
  overlaps.rule.geneAssignment = NULL,
  overlaps.TranscriptRegions = "frequency",
  overlaps.rule.TranscriptRegions = NULL,
  stf.flank = "bs",
  stf.flank.size = NULL,
  match.score = "score",
  match.geneID = "gene_id",
  match.geneName = "gene_name",
  match.geneType = "gene_type",
  match.ranges.score = NULL,
  match.option.score = "max",
  anno.annoDB = NULL,
  anno.genes = NULL,
  anno.transcriptRegionList = NULL,
  quiet = FALSE,
  veryQuiet = FALSE,
  ...
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with stored ranges}

\item{bsSize}{an odd integer value specifying the size of the output
binding sites}

\item{cutoff.geneWiseFilter}{numeric; defines the cutoff for which sites to
remove in in function \code{\link{pureClipGeneWiseFilter}}. The smallest step
is 1\% (0.01). A cutoff of 5\% will remove the lowest 5\% sites, given their
score, on each gene, thus keeping the strongest 95\%.}

\item{cutoff.globalFilter}{numeric; defines the cutoff for which sites to
keep, the smallest step is 1\% (0.01) in function
\code{\link{pureClipGlobalFilter}}}

\item{est.bsResolution}{character; level of resolution of the binding site
width in function \code{\link{estimateBsWidth}}}

\item{est.geneResolution}{character; level of resolution of the gene-wise
filtering in function \code{\link{estimateBsWidth}}}

\item{est.maxBsWidth}{numeric; the largest binding site width which should
considered in the testing}

\item{est.minimumStepGain}{numeric; the minimum additional gain in the score
in percent the next binding site width has to have, to be selected as best option}

\item{est.maxSites}{numeric; maximum number of PureCLIP sites that are used}

\item{est.subsetChromosome}{character; define on which chromosome the
estimation should be done in function \code{\link{estimateBsWidth}}}

\item{est.minWidth}{the minimum size of regions that are subjected to the
iterative merging routine, after the initial region concatenation.}

\item{est.offset}{constant added to the flanking count in the signal-to-flank
ratio calculation to avoid division by Zero}

\item{est.sensitive}{logical; whether to enable sensitive pre-filtering before
binding site merging or not}

\item{est.sensitive.size}{numeric; the size (in nucleotides) of the merged
sensitive region}

\item{est.sensitive.minWidth}{numeric; the minimum size (in nucleoties) of the
merged sensitive region}

\item{merge.minWidth}{the minimum size of regions that are subjected to the
iterative merging routine, after the initial region concatenation.}

\item{merge.minCrosslinks}{the minimal number of positions to overlap with at least
one crosslink event in the final binding sites}

\item{merge.minClSites}{the minimal number of crosslink sites that have to
overlap a final binding site}

\item{merge.CenterIsClSite}{logical, whether the center of a final binding
site must be covered by an initial crosslink site}

\item{merge.CenterIsSummit}{logical, whether the center of a final binding
site must exhibit the highest number of crosslink events}

\item{repro.cutoff}{numeric; percentage cutoff to be used for the
reproducibility quantile filtering}

\item{repro.nReps}{numeric; number of replicates that must meet the cutoff
defined in \code{repro.cutoff} for a binding site to be called reproducible.
Defaults to N-1.}

\item{repro.minCrosslinks}{numeric; minimal number of crosslinks a binding
site needs to have to be called reproducible. Acts as a lower boundary for
\code{repro.cutoff}. Defaults to 1.}

\item{overlaps.geneWiseFilter}{character; how overlaps should be handled in
\code{\link{pureClipGeneWiseFilter}}}

\item{overlaps.geneAssignment}{character; how overlaps should be handled in
\code{\link{assignToGenes}}}

\item{overlaps.rule.geneAssignment}{character vector; a vector of gene types
that should be used to handle overlaps if option 'hierarchy' is selected
for \code{\link{assignToGenes}}. The order of the vector is the order of
the hierarchy.}

\item{overlaps.TranscriptRegions}{character; how overlaps should be handled in
\code{\link{assignToTranscriptRegions}}}

\item{overlaps.rule.TranscriptRegions}{character vector; a vector of gene types
that should be used to handle overlaps if option 'hierarchy' is selected
for \code{\link{assignToTranscriptRegions}}. The order of the vector is the order of
the hierarchy.}

\item{stf.flank}{character; how the flanking region shoule be set. Options are
'bs', 'manual'}

\item{stf.flank.size}{numeric; if flank='manual' provide the desired flanking size}

\item{match.score}{character; meta column name of the crosslink site}

\item{match.geneID}{character; meta column name of the genes}

\item{match.geneName}{character; meta column name of the gene name}

\item{match.geneType}{character; meta column name of the gene type}

\item{match.ranges.score}{a GRanges object, with numeric column for the score
to match in function \code{\link{annotateWithScore}}}

\item{match.option.score}{character; meta column name of the crosslink site
in function \code{\link{annotateWithScore}}}

\item{anno.annoDB}{an object of class \code{OrganismDbi} that contains
the gene annotation !!! Experimental !!!}

\item{anno.genes}{an object of class \code{\link{GenomicRanges}} that represents
the gene ranges directly}

\item{anno.transcriptRegionList}{an object of class \code{\link{CompressedGRangesList}}
that holds an ranges for each transcript region}

\item{quiet}{logical; whether to print messages}

\item{veryQuiet}{logical; whether to suppress all messages}

\item{...}{additional arguments passed to \code{\link{estimateBsWidth}},
\code{\link{makeBindingSites}} and \code{\link{reproducibilityFilter}}}
}
\value{
an object of class \code{\link{BSFDataSet}} with ranges merged into
binding sites given the inputs.
}
\description{
This is the main function that performs the binding site definition analysis
through the following steps:
\enumerate{
\item Filter PureCLIP sites by their score distribution: \code{\link{pureClipGlobalFilter}}
\item Estimate the appropriate binding site width together with the optimal gene-wise filter level: \code{\link{estimateBsWidth}}
\item Filter PureCLIP sites by their score distribution per gene: \code{\link{pureClipGeneWiseFilter}}
\item Define equally sized binding sites: \code{\link{makeBindingSites}}
\item Perform replicate reproducibility filter: \code{\link{reproducibilityFilter}}
\item Assign binding sites to their hosting genes: \code{\link{assignToGenes}}
\item Assign binding sites to their hosting transcript regions: \code{\link{assignToTranscriptRegions}}
\item Re-assign PureCLIP scores to binding sites: \code{\link{annotateWithScore}}
\item Calculation of signal-to-flank ratio: \code{\link{calculateSignalToFlankScore}}
}
}
\details{
If only the annotation is provided (\code{anno.genes} and
\code{anno.transcriptRegionList}), then binding sites size (\code{bsSize})
and gene-wise cutoff (\code{cutoff.geneWiseFilter}) are estimated using
\code{\link{estimateBsWidth}}. To avoid this behavior one has to provide
input values for the arguments \code{bsSize} and \code{cutoff.geneWiseFilter}.

If no binding site size is provided through \code{bsSize}, then
\code{\link{estimateBsWidth}} is called to estimate the optimal size for the
given data-set. The result of this estimation can be looked at with
\code{\link{estimateBsWidthPlot}} and arguments can be adjusted if needed.

Use the \code{\link{processingStepsFlowChart}} function to get an overview
of all steps carried out by the function.

For complete details on each step, see the manual pages of the respective
functions. The \code{BSFind} function returns a \code{\link{BSFDataSet}}
with ranges merged into binding sites. A full flowchart for the entire process
can be visualized with \code{\link{processingStepsFlowChart}}. For each of the
individual steps dedicated diagnostic plots exists. Further information can be
found in our Bioconductor vignette:
\url{https://www.bioconductor.org/packages/release/bioc/html/BindingSiteFinder.html}
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
# Load genes
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
# load transcript regions
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
BSFind(object = bds, bsSize = 9, anno.genes = gns,
 anno.transcriptRegionList = regions, est.subsetChromosome = "chr22")

}
\seealso{
\code{\link{BSFDataSet}}, \code{\link{estimateBsWidth}},
\code{\link{pureClipGlobalFilter}}, \code{\link{pureClipGeneWiseFilter}},
\code{\link{assignToGenes}}, \code{\link{assignToTranscriptRegions}},
\code{\link{annotateWithScore}}, \code{\link{reproducibilityFilter}},
\code{\link{calculateSignalToFlankScore}},
\code{\link{processingStepsFlowChart}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{reproducibilityScatterPlot}
\alias{reproducibilityScatterPlot}
\title{Plot that shows binding site reproducibility as scatter}
\usage{
reproducibilityScatterPlot(object, quiet = FALSE)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class \code{ggplot2}
}
\description{
Function compute the number of crosslinks per binding site on a log2 scale
for each sample. Samples are pairwise correlated as a scatter and pairwise
pearson correlation is shown.
}
\details{
Unlike most plotting functions, this function is actively calculating the
values to plot.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 3, sub.chr = "chr22")
reproducibilityScatterPlot(bds)

}
\seealso{
\code{\link{BSFind}},
\code{\link{reproducibilityFilter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exports.R
\name{collapseReplicates}
\alias{collapseReplicates}
\title{Collapse signal from replicates}
\usage{
collapseReplicates(object, collapseAll = FALSE)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}

\item{collapseAll}{TRUE/FALSE, if all samples should be collapsed (TRUE), or
if samples should be kept separate by condition (FALSE)}
}
\value{
object of type \code{\link{BSFDataSet}} with updated signal
}
\description{
Collapses all replicates merges all samples from a \link{BSFDataSet} object
into a single signal stream, only split by minus and plus strand.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

bdsNew = collapseReplicates(bds)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coveragePlots.R
\name{bindingSiteCoveragePlot}
\alias{bindingSiteCoveragePlot}
\title{Plot signal coverage of selected ranges}
\usage{
bindingSiteCoveragePlot(
  object,
  plotIdx,
  flankPos,
  shiftPos = NULL,
  mergeReplicates = FALSE,
  autoscale = FALSE,
  highlight = TRUE,
  showCentralRange = TRUE,
  customRange = NULL,
  customRange.name = "custom",
  customAnnotation = NULL,
  customAnnotation.name = "anno",
  title = NULL,
  colorPalette = NULL
)
}
\arguments{
\item{object}{a \link{BSFDataSet} object}

\item{plotIdx}{integer, index of the range to plot}

\item{flankPos}{numeric, number of nucleotides by which the plotting frame is
symmetrically extended}

\item{shiftPos}{numeric, nucleotide positions by which the current plotting
range should be shifted}

\item{mergeReplicates}{logical, if replicates should be merge per
condition (TRUE) or if every replicates should be shown separately (FALSE)}

\item{autoscale}{logical, if y-axis should be scaled to the maximum for all
replicates (TRUE), or not (FALSE)}

\item{highlight}{logical, if the central range should be highlighted (TRUE),
or not (FALSE)}

\item{showCentralRange}{logical, if the central range should be shown (TRUE),
or not (FALSE)}

\item{customRange}{\code{GenomicRanges}, a custom range object to be shown
underneath the coverage tracks}

\item{customRange.name}{character, the name of the customRange track}

\item{customAnnotation}{\code{GenomicRanges} or \code{TxDb}, a custom annotation
for eg. gene, exons, etc. to be shown underneath the coverage tracks}

\item{customAnnotation.name}{character, the name of the customAnnotation track}

\item{title}{character, set plotting title}

\item{colorPalette}{vector, hex colors used for the conditions}
}
\value{
an object of class \code{GVIZ}
}
\description{
Function plots the coverage of the CLIP data in the signal slot and plots it
as coverage. The plot is centered around a given binding site, which can be
selected by an index.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

bindingSiteCoveragePlot(bds, plotIdx = 3, flankPos = 10)

}
\seealso{
\code{\link{BSFDataSet}}, \code{\link{BSFind}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{rangeCoveragePlot}
\alias{rangeCoveragePlot}
\title{Plot crosslink event coverage over binding site range}
\usage{
rangeCoveragePlot(
  object,
  width = 20,
  show.samples = FALSE,
  subset.chromosome = "chr1",
  quiet = TRUE
)
}
\arguments{
\item{object}{a BSFDataSet, or a list of BSFDataSet}

\item{width}{numeric; set the plotting range to show (in nt)}

\item{show.samples}{logical; to show individual samples as lines}

\item{subset.chromosome}{character; subset by a all ranges on the indicated
chromosome. Can also be a vector with multiple chromosomes. If NULL then all
ranges are being used.}

\item{quiet}{logical; whether to print messages}
}
\value{
a plot of type \code{ggplot2} displaying the crosslink coverage over
the ranges of the given \code{\link{BSFDataSet}}
}
\description{
A diagnostic plot function that allows to check the coverage of crosslink
events over different merged regions. The coverage is shown as mean over all
replicates and conditions, with a standard deviation corridor.
}
\details{
If \code{object} is a single BSFDataObject a single coverage plot will be
drawn, whereas if it is a list of BSFDataObjects, then faceting is used to
make a plot for each list element.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

# plotting a single object
bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1)
rangeCoveragePlot(bds, subset.chromosome = "chr22")

# plotting multiple objects
bds1 <- makeBindingSites(object = bds, bsSize = 3, minWidth = 2,
minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")
bds2 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")
l = list(`1. bsSize = 3` = bds1, `2. bsSize = 9` = bds2)
rangeCoveragePlot(l, subset.chromosome = "chr22")

}
\seealso{
\code{\link{BSFDataSet}}, \code{\link{makeBindingSites}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{add-BSFDataSet}
\alias{add-BSFDataSet}
\alias{+,BSFDataSet,BSFDataSet-method}
\title{Add two \code{\link{BSFDataSet}} objects}
\usage{
\S4method{+}{BSFDataSet,BSFDataSet}(e1, e2)
}
\arguments{
\item{e1}{BSFDataSet; the first dataset}

\item{e2}{BSFDataSet; the second dataset}
}
\value{
A \link{BSFDataSet} object with ranges, signal and meta data from
both inputs.
}
\description{
Use '+' to add two objects of type \code{\link{BSFDataSet}} to each other.
}
\details{
Meta data is extended by binding both input tables together. Ranges are
added by re-centering partially overlapping ranges according to their combined
coverage maximum.

Input ranges must be of the same size. Differently size objects cannot be
added. For this and other usecases please see \code{\link{combineBSF}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

# make binding sites
bds = makeBindingSites(bds, bsSize = 7)

# split ranges in two groups
allRanges = getRanges(bds)
set.seed(1234)
idx = sample(1:length(allRanges), size = length(allRanges)/2, replace = FALSE)
r1 = allRanges[idx]
r2 = allRanges[-idx]

# splite meta data
allMeta = getMeta(bds)
m1 = allMeta[1:2,]
m2 = allMeta[3:4,]

# create new objects
bds1 = setRanges(bds, r1)
bds2 = setRanges(bds, r2)
bds1 = setMeta(bds1, m1)
bds2 = setMeta(bds2, m2)
bds1 = setName(bds1, "test1")
bds2 = setName(bds2, "test2")

# merge two objects with '+' operator
c1 = bds1 + bds2

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{supportRatioPlot}
\alias{supportRatioPlot}
\title{Plot that shows the binding site support ratio}
\usage{
supportRatioPlot(object, bsWidths, bsFlank = NA, ...)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{bsWidths}{a numeric vector indicating the different binding site
width to compute the ratio for}

\item{bsFlank}{optional; a numeric vector of the same length as
\code{bsWidth} used to specify the width of the flanking regions}

\item{...}{further arguments passed to \code{makeBindingSites}}
}
\value{
an object of class \code{ggplot2}
}
\description{
Function that shows a ratio to determine how well a given binding
site with is supported by the crosslink coverage of the data.
Ratios are computed using the \code{\link{supportRatio}} function.
}
\details{
The higher the ratio, the more does the given binding site width captures
the enrichment of crosslinks compared the the local surrounding. A ratio
equal to 1 would mean no enrichemnt at all.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

suppressWarnings(supportRatioPlot(bds, bsWidths = c(3,7),
minWidth = 1, minClSites = 1, minCrosslinks = 2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/differentialFunctions.R
\name{filterBsBackground}
\alias{filterBsBackground}
\title{Filter for genes not suitable for differential testing}
\usage{
filterBsBackground(
  object,
  minCounts = TRUE,
  minCounts.cutoff = 100,
  balanceBackground = TRUE,
  balanceBackground.cutoff.bs = 0.3,
  balanceBackground.cutoff.bg = 0.7,
  balanceCondition = TRUE,
  balanceCondition.cutoff = 0.05,
  match.geneID = "geneID",
  flag = FALSE,
  quiet = FALSE,
  veryQuiet = FALSE
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with computed count data
for binding sites and background regions}

\item{minCounts}{logical; whether to use the minimum count filter}

\item{minCounts.cutoff}{numeric; the minimal number of crosslink
per gene over all samples for the gene to be retained (default = 100)}

\item{balanceBackground}{logical; whether to use the counts balancing filter
between binding sites and background}

\item{balanceBackground.cutoff.bs}{numeric; the maximum fraction of the total signal
per gene that can be within binding sites (default = 0.2)}

\item{balanceBackground.cutoff.bg}{numeric; the minimum fraction of the total signal
per gene that can be within the background (default = 0.8)}

\item{balanceCondition}{logical; whether to use the counts balancing filter
between conditions}

\item{balanceCondition.cutoff}{numeric; the maximum fraction of the total signal
that can be attributed to only one condition}

\item{match.geneID}{character; the name of the column with the gene ID
in the binding sites meta columns used for matching binding sites to genes}

\item{flag}{logical; whether to remove or flag binding sites from genes that do not
pass any of the filters}

\item{quiet}{logical; whether to print messages or not}

\item{veryQuiet}{logical; whether to print messages or not}
}
\value{
an object of class \code{\link{BSFDataSet}} with biniding sites filtered
or flagged by the above filter options
}
\description{
This function removes genes where the differential testing protocol can
not be applied to, using count coverage information on the binding sites
and background regions per gene, through the following steps:
\enumerate{
\item Remove genes with overall not enough crosslinks: \code{minCounts}
\item Remove genes with a disproportion of counts in binding sites vs. the
background: \code{balanceBackground}
\item Remove genes where the expression between conditions is too much off
balance: \code{balanceCondition}
}
}
\details{
To remove genes with overall not enough crosslinks (\code{minCounts})
all counts are summed up per gene across all samples and compared to the
minimal count threshold (\code{minCounts.cutoff}).

To remove genes with a count disproportion between binding sites and
background regions crosslinks are summed up for binding sites and background
per gene. These sums are combined in a ratio. Genes where eg. 50\% of all
counts are within binding sites would be removed
(see \code{balanceBackground.cutoff.bs} and \code{balanceBackground.cutoff.bg}).

To remove genes with very large expression differences between conditions,
crosslinks are summed up per gene for each condition. If now eg. the total
number of crosslinks is for 98\% in one condition and only 2\% of the
combined signal is in the second condition, expression levels are too
different for a reliable comparisson (see \code{balanceCondition.cutoff}).

This function is intended to be used right after a call of \code{\link{calculateBsBackground}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

# make testset
bds = makeBindingSites(bds, bsSize = 7)
bds = assignToGenes(bds, anno.genes = gns)
bds = imputeBsDifferencesForTestdata(bds)
bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)

# use all filters and remove binding sites that fail (default settings)
f0 = filterBsBackground(bds)

# do not use the condition balancing filter
f1 = filterBsBackground(bds, balanceCondition = FALSE)

# use only the minimum count filter and flag binding sites instead of
# removing them
f3 = filterBsBackground(bds, flag = TRUE, balanceCondition = FALSE,
 balanceBackground = FALSE)

}
\seealso{
\code{\link{calculateBsBackground}}, \code{\link{plotBsBackgroundFilter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{globalScorePlot}
\alias{globalScorePlot}
\title{Plot the PureCLIP score distribution after re-assignment}
\usage{
globalScorePlot(object)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the PureCLIP score distribution on a log2
scale after the re-assignment on binding site level.
The function \code{\link{annotateWithScore}} is expected to be executed
prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds1 = makeBindingSites(object = bds, bsSize = 9)
bds1 = annotateWithScore(bds1, match.ranges = getRanges(bds))
globalScorePlot(bds1)

}
\seealso{
\code{\link{annotateWithScore}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{getRanges}
\alias{getRanges}
\alias{getRanges,BSFDataSet-method}
\title{Accessor method for the ranges of the BSFDataSet object}
\usage{
getRanges(object)

\S4method{getRanges}{BSFDataSet}(object)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}
}
\value{
returns the genomic ranges (GRanges) of the associated ranges
}
\description{
The ranges slot holds the genomic ranges information of the sites currently
in the object. They are encoded as a GRanges object with each binding site
having a single ranges entry.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

getRanges(bds)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{BSFDataSet}
\alias{BSFDataSet}
\alias{BSFDataSet,}
\alias{BSFDataSet-class,}
\alias{BSFDataSetFromBigWig}
\title{BSFDataSet object and constructors}
\usage{
BSFDataSet(ranges, meta, signal, dropSeqlevels = TRUE, silent = FALSE)

BSFDataSet(ranges, meta, signal, dropSeqlevels = TRUE, silent = FALSE)

BSFDataSetFromBigWig(ranges, meta, silent = FALSE, dropSeqlevels = TRUE)
}
\arguments{
\item{ranges}{a \code{GenomicRanges} with the desired ranges to process. The
strand slot must be either + or -.}

\item{meta}{a \code{data.frame} with at least two columns. The first column
should be a unique numeric id. The second column holds sample type
information, such as the condition.}

\item{signal}{a \code{list} with the two entries 'signalPlus' and
'signalMinus', following a special representation of \code{SimpleRleList}
for counts per replicates (see details for more information).}

\item{dropSeqlevels}{enforce seqnames to be the same in ranges and signal,
by dropping unused seqlevels which is required for most downstream functions
such as \code{coverageOverRanges}}

\item{silent}{suppress messages but not warnings (TRUE/ FALSE)}
}
\value{
A BSFDataSet object.
}
\description{
\code{BSFDataSet} contains the class \code{GenomicRanges}, which is used to
store input ranges. Alongside with the iCLIP signal in \code{list} structure
and additional meta data as \code{data.frame}.
}
\details{
The ranges are enforced to have to have a "+" or "-" strand
annotation,"*" is not allowed. They are expected to be of the same width
and a warning is thrown otherwise.

The meta information is stored as \code{data.frame} with at least two
required columns, 'id' and 'condition'. They are used to build the unique
identifier for each replicate split by '_' (eg. id = 1 and condition = WT
will result in 1_WT).

The meta data needs to have the additional columns 'clPlus' and 'clMinus'
to be present if \code{BSFDataSetFromBigWig} is called.
It is used to provide the location to the iCLIP coverage
files to the import function. On object initialization these files are
loaded and internally represented in the signal slot of the object
(see \code{\link{BSFDataSet}}).

The iCLIP signal is stored in a special list structure.
At the lowest level crosslink counts per nucleotide are stored as
\code{Rle} per chromosome summarized as a \code{SimpleRleList}. Such a list
exits for each replicate and must be named by the replicate identifier
(eg. 1_WT). Therefore this list contains always exactly the same number of
entries as the number of replicates in the dataset. Since we handle strands
initially seperated from each other this list must be given twice, once
for each strand. The strand specific entries must be named 'signalPlus' and
'signalMinus'.

The option \code{dropSeqlevels} forces the seqnames of the ranges and
the signal to be the same. If for a specific chromosome in the ranges
no respective entry in the signal list can be found, then entries with
that chromosome are dropped This behavior is needed to keep the
\link{BSFDataSet} object in sync, which is required for downstream functions
such as \code{coverageOverRanges}
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
rng = getRanges(bds)
sgn = getSignal(bds)
mta = getMeta(bds)
bdsNew = BSFDataSet(ranges = rng, signal = sgn, meta = mta)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{reproducibilityFilterPlot}
\alias{reproducibilityFilterPlot}
\title{Plot to that shows the crosslink site distribution per replicate}
\usage{
reproducibilityFilterPlot(object, plotRange = 20)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{plotRange}{numeric; number of crosslinks per sites that should be shown
before summing them up}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the number of crosslinks sites over the number
of crosslink in these sites and highlights the replicate specific
reproducibility cutoff that is derived from that distribution.
The function \code{\link{reproducibilityFilter}} is expected to be executed prior
to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds = makeBindingSites(object = bds, bsSize = 9)
bds = reproducibilityFilter(bds)
reproducibilityFilterPlot(bds)

}
\seealso{
\code{\link{reproducibilityFilter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/differentialFunctions.R
\name{calculateBsBackground}
\alias{calculateBsBackground}
\title{Compute background coverage for binding sites per gene}
\usage{
calculateBsBackground(
  object,
  anno.annoDB = NULL,
  anno.genes = NULL,
  blacklist = NULL,
  use.offset = TRUE,
  ranges.offset = NULL,
  match.geneID.gene = "gene_id",
  match.geneID.bs = "geneID",
  match.geneID.blacklist = "geneID",
  generate.geneID.bs = FALSE,
  generate.geneID.blacklist = FALSE,
  uniqueID.gene = "gene_id",
  uniqueID.bs = "bsID",
  uniqueID.blacklist = "bsID",
  force.unequalSites = FALSE,
  quiet = FALSE,
  veryQuiet = TRUE,
  ...
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with two conditions}

\item{anno.annoDB}{an object of class \code{OrganismDbi} that contains
the gene annotation.}

\item{anno.genes}{an object of class \code{\link{GenomicRanges}} that represents
the gene ranges directly}

\item{blacklist}{GRanges; genomic ranges where the signal should be
excluded from the background}

\item{use.offset}{logical; if an offset region around the binding sites should
be used on which the signal is excluded from the background}

\item{ranges.offset}{numeric; number of nucleotides the offset window around
each binding site should be wide (defaults to 1/2 binding site width - NULL)}

\item{match.geneID.gene}{character; the name of the column with the gene ID
in the genes meta columns used for matching binding sites to genes}

\item{match.geneID.bs}{character; the name of the column with the gene ID
in the binding sites meta columns used for matching binding sites to genes}

\item{match.geneID.blacklist}{character; the name of the column with the gene
ID in the blacklist meta columns used for matching the blacklist regions
with the genes}

\item{generate.geneID.bs}{logical; if the binding site to gene matching
should be performed if no matching gene ID is provided}

\item{generate.geneID.blacklist}{logical; if the blacklist to gene matching
should be performed if no matching gene ID is provided}

\item{uniqueID.gene}{character; column name of a unique ID for all genes}

\item{uniqueID.bs}{character; column name of a unique ID for all binding sites}

\item{uniqueID.blacklist}{character; column name of a unique ID for all
blacklist regions}

\item{force.unequalSites}{logical; if binding sites of not identical width
should be allowed or not}

\item{quiet}{logical; whether to print messages or not}

\item{veryQuiet}{logical; whether to print messages or not}

\item{...}{additional arguments; passed to \code{\link{assignToGenes}}}
}
\value{
an object of class \code{\link{BSFDataSet}} with counts for binding
sites, background and total gene added to the meta column of the ranges
}
\description{
This function computes the background coverage used for the differential
binding analysis to correct for transcript level changes. Essentially,
the crosslink signal on each gene is split into crosslinks that can be
attributed to the binding sites and all other signal that can be attributed
to the background.
}
\details{
To avoid that crosslinks from binding sites contaminate the background counts
a protective region around each binding sites can be spanned with
\code{use.offset} the default width of the offset region is half of the
binding site width, but can also be changed with the \code{ranges.offset}
parameter.

Additional region that one wants to exclude from contributing to the
background signal can be incorporated as \code{GRanges} objects through
the \code{blacklist} option.

It is expected that binding sites are assigned to hosting genes prior to
running this funciton (see \code{\link{BSFind}}). This means a unique gene ID
is present in the meta columns of each binding site ranges. If this is not the
case one can invoce the binding site to gene assignment with
\code{generate.geneID.bs}. The same is true for the blacklist regions with
option \code{generate.geneID.blacklist}.

It is expected that all binding sites are of the same size
(See \code{\link{BSFind}} on how to achieve this). If this is however not
the case and one wants to keep binding sites of different with then option
\code{force.unequalSites} can be used.

This function is intended to be used for the generation of the count matrix
used for the differential binding analysis. It is usually preceded by
\code{\link{combineBSF}} and followed by \code{\link{filterBsBackground}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

# make binding sites
bds = makeBindingSites(bds, bsSize = 7)
bds = assignToGenes(bds, anno.genes = gns)

# change meta data
m = getMeta(bds)
m$condition = factor(c("WT", "WT", "KO", "KO"), levels = c("WT", "KO"))
bds = setMeta(bds, m)

# change signal
s = getSignal(bds)
names(s$signalPlus) = paste0(m$id, "_", m$condition)
names(s$signalMinus) = paste0(m$id, "_", m$condition)
bds = setSignal(bds, s)

# make example blacklist region
myBlacklist = getRanges(bds)
set.seed(1234)
myBlacklist = sample(myBlacklist, size = 500) + 4

# make background
bds.b1 = calculateBsBackground(bds, anno.genes = gns)

# make background - no offset
bds.b2 = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)

# make background - use blacklist
bds.b3 = calculateBsBackground(bds, anno.genes = gns, blacklist = myBlacklist)

}
\seealso{
\code{\link{combineBSF}},
\code{\link{filterBsBackground}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{mergeSummaryPlot}
\alias{mergeSummaryPlot}
\title{Plot summarized results of the different binding site merging and filtering
steps}
\usage{
mergeSummaryPlot(
  object,
  select = c("all", "filter", "inputRanges", "minClSites", "mergeCrosslinkSites",
    "minCrosslinks", "centerIsClSite", "centerIsSummit"),
  ...
)
}
\arguments{
\item{object}{a BSFDataObject, with the makeBindingSites function already run}

\item{select}{one of "all", "filter", "inputRanges",
"minCLSites", "mergeCrosslinkSites", "minCrosslinks",
"centerIsClSite" or "centerIsSummit".
Defines which parameter is selected for plotting.}

\item{...}{further arguments passed to ggplot}
}
\value{
a plot of type ggplot after the \code{\link{makeBindingSites}}
function was run
}
\description{
Bar charts produced for the different filter steps in the binding site
merging routine. Depending on the selected option (\code{select}) all or
only a user defined filter can be shown.
}
\details{
If \code{object} is a single BSFDataObject a single coverage plot will be
drawn, whereas if it is a list of BSFDataObjects, then faceting is used to
make a plot for each list element.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

# plotting a single object
bds0 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1)
mergeSummaryPlot(bds0)

# plotting mulitple obejcts
bds1 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")
bds2 <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 3, sub.chr = "chr22")
l = list(`1. bsSize = 3` = bds1, `2. bsSize = 9` = bds2)
mergeSummaryPlot(l, width = 20)

}
\seealso{
\code{\link{makeBindingSites}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoverageFunctions.R
\name{coverageOverRanges}
\alias{coverageOverRanges}
\title{Coverage function for BSFDataSet objects}
\usage{
coverageOverRanges(
  object,
  returnOptions = c("merge_ranges_keep_positions", "merge_replicates_per_condition",
    "merge_all_replicates", "merge_positions_keep_replicates"),
  method = "sum",
  allowNA = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{returnOptions}{one of merge_ranges_keep_positions,
merge_replicates_per_condition, merge_all_replicates,
merge_positions_keep_replicates}

\item{method}{sum/ mean, select how replicates/ ranges should be summarized}

\item{allowNA}{TRUE/ FALSE, allow NA values in the output if input ranges
are of different width}

\item{quiet}{logical, whether to print messages}
}
\value{
an object of class specified in \code{returnOptions}
}
\description{
The crosslink coverage is computed for all ranges in the the given
\code{BSFDataSet} object (see \code{\link{BSFDataSet}} for details).
Depending on the \code{returnOptions} the resulting coverage information is
summarized, suitable for diverse computation and plotting tasks. The
coverage can only be compute for objects with identical ranges.
}
\details{
If \code{returnOptions} is set to merge_ranges_keep_positions:
Returns a matrix with ncol being the nucleotides of the ranges (equal to the
width of the input ranges) and nrow being the number of replicates in the
meta information.

If \code{returnOptions} is set to merge_replicates_per_condition:
Returns a list of matrices. Each list corresponds to one condition set in
the meta information. The matrix in each entry has ncols equal to the
ranges width and nrow equal to the number of ranges. Counts per ranges
and position are summed.

If \code{returnOptions} is set to merge_all_replicates:
Returns a matrix with ncols equal to the range width and nrow equal to the
number of ranges. Counts per range and position are summed.

If \code{returnOptions} is set to merge_positions_keep_replicates:
Returns a GRanges object where the counts are summed for each replicate
and added to the original granges object.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

rng = coverageOverRanges(bds, returnOptions = "merge_ranges_keep_positions")
rng = coverageOverRanges(bds, returnOptions = "merge_replicates_per_condition")
rng = coverageOverRanges(bds, returnOptions = "merge_all_replicates")
rng = coverageOverRanges(bds, returnOptions = "merge_positions_keep_replicates")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/differentialPlots.R
\name{plotBsBackgroundFilter}
\alias{plotBsBackgroundFilter}
\title{Diagnostic plots for the differential binding background}
\usage{
plotBsBackgroundFilter(
  object,
  filter = c("minCounts", "balanceBackground", "balanceCondition")
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object with background counts filtered
by \code{\link{filterBsBackground}}}

\item{filter}{character; which filter to display in the plot (one of:
'minCounts', 'balanceBackground', 'balanceCondition')}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
To perform differential binding analysis between two conditions the
\code{\link{calculateBsBackground}} function groups crosslinks per gene
into those from binding sites and those from background regions.
The \code{\link{filterBsBackground}} function perfroms certain
filtering operations on that background to ensure that it's suitable
for differential testing. This function visually displays the effect
of these filtering operations.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

# make testset
bds = makeBindingSites(bds, bsSize = 7)
bds = assignToGenes(bds, anno.genes = gns)
bds = imputeBsDifferencesForTestdata(bds)
bds = calculateBsBackground(bds, anno.genes = gns, use.offset = FALSE)

# use all filters and remove binding sites that fail (default settings)
bds = filterBsBackground(bds)

# display minCount filter
plotBsBackgroundFilter(bds, filter = "minCounts")

# display balance background filter
plotBsBackgroundFilter(bds, filter = "balanceBackground")

# display balance condition filter
plotBsBackgroundFilter(bds, filter = "balanceCondition")

}
\seealso{
\code{\link{calculateBsBackground}}

\code{\link{filterBsBackground}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Functions.R
\name{calculateSignalToFlankScore}
\alias{calculateSignalToFlankScore}
\title{Calculate signal-to-flank score}
\usage{
calculateSignalToFlankScore(
  object,
  flank = c("bs", "manual"),
  flank.size = NULL,
  quiet = FALSE
)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{flank}{character; how the flanking region shoule be set. Options are
'bs', 'manual'}

\item{flank.size}{numeric; if flank='manual' provide the desired flanking size}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of class \code{\link{BSFDataSet}} with signal-to-flank ratios
added to the meta column of the ranges.
}
\description{
This function calculates the signal-to-flank ratio for all present binding
sites.
}
\details{
Each input range is treated as a binding site. For a particular binding site
all overlapping crosslinks are summed up and divided by the normalized sum of
the crosslinks in the two adjecent regions of the same size. This is done
for all bining sites and the ratio is reported as a score.

The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
bds = makeBindingSites(bds, bsSize = 5)
bds = calculateSignalToFlankScore(bds)

}
\seealso{
\code{\link{BSFind}}, \code{\link{bindingSiteDefinednessPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bindingsites.R
\name{makeBindingSites}
\alias{makeBindingSites}
\title{Define equally sized binding sites from peak calling results and iCLIP
crosslink events.}
\usage{
makeBindingSites(
  object,
  bsSize = NULL,
  minWidth = 2,
  minCrosslinks = 2,
  minClSites = 1,
  centerIsClSite = TRUE,
  centerIsSummit = TRUE,
  sub.chr = NA,
  quiet = FALSE
)
}
\arguments{
\item{object}{a BSFDataSet object (see \code{\link{BSFDataSet}})}

\item{bsSize}{an odd integer value specifying the size of the output
binding sites}

\item{minWidth}{the minimum size of regions that are subjected to the
iterative merging routine, after the initial region concatenation.}

\item{minCrosslinks}{the minimal number of positions to overlap with at least
one crosslink event in the final binding sites}

\item{minClSites}{the minimal number of crosslink sites that have to
overlap a final binding site}

\item{centerIsClSite}{logical, whether the center of a final binding
site must be covered by an initial crosslink site}

\item{centerIsSummit}{logical, whether the center of a final binding
site must exhibit the highest number of crosslink events}

\item{sub.chr}{chromosome identifier (eg, chr1, chr2) used for subsetting the
BSFDataSet object. This option can be used for testing different
parameter options}

\item{quiet}{logical, whether to print info messages}
}
\value{
an object of type BSFDataSet with modified ranges
}
\description{
This function performs the merging of single nucleotide crosslink sites into
binding sites of a user defined width (\code{bsSize}). Depending on the
desired output width crosslink sites with a distance closer than
\code{bsSize} -1 are concatenated. Initially all input regions are
concatenated and then imperatively merged and extended. Concatenated regions
smaller than \code{minWidth} are removed prior to the merge and extension
routine. This prevents outlier crosslink pileup, eg. mapping artifacts
to be integrated into the final binding sites. All remaining regions are
further processed and regions larger than the desired output width are
interactively split up by setting always the position with the highest
number of crosslinks as center. Regions smaller than the desired width are
symmetrically extended. Resulting binding sites are then filtered by the
defined constraints.
}
\details{
The \code{bsSize} argument defines the final output width of the merged
binding sites. It has to be an odd number, to ensure that a binding site
has a distinct center.

The \code{minWidth} parameter is used to describe the minimum width a ranges
has to be after the initial concatenation step. For example:
Consider bsSize = 9 and minWidth = 3. Then all initial crosslink sites that
are closer to each other than 8 nucleotides (bsSize -1) will be concatenated.
Any of these ranges with less than 3 nucleotides of width will be removed,
which reflects about 1/3 of the desired binding site width.

The argument \code{minCrosslinks} defines how many positions of the binding
sites are covered with at least one crosslink event. This threshold has to
be defined in conjunction with the binding site width. A default value of 3
with a binding site width of 9 means that 1/3 of all positions in the final
binding site must be covered by a crosslink event. Setting this filter to 0
deactivates it.

The \code{minClSites} argument defines how many positions of the binding site
must have been covered by the original crosslink site input. If the input was
based on the single nucleotide crosslink positions computed by PureCLIP than
this filter checks for the number of positions originally identified by
PureCLIP in the computed binding sites. The default of \code{minClSites} = 1
essentially deactivates this filter. Setting this filter to 0 deactivates it.

The options \code{centerIsClSite} and \code{centerIsSummit} ensure that the
center of each binding site is covered by an initial crosslink site and
represents the summit of crosslink events in the binding site, respectively.

The option \code{sub.chr} allows to run the binding site merging on a
smaller subset (eg. "chr1") for improoved computational speed when testing
the effect of various binding site width and filtering options.
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

# standard options, no subsetting
bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1)

# standard options, with subsetting
bds <- makeBindingSites(object = bds, bsSize = 9, minWidth = 2,
minCrosslinks = 2, minClSites = 1, sub.chr = "chr22")

}
\seealso{
\code{\link{BSFDataSet}}, \code{\link{BSFind}},
\code{\link{mergeCrosslinkDiagnosticsPlot}},
\code{\link{makeBsSummaryPlot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{bindingSiteDefinednessPlot}
\alias{bindingSiteDefinednessPlot}
\title{Binding site definedness plot}
\usage{
bindingSiteDefinednessPlot(
  object,
  by = c("all", "transcript_region", "gene_type"),
  showN.genes = 5,
  show.others = FALSE
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{by}{character; the option by which the plot should be grouped by.
Options are: "all", "transcript_region", "gene_type"}

\item{showN.genes}{numeric; if \code{by} is `gene_type`, then this argument
set the maximum number of groups to be shown in the plot}

\item{show.others}{logical; whether to show 'others' category.}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
Binding site definedness is given by the percent of crosslinks that fall
diretly inside the binding site compare to those around the binding site.
This plotting function shows the distribution of those percentage values
grouped by what is indicated in the \code{by} argument.
}
\details{
If \code{by} = 'all', then all binding site are grouped into one distribution.
For options 'transcript_region' and 'gene_type' binding sites are split into
groups according to the respective assignment. This requires that the respective
assignment function was executed on the dataset prior to calling this plot function.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
bds = BSFind(bds, anno.genes = gns, anno.transcriptRegionList = regions,
 est.subsetChromosome = "chr22")
bds = calculateSignalToFlankScore(bds)
bindingSiteDefinednessPlot(bds)

}
\seealso{
\code{\link{BSFind}}, \code{\link{calculateSignalToFlankScore}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{processingStepsFlowChart}
\alias{processingStepsFlowChart}
\title{Step-wise flowchart plot}
\usage{
processingStepsFlowChart(object, size.all = 3)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{size.all}{numeric; size of all numbers}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
An overview plot that shows all workflow functions that were executed on the
current object, with all input and output binding site numbers and major options
that were used. The function can be called at any time in the analysis. Most
optimal usage is after a full run of the wrapper function \code{\link{BSFind}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])
load(list.files(files, pattern = ".rds$", full.names = TRUE)[2])
bds = BSFind(bds, anno.genes = gns, anno.transcriptRegionList = regions,
 est.subsetChromosome = "chr22")
processingStepsFlowChart(bds)

}
\seealso{
\code{\link{BSFind}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{getName}
\alias{getName}
\alias{getName,BSFDataSet-method}
\title{Accessor method for the name of the BSFDataSet object}
\usage{
getName(object)

\S4method{getName}{BSFDataSet}(object)
}
\arguments{
\item{object}{a \code{BSFDataSet} object}
}
\value{
returns the name of the dataset
}
\description{
The name slot holds the name of the dataset
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

getName(bds)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFunction.R
\name{pureClipGlobalFilterPlot}
\alias{pureClipGlobalFilterPlot}
\title{Plot the PureCLIP score distribution with global cutoff indicator}
\usage{
pureClipGlobalFilterPlot(object)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}
}
\value{
a plot of type \code{\link{ggplot}}
}
\description{
A diagnostic function that plots the PureCLIP score distribution on a log2
scale. The function \code{\link{pureClipGlobalFilter}} is expected to be
executed prior to calling this plot function.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
# apply 5\% filter
bds = pureClipGlobalFilter(object = bds, cutoff = 0.05)
pureClipGlobalFilterPlot(bds)

}
\seealso{
\code{\link{pureClipGlobalFilter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/methods.R
\docType{methods}
\name{setMeta}
\alias{setMeta}
\alias{setMeta,BSFDataSet-method}
\title{Setter method for the meta data of the BSFDataSet object}
\usage{
setMeta(object, ...)

\S4method{setMeta}{BSFDataSet}(object, newMeta)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{...}{additional arguments}

\item{newMeta}{the replacement meta data table}
}
\value{
an object of type \code{\link{BSFDataSet}} with updated meta data
}
\description{
Meta data is stored as a \code{data.frame} and must contain the columns
"condition", "clPlus" and "clMinus".
}
\examples{

# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

nMeta = getMeta(bds)
setMeta(bds, nMeta)

}
\seealso{
\code{\link{BSFDataSet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/differentialFunctions.R
\name{calculateBsFoldChange}
\alias{calculateBsFoldChange}
\title{Compute fold-changes per binding site}
\usage{
calculateBsFoldChange(
  object,
  fitType = "local",
  sfType = "ratio",
  minReplicatesForReplace = 10,
  independentFiltering = TRUE,
  alpha = 0.05,
  pAdjustMethod = "BH",
  minmu = 0.5,
  filterFun = NULL,
  use.lfc.shrinkage = FALSE,
  type = c("ashr", "apeglm", "normal"),
  svalue = FALSE,
  apeAdapt = TRUE,
  apeMethod = "nbinomCR",
  match.geneID = "geneID",
  quiet = TRUE,
  veryQuiet = FALSE,
  replaceNegative = FALSE,
  removeNA = FALSE
)
}
\arguments{
\item{object}{a \code{\link{BSFDataSet}} object}

\item{fitType}{either "parametric", "local", "mean", or "glmGamPoi" for the
type of fitting of dispersions to the mean intensity.
See \code{\link[DESeq2]{DESeq}} for more details.}

\item{sfType}{either "ratio", "poscounts", or "iterate" for the type of size
factor estimation. See \code{\link[DESeq2]{DESeq}} for more details.}

\item{minReplicatesForReplace}{the minimum number of replicates required in
order to use replaceOutliers on a sample. See \code{\link[DESeq2]{DESeq}} for
more details.}

\item{independentFiltering}{logical, whether independent filtering should be
applied automatically. See \code{\link[DESeq2]{results}} for more details.}

\item{alpha}{the significance cutoff used for optimizing the independent
filtering. See \code{\link[DESeq2]{results}} for more details.}

\item{pAdjustMethod}{he method to use for adjusting p-values. See
\code{\link[DESeq2]{results}} for more details.}

\item{minmu}{lower bound on the estimated count. See
\code{\link[DESeq2]{results}} for more details.}

\item{filterFun}{an optional custom function for performing independent
filtering and p-value adjustment. See \code{\link[DESeq2]{results}} for more
details.}

\item{use.lfc.shrinkage}{logical; whether to compute shrunken log2 fold
changes for the DESeq results. See \code{\link[DESeq2]{lfcShrink}} for more
details.}

\item{type}{if 'ashr', 'apeglml' or 'normal' should be used for fold change
shrinkage. See \code{\link[DESeq2]{lfcShrink}} for more details.}

\item{svalue}{logical, should p-values and adjusted p-values be replaced with
s-values when using apeglm or ashr. See \code{\link[DESeq2]{lfcShrink}}
for more details.}

\item{apeAdapt}{logical, should apeglm use the MLE estimates of LFC to adapt
the prior. See \code{\link[DESeq2]{lfcShrink}} for more details.}

\item{apeMethod}{what method to run apeglm, which can differ in terms of
speed. See \code{\link[DESeq2]{lfcShrink}} for more details.}

\item{match.geneID}{character; the name of the column with the gene ID
in the binding sites meta columns used for matching binding sites to genes}

\item{quiet}{logical; whether to print messages or not}

\item{veryQuiet}{logical; whether to print messages or not}

\item{replaceNegative}{logical; force negative counts to be replaces by 0.
Be careful when using this, having negative counts can point towards problems
with the gene annotation in use.}

\item{removeNA}{logical; force binding sites with any NA value to be removed.
Be careful when using this, having negative counts can point towards problems
with the gene annotation in use.}
}
\value{
a \code{\link{BSFDataSet}} object, with results from the
\code{\link[DESeq2]{DESeq}} analysis added to the meta columns of the
binding site ranges.
}
\description{
Given count data for binding sites and background regions this function will
compute fold-changes between two condition for each binding site. Computation
is based on \code{\link[DESeq2]{DESeq}} using the Likelihood ratio test to
disentangle transcription level changes from binding site level changes.
}
\details{
Fold-changes per binding sites are corrected for transcript level changes.
Essentially, background counts are used to model transcript level changes,
which are then used to compute fold-changes per binding site, which are
corrected for the observed transcript level changes. This is done by using a
Likelihood ratio test comparing the full model
(~condition + type + condition:type) to the reduced model (~condition + type).

Fold-changes for the transcript level changes are modeled explicitly in a
second round of the \code{\link[DESeq2]{DESeq}} workflow, using the default
Wald test to compare changes between the conditions (~condition). Counts
attributed to binding sites are removed from the gene level counts.

Results from both calculation rounds can be filtered and further manipulated
with parameters given in the DESeq2 framework
(see \code{\link[DESeq2]{results}}, \code{\link[DESeq2]{lfcShrink}}).

This function is intended to be used right after a call of \code{\link{filterBsBackground}}.
}
\examples{
# load clip data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))
load(list.files(files, pattern = ".rds$", full.names = TRUE)[1])

# make example dataset
bds = makeBindingSites(bds, bsSize = 7)
bds = assignToGenes(bds, anno.genes = gns)
bds = imputeBsDifferencesForTestdata(bds)
bds = calculateBsBackground(bds, anno.genes = gns)

# calculate fold changes - no shrinkage
bds = calculateBsFoldChange(bds)

# calculate fold changes - with shrinkage
bds = calculateBsFoldChange(bds, use.lfc.shrinkage = TRUE)

}
\seealso{
\code{\link{calculateBsBackground}},
\code{\link{filterBsBackground}},
\code{\link{plotBsBackgroundFilter}},
\code{\link[DESeq2]{DESeq}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Functions.R
\name{reproducibilityFilter}
\alias{reproducibilityFilter}
\title{Replicate reproducibility filter function}
\usage{
reproducibilityFilter(
  object,
  cutoff = NULL,
  nReps = NULL,
  minCrosslinks = 1,
  returnType = c("BSFDataSet", "data.frame"),
  n.reps = lifecycle::deprecated(),
  min.crosslinks = lifecycle::deprecated(),
  quiet = FALSE
)
}
\arguments{
\item{object}{a BSFDataSet object}

\item{cutoff}{numeric; percentage cutoff to be used for the
reproducibility quantile filtering}

\item{nReps}{numeric; number of replicates that must meet the cutoff
defined in \code{cutoff} for a binding site to be called reproducible.
Defaults to N-1.}

\item{minCrosslinks}{numeric; minimal number of crosslinks a binding
site needs to have to be called reproducible. Acts as a lower boundary for
\code{cutoff}. Defaults to 1.}

\item{returnType}{one of "BSFDataSet" or "data.frame". "BSFDataSet" is the
default and "matrix" can be used for easy plotting.}

\item{n.reps}{deprecated -> use nReps instead}

\item{min.crosslinks}{deprecated -> use minCrosslinks instead}

\item{quiet}{logical; whether to print messages}
}
\value{
an object of type BSFDataSet
}
\description{
For each replicate the number of binding sites with a certain number of
crosslinks is calculated. A quantile based threshold (\code{cutoff}) is
applied to each replicate. This indicates how many of the merged binding
sites are supported by crosslinks from the respective replicate. Next, one
can specify how many replicates need to pass the defined threshold for a
binding site to be considered reproducible.
}
\details{
If \code{cutoff} is a single number then the indicated cutoff will be
applied to all replicates. If it is a vector then each element in the vector
is applied to all replicates of the respective condition. The order is
hereby given by the levels of the condition column of the meta data
(see \code{\link{BSFDataSet}},\code{\link{getMeta}}). If the condition
specific filter is applied, a meta column is added to the GRanges of the
\code{BSFDataSet} object, indicating the support for each condition.

If \code{nReps} is a single number then this number is used as treshold for
all binding sites. If it is a vector then it is applied to the replicates of
the respective condition (like in \code{cutoff}). This allows the
application of different thresholds for experiments of different
experimental conditions. If the condition specific filter is applied, a meta
column is added to the GRanges of the \code{BSFDataSet} object,
indicating the support for each condition.

The function is part of the standard workflow performed by \code{\link{BSFind}}.
}
\examples{
# load data
files <- system.file("extdata", package="BindingSiteFinder")
load(list.files(files, pattern = ".rda$", full.names = TRUE))

# merge binding sites
bds <- makeBindingSites(object = bds, bsSize = 9)

# use default return with condition specific threshold
bds = reproducibilityFilter(bds, cutoff = 0.1, nReps = 1)

}
\seealso{
\code{\link{BSFind}},
\code{\link{reproducibilityFilterPlot}},
\code{\link{reproducibilitySamplesPlot}},
\code{\link{reproducibilityScatterPlot}}
}
